/*
 * channel.h
 * Copyright (C) 1999 Patrick Alken
 *
 * $Id: channel.h,v 1.00 1999/08/21 13:57:38 wnder Exp $
 */

typedef struct ChannelBanInfo
{
	struct ChannelBanInfo *next, *prev;
	char *who; /* who set the ban */
	time_t when; /* when the ban was made */
	char *mask; /* hostmask of the ban */
} BanInfo;

typedef struct Exception
{
	struct Exception *next, *prev;
	char *who; /* who set the exception */
	time_t when; /* when it was set */
	char *mask; /* exception hostmask */
} Except;

struct c_user
{
	struct c_user *next;
	Luser *lptr; /* pointer to user structure */
	int flags; /* flags such as opped/voiced */
};

/* Stores info for network channels */
struct ChannelInfo

{
	struct ChannelInfo *next, *prev, *hnext;

#ifdef BLOCK_ALLOCATION
	char name[CHANNELLEN + 1]; /* channel name */
	char key[KEYLEN + 1];
#else
	char *name; /* channel name */
	char *key; /* NULL if no key */
#endif /* BLOCK_ALLOCATION */

	int numusers; /* number of users in the channel */
	int modes; /* channel modes */
	int limit; /* 0 if no limit */
	ChanUser *firstuser; /* pointer to first user in channel */
	time_t since; /* when the channel was created (TS) */
	BanInfo *firstban; /* pointer to first ban */
	Except *exceptlist; /* pointer to first ban exception */

	/*
	 * flood_ts[0] is the TS of the first time a *Serv was kicked;
	 * flood_ts[1] is the TS of the last time
	 */
	time_t flood_ts[2];
	int floodcnt; /* how many times a *Serv was kicked/deoped */
};

/*
 * Prototypes
 */

void AddBan(char *, Channel *, char *);
void DeleteBan(Channel *, char *);
void AddException(char *, Channel *, char *);
void DeleteException(Channel *, char *);
Channel *AddChannel(char **, int, char **);
void DeleteChannel(Channel *);
void AddToChannel(Channel *, char *);
void RemoveNickFromChannel(char *, char *);
void RemoveFromChannel(Channel *, Luser *);
void UpdateChanModes(char *source, Channel *chptr, char *modes);
UserChan *FindChannelByUser(Luser *lptr, Channel *chptr);
ChanUser *FindUserByChannel(Channel *chptr, Luser *lptr);
void DoMode(Channel *, char *, int);
void SetModes(char *source, int plus, char mode, Channel *chptr, char *args);
void KickBan(int, char *, Channel *, char *, char *);
Channel *IsChan(char *channel);
int IsChannelMember(Channel *chptr, Luser *lptr);
int IsChannelVoice(Channel *chptr, Luser *lptr);
int IsChannelOp(Channel *chptr, Luser *lptr);
BanInfo *MatchBan(Channel *, char *);
BanInfo *FindBan(Channel *, char *);
Except *MatchException(Channel *, char *);
Except *FindException(Channel *, char *);
