/*
 * hash.h
 * Copyright (C) 1998-1999 Patrick Alken
 */

/*
 * size of client hash table
 */
#define HASHCLIENTS     8192

/*
 * size of channel hash table
 */
#define HASHCHANNELS    4096

/*
 * size of server hash table
 */
#define HASHSERVERS     16

/*
 * size of NickServ hash table
 */
#define NICKLIST_MAX    256

/*
 * size of ChanServ hash table
 */
#define CHANLIST_MAX    256

/*
 * size of MemoServ hash table
 */
#define MEMOLIST_MAX    256

typedef	struct hashentry
{
	void *list; /* pointer to first element in the bucket */
} aHashEntry;

/*
 * Function prototypes
 */

Luser *FindClient(const char *name);
void ClearHashes(int rehash);
int HashUhost(char *userhost);
int CloneMatch(Luser *lptr1, Luser *lptr2);
int IsClone(Luser *lptr);
Luser *HashAddClient(Luser *lptr, int nickchange);
int HashDelClient(Luser *lptr, int nickchange);
Channel *FindChannel(const char *name);
int HashAddChan(Channel *chptr);
int HashDelChan(Channel *chptr);
Server *FindServer(const char *name);
int HashAddServer(Server *sptr);
int HashDelServer(Server *sptr);

#ifdef NICKSERVICES

unsigned int NSHashNick(const char *);

#ifdef CHANNELSERVICES
unsigned int CSHashChan(const char *);
#endif

#ifdef MEMOSERVICES
unsigned int MSHashMemo(const char *);
#endif

#endif /* NICKSERVICES */

/*
 * Extern declarations
 */

extern aHashEntry cloneTable[HASHCLIENTS];

#ifdef STATSERVICES
extern aHashEntry hostTable[HASHCLIENTS];
#endif

#ifdef NICKSERVICES
extern NickInfo *nicklist[NICKLIST_MAX];

#ifdef CHANNELSERVICES
extern ChanInfo *chanlist[CHANLIST_MAX];
#endif

#ifdef MEMOSERVICES
extern MemoInfo *memolist[MEMOLIST_MAX];
#endif

#endif /* NICKSERVICES */
