/*
 * proto.h
 * Copyright (C) 1998-1999 Patrick Alken
 */

/********************* (Global) PROTOTYPES ************************/

/* Defined in chanserv.c */

void cs_process(char *, char *);
void cs_join(ChanInfo *);
void cs_join_ts_minus_1(ChanInfo *);
void cs_part(Channel *);
void cs_CheckOp(Channel *, ChanInfo *, char *);
void cs_CheckJoin(Channel *, ChanInfo *, char *);
void cs_CheckSjoin(Channel *, ChanInfo *, int, char **, int);
void cs_CheckModes(Luser *, ChanInfo *, int, int, Luser *);
void cs_CheckTopic(char *, char *);
int cs_ShouldBeOnChan(ChanInfo *);
void cs_RejoinChannels();
void ExpireChannels(time_t);

#ifndef HYBRID_ONLY
void CheckEmptyChans();
#endif /* !HYBRID_ONLY */

ChanInfo *FindChan(char *);
void DelChan(ChanInfo *);
void DeleteChan(ChanInfo *);
void RemFounder(Luser *, ChanInfo *);
int HasAccess(ChanInfo *, Luser *, int);

/* Defined in helpserv.c */

#ifdef HELPSERVICES

void hs_process(char *, char *);

#endif /* HELPSERVICES */

void GiveHelp(char *, char *, char *, int);

/* Defined in hybserv.c */

int writesocket(int, char *);

void toserv(char *, ...);
void tosock(int, char *, ...);
void debug(char *, ...);
void fatal(int, char *, ...);

void ProcessInfo(int, char **);

int ConnectToServer(char *, int);
int SplitBuf(char *, char ***);
void CycleServers();

/* Defined in list.c */

Server *AddServer(int, char **);
void DeleteServer(Server *);

void AddHostLimit(char *, int);
void AddUser(char *, char *, char *, char *);
void AddServ(char *, char *, int);
void AddMyChan(char *);
void AddBot(char *, char *, char *, int);
void AddRemoteBot(char *, char *, char *);
void AddPort(int, char *, char *);
void ClearConfLines();

Chanlist *IsChannel(char *);
Servlist *IsServLine(char *, char *);
PortInfo *IsListening(int);
PortInfo *IsPort(int);
Botlist *IsBot(char *);

/* Defined in memoserv.c */

#if defined(NICKSERVICES) && defined(MEMOSERVICES)

void ms_process(char *, char *);
void DelMemo(MemoInfo *);
int PurgeMemos(MemoInfo *);
void ExpireMemos(time_t);
void PurgeCheck();
int sendmemos(char *, char *, Luser *);
void MoveMemos(NickInfo *, NickInfo *);
MemoInfo *FindMemoList(char *);

#endif /* defined(NICKSERVICES) && defined(MEMOSERVICES) */

/* Defined in nickserv.c */

void ns_process(char *, char *);
int CheckNick(char *);
void CheckOper(char *);
void ExpireNicknames(time_t);

#ifdef NICKSERVICES
NickInfo *FindNick(char *);
NickInfo *GetLink(char *);
void DelNick(NickInfo *);
void DeleteNick(NickInfo *);
int HasFlag(char *, int);
#endif /* NICKSERVICES */

void collide(char *);
void release(char *);
void CollisionCheck(time_t);

#ifdef LINKED_NICKNAMES
int IsLinked(NickInfo *, NickInfo *);
#endif /* LINKED_NICKNAMES */

#ifdef NICKSERVICES
NickInfo *GetMaster(NickInfo *);
#endif

/* Defined in operserv.c */

void os_process(char *, char *, int);
void os_join(Channel *);
void os_join_ts_minus_1(Channel *);
void os_part(Channel *);
void check_ftarget(Luser *, char *);
char *modestr(int, char);

Ignore *OnIgnoreList(char *);
void AddIgnore(char *, time_t);
int DelIgnore(char *);
void ExpireIgnores(time_t);

/* Defined in proxy.c */

#ifdef HAVE_PTHREADS
void Proxy(Luser *);
int IsProxy(Luser *);
#endif /* HAVE_PTHREADS */

/* Defined in server.c */

char *Substitute(char *, char *, int);
void CheckChans();
void CheckServers();
void CheckListenPorts();

void do_squit(char *, char *);

void notice(char *, char *, char *, ...);
void RecordCommand(char *, ...);

void ServNotice(char *, char *, int);
void do_shutdown(char *, char *);
void ClearUsers();
void ClearServs();
void ClearChans();

void s_pass(int, char **);
void s_ping(int, char **);
void s_server(int, char **);
void s_nick(int, char **);
void s_squit(int, char **);
void s_privmsg(int, char **);
void s_quit(int, char **);
void s_kill(int, char **);
void s_kick(int, char **);
void s_mode(int, char **);
void s_sjoin(int, char **);
void s_part(int, char **);
void s_error(int, char **);
#if defined(NICKSERVICES) && defined(CHANNELSERVICES)
void s_topic(int, char **);
#endif
#ifdef STATSERVICES
void s_pong(int, char **);
#endif

Userlist *GetUser(int, char *, char *, char *);
Luser *GetService(char *);
Luser *FindService(Luser *);
int pwmatch(char *, char *);
int operpwmatch(char *, char *);
int IsInNickArray(int, char **, char *);
int IsNum(char *);
Command *GetCommand(Command *, char *);
int CheckAccess(Userlist *, char);
int IsProtectedHost(char *, char *);
rHost *IsRestrictedHost(char *, char *);

int MatchesAdmin(char *);
char *HostToMask(char *, char *);

/* Defined in statserv.c */

#ifdef STATSERVICES

void ss_process(char *, char *);
void ExpireStats(time_t);
void DoPings();
HostHash *FindHost(char *);
HostHash *FindDomain(char *);
char *GetDomain(char *);

#endif /* STATSERVICES */

/* Defined in time.c */

char *timeago(time_t, int);
long timestr(char *);
struct timeval *GetTime(struct timeval *);

/* Defined in timer.c */

#ifdef HAVE_PTHREADS
void *p_CheckSignals();
void *p_CheckTime();
#endif

void DoTimer(time_t);
