/*
 * HybServ TS Services, Copyright (C) 1998-1999 Patrick Alken
 * This program comes with absolutely NO WARRANTY
 *
 * Should you choose to use and/or modify this source code, please
 * do so under the terms of the GNU General Public License under which
 * this program is distributed.
 */

static char *rcsid = "$Id: channel.c,v 1.00 1999/08/07 20:56:13 wnder Exp $";

#include	"headers.h"

/*
AddBan()
	args: char *who, Channel *cptr, char *ban
	purpose: add 'ban' set by 'who' to channel 'cptr'
	return: none
*/

void
AddBan(char *who, Channel *cptr, char *ban)

{
	time_t CurrTime = time(NULL);
	BanInfo	*tempban;

	tempban = (BanInfo *)MyMalloc(sizeof(BanInfo));

	if (who)
		tempban->who = MyStrdup(who);
	else
		tempban->who = (char *) NULL;

	tempban->mask = MyStrdup(ban);
	tempban->when = CurrTime;
	tempban->next = cptr->firstban;
	tempban->prev = (BanInfo *) NULL;

	if (tempban->next)
		tempban->next->prev = tempban;
	cptr->firstban = tempban;
} /* AddBan() */

/*
DeleteBan()
	args: Channel *cptr, char *ban
	purpose: remove 'ban' from channel 'cptr'
	return: none
*/

void
DeleteBan(Channel *cptr, char *ban)

{
	BanInfo	*bptr;

	if (!(bptr = FindBan(cptr, ban)))
		return;

	MyFree(bptr->mask);
	if (bptr->who)
		MyFree(bptr->who);

	if (bptr->prev)
		bptr->prev->next = bptr->next;
	else
	{
		/*
		 * We're deleting the first ban in the list, reset
		 * cptr->firstban
		 */
		cptr->firstban = bptr->next;
	}

	if (bptr->next)
		bptr->next->prev = bptr->prev;

	MyFree(bptr);
} /* DeleteBan() */

/*
AddException()
 Add hostmask 'mask' set by 'who' to channel exception list
*/

void
AddException(char *who, Channel *cptr, char *mask)

{
	Except *tempe;

	tempe = (Except *) MyMalloc(sizeof(Except));

	if (who)
		tempe->who = MyStrdup(who);
	else
		tempe->who = (char *) NULL;

	tempe->mask = MyStrdup(mask);
	tempe->when = time(NULL);

	tempe->next = cptr->exceptlist;
	tempe->prev = (Except *) NULL;
	if (tempe->next)
		tempe->next->prev = tempe;

	cptr->exceptlist = tempe;
} /* AddException() */

/*
DeleteException()
 Remove hostmask 'mask' from cptr's exception list
*/

void
DeleteException(Channel *cptr, char *mask)

{
	Except *tempe;

	if (!(tempe = FindException(cptr, mask)))
		return;

	MyFree(tempe->mask);
	if (tempe->who)
		MyFree(tempe->who);

	if (tempe->prev)
		tempe->prev->next = tempe->next;
	else
		cptr->exceptlist = tempe->next;

	if (tempe->next)
		tempe->next->prev = tempe->prev;

	MyFree(tempe);
} /* DeleteException() */

/*
MatchBan()
 Same as FindBan() but use match() to compare bans to allow
for wildcards
*/

BanInfo *
MatchBan(Channel *cptr, char *ban)

{
	BanInfo *tempban;

	tempban = cptr->firstban;
	while (tempban && (match(ban, tempban->mask) == 0))
		tempban = tempban->next;

	return (tempban);
} /* MatchBan() */

/*
FindBan()
	args: Channel *cptr, char *ban
	purpose: determine if 'ban' is on 'cptr's ban list
	return: pointer to ban
*/

BanInfo *
FindBan(Channel *cptr, char *ban)

{
	BanInfo 	*tempban;

	if (!cptr || !ban)
		return ((BanInfo *) NULL);

	tempban = cptr->firstban;
	while (tempban && (strcasecmp(tempban->mask, ban) != 0))
		tempban = tempban->next;

	return (tempban);
} /* FindBan() */

/*
MatchException()
 Same as FindException() but use match() to compare bans to allow
for wildcards
*/

Except *
MatchException(Channel *cptr, char *ban)

{
	Except *tempe;

	tempe = cptr->exceptlist;
	while (tempe && (match(ban, tempe->mask) == 0))
		tempe = tempe->next;

	return (tempe);
} /* MatchException() */

/*
FindException()
 Return a pointer to occurence of 'mask' on cptr's ban exception
list
*/

Except *
FindException(Channel *cptr, char *mask)

{
	Except *tempe;

	if (!cptr || !mask)
		return ((Except *) NULL);

	tempe = cptr->exceptlist;
	while (tempe && (strcasecmp(tempe->mask, mask) != 0))
		tempe = tempe->next;

	return (tempe);
} /* FindException() */

/*
AddChannel()
	args: char **line, int nickcnt, char **nicks
	purpose: add channel info in 'line' to channel list
	return: pointer to channel structure

		line[2] = channel origin
		line[3] = channel name
		line[4] = channel modes
		line[5] = nicks in channel (or new nick who's joining)

s_sjoin() already splits the nicknames up into a nice array,
so if nickcnt > 0, use the provided nick array, otherwise
create one from the nicks in 'line'
*/

Channel *
AddChannel(char **line, int nickcnt, char **nicks)

{
	char *names;
	char **anames;
	char *currnick;
	char modes[MAXLINE];
	Channel *chname, *cptr;
	Channel *tempchan;
	int ii, ncnt, acnt;

	ncnt = 5; /* default position for channel nicks, if no limit/key */
	strcpy(modes, line[4]);
	if (line[5][0] != ':') /* names list *should* start w/ a :         */
	{                      /* if it doesn't, theres a limit and/or key */
		strcat(modes, " ");
		strcat(modes, line[5]);
		ncnt++;
		if (line[6][0] != ':')
		{
			strcat(modes, " ");
			strcat(modes, line[6]);
			ncnt++;
		}
	}

	if (nickcnt > 0)
	{
		acnt = nickcnt;
		anames = nicks;
	}
	else
	{
		names = line[ncnt];
		names++; /* point past the leading : */

		ii = strlen(names);

		/* kill the \n char on the end */
		if (IsSpace(names[ii - 2]))
			names[ii - 2] = '\0';
		else if (IsSpace(names[ii - 1]))
			names[ii - 1] = '\0';

		acnt = SplitBuf(names, &anames);
	}

	if (!(cptr = FindChannel(line[3])))
	{
	#ifdef BLOCK_ALLOCATION
		tempchan = (Channel *) BlockSubAllocate(ChannelHeap);
		strncpy(tempchan->name, line[3], CHANNELLEN);
		tempchan->key[0] = '\0';
	#else
		tempchan = (Channel *) MyMalloc(sizeof(Channel));
		tempchan->name = MyStrdup(line[3]);
		tempchan->key = (char *) NULL;
	#endif /* BLOCK_ALLOCATION */

		tempchan->modes = 0;
		tempchan->limit = 0;
		tempchan->since = atol(line[2]);
		tempchan->firstuser = (ChanUser *) NULL;
		tempchan->numusers = acnt;
		tempchan->firstban = (BanInfo *) NULL;
		tempchan->exceptlist = (Except *) NULL;
		tempchan->flood_ts[0] = tempchan->flood_ts[1] = 0;
		tempchan->floodcnt = 0;
		tempchan->next = ChannelList;
		tempchan->prev = (Channel *) NULL;
		if (tempchan->next)
			tempchan->next->prev = tempchan;

		HashAddChan(tempchan);

		ChannelList = tempchan;
		chname = ChannelList;

		Network->TotalChannels++;
	#ifdef STATSERVICES
		if (Network->TotalChannels > Network->MaxChannels)
		{
			Network->MaxChannels = Network->TotalChannels;
			Network->MaxChannels_ts = time(NULL);

			if ((Network->MaxChannels % 10) == 0)
			{
				/* notify +y people about new max channel count */
				SendUmode(OPERUMODE_Y,
					"*** New Max Channel Count: %ld",
					Network->MaxChannels);
			}
		}
		if (Network->TotalChannels > Network->MaxChannelsT)
		{
			Network->MaxChannelsT = Network->TotalChannels;
			Network->MaxChannelsT_ts = time(NULL);
		}
	#endif /* STATSERVICES */
	}
	else /* it's an existing channel, but someone has joined it */
	{
		cptr->numusers += acnt;
		chname = cptr;
	}

	/* Add the channel to each nick's channel list */
	for (ii = 0; ii < acnt; ii++)
	{
		currnick = GetNick(anames[ii]);
		if (!currnick)
			continue;

		if (!IsChannelMember(chname, FindClient(currnick)))
		{
			/*
			 * Use anames[ii] instead of currnick here so we get
			 * the @/+ flags
			 */
			AddToChannel(chname, anames[ii]);
		}
		else
			chname->numusers--;
	}
	
	/* finally, add the modes for the channel */
	UpdateChanModes((char *) NULL, chname, modes);

	/*
	 * Only free anames[] if there was no nick list
	 * given
	 */
	if (!nickcnt)
		MyFree(anames);

	return (chname);
} /* AddChannel() */

/*
AddToChannel()
	args: Channel *cptr, char *nick
	purpose: add 'cptr' to the channels that 'nick' is in
					 (when 'nick' joins a new channel, keep the channel list
					 updated for 'nick')
					 'nick' may contain an @ or a + indicating op/voice status
	return: none
*/

void
AddToChannel(Channel *cptr, char *nick)

{
	Luser	*lptr;
	UserChan *ptr;
	ChanUser *ptr2;
	char *tmpnick;

	tmpnick = GetNick(nick);

	lptr = FindClient(tmpnick);

	if (!lptr || !cptr)
		return;

	ptr = (UserChan *) MyMalloc(sizeof(UserChan));
	ptr->chptr = cptr;

	ptr->flags = 0;
	if ((nick[0] == '+') || (nick[1] == '+'))
		ptr->flags |= CH_VOICED;
	if ((nick[0] == '@') || (nick[1] == '@'))
		ptr->flags |= CH_OPPED;

	ptr->next = lptr->firstchan;
	lptr->firstchan = ptr;

	ptr2 = (ChanUser *) MyMalloc(sizeof(ChanUser));
	ptr2->lptr = lptr;
	ptr2->flags = 0;
	if ((nick[0] == '+') || (nick[1] == '+'))
		ptr2->flags |= CH_VOICED;
	if ((nick[0] == '@') || (nick[1] == '@'))
		ptr2->flags |= CH_OPPED;
	ptr2->next = cptr->firstuser;
	cptr->firstuser = ptr2;
} /* AddToChannel() */

/*
DeleteChannel()
	args: Channel *cptr
	purpose: remove 'cptr' from channel list
	return: none
*/

void
DeleteChannel(Channel *cptr)

{
	ChanUser *cnext;
	BanInfo  *bnext;
	Except   *enext;

	if (!cptr)
		return;

	HashDelChan(cptr);

	/*
	 * right now, having a loop to free the members of the channel
	 * is unnecessary, because this function is only called either
	 * from RemoveFromChannel(), which already took care of it,
	 * or from ClearChans(), which is called after ClearUsers() - 
	 * ClearUsers() would have called RemoveFromChannel() to
	 * clear each user's channel list - thus cptr->firstuser will
	 * ALWAYS be null, however, this could possibly develop into a
	 * memory leak, if DeleteChannel() were ever to be called with
	 * users still in the channel, so this algorithm will take care
	 * of it.
	 */
	while (cptr->firstuser)
	{
		cnext = cptr->firstuser->next;
		MyFree(cptr->firstuser);
		cptr->firstuser = cnext;
	}

	/*
	 * clear channel bans
	 */
	while (cptr->firstban)
	{
		bnext = cptr->firstban->next;
		if (cptr->firstban->who)
			MyFree(cptr->firstban->who);
		MyFree(cptr->firstban->mask);
		MyFree(cptr->firstban);
		cptr->firstban = bnext;
	}

	/*
	 * clear channel exceptions
	 */
	while (cptr->exceptlist)
	{
		enext = cptr->exceptlist->next;
		MyFree(cptr->exceptlist->who);
		MyFree(cptr->exceptlist->mask);
		MyFree(cptr->exceptlist);
		cptr->exceptlist = enext;
	}

#ifndef BLOCK_ALLOCATION
	MyFree(cptr->name);
	if (cptr->key)
		MyFree(cptr->key);
#endif /* BLOCK_ALLOCATION */

	if (cptr->prev)
		cptr->prev->next = cptr->next;
	else
		ChannelList = cptr->next;

	if (cptr->next)
		cptr->next->prev = cptr->prev;

#ifdef BLOCK_ALLOCATION

	BlockSubFree(ChannelHeap, cptr);

#else

	MyFree(cptr);

#endif /* BLOCK_ALLOCATION */

	Network->TotalChannels--;
} /* DeleteChannel() */

/*
RemoveNickFromChannel()
 Similar to RemoveFromChannel(), except accept string
arguements for the nickname and channel
*/

void
RemoveNickFromChannel(char *channel, char *nickname)

{
	Channel *cptr;
	Luser *lptr;
	char *tmp;

	tmp = channel;

	if (IsNickPrefix(*tmp))
	{
		if (IsNickPrefix(*(++tmp)))
			++tmp;
	}

	if (!(cptr = FindChannel(tmp)))
		return;

	tmp = GetNick(nickname);
	if (!(lptr = FindClient(tmp)))
		return;

	RemoveFromChannel(cptr, lptr);
} /* RemoveNickFromChannel() */

/*
RemoveFromChannel()
 args: Channel *cptr, Luser *lptr
 purpose: remove lptr from list of nicks in cptr (when a user
          leaves a channel)
 return: none
*/

void
RemoveFromChannel(Channel *cptr, Luser *lptr)

{
	UserChan	*tempchan, *prev = (UserChan *) NULL;
	ChanUser	*tempuser, *prev2 = (ChanUser *) NULL;

	if (!cptr || !lptr)
		return;

	SendUmode(OPERUMODE_P,
		"*** Channel part: %s (%s)",
		lptr->nick,
		cptr->name);

	/* remove cptr from lptr's chan list */
	for (tempchan = lptr->firstchan; tempchan; tempchan = tempchan->next)
	{
		if (cptr == tempchan->chptr)
		{
			if (prev)
				prev->next = tempchan->next;
			else
				lptr->firstchan = tempchan->next;
			MyFree(tempchan);
			tempchan = (UserChan *) NULL;
			break;
		}
		prev = tempchan;
	}

	/* remove lptr from cptr's nick list */
	for (tempuser = cptr->firstuser; tempuser; tempuser = tempuser->next)
	{
		if (lptr == tempuser->lptr)
		{
			if (prev2)
				prev2->next = tempuser->next;
			else
				cptr->firstuser = tempuser->next;
			MyFree(tempuser);
			tempuser = (ChanUser *) NULL;
			cptr->numusers--;
			break;
		}
		prev2 = tempuser;
	}

	if (cptr->numusers == 0)
		DeleteChannel(cptr); /* the last nick left the chan, erase it */
} /* RemoveFromChannel() */

/*
UpdateChanModes()
	args: char *who, Channel *cptr, char *modes
	purpose: keep modes field of chan structure updated when theres
					 a mode change on a channel
	return: none
*/

void
UpdateChanModes(char *who, Channel *cptr, char *modes)

{
	int		minus, /* 1 if the mode was a -, 0 if + */
				ii,
				mlength;
	char	*tmp;
	Luser	*lptr,
				*userptr;
#if defined(NICKSERVICES) && defined(CHANNELSERVICES)
	int		cs_deoped = 0; /* was chanserv deoped? */
#endif

	char	**modeargs; /* arguements to +l/k/o/v modes */
	char	tempargs[MAXLINE];
	int		argcnt; /* number of arguements */
	int		argidx; /* current index in modeargs[] */

#ifndef SAVE_TS
	char	sendstr[MAXLINE];
#endif

	if (!cptr)
		return;

	lptr = FindClient(who);

	if ((tmp = strchr(modes, ' ')))
		strcpy(tempargs, *(tmp + 1) ? tmp + 1 : "");
	else
		tempargs[0] = '\0';

	argcnt = SplitBuf(tempargs, &modeargs);

	/*
	 * This routine parses the given channel modes and keeps
	 * the corresponding lists correctly updated - also make
	 * sure OperServ and ChanServ remain opped
	 */

	minus = 0;
	argidx = (-1);
	mlength = strlen(modes);

	for (ii = 0; ii < mlength; ii++)
	{
		if (IsSpace(modes[ii]))
			break;

		switch (modes[ii])
		{
			case ' ':
			case '\n':
			case '\r': break;

			case '-':
			{
				minus = 1;
				break;
			}
			case '+':
			{
				minus = 0;
				break;
			}

			case 'o': /* someone was opped or deopped */
			{
				argidx++;
				if (argidx >= argcnt)
				{
					/*
					 * hmm..more 'o' flags than there are nicknames, just
					 * break
					 */
					break;
				}

				if (!(userptr = FindClient(modeargs[argidx])))
					break;

				if (minus)
				{
					if (!strcasecmp(modeargs[argidx], n_OperServ))
					{
						if (!FloodCheck(cptr, lptr, Me.osptr, 0))
						{
							#ifdef SAVE_TS
								os_part(cptr);
								os_join(cptr);
							#else
								toserv(":%s MODE %s +o %s\n",
									Me.name,
									cptr->name,
									n_OperServ);
							#endif
						}

						if (!lptr)
							putlog(LOG1, "%s: %s attempted to deop %s",
								cptr->name,
								who,
								n_OperServ);
						else
							putlog(LOG1, "%s: %s!%s@%s attempted to deop %s",
								cptr->name,
								lptr->nick,
								lptr->username,
								lptr->hostname,
								n_OperServ);
					}
					else
					{
					#if defined(NICKSERVICES) && defined(CHANNELSERVICES)
						if (!strcasecmp(modeargs[argidx], n_ChanServ))
						{
							cs_deoped = 1;
						}
						else /* someone was deoped (not a *Serv) */
						{
					#endif /* defined(NICKSERVICES) && defined(CHANNELSERVICES) */

							UserChan *tempc;
							ChanUser *tempu;

							if (IsChannelOp(cptr, userptr))
							{
								/* remove @ from nick in channel nick list */

								tempu = FindUserByChannel(cptr, userptr);
								if (tempu)
									tempu->flags &= ~CH_OPPED;

								/* remove @ from channel in nick's channel list */
								tempc = FindChannelByUser(userptr, cptr);
								if (tempc)
									tempc->flags &= ~CH_OPPED;

							#if defined(NICKSERVICES) && defined(CHANNELSERVICES)
								cs_CheckModes(lptr, FindChan(cptr->name), minus, MODE_O, userptr);
							#endif
							} /* if (IsChannelOp(cptr, userptr)) */

					#if defined(NICKSERVICES) && defined(CHANNELSERVICES)
						}
					#endif

					} /* else */

				#ifdef STATSERVICES
					if (lptr)
						lptr->numdops++;
				#endif

				} /* if (minus) */
				else /* someone was oped..update '@' signs :-) */
				{
					UserChan *tempc;
					ChanUser *tempu;

					if (!IsChannelOp(cptr, userptr))
					{
						/* add @ to nick in channel nick list */
						tempu = FindUserByChannel(cptr, userptr);
						if (tempu)
							tempu->flags |= CH_OPPED;

						/* add @ to channel in nick's channel list */
						tempc = FindChannelByUser(userptr, cptr);
						if (tempc)
							tempc->flags |= CH_OPPED;

					#if defined(NICKSERVICES) && defined(CHANNELSERVICES)
						cs_CheckModes(lptr, FindChan(cptr->name), minus, MODE_O, userptr);
					#endif
					} /* if (!IsOp) */

				#ifdef STATSERVICES
					if (lptr)
						lptr->numops++;
				#endif
				}

				break;
			} /* case 'o' */

			case 'v': /* someone was voiced/devoiced */
			{
				argidx++;
				if (argidx >= argcnt)
					break;

				if (!(userptr = FindClient(modeargs[argidx])))
					break;

				if (minus)
				{
					UserChan *tempc;
					ChanUser *tempu;

					if (IsChannelVoice(cptr, userptr))
					{
						/* remove + from nick in channel list */
						tempu = FindUserByChannel(cptr, userptr);
						if (tempu)
							tempu->flags &= ~CH_VOICED;

						/* remove + from channel in nick's channel list */
						tempc = FindChannelByUser(userptr, cptr);
						if (tempc)
							tempc->flags &= ~CH_VOICED;

					#if defined(NICKSERVICES) && defined(CHANNELSERVICES)
						cs_CheckModes(lptr, FindChan(cptr->name), minus, MODE_V, userptr);
					#endif
					} /* if (IsChannelVoice) */

				#ifdef STATSERVICES
					if (lptr)
						lptr->numdvoices++;
				#endif
				} /* if (minus) */
				else /* someone was voiced..update '+' signs :-) */
				{
					UserChan	*tempc;
					ChanUser	*tempu;

					if (!IsChannelVoice(cptr, userptr))
					{
						/* add + to nick in channel's nick list */
						tempu = FindUserByChannel(cptr, userptr);
						if (tempu)
							tempu->flags |= CH_VOICED;

						/* add + to channel in nick's channel list */
						tempc = FindChannelByUser(userptr, cptr);
						if (tempc)
							tempc->flags |= CH_VOICED;

					#if defined(NICKSERVICES) && defined(CHANNELSERVICES)
						cs_CheckModes(lptr, FindChan(cptr->name), minus, MODE_V, userptr);
					#endif
					} /* if (!IsChannelVoice) */

				#ifdef STATSERVICES
					if (lptr)
						lptr->numvoices++;
				#endif
				}

				break;
			} /* case 'v' */

			case 'l': /* channel limit */
			{
				if (!minus)
				{
					argidx++;
					if (argidx >= argcnt)
						break;

					cptr->limit = atoi(modeargs[argidx]);
				}
				else
					cptr->limit = 0;

			#if defined(NICKSERVICES) && defined(CHANNELSERVICES)
				cs_CheckModes(lptr, FindChan(cptr->name), minus, MODE_L, (Luser *) NULL);
			#endif

				break;
			} /* case 'l' */

			case 'k': /* channel key */
			{
				argidx++;
				if (argidx >= argcnt)
					break;

			#ifndef BLOCK_ALLOCATION
				if (cptr->key)
					MyFree(cptr->key);
			#endif

				if (minus)
				{
				#ifdef BLOCK_ALLOCATION
					cptr->key[0] = '\0';
				#else
					cptr->key = (char *) NULL;
				#endif /* BLOCK_ALLOCATION */
				}
				else
				{
				#ifdef BLOCK_ALLOCATION
					strncpy(cptr->key, modeargs[argidx], KEYLEN);
				#else
					cptr->key = MyStrdup(modeargs[argidx]);
				#endif /* BLOCK_ALLOCATION */
				}

			#if defined(NICKSERVICES) && defined(CHANNELSERVICES)
				cs_CheckModes(lptr, FindChan(cptr->name), minus, MODE_K, (Luser *) NULL);
			#endif

				break;
			} /* case 'k' */

			case 'b': /* channel ban */
			{
				argidx++;
				if (argidx >= argcnt)
					break;

				if (!minus)
					AddBan(who, cptr, modeargs[argidx]);
				else
					DeleteBan(cptr, modeargs[argidx]);

				break;
			} /* case 'b' */

			case 'e': /* channel ban exception */
			{
				argidx++;
				if (argidx >= argcnt)
					break;

				if (!minus)
					AddException(who, cptr, modeargs[argidx]);
				else
					DeleteException(cptr, modeargs[argidx]);

				break;
			} /* case 'e' */

			default:
			{
				int		modeflag = 0;

				if (modes[ii] == 's')
					modeflag = MODE_S;
				else if (modes[ii] == 'p')
					modeflag = MODE_P;
				else if (modes[ii] == 'n')
					modeflag = MODE_N;
				else if (modes[ii] == 't')
					modeflag = MODE_T;
				else if (modes[ii] == 'm')
					modeflag = MODE_M;
				else if (modes[ii] == 'i')
					modeflag = MODE_I;

				if (modeflag)
				{
					if (minus)
						cptr->modes &= ~modeflag;
					else
						cptr->modes |= modeflag;
				}

			#if defined(NICKSERVICES) && defined(CHANNELSERVICES)
				if (modeflag)
					cs_CheckModes(lptr, FindChan(cptr->name), minus, modeflag, (Luser *) NULL);
			#endif

				break;
			} /* default: */
		} /* switch (modes[ii]) */
	} /* for (ii = 0; ii < mlength; ii++) */

	MyFree(modeargs);

#if defined(NICKSERVICES) && defined(CHANNELSERVICES)
	if ((cs_deoped) && (!FloodCheck(cptr, lptr, Me.csptr, 0)))
	{
		/* reop ChanServ */
		#ifdef SAVE_TS
			cs_part(cptr);
			cs_join(FindChan(cptr->name));
		#else
			toserv(":%s MODE %s +o %s\n",
				Me.name,
				cptr->name,
				n_ChanServ);
		#endif

		if (!lptr)
			putlog(LOG1, "%s: %s attempted to deop %s",
				cptr->name,
				who,
				n_ChanServ);
		else
			putlog(LOG1, "%s: %s!%s@%s attempted to deop %s",
				cptr->name,
				lptr->nick,
				lptr->username,
				lptr->hostname,
				n_ChanServ);
	}
#endif /* defined(NICKSERVICES) && defined(CHANNELSERVICES) */

} /* UpdateChanModes() */

/*
FindChannelByUser()
	return a pointer to channel 'channel' in lptr's channel list
*/

UserChan *
FindChannelByUser(Luser *lptr, Channel *chptr)

{
	UserChan *tempchan;

	if (!lptr || !chptr)
		return ((UserChan *) NULL);

	for (tempchan = lptr->firstchan; tempchan; tempchan = tempchan->next)
		if (tempchan->chptr == chptr)
			return (tempchan);

	return ((UserChan *) NULL);
} /* FindChannelByUser() */

/*
FindUserByChannel()
	return pointer to client 'lptr' in channel chptr's nick list
*/

ChanUser *
FindUserByChannel(Channel *chptr, Luser *lptr)

{
	ChanUser *tempuser;

	if (!chptr || !lptr)
		return ((ChanUser *) NULL);

	for (tempuser = chptr->firstuser; tempuser; tempuser = tempuser->next)
		if (tempuser->lptr == lptr)
			return (tempuser);

	return ((ChanUser *) NULL);
} /* FindUserByChannel() */

/*
IsChannelVoice()
	args: Channel *chptr, Luser *lptr
	purpose: determine if 'lptr' is currently voiced on 'chptr'
	return: 1 if voiced, 0 if not
*/

int
IsChannelVoice(Channel *chptr, Luser *lptr)

{
	UserChan *tempchan;

	if (!chptr || !lptr)
		return 0;

	for (tempchan = lptr->firstchan; tempchan; tempchan = tempchan->next)
		if (tempchan->chptr == chptr)
			return (tempchan->flags & CH_VOICED);

	return 0;
} /* IsChannelVoice() */

/*
IsChannelOp()
	args: char *channel, char *nick
	purpose: determine if 'nick' is currently oped on 'channel'
	return: 1 if oped, 0 if not
*/

int
IsChannelOp(Channel *chptr, Luser *lptr)

{
	UserChan *tempchan;

	if (!chptr || !lptr)
		return 0;

	for (tempchan = lptr->firstchan; tempchan; tempchan = tempchan->next)
		if (tempchan->chptr == chptr)
			return (tempchan->flags & CH_OPPED);

	return 0;
} /* IsChannelOp() */

/*
DoMode()
	Set 'modes' on 'chptr'
	if 'joinpart' is != 1, do not join/part because the channel is being
	secured - so there might be more modes to come.
*/

void
DoMode(Channel *chptr, char *modes, int joinpart)

{
#ifdef SAVE_TS
	Chanlist *chanptr;
#endif

	if (!chptr | !modes)
		return;

#ifdef SAVE_TS

	chanptr = IsChannel(chptr->name);

	if ((joinpart == YES) && !chanptr)
		/* make sure its not already a monitored channel */
		os_join(chptr);

	toserv(":%s MODE %s %s\n",
		n_OperServ,
		chptr->name,
		modes);
	UpdateChanModes(n_OperServ, chptr, modes);

	if ((joinpart == YES) && !chanptr)
		os_part(chptr);

#else

	toserv(":%s MODE %s %s\n",
		Me.name,
		chptr->name,
		modes);
	UpdateChanModes(Me.name, chptr, modes);

#endif /* SAVE_TS */
} /* DoMode() */

/*
IsChannelMember()
	args: Channel *cptr, Luser *lptr
	purpose: determine if 'lptr' is currently on channel 'cptr'
	return: 1 if nick is on the channel, 0 if not
*/

int
IsChannelMember(Channel *cptr, Luser *lptr)

{
	ChanUser *tempuser;

	if (!cptr || !lptr)
		return 0;

	for (tempuser = cptr->firstuser; tempuser; tempuser = tempuser->next)
	{
		if (tempuser->lptr == lptr)
			return 1;
	}

	return 0;
} /* IsChannelMember() */

/*
SetModes()
	Set mode +/-<mode> <args> on <channel> from <source>
*/

void
SetModes(char *source, int plus, char mode, Channel *chptr, char *args)

{
	int		acnt, mcnt, ii;
	char	done[MAXLINE], sendstr[MAXLINE];
	char	**av, *temp, *mtmp;

	if (!source || !chptr || !args)
		return;

	temp = MyStrdup(args);
	acnt = SplitBuf(temp, &av);
	memset(&done, 0, MAXLINE);
	mcnt = 1;
	for (ii = 0; ii < acnt; ii++)
	{
		strcat(done, av[ii]);
		strcat(done, " ");
		if (mcnt == MaxModes)
		{
			mcnt = 0;
			mtmp = modestr(MaxModes, mode);
			sprintf(sendstr, "%s%s %s",
				plus ? "+" : "-",
				mtmp,
				done);
			toserv(":%s MODE %s %s\n",
				source,
				chptr->name,
				sendstr);
			UpdateChanModes((char *) NULL, chptr, sendstr);
			MyFree(mtmp);
			memset(&done, 0, MAXLINE);
		}
		mcnt++;
	}

	if (done[0] != '\0')
	{
		mtmp = modestr(mcnt - 1, mode);
		sprintf(sendstr, "%s%s %s",
			plus ? "+" : "-",
			mtmp,
			done);
		toserv(":%s MODE %s %s\n",
			source,
			chptr->name,
			sendstr);
		UpdateChanModes((char *) NULL, chptr, sendstr);
		MyFree(mtmp);
	}
	MyFree(temp);
	MyFree(av);
} /* SetModes() */

/*
KickBan()
	KickBan 'nicks' on 'channel'
*/

void
KickBan(int ban, char *source, Channel *channel, char *nicks, char *reason)

{
	char	*mask, *tempnix, **av, *bans;
	char	temp[MAXLINE];
	int		ac, ii;
	Luser	*lptr;

	if (!source || !channel || !nicks)
		return;

	tempnix = MyStrdup(nicks);
	ac = SplitBuf(tempnix, &av);

	if (ban)
	{
		bans = (char *) MyMalloc(CHARSIZE);
		bans[0] = '\0';
		for (ii = 0; ii < ac; ii++)
		{
			if (!(lptr = FindClient(av[ii])))
				continue;
			mask = HostToMask(lptr->username, lptr->hostname);
			sprintf(temp, "*!%s", mask);
			bans = (char *) MyRealloc(bans, strlen(bans) + strlen(temp) + (2 * CHARSIZE));
			strcat(bans, temp);
			strcat(bans, " ");
			MyFree(mask);
		}

		SetModes(source, 1, 'b', channel, bans);
		MyFree(bans);
	}

	for (ii = 0; ii < ac; ii++)
	{
		toserv(":%s KICK %s %s :%s\n",
			source,
			channel->name,
			av[ii],
			reason ? reason : "");
		RemoveFromChannel(channel, FindClient(av[ii]));
	}

	MyFree(tempnix);
	MyFree(av);
} /* KickBan() */
