/*
 * HybServ TS Services, Copyright (C) 1998-1999 Patrick Alken
 * This program comes with absolutely NO WARRANTY
 *
 * Should you choose to use and/or modify this source code, please
 * do so under the terms of the GNU General Public License under which
 * this program is distributed.
 */

static char *rcsid = "$Id: dcc.c,v 2.45 1998/10/26 18:16:11 wnder Exp $";

#include	"headers.h"

#include	<sys/socket.h>
#include 	<netinet/in.h>
#include	<arpa/inet.h>
#include	<netdb.h>
#include	<fcntl.h>

static int IsAuth(DccUser *);
static int request_ident(DccUser *, struct in_addr *);

/*
SendMotd
	args: int sockfd
	purpose: output message of the day to the socket 'sockfd'
	return: none
*/

void
SendMotd(int sockfd)

{
	char	line[MAXLINE];
	char	*final;
	FILE	*fp;

	if ((fp = fopen(DccMotdFile, "r")) == (FILE *) NULL)
	{
		writesocket(sockfd, "MOTD file missing\n");
		return;
	}

	while (fgets(line, MAXLINE - 1, fp))
	{
		final = Substitute((char *) NULL, line, sockfd);
		if (final && (final != (char *) -1))
		{
			writesocket(sockfd, final);
			MyFree(final);
		}
	}

	fclose(fp);
} /* SendMotd() */

/*
IsAuth()
 Determine if authorization checks are complete for dccptr.
Such checks include ident request, and correct nickname/password
for telnet clients.
Return: 1 if client is authorized
        0 if not
*/

static int
IsAuth(DccUser *dccptr)

{
	if (!dccptr)
		return (0);

	if ((dccptr->socket == NOSOCKET) ||
			(dccptr->flags & SOCK_NEEDID) ||
			(dccptr->flags & SOCK_PENDING))
		return (0);

	return (1);
} /* IsAuth() */

/*
BroadcastDcc()
 Send the message (specified in 'format') to users who match
the integer 'towho'. 'towho' can be either DCCALL (all users),
or DCCOPS (opers and above only)

 return: none
*/

void
BroadcastDcc(int towho, char *format, ...)

{
	char	tcmbuf[MAXLINE],
				buffer[MAXLINE];
	DccUser	*dccptr;
	va_list	args;

	va_start(args, format);

#ifdef HAVE_VSNPRINTF
	vsnprintf(buffer, sizeof(buffer) - 1, format, args);
	buffer[sizeof(buffer) - 1] = '\0';
#else
	vsprintf(buffer, format, args);
#endif /* HAVE_VSNPRINTF */

	va_end(args);

	/*
	 * Construct the string we want to send
	 */
	tcmbuf[0] = '\0';
	if (towho == DCCALL)
	{
		/*
		 * Make the tcm buffer look like:
		 * (OperServ) Message
		 */
		Snprintf(tcmbuf, sizeof(tcmbuf),
			"(%s) %s",
			n_OperServ,
			buffer);
	}

	for (dccptr = connections; dccptr; dccptr = dccptr->next)
	{
		/*
		 * Don't broadcast to a client who hasn't entered their
		 * password yet, or who hasn't responded to an ident
		 * request yet
		 */
		if (!IsAuth(dccptr))
			continue;

		switch (towho)
		{
			case DCCALL:
			{
				/* send it to everybody */

				if ((dccptr->flags & SOCK_TCMBOT) && (tcmbuf[0]))
				{
					/*
					 * This client is a tcm bot, so send it tcmbuf[]
					 */

					writesocket(dccptr->socket, tcmbuf);
				}
				else
				{
					/*
					 * This is a user connection, send them buffer[]
					 */
					writesocket(dccptr->socket, buffer);
				}

				break;
			} /* case DCCALL */

			case DCCOPS:
			{
				Userlist		*tempuser;

				/*
				 * send it only to operators (must have an "o" in their
				 * O: line)
				 */

				/*
				 * Assume dccptr has already been verified to come
				 * from the correct host etc, so allow GetUser()
				 * to match only nicknames if need be (they might
				 * be telnetting from an unknown host, or services
				 * might be squit'd from the network, and they're
				 * dcc chatting)
				 */
				tempuser = DccGetUser(dccptr);

				if (!tempuser)
				{
					/*
					 * One reason tempuser might be null here is if the user
					 * connected to the partyline, but then his/her O: was
					 * commented out of hybserv.conf and rehashed
					 */
					break;
				}

				if (dccptr->flags & SOCK_TCMBOT)
					break;

				if (IsOper(tempuser) &&
						IsRegistered((Luser *) NULL, dccptr->socket))
					writesocket(dccptr->socket, buffer);

				break;
			} /* case DCCOPS */

			default:
				break;
		} /* switch (towho) */
	} /* for (dccptr = connections; dccptr; dccptr = dccptr->next) */
} /* BroadcastDcc() */

/*
SendDccMessage()
 Called when a dcc or telnet client sends a message to the partyline.
Send the message to all other connected clients in the format:

<Nickname@OperServ> message

*/

void
SendDccMessage(DccUser *from, char *message)

{
	char final[MAXLINE];
	DccUser *dccptr;

	if (!from || !message)
		return;

	Snprintf(final, sizeof(final) - 1,
		"<%s@%s> %s",
		from->nick,
		n_OperServ,
		message);
	strcat(final, "\n");

	/* now send the string to all clients */
	for (dccptr = connections; dccptr; dccptr = dccptr->next)
	{
		/*
		 * Don't broadcast to a client who hasn't entered their
		 * password yet, or who hasn't responded to an ident
		 * request yet
		 */
		if (!IsAuth(dccptr))
			continue;

		writesocket(dccptr->socket, final);
	}
} /* SendDccMessage() */

/*
SendUmode()
 Send message 'format' to all users on the partyline that
currently have 'umode' set in their usermodes

Note to self: don't put debug statements in here or an infinite
loop will occur, since debug() calls SendUmode() to broadcast
to +d people
*/

void
SendUmode(int umode, char *format, ...)

{
	va_list		args;
	char		buffer[MAXLINE];
	DccUser		*dccptr;
	Userlist	*tempuser;
	struct tm *tmp_tm;
	time_t	current_ts;
	char	timestr[MAXLINE];

	if (ActiveFlood)
		return;

	current_ts = time(NULL);

	if (umode != OPERUMODE_D)
		if (IsFlood())
			ActiveFlood = current_ts;

	if (!ActiveFlood)
	{
		va_start(args, format);

	#ifdef HAVE_VSNPRINTF
		vsnprintf(buffer, sizeof(buffer) - 1, format, args);
		buffer[sizeof(buffer) - 1] = '\0';
	#else
		vsprintf(buffer, format, args);
	#endif /* HAVE_VSNPRINTF */

		va_end(args);
	} /* if (!ActiveFlood) */

	tmp_tm = localtime(&current_ts);

	sprintf(timestr, "[%02d:%02d:%02d]",
		tmp_tm->tm_hour,
		tmp_tm->tm_min,
		tmp_tm->tm_sec);

	for (dccptr = connections; dccptr; dccptr = dccptr->next)
	{
		if (!IsAuth(dccptr))
			continue;

		if ((tempuser = DccGetUser(dccptr)))
		{
			if (!IsOper(tempuser))
				continue;

			if (ActiveFlood)
			{
				/*
				 * Send +y users a notice informing them of the detected
				 * flood
				 */
				if (tempuser->umodes & OPERUMODE_Y)
				{
					tosock(dccptr->socket,
						"%s *** Broadcast Flood detected, halting broadcasts for 20 seconds\n",
						timestr);
				}
			}
			else if (tempuser->umodes & umode)
			{
				tosock(dccptr->socket,
					"%s %s\n",
					timestr,
					buffer);
			}
		}
	}
} /* SendUmode() */

/*
ConnectClient()
	purpose: accept remote connection on portptr->socket
	return: none 
*/

void
ConnectClient(PortInfo *portptr)

{
	struct sockaddr_in RemoteAddr;
	struct hostent *RemoteHost;
	int addrlen, goodid = 1;
	DccUser		*tempconn;

	if (!portptr)
		return;

	tempconn = (DccUser *) MyMalloc(sizeof(DccUser));

	addrlen = sizeof(struct sockaddr_in);
	if ((tempconn->socket = accept(portptr->socket,
		 (struct sockaddr *)&RemoteAddr, &addrlen)) < 0 )
	{
		putlog(LOG1, "Error in accept on socket %d", portptr->socket);
		MyFree(tempconn);
		return;
	}

	if (MaxConnections && (Network->TotalConns >= MaxConnections))
	{
		/* maximum users reached */
		writesocket(tempconn->socket, "Maximum connections reached, try again later\n");
		close(tempconn->socket);
		MyFree(tempconn);
		return;
	}

	RemoteHost = gethostbyaddr((char *)&RemoteAddr.sin_addr.s_addr,
															4,AF_INET);
	
	if (RemoteHost)
		tempconn->hostname = MyStrdup(RemoteHost->h_name);
	else
		tempconn->hostname = MyStrdup(inet_ntoa(RemoteAddr.sin_addr));

	tempconn->nick = (char *) NULL;
	tempconn->username = MyStrdup("unknown");
	tempconn->port = ntohs(RemoteAddr.sin_port);
	tempconn->idle = time(NULL);

	set_nonblocking(tempconn->socket);
	/* fcntl(tempconn->socket, F_SETFL, O_NONBLOCK); */

	switch (portptr->type)
	{
		case PRT_TCM:
		{
			int		sock;

		#if 0
			/*
			 * This is done in BotProcess() now
			 */

			/* Check if the tcm bot has an L: line */
			if (!GoodTCM(tempconn))
			{
				putlog(LOG1, "Illegal tcm connection attempt on port %d from %s:%d",
					portptr->port,
					tempconn->hostname,
					tempconn->port);

				SendUmode(OPERUMODE_B,
					"*** Unauthorized tcm connection attempt from [%s:%d] on port %d",
					tempconn->hostname,
					tempconn->port,
					portptr->port);

				close(tempconn->socket);
				MyFree(tempconn->hostname);
				MyFree(tempconn);
				return;
			}
		#endif /* 0 */

			tempconn->flags = (SOCK_TCMBOT | SOCK_PENDING | SOCK_NEEDID);

			SendUmode(OPERUMODE_B,
				"*** Received tcm connection from [%s:%d] on port %d",
				tempconn->hostname,
				tempconn->port,
				portptr->port);

			sock = request_ident(tempconn, &RemoteAddr.sin_addr);
			if (sock < 0)
				goodid = 0;

			break;
		}

		case PRT_USERS:
		{
			int		sock;

			if (portptr->host)
			{
				/* Check if incoming connection has an authorized host */
				if (match(portptr->host, tempconn->hostname) == 0)
				{
					putlog(LOG1,
						"Illegal connection attempt on port %d from [%s:%d]",
						portptr->port,
						tempconn->hostname,
						tempconn->port);

					SendUmode(OPERUMODE_B,
						"*** Unauthorized connection attempt from [%s:%d] on port %d",
						tempconn->hostname,
						tempconn->port,
						portptr->port);

					close(tempconn->socket);
					MyFree(tempconn->hostname);
					MyFree(tempconn);
					return;
				}
			}

			tempconn->flags = SOCK_NEEDID;

			SendUmode(OPERUMODE_Y,
				"*** Received connection from [%s:%d] on port %d",
				tempconn->hostname,
				tempconn->port,
				portptr->port);

			sock = request_ident(tempconn, &RemoteAddr.sin_addr);
			if (sock < 0)
				goodid = 0;

			break;
		}

		default:
		{
			/* something is really screwed */
			putlog(LOG1,
				"Invalid connection type specified on port [%d] (%s:%d)",
				portptr->port,
				tempconn->hostname,
				tempconn->port);

			close(tempconn->socket);
			MyFree(tempconn->hostname);
			MyFree(tempconn);
			return;
		}
	}

	if (!goodid)
	{
		/* something went wrong while connecting to their ident port */

		close(tempconn->authfd);
		tempconn->authfd = NOSOCKET;
		tempconn->flags &= ~SOCK_NEEDID;

		/*
		 * Their username will remain "unknown" from above
		 */

		if (!(tempconn->flags & SOCK_TCMBOT))
			telnet_greet(tempconn);
	}

	AddConn(tempconn);

#ifdef HAVE_PTHREADS
	pthread_exit(NULL);
#endif
} /* ConnectClient() */

/*
readauth()
	dccptr->authfd is active - read the ident reply if there is one
*/

void
readauth(DccUser *dccptr)

{
	char		buffer[MAXLINE], ident[USERLEN + 1];
	int			length;

	if (dccptr->authfd == NOSOCKET)
		return;

	memset(&buffer, 0, MAXLINE);
	length = recv(dccptr->authfd, buffer, sizeof(buffer), 0);
	if (length > 0)
	{
		u_short	remport = 0,
						locport = 0;
		int num;

		num = sscanf(buffer,
						"%hd , %hd : USERID : %*[^:]: %10s",
						&remport,
						&locport,
						ident);

		/*
		 * If num != 3, it's possible the remote identd returned
		 * <remoteport> , <localport> : ERROR : NO-USER
		 * so make the ident "unknown"
		 */
		if (num != 3)
			strcpy(ident, "unknown");
	}
	else
	{
		/* connection was closed */
		strcpy(ident, "unknown");
	}

	close(dccptr->authfd);
	dccptr->authfd = NOSOCKET;

	dccptr->flags &= ~SOCK_NEEDID;

	if (dccptr->username)
		MyFree(dccptr->username);
	dccptr->username = MyStrdup(ident);

	if (!(dccptr->flags & SOCK_TCMBOT))
		telnet_greet(dccptr);
} /* readauth() */

/*
writeauth()
	Request ident reply for dccptr->authfd
*/

void
writeauth(DccUser *dccptr)

{
	char		idstr[MAXLINE];
	struct	sockaddr_in	local, remote;
	int			len;

	len = sizeof(local);
	if (getsockname(dccptr->socket, (struct sockaddr *) &local, &len) ||
			getpeername(dccptr->socket, (struct sockaddr *) &remote, &len))
	{
		/* something went wrong */
		putlog(LOG1,
			"writeauth: get{sock,peer}name error for %d: %s",
			dccptr->socket,
			strerror(errno));
		return;
	}

	/* send ident request */
	sprintf(idstr, "%u , %u\n",
		(unsigned int) ntohs(remote.sin_port),
		(unsigned int) ntohs(local.sin_port));
	writesocket(dccptr->authfd, idstr);

	dccptr->flags &= ~SOCK_WRID;
} /* writeauth() */

void
set_nonblocking(int socket)

{
	int		flags = 0;

	if ((flags = fcntl(socket, F_GETFL, 0)) == -1)
		putlog(LOG1,
			"set_nonblocking: fcntl(%d, F_GETFL, 0) failed: %s",
			socket,
			strerror(errno));
	else
		if (fcntl(socket, F_SETFL, flags | O_NONBLOCK) == -1)
			putlog(LOG1,
				"set_nonblocking: fcntl(%d, F_SETFL, %d) failed: %s",
				socket,
				flags | O_NONBLOCK,
				strerror(errno));
} /* set_nonblocking() */

/*
SetSocketOptions()
	set various socket options on 'socket'
*/

void
SetSocketOptions(int socket)

{
	int		option;

	option = 1;

	/*
	 * SO_REUSEADDR will enable immediate local address reuse of
	 * the port this socket is bound to
	 */
	if (setsockopt(socket, SOL_SOCKET, SO_REUSEADDR, (char *) &option, sizeof(option)) < 0)
	{
		putlog(LOG1, "SetSocketOptions: setsockopt(SO_REUSEADDR) failed: %s",
			strerror(errno));
		return;
	}
} /* SetSocketOptions() */

/*
request_ident()
	Connect to the ident port of dccptr
*/

static int
request_ident(DccUser *dccptr, struct in_addr *in)

{
	int			length;
	struct	sockaddr_in	localaddr;
	struct	sockaddr_in	remote;

	if ((dccptr->authfd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
	{
		putlog(LOG1,
			"Unable to open stream socket for connection %s/%d: %s",
			dccptr->hostname,
			dccptr->port,
			strerror(errno));
		return (-1);
	}

	set_nonblocking(dccptr->authfd);
	/*fcntl(dccptr->authfd, F_SETFL, O_NONBLOCK);*/

	length = sizeof(struct sockaddr_in);
 	memset((void *) &localaddr, 0, length);
 	/* getsockname(dccptr->socket, (struct sockaddr *) &localaddr, &length); */
 	localaddr.sin_port = htons(0);

 	if (bind(dccptr->authfd, (struct sockaddr *)&localaddr,
			sizeof(localaddr)) == -1)
	{
		putlog(LOG1,
 	 		"Error binding ident stream socket %d: %s",
 	 		dccptr->authfd,
 	 		strerror(errno));
		return (-1);
	}

	memset((void *) &remote, 0, sizeof(struct sockaddr_in));
	memcpy((void *) &remote.sin_addr, (void *) in,
				sizeof(struct in_addr));

	remote.sin_port = htons(113);
	remote.sin_family = AF_INET;

	if ((connect(dccptr->authfd, (struct sockaddr *)&remote,
			sizeof(remote)) == -1) && (errno != EINPROGRESS))
	{
		putlog(LOG1,
			"Unable to connect to ident port of %s: %s",
			dccptr->hostname,
			strerror(errno));
		return (-1);
	}

	dccptr->flags |= SOCK_WRID;

	return (dccptr->authfd);
} /* request_ident() */

/*
telnet_greet()
	Send new telnet connection version and "enter nickname" lines
*/

void
telnet_greet(DccUser *dccptr)

{
	char	ver[MAXLINE];

	if (!dccptr)
		return;

	if (dccptr->socket == NOSOCKET)
		return;

	dccptr->flags |= SOCK_PENDING;

	sprintf(ver, "\nHybServ (TS Services v%s)\n", hVersion);
	writesocket(dccptr->socket, ver);

	writesocket(dccptr->socket, "\nEnter nickname\n");
} /* telnet_greet() */

/*
MakeConn()
	args: char *host, int port, Luser *lptr
	purpose: create a DCC CHAT session for 'lptr->nick'
	return: 1 if connection was successful, 0 if not
*/

int
MakeConn(char *host, int port, Luser *lptr)

{
	char prefix[MAXLINE],
	     sendstr[MAXLINE];
	Userlist *tempuser;
	DccUser *dccptr;
	time_t CurrTime = time(NULL);
	struct tm *motd_tm;

	if (!host || !lptr)
		return 0;

	if (MaxConnections && (Network->TotalConns >= MaxConnections))
	{
		/* maximum users reached */
		notice(n_OperServ, lptr->nick,
			"Maximum DCC connections have been reached, try again later");
		return 0;
	}

	if (lptr->flags & L_OSREGISTERED)
		tempuser = GetUser(1, lptr->nick, lptr->username, lptr->hostname);
	else
		tempuser = GetUser(0, lptr->nick, lptr->username, lptr->hostname);

	if (!tempuser)
		return 0;

	dccptr = (DccUser *) MyMalloc(sizeof(DccUser));

	dccptr->socket = ConnectToServer(host, port);
	if (dccptr->socket == -1)
	{
		MyFree(dccptr);
		return 0;
	}

	dccptr->flags = SOCK_DCC;
	dccptr->port = port;
	dccptr->idle = time(NULL);
	dccptr->authfd = NOSOCKET;

	dccptr->nick = MyStrdup(tempuser->nick);
	dccptr->username = MyStrdup(lptr->username);
	dccptr->hostname = MyStrdup(lptr->hostname);

	/* add them to list */
	AddConn(dccptr);

	motd_tm = localtime(&CurrTime);
	sprintf(sendstr, "HybServ %s (%d/%d/%d %d:%02d)\n", 
		hVersion,
		1900 + motd_tm->tm_year,
		motd_tm->tm_mon + 1,
		motd_tm->tm_mday,
		motd_tm->tm_hour,
		motd_tm->tm_min);
	writesocket(dccptr->socket, sendstr);

	SendMotd(dccptr->socket); /* give new user the motd */

	if (!(lptr->flags & L_OSREGISTERED))
	{
		writesocket(dccptr->socket, 
			"You must .identify before you may use commands\n\n");
		dccptr->flags |= SOCK_PENDING;
	}

	if (IsServicesAdmin(tempuser))
		strcpy(prefix, "Services Admin");
	else if (IsAdmin(tempuser))
		strcpy(prefix, "Admin");
	else if (IsOper(tempuser))
		strcpy(prefix, "Oper");
	else
		strcpy(prefix, "User");

	/* inform everyone of new user */
	BroadcastDcc(DCCALL, "%s %s (%s@%s) has connected\n",
		prefix,
		dccptr->nick,
		dccptr->username,
		dccptr->hostname);

	return 1;
} /* MakeConn() */

/*
onctcp()
	args: char *nick, char *msg
	purpose: handle ctcp message 'msg' sent by 'nick' to 'target'
	return: none
*/

void
onctcp(char *nick, char *target, char *msg)

{
	char	temp[MAXLINE];
	int		goodDCC;
	Luser	*lptr;
	Userlist	*tempuser;

	if (!msg)
		return;

	msg++;
	if (!msg)
		return;
	if (!strlen(msg))
		return;

	lptr = FindClient(nick);
	if (lptr == (Luser *) NULL)
		return;

	if (!strncasecmp(msg, "ACTION", 6))
		return; /* don't log ctcp actions */

	if (strncasecmp(msg, "PING", 4) == 0)
	{
		notice(target, nick, msg - 1);
		SendUmode(OPERUMODE_Y,
			"%s: CTCP PING received from %s!%s@%s",
			target,
			lptr->nick,
			lptr->username,
			lptr->hostname);
		return;
	}

	if (strncasecmp(msg, "VERSION", 7) == 0)
	{
		notice(target, nick,
			"\001VERSION HybServ TS Services version %s\001",
			hVersion); 
		notice(target, nick,
			"\001VERSION Administrators: %s\001",
			Me.admin);

		SendUmode(OPERUMODE_Y,
			"%s: CTCP VERSION received from %s!%s@%s",
			target,
			lptr->nick,
			lptr->username,
			lptr->hostname);
		return;
	}

	if (strncasecmp(msg, "DCC CHAT", 8) == 0)
	{
		int		acnt;
		char	**av;
		char	buff[MAXLINE];

		if (strcasecmp(target, n_OperServ) != 0)
			return; /* only let n_OperServ accept dcc chats */

		goodDCC = 1;
		if (lptr == (Luser *) NULL)
			goodDCC = 0;

	if (lptr->flags & L_OSREGISTERED)
	{
		/*
		 * If they have already registered through /msg, allow
		 * them to be from a different host, by passing 1 to
		 * GetUser() so it can only match nicknames and
		 * not hostnames if need be - if they are not registered,
		 * make sure hostnames match, otherwise anyone could just
		 * /nick <Oline-nickname> and gain dcc chat access
		 */
		tempuser = GetUser(1, lptr->nick, lptr->username, lptr->hostname);
	}
	else
		tempuser = GetUser(0, lptr->nick, lptr->username, lptr->hostname);

		if (tempuser)
		{
			if (CanChat(tempuser))
			{
				goodDCC = 1;

				if (OpersOnlyDcc && !IsOper(tempuser))
					goodDCC = 0;
			}
			else
				goodDCC = 0;
			if (IsAdmin(tempuser))
				goodDCC = 1;
		} /* if (tempuser) */
		else
			goodDCC = 0;

		if (goodDCC == 0)
		{
			notice(n_OperServ, nick,
				"You are not authorized to use this service");

			SendUmode(OPERUMODE_Y,
				"*** Unauthorized Dcc Chat attempt from %s (%s@%s)",
				nick,
				lptr->username,
				lptr->hostname);

			notice(n_OperServ, nick, "\001DCC REJECT chat chat\001");

			putlog(LOG1, "Unauthorized Dcc Chat attempt from %s!%s@%s", 
				nick,
				lptr->username,
				lptr->hostname);

			return;
		}

		strncpy(temp, msg, strlen(msg) - 1);
		temp[strlen(msg) - 1] = '\0';

		acnt = SplitBuf(temp, &av);
		if (acnt < 5)
		{
			notice(n_OperServ, nick, "DCC CHAT connection failed");
			notice(n_OperServ, nick, "\001DCC REJECT chat chat\001");
			MyFree(av);
			return;
		}

		strcpy(buff, "#");
		strcat(buff, av[3]);

		if (atoi(av[4]) < 1024)
		{
			notice(n_OperServ, nick, "Invalid DCC CHAT port specified.  Nice Try.");
			notice(n_OperServ, nick, "\001DCC REJECT chat chat\001");
			putlog(LOG1, "Bogus DCC CHAT port (%s) specified by %s!%s@%s",
				av[4],
				lptr->nick,
				lptr->username,
				lptr->hostname);
		}
		else 
		{
			if (!MakeConn(buff, atoi(av[4]), lptr))
			{
				notice(n_OperServ, nick, "DCC CHAT connection failed");
				notice(n_OperServ, nick, "\001DCC REJECT chat chat\001");
			}
		}

		MyFree(av);
		return;
	}
	else
	{
		strncpy(temp, msg, strlen(msg) - 1); /* chop off ending \001 */
		temp[strlen(msg) - 1] = '\0';
		SendUmode(OPERUMODE_Y,
			"%s: CTCP [%s] received from %s!%s@%s",
			target,
			temp,
			lptr->nick,
			lptr->username,
			lptr->hostname);
	}
} /* onctcp() */

/*
ConnectToTCM()
	args: char *nick, int bindex
	purpose: attempt to connect to tcm bot 'bot[bindex].BotName'
	return: 1 on a successful connection, 0 if not
*/

int
ConnectToTCM(char *nick, Botlist *bptr)

{
	char	sendstr[MAXLINE];
	Luser	*lptr;
	DccUser	*dccptr, *nptr, *tmp;

	if (!nick)
	{
		/*
		 * nick will be null if we are trying to connect to all tcms
		 * at once (AutoLinkFreq)
		 */
		nptr = (DccUser *) NULL;
	}
	else
		if (!(nptr = IsOnDcc(nick)))
			return 0;

	dccptr = (DccUser *) MyMalloc(sizeof(DccUser));
	dccptr->socket = ConnectToServer(bptr->hostname, bptr->port);

	if (dccptr->socket == -1)
	{
		SendUmode(OPERUMODE_B,
			"*** Unable to connect to tcm %s [%s:%d]: %s",
			bptr->name,
			bptr->hostname,
			bptr->port,
			strerror(errno));

		if (nptr)
		{
			sprintf(sendstr, "Unable to connect to port %d of %s: %s\n", 
				bptr->port,
				bptr->hostname,
				strerror(errno));
			writesocket(nptr->socket, sendstr);
		}
		MyFree(dccptr);
		return (0);
	}

	/* 
		The fields for linking to a tcm are in order:
		<remote tcm's nick> <connecting bot's nick> <password>
	*/
	sprintf(sendstr, "%s %s %s\n",
		bptr->name, 
		n_OperServ,
		bptr->password);
	writesocket(dccptr->socket, sendstr);

	dccptr->flags = (SOCK_TCMBOT | SOCK_BOTHUB);
	dccptr->authfd = NOSOCKET;
	dccptr->nick = MyStrdup(bptr->name);

	lptr = FindClient(dccptr->nick); /* check if bot is on network */
	if (!lptr)
	{
		/* just make the bot's hostname ip:port */
		dccptr->username = MyStrdup("unknown");

		sprintf(sendstr, "%s:%d",
			bptr->hostname,
			bptr->port);
		dccptr->hostname = MyStrdup(sendstr);
	}
	else /* give the bot a user@host if possible */
	{
		dccptr->username = MyStrdup(lptr->username);
		dccptr->hostname = MyStrdup(lptr->hostname);
	}

	AddConn(dccptr);

	/* inform everyone of new bot */
	for (tmp = connections; tmp; tmp = tmp->next)
	{
		if (!IsAuth(tmp))
			continue;

		if (tmp == dccptr)
			continue;

		if (tmp->flags & SOCK_TCMBOT)
		{
			sprintf(sendstr, "(%s) Linked to %s [%s@%s]\n",
				n_OperServ,
				bptr->name,
				dccptr->username,
				dccptr->hostname);
			writesocket(tmp->socket, sendstr);
		}
		else
		{
			sprintf(sendstr, "*** Linked to %s [%s@%s]\n",
				bptr->name,
				dccptr->username,
				dccptr->hostname);
			writesocket(tmp->socket, sendstr);
		}
	}

	return (1);
} /* ConnectToTCM */

/*
LinkBots()
 Called from DoTimer() to link all tcm bots who are not
already linked
*/

void
LinkBots()

{
	Botlist		*bptr;
	DccUser		*dccptr;
	int		goodlink;

	if (botlist)
	{
		goodlink = 1;
		for (bptr = botlist; bptr; bptr = bptr->next)
		{
			for (dccptr = connections; dccptr; dccptr = dccptr->next)
			{
				if ((dccptr->flags & SOCK_TCMBOT) &&
						(!strcasecmp(dccptr->nick, bptr->name)))
				{
					/* bot is already linked */
					goodlink = 0;
					break;
				}
			}

			if (goodlink)
				ConnectToTCM((char *) NULL, bptr);
			else
				goodlink = 1;
		}
	} /* if (botlist) */
} /* LinkBots() */

/*
AddConn()
	Insert 'dccptr' into connections list
*/

void
AddConn(DccUser *dccptr)

{
	dccptr->prev = (DccUser *) NULL;
	dccptr->next = connections;
	if (dccptr->next)
		dccptr->next->prev = dccptr;
	connections = dccptr;

	Network->TotalConns++;
} /* AddConn() */

/*
CloseConnection()
	args: DccUser *dccptr
	purpose: close socket of DCC CHAT client specified by 'cindex'
	return: none
*/

void
CloseConnection(DccUser *dccptr)

{
	char prefix[MAXLINE];
	Userlist *tempuser;

	if (!dccptr)
		return;

	close(dccptr->socket);
	dccptr->socket = NOSOCKET;
	Network->TotalConns--;

	if ((dccptr->flags & SOCK_DCC) || !(dccptr->flags & SOCK_PENDING))
	{
		prefix[0] = '\0';

		if (dccptr->flags & SOCK_TCMBOT)
			strcpy(prefix, "Bot");
		else
		{
			tempuser = DccGetUser(dccptr);

			if (IsServicesAdmin(tempuser))
				strcpy(prefix, "Services Admin");
			else if (IsAdmin(tempuser))
				strcpy(prefix, "Admin");
			else if (IsOper(tempuser))
				strcpy(prefix, "Oper");
			else
				strcpy(prefix, "User");
		}

		BroadcastDcc(DCCALL,
			"%s %s (%s@%s) has disconnected\n",
			prefix,
			dccptr->nick,
			dccptr->username,
			dccptr->hostname);
	}
	else
	{
		/*
		 * they connected through telnet, but closed the connection
		 * before sending their password
		 */
		SendUmode(OPERUMODE_Y,
			"*** Lost connection to [%s@%s:%d]",
			dccptr->username,
			dccptr->hostname,
			dccptr->port);
	}

	if (dccptr->nick)
		MyFree(dccptr->nick);
	if (dccptr->username)
		MyFree(dccptr->username);
	if (dccptr->hostname)
		MyFree(dccptr->hostname);

	if (!dccptr->prev)
	{
		if (!dccptr->next)
			connections = (DccUser *) NULL;
		else
		{
			/* its the first in the list */
			dccptr->next->prev = (DccUser *) NULL;
			connections = dccptr->next;
		}
	}
	else
	{
		if (!dccptr->next) /* it's last in the list */
			dccptr->prev->next = (DccUser *) NULL;
		else /* it's in the middle somewhere */
		{
			dccptr->next->prev = dccptr->prev;
			dccptr->prev->next = dccptr->next;
		}
	}

	MyFree(dccptr);
	dccptr = (DccUser *) NULL;

	DccError = 1;

} /* CloseConnection() */

/*
ExpireIdent()
 Check if any ident requests have timed out - if so, make
their username "unknown", and allow them to connect
*/

void
ExpireIdent(time_t unixtime)

{
	DccUser *dccptr;

	for (dccptr = connections; dccptr; dccptr = dccptr->next)
	{
		if ((dccptr->flags & SOCK_NEEDID) &&
				(dccptr->authfd != NOSOCKET) &&
				!(dccptr->flags & SOCK_DCC) &&
				((unixtime - dccptr->idle) >= IdentTimeout))
		{
			close (dccptr->authfd);
			dccptr->authfd = NOSOCKET;
			dccptr->flags &= ~SOCK_NEEDID;

			if (dccptr->username)
				MyFree(dccptr->username);
			dccptr->username = MyStrdup("unknown");

			/* reset idle time for TelnetTimeout */
			dccptr->idle = time(NULL);

			writesocket(dccptr->socket, "\nIdent-request timed out\n");
			if (!(dccptr->flags & SOCK_TCMBOT))
				telnet_greet(dccptr);
		}
	}
} /* ExpireIdent() */

/*
ExpireTelnet()
 Go through dcc list and determine if any clients have taken
longer than TelnetTimeout seconds to enter their password -
if so, close the connection
*/

void
ExpireTelnet(time_t unixtime)

{
	DccUser *dccptr, *prev;

	prev = (DccUser *) NULL;
	for (dccptr = connections; dccptr; )
	{
		if ((dccptr->flags & SOCK_PENDING) &&
				!(dccptr->flags & SOCK_DCC) &&
				((unixtime - dccptr->idle) >= TelnetTimeout))
		{
			if (prev)
			{
				prev->next = dccptr->next;
				CloseConnection(dccptr);
				dccptr = prev;
			}
			else
			{
				/*
				 * We're deleting the first element in the list
				 */
				connections = dccptr->next;
				CloseConnection(dccptr);
				dccptr = (DccUser *) NULL;
			}
		}

		prev = dccptr;

		/*
		 * If dccptr is NULL, we deleted the first element in
		 * the list, and connections was set to connections->next,
		 * so set dccptr to the new connections, to check what
		 * was previously the second element, but now is the first
		 */

		if (dccptr)
			dccptr = dccptr->next;
		else
			dccptr = connections;
	}
} /* ExpireTelnet() */

/*
CheckEOF()
 Called from DoTimer() to close any connections that have
EOF'd
*/

void
CheckEOF()

{
	DccUser *dccptr, *prev;

	prev = (DccUser *) NULL;
	for (dccptr = connections; dccptr; )
	{
		if (dccptr->flags & SOCK_EOF)
		{
			if (prev)
			{
				prev->next = dccptr->next;
				CloseConnection(dccptr);
				dccptr = prev;
			}
			else
			{
				connections = dccptr->next;
				CloseConnection(dccptr);
				dccptr = (DccUser *) NULL;
			}
		}

		prev = dccptr;

		if (dccptr)
			dccptr = dccptr->next;
		else
			dccptr = connections;
	}
} /* CheckEOF() */

/*
DoBinds()
 Called from DoTimer() to bind P: lined ports that we were
unable to bind on startup
*/

void
DoBinds()

{
	PortInfo	*pptr;

	for (pptr = portlist; pptr; pptr = pptr->next)
	{
		if ((pptr->tries < MaxBinds) && (pptr->socket == NOSOCKET))
		{
			do_listen(pptr);

			if (pptr->tries >= MaxBinds)
			{
				putlog(LOG2,
					"Giving up attempt to bind port [%d]",
					pptr->port);

				SendUmode(OPERUMODE_Y,
					"Giving up attempt to bind port [%d]",
					pptr->port);
			}
		}
	}
} /* DoBinds() */

/*
GoodTCM()
	args: DccUser *dccptr
	purpose: determine if dccptr matches an L: line for TCM bots
	return: pointer to bot structure in list
*/

Botlist *
GoodTCM(DccUser *dccptr)

{
	Botlist *tempbot;
	char *user,
	     *host;

	if (!dccptr)
		return ((Botlist *) NULL);

	if (dccptr->flags & SOCK_NEEDID)
		user = (char *) NULL;
	else
		user = dccptr->username;

	host = dccptr->hostname;

	for (tempbot = remotebots; tempbot; tempbot = tempbot->next)
	{
		if (user && tempbot->username)
			if (match(tempbot->username, user) == 0)
				continue;

		if (match(tempbot->hostname, host))
			return (tempbot);
	}

	return ((Botlist *) NULL);
} /* GoodTCM() */

DccUser *
GetBot(char *bname)

{
	DccUser	*botptr;

	for (botptr = connections; botptr; botptr = botptr->next)
		if ((botptr->flags & SOCK_TCMBOT) &&
				(!strcasecmp(botptr->nick, bname)))
			return(botptr);

	return ((DccUser *) NULL);
} /* GetBot() */

/*
IsDccSock()
	args: int sockfd
	purpose: determine if 'sockfd' is being used by a DCC CHATter
	return: ptr to structure in list connections
*/

DccUser *
IsDccSock(int sockfd)

{
	DccUser		*tmp;

	for (tmp = connections; tmp; tmp = tmp->next)
		if (tmp->socket == sockfd)
			return (tmp);

	return ((DccUser *) NULL);
} /* IsDccSock() */

/*
IsOnDcc()
	args: char *nick
	purpose: determine if 'nick' is currently on DCC CHAT
	return: ptr to structure containing 'nick'
*/

DccUser *
IsOnDcc(char *nick)

{
	DccUser	*tmp;

	if (!nick)
		return ((DccUser *) NULL);

	for (tmp = connections; tmp; tmp = tmp->next)
	{
		if (((tmp->flags & SOCK_PENDING) && !(tmp->flags & SOCK_DCC)) ||
				(tmp->flags & SOCK_NEEDID))
			continue;

		if (!strcasecmp(tmp->nick, nick))
			return (tmp);
	}

	return ((DccUser *) NULL);
} /* IsOnDcc() */

/*
DccGetUser()
 Return O: line structure matching dccptr->nick/username/hostname
*/

Userlist *
DccGetUser(DccUser *dccptr)

{
	Userlist *userptr;

	if (!dccptr)
		return ((Userlist *) NULL);

	/*
	 * Give the arguement 1 to match nicknames if it can't
	 * find a hostname match, because the client may be
	 * a telnet client from a different host, but they
	 * still entered the correct username/password
	 */
	userptr = GetUser(1, dccptr->nick, dccptr->username, dccptr->hostname);

	if (userptr && !(dccptr->flags & SOCK_DCC))
	{
		/*
		 * dccptr is a telnet client, not dcc
		 */

		/*
		 * Check if dccptr->nick matches userptr->nick. If it does
		 * not, the problem is most likely the following scenario:
		 * There are two O: lines:
		 *        UserA *@a.com
		 *        UserB *@b.com
		 * Now, UserA connects to HybServ (telnet) from *@b.com,
		 * so the above call to GetUser() would have returned
		 * UserB's structure, since hostnames are a higher
		 * priority than nicknames. However, since this is a
		 * telnet connection, the nicknames must also match since
		 * the user is forced to enter their nickname when
		 * they connect. Thus, call GetUser() again, telling
		 * it to match only nicknames, since matching hostnames
		 * has already failed. Otherwise UserA would incorrectly
		 * get UserB's privileges.
		 */
		if (strcasecmp(dccptr->nick, userptr->nick) != 0)
			userptr = GetUser(1, dccptr->nick, (char *) NULL, (char *) NULL);
	}

	return (userptr);
} /* DccGetUser() */

/*
telnet_check()
	Check if any characters of 'buf' are telnet codes, and respond
accordingly.  Return 1 if telnet codes found, 0 if not

 Telnet Codes
 Type  Decimal Octal

 IAC:  255     377
 WILL: 251     373     
 DO:   253     375
 AYT:  246     366
*/

int
telnet_check(int socket, char *buf)

{
	int		ii;
	int		ret = 0;

	if (!buf)
		return 0;

	for (ii = 0; ii < strlen(buf); ii++)
	{
		if ((buf[ii] == '\377') ||
				(buf[ii] == '\373') ||
				(buf[ii] == '\375'))
		{
			ret = 1;
		}
		if (buf[ii] == '\366')
		{
			/* Are you there? */
			writesocket(socket, "\r\nAYT Received\r\n");
			ret = 1;
			break;
		}
	}

	return (ret);
} /* telnet_check() */

/*
DccProcess()
	args: int cindex, char *command
	purpose: process 'command' sent by DCC client specified by 'dccptr'
	return: none
*/

void
DccProcess(DccUser *dccptr, char *command)

{
	if (!dccptr || !command)
		return;

	Network->RecvB += (command ? strlen(command) : 0);

	/* reset idle time */
	dccptr->idle = time(NULL);

	if (!(dccptr->flags & SOCK_DCC) && (dccptr->flags & SOCK_PENDING))
	{
		Userlist *tempuser;
		char prefix[MAXLINE];
		int goodconn; /* are they allowed to connect? */

		/*
		 * dccptr is an incoming telnet connection - they should
		 * either be entering their nickname or password right now
		 */

		if (!dccptr->nick)
		{
			/*
			 * they should be entering their nickname now
			 */
			dccptr->nick = MyStrdup(command);

			/*
			 * \377\373\001 (IAC WILL ECHO) - turn off remote telnet echo
			 * for them to enter their password
			 */
			writesocket(dccptr->socket, "Enter password\377\373\001\r\n");

			return;
		}

		if (telnet_check(dccptr->socket, command))
			return;

		/* Check if they have an O: line */
		if (!(tempuser = DccGetUser(dccptr)))
		{
			writesocket(dccptr->socket, "Invalid password\r\n");

			SendUmode(OPERUMODE_Y,
				"*** Unauthorized connection attempt from %s@%s:%d [%s]",
				dccptr->username,
				dccptr->hostname,
				dccptr->port,
				dccptr->nick);
			CloseConnection(dccptr);
			return;
		}

		if (CanChat(tempuser))
		{
			goodconn = 1;

			if (OpersOnlyDcc && !IsOper(tempuser))
				goodconn = 0;
		}
		else
			goodconn = 0;

		if (goodconn == 0)
		{
			writesocket(dccptr->socket,
				"You are not authorized to use this service\r\n");

			SendUmode(OPERUMODE_Y,
				"Unauthorized connection attempt from %s@%s:%d [%s]",
				dccptr->username,
				dccptr->hostname,
				dccptr->port,
				tempuser->nick);
			CloseConnection(dccptr);
			return;
		}

		/* check if 'command' matches their password */

		if (!operpwmatch(tempuser->password, command))
		{
			writesocket(dccptr->socket, "Invalid password\r\n");
			SendUmode(OPERUMODE_Y,
				"*** Invalid password on connection from %s@%s:%d [%s]",
				dccptr->username,
				dccptr->hostname,
				dccptr->port,
				tempuser->nick);
			CloseConnection(dccptr);
			return;
		}
		else
		{
			MyFree(dccptr->nick);
			dccptr->nick = MyStrdup(tempuser->nick);

			dccptr->flags &= ~SOCK_PENDING;

			if (IsServicesAdmin(tempuser))
				strcpy(prefix, "Services Admin");
			else if (IsAdmin(tempuser))
				strcpy(prefix, "Admin");
			else if (IsOper(tempuser))
				strcpy(prefix, "Oper");
			else
				strcpy(prefix, "User");

			/* turn ECHO back on */
			writesocket(dccptr->socket, "\377\374\001\n");

      /* send them the motd */
      SendMotd(dccptr->socket);

			/* inform everyone of new user */
			BroadcastDcc(DCCALL, "%s %s (%s@%s) has connected\n",
				prefix,
				dccptr->nick,
				dccptr->username,
				dccptr->hostname);
		}

		return;
	}

	if (!(dccptr->flags & SOCK_DCC))
		if (telnet_check(dccptr->socket, command))
			return;

	if (*command == '.')
	{
		int		clength;
		char	*tmp;

		/* its a command, not a message to the partyline */

		command++;
		if (IsRegistered((Luser *) NULL, dccptr->socket))
		{
			os_process(dccptr->nick, command, dccptr->socket);
			return;
		}

		clength = 0;
		for (tmp = command; *tmp; tmp++)
		{
			if (*tmp == ' ')
				break;
			clength++;
		}

		if ((!strncasecmp(command, "REGISTER", clength)) ||
				(!strncasecmp(command, "IDENTIFY", clength)) ||
				(!strncasecmp(command, "HELP", clength)))
			os_process(dccptr->nick, command, dccptr->socket);
		else
			writesocket(dccptr->socket, "You must identify first\n");

	}
	else
	{
		/* it was not a command, broadcast it to dcc users */
		if (strlen(command))
			SendDccMessage(dccptr, command);
	}
} /* DccProcess() */

/*
BotProcess()
	args: int bindex, char *line
	purpose: send 'line' to all other bots/users on botnet coming from
					 botptr
	return: none
*/

void
BotProcess(DccUser *botptr, char *line)

{
	DccUser	*tmp;

	if (!line)
		return;

	Network->RecvB += (line ? strlen(line) : 0);

	if (botptr->flags & SOCK_PENDING)
	{
		char	*mynick;
		char	*tcmnick;
		char	*password;
		Botlist		*bptr;
		int		goodconn = 1;
		char	sendstr[MAXLINE];

		if (!(bptr = GoodTCM(botptr)))
		{
			SendUmode(OPERUMODE_B,
				"*** Unauthorized tcm connection received from %s@%s:%d",
				botptr->username,
				botptr->hostname,
				botptr->port);
			putlog(LOG1,
				"Unauthorized tcm connection received from %s@%s:%d",
				botptr->username,
				botptr->hostname,
				botptr->port);
			goodconn = 0;
		}

		if (goodconn)
		{
			while (1)
			{
				/*
				 * this loop is needed so we can break if the tcm gives
				 * an incorrect pass/myname etc
				 */

				mynick = strtok(line, " ");
				tcmnick = strtok((char *) NULL, " ");
				password = strtok((char *) NULL, "\r\n");

				if (!mynick || 
						(strcasecmp(mynick, n_OperServ) != 0))
				{
					SendUmode(OPERUMODE_B,
						"*** Invalid myname on tcm connection attempt from %s@%s:%d",
						botptr->username,
						botptr->hostname,
						botptr->port);
					goodconn = 0;
					break;
				}
	
				if (!tcmnick ||
						(strcasecmp(tcmnick, bptr->name) != 0))
				{
					SendUmode(OPERUMODE_B,
						"*** Invalid bot name specified on tcm connection attempt from %s@%s:%d",
						botptr->username,
						botptr->hostname,
						botptr->port);
					goodconn = 0;
					break;
				}

				if (!password ||
						(strcasecmp(password, bptr->password) != 0))
				{
					SendUmode(OPERUMODE_B,
						"*** Invalid password specified on tcm connection attempt from %s@%s:%d",
						botptr->username,
						botptr->hostname,
						botptr->port);
					goodconn = 0;
					break;
				}

				botptr->nick = MyStrdup(tcmnick);
				botptr->flags &= ~SOCK_PENDING;
				SendUmode(OPERUMODE_B,
					"*** Connection with %s [%s@%s] established\n",
					botptr->nick,
					botptr->username,
					botptr->hostname);
	
				for (tmp = connections; tmp; tmp = tmp->next)
				{
					if (tmp == botptr)
						continue;
	
					if (tmp->flags & SOCK_TCMBOT)
					{
						sprintf(sendstr, "(%s) Linked to %s [%s@%s]\n",
							n_OperServ,
							botptr->nick,
							botptr->username,
							botptr->hostname);
						writesocket(tmp->socket, sendstr);
					}
					else
					{
						sprintf(sendstr, "*** Linked to %s [%s@%s]\n",
							botptr->nick,
							botptr->username,
							botptr->hostname);
						writesocket(tmp->socket, sendstr);
					}
				}
				break;
			} /* while (1) */
		} /* if (goodconn) */

		if (!goodconn)
		{
			DccUser	*prv;

			close(botptr->socket);
			MyFree(botptr->username);
			MyFree(botptr->hostname);

			if (connections)
			{
				if (connections == botptr)
				{
					connections = connections->next;
					MyFree(botptr);
				}
				else
				{
					prv = connections;
					for (tmp = connections->next; tmp; )
					{
						if (tmp == botptr)
						{
							prv->next = tmp->next;
							MyFree(tmp);
							break;
						}
						else
						{
							prv = tmp;
							tmp = tmp->next;
						}
					}
				}
			} /* if (connections) */
			DccError = 1;

			Network->TotalConns--;
		}

		return;
	}

	if (line[0] == '<') /* <nick@bot> said something */
	{
		for (tmp = connections; tmp; tmp = tmp->next)
		{
			if (botptr == tmp)
				continue; /* don't send it back to the bot it came from */
			writesocket(tmp->socket, line);
			writesocket(tmp->socket, "\n");
		}
	}
} /* BotProcess() */
