/*
 * HybServ TS Services, Copyright (C) 1998-1999 Patrick Alken
 * This program comes with absolutely NO WARRANTY
 *
 * Should you choose to use and/or modify this source code, please
 * do so under the terms of the GNU General Public License under which
 * this program is distributed.
 */

static char *rcsid = "$Id: hybserv.c,v 3.40 1998/07/22 00:34:37 wnder Exp $";

#include	"headers.h"
#include  "settings.h"

#define	SERVTAB
#include	"cmds.h"
#undef 	SERVTAB

#include	<netdb.h>
#include	<sys/socket.h>
#include	<netinet/in.h>
#include 	<arpa/inet.h>

int				HubSock; /* socket fd of current hub */
int				DccError;
time_t		TimeStarted; /* unixtime of when services was started */
long			gmt_offset; /* offset from UTC time */
Luser			*ClientList;
aNetwork	*Network;
Channel		*ChannelList;
Server		*ServerList;
MyInfo    Me; /* structure containing services' info */
char			*LocalHostName; /* virtual hostname */
struct	sockaddr_in	LocalAddr;

#ifdef HIGHTRAFFIC_MODE
int				HTM; /* high traffic mode */
time_t		HTM_ts; /* when high traffic mode was activated */
int				ReceiveLoad; /* when to switch to HTM */
#endif /* HIGHTRAFFIC_MODE */

#ifdef ALLOW_JUPES
Jupe			*jupelist;
#endif

#ifdef ALLOW_GLINES
Gline			*glinelist;
#endif

rHost			*rhostlist;
Userlist	*userlist; /* linked list of privileged users */
Servlist	*serverlist; /* list of hub servers */
Chanlist	*mychanlist; /* list of channels to monitor */
Botlist		*botlist, /* list of tcm bots that can be linked to */
					*remotebots; /* list of tcm bots authorized to connect */
DccUser		*connections; /* dcc connections */
PortInfo	*portlist; /* list of ports to listen on */

Ignore		*ignorelist; /* list of ignored hosts */

Userlist	*GenericOper;

#ifdef HAVE_PTHREADS
pthread_t	timerid; /* thread ID for p_CheckTime() thread */
int				GoodTimer; /* 1 if ok to call DoTimer() */
#endif /* HAVE_PTHREADS */

/*
 * Certain functions, like SendUmode() could possibly
 * be called very rapidly for certain events like netjoins/splits
 * where +d users would be notified of every client/server quit -
 * we don't want a +d user to be flooded with a couple thousand
 * messages, so keep FloodTable[] to record how many times
 * these functions were called in the last few seconds.
 * FloodTable[0] will be set to the TS when one of these
 * functions is first called. After that, FloodTable[1] will
 * be used. FloodHits gets incremented everytime the function
 * is called. IsFlood() will determine whether the function
 * is being called too much in too little time.
 */
time_t	FloodTable[2];
int			FloodHits;
int			ActiveFlood; /* is a flood currently taking place? */

/* Contains data read from socket descriptors */
char		buffer[BUFSIZE + 1];

/* array of pointers to each parameter the hub sends */
static	char	*param[MAXPARAM + 1];
int			paramc = 0; /* parameter count */
char		*nextparam = NULL; /* address of next parameter */

/*
 * If recv() fills buffer[] completely, splitting the last
 * line off in the middle somewhere, this array will keep
 * the "spilled over" characters - the partial line that
 * buffer wasn't big enough to hold
 */
char	spill[MAXLINE];
int		offset = 0; /* index of spill where we left off */

char	dccspill[MAXLINE];
int		dccoffset = 0; /* index of dccspill where we left off */

/* 
 * pointer to structure containing the info for the hub services
 * is currently connected to
 */
Servlist	*currenthub;

/*
 * Set to 1 "ConnectBurst" seconds after a successful
 * hub connection, so we don't send bursts of client/clone
 * connections to +y users etc, during an initial connect.
 */
int       SafeConnect;

int       HubCount; /* number of S: lines in hybserv.conf */

static void ReadSocketInfo(void);
static int ReadHub(void);
static int ReadSock(DccUser *);

/*
writesocket()
	args: sockfd, writestr
	purpose: write 'writestr' to socket designated by 'sockfd'
	return: amount of bytes written
*/

int
writesocket(int sockfd, char *writestr)

{
	int		ii;

	if (sockfd == NOSOCKET)
		return (0);

	/*send(sockfd, writestr, strlen(writestr), 0);*/

#ifdef DEBUGMODE
	fprintf(stderr, "Writing: %s",
		writestr);
#endif /* DEBUGMODE */

	ii = write(sockfd, writestr, strlen(writestr));

	Network->SendB += (writestr ? strlen(writestr) : 0);

	if ((ii < 0) && (errno != EAGAIN))
	{
		DccUser		*conn;
		/* 
		 * EOF writing to the socket - check if we were writing to a 
		 * partyline socket, if so, mark the user with SOCK_EOF to
		 * kill the connection later
		 */
		if ((conn = IsDccSock(sockfd)))
			conn->flags |= SOCK_EOF;
	}

	return (ii);
} /* writesocket() */

/*
toserv()
	sends 'str' to hub server
*/

void
toserv(char *format, ...)

{
	char	buf[MAXLINE];
	int ii;
	va_list	args;

	if (HubSock == NOSOCKET)
		return;

	va_start(args, format);

#ifdef HAVE_VSNPRINTF
	vsnprintf(buf, sizeof(buf) - 1, format, args);
	buf[sizeof(buf) - 1] = '\0';
#else
	vsprintf(buf, format, args);
#endif /* HAVE_VSNPRINTF */

	va_end(args);

	/* send the string to the server */
	ii = writesocket(HubSock, buf);

	/*
	 * If format contains a trailing \n, but buf was too large
	 * to fit 'format', make sure the \n gets through
	 */
	if ((format[strlen(format) - 1] == '\n') &&
			(buf[ii - 1] != '\n'))
		writesocket(HubSock, "\n");
} /* toserv() */

/*
tosock()
	sends string to specified socket
*/

void
tosock(int sockfd, char *format, ...)

{
	char buf[MAXLINE];
	va_list args;

	va_start(args, format);

#ifdef HAVE_VSNPRINTF
	vsnprintf(buf, sizeof(buf) - 1, format, args);
	buf[sizeof(buf) - 1] = '\0';
#else
	vsprintf(buf, format, args);
#endif /* HAVE_VSNPRINTF */

	va_end(args);

	/* send the string to the socket */
	writesocket(sockfd, buf);
} /* tosock() */

/*
debug()
	Output a debug message to stderr
*/

void
debug(char *format, ...)

#ifdef DEBUGMODE

{
	va_list	args;

	va_start(args, format);

	fprintf(stderr, "DEBUG> ");
	vfprintf(stderr, format, args);

	va_end(args);
} /* debug() */

#else /* !DEBUGMODE */

{
	va_list	args;
	char	buf[MAXLINE];

	va_start(args, format);

#ifdef HAVE_VSNPRINTF
	vsnprintf(buf, sizeof(buf) - 1, format, args);
	buf[sizeof(buf) - 1] = '\0';
#else
	vsprintf(buf, format, args);
#endif

	va_end(args);

	/* send the string to +d users */
	SendUmode(OPERUMODE_D, "DEBUG: %s",
		buf);
} /* debug() */

#endif /* DEBUGMODE */

/*
fatal()
	Fatal error occured - log it and exit
*/

void
fatal(int keepgoing, char *format, ...)

{
	char	buf[MAXLINE];
	va_list	args;
	int oldlev;

	va_start(args, format);

#ifdef HAVE_VSNPRINTF
	vsnprintf(buf, sizeof(buf) - 1, format, args);
	buf[sizeof(buf) - 1] = '\0';
#else
	vsprintf(buf, format, args);
#endif /* HAVE_VSNPRINTF */

	va_end(args);

	/*
	 * If this is being called from dparse(), it's possible
	 * LogLevel hasn't been set yet. So, make sure this
	 * error gets logged, whether LogLevel is set or not
	 */
	oldlev = LogLevel;
	LogLevel = 1;

	/* log the error */
	if (!keepgoing)
	{
		putlog(LOG1, "FATAL: %s", buf);
		exit(1);
	}
	else
		putlog(LOG1, "Warning: %s", buf);

	LogLevel = oldlev;
} /* fatal() */

/*
ConnectToServer()
	args: hostname, port
	purpose: bind socket and establish connection to hostname at port
	return: socket file descriptor if successful connect, -1 if not
*/

int
ConnectToServer(char *hostname, int port)

{
	struct  sockaddr_in	serv_addr;
	char    ServHostIP[25]; /* Internet address for server hostname */
	char		host[MAXLINE];
	register struct hostent	*hostptr;
	struct 	in_addr	*ptr;
	int			socketfd; /* socket file descriptor */

	memset((void *) &serv_addr, 0, sizeof(serv_addr));

	strcpy(host, hostname);
	if (*hostname == '#') /* DCC CHAT host */
	{
		serv_addr.sin_addr.s_addr = htonl(strtoul(hostname + 1,(char **) NULL, 10));
#ifdef DEBUGMODE
		printf("Connecting to port %d of %s (DCC CHAT)\n", port, inet_ntoa(serv_addr.sin_addr));
#endif
		strcpy(host, inet_ntoa(serv_addr.sin_addr));
	}
	else
	{
		if ((hostptr = gethostbyname(hostname)) == NULL)
		{
		#ifdef DEBUGMODE
			printf("Cannot connect to port %d of %s: Unknown host\n", port, hostname);
		#endif
			return(-1);
		}
		else
		{
			ptr = (struct in_addr *) *hostptr->h_addr_list;
			strcpy(ServHostIP, inet_ntoa(*ptr));
			serv_addr.sin_addr.s_addr = inet_addr(ServHostIP);
		#ifdef DEBUGMODE
			printf("Connecting to port %d of %s (%s)\n", port, hostptr->h_name, ServHostIP); 
		#endif
		}
	}

	serv_addr.sin_family = AF_INET;
	serv_addr.sin_port = htons(port);

	if ((socketfd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
	{
	#ifdef DEBUGMODE
		printf("Unable to open stream socket\n");
	#endif
		putlog(LOG1, "Unable to open stream socket");
		close(socketfd);
		return(-1);
	}

	if (LocalHostName)
	{
		/* bind to virtual host */
		if ((bind(socketfd, (struct sockaddr *) &LocalAddr, sizeof(LocalAddr))) < 0)
			putlog(LOG1, "Unable to bind virtual host %s[%s]: %s",
				LocalHostName,
				inet_ntoa(LocalAddr.sin_addr),
				strerror(errno));
	}

	if (connect(socketfd, (struct sockaddr *) &serv_addr, sizeof(serv_addr)) < 0)
	{
#ifdef DEBUGMODE
		printf("Cannot connect to port %d of %s: %s\n",
			port,
			host,
			strerror(errno));
#endif
		close(socketfd);
		return(-1);
	}

#ifdef DEBUGMODE
	printf("Connection established\n");
#endif

	return(socketfd);
} /* ConnectToServer() */

/*
ProcessInfo()
	args: int arc, char **arv
	purpose: process string sent by hub server (stored in 'arv')
	return: none
*/

void
ProcessInfo(int arc, char **arv)

{
	ServCmd	*stptr;
	int index;

	if ((arc >= 2) && (*arv[0] == ':'))
		index = 1;
	else
		index = 0;

	for (stptr = servtab; stptr->cmd; stptr++)
	{
		if (!strcasecmp(arv[index], stptr->cmd))
		{
			(*stptr->func)(arc, arv);
			break;
		}
	}
} /* ProcessInfo() */

/*
SplitBuf()
 Take string "buff" and insert NULLs in the spaces between words.
Keep pointers to the beginning of each word, and store them
in "array".
 Returns the number of words in "buff"
*/

int
SplitBuf(char *buff, char ***array)

{
	int argsize = 8;
	int acnt,
	    ii;
	char *temp1, *tempbuf;

	tempbuf = buff;

	ii = strlen(tempbuf);

	/* try to kill trailing \r or \n chars */
	if (IsEOL(tempbuf[ii - 1]))
		tempbuf[ii - 1] = '\0';

	/*
	 * When hybrid sends channel bans during a netjoin, it leaves
	 * a preceding space (not sure why) - just make sure there
	 * are no preceding spaces
	 */
	while (IsSpace(*tempbuf))
		tempbuf++;

	*array = (char **) MyMalloc(sizeof(char *) * argsize);
	acnt = 0;
	while (*tempbuf)
	{
		if (acnt == argsize)
		{
			argsize += 8;
			*array = (char **) MyRealloc(*array, sizeof(char *) * argsize);
		}

		if ((*tempbuf == ':') && (acnt != 0))
		{
			(*array)[acnt++] = tempbuf;
			/* (*array)[acnt++] = tempbuf + 1; */
			tempbuf = "";
		}
		else
		{
			temp1 = strpbrk(tempbuf, " ");
			if (temp1)
			{
				*temp1++ = 0;
				while (IsSpace(*temp1))
					temp1++;
			}
			else
				temp1 = tempbuf + strlen(tempbuf);

			(*array)[acnt++] = tempbuf;
			tempbuf = temp1;
		}
	}

	return (acnt);
} /* SplitBuf() */

/*
ReadSocketInfo()
	args: none
	purpose: wait for information from sockets; once an entire
					 line has been sent, call ProcessInfo() to process the line
	return: none
*/

static void
ReadSocketInfo(void)

{
	int		SelectResult;
	struct	timeval	TimeOut;

	DccUser	*tempconn;
	PortInfo	*tempport;
	fd_set		readfds, writefds;

#if defined(HIGHTRAFFIC_MODE) || !defined(HAVE_PTHREADS)
	time_t	curr_ts;
#endif

#ifndef HAVE_PTHREADS
	time_t	last_ts = time(NULL);
#endif

	spill[0] = '\0';
	dccspill[0] = '\0';

	for (;;)
	{

	#ifndef HAVE_PTHREADS
		/*
		 * If pthreads are not enabled, DoTimer() will be called
		 * from this code, and so we need curr_ts here
		 */
		curr_ts = time(NULL);
	#endif

#ifdef HIGHTRAFFIC_MODE

	#ifdef HAVE_PTHREADS
		/*
		 * Since pthreads are enabled, HTM mode is the only thing
		 * that needs curr_ts, since DoTimer() is called from
		 * a separate thread
		 */
		curr_ts = time(NULL);
	#endif

		if (HTM)
		{
			int		htm_count;

		#ifndef HAVE_PTHREADS

			if (last_ts != curr_ts)
			{
				/*
				 * So DoTimer() doesn't get called a billion times when
				 * HTM is turned off
				 */
				last_ts = curr_ts;

				/*
				 * DoTimer() won't do anything other than update
				 * Network->CheckRecvB since we are in HTM
				 */
				DoTimer(curr_ts);
			} /* if (last_ts != curr_ts) */

		#endif /* HAVE_PTHREADS */

			/* Check if HTM should be turned off */
			if ((curr_ts - HTM_ts) >= HTM_TIMEOUT)
			{
				float		currload;

				currload = ((float) (Network->RecvB - Network->CheckRecvB) / 
										(float) 1024) / (float) HTM_INTERVAL;

				if (currload >= (float) ReceiveLoad)
				{
					HTM_ts = curr_ts;
					putlog(LOG1, "Continuing high-traffic mode (%0.2f K/s)",
						currload);
					SendUmode(OPERUMODE_Y,
						"*** Continuing high-traffic mode (%0.2f K/s)",
						currload);
				}
				else
				{
					HTM = HTM_ts = 0;
					putlog(LOG1, "Resuming standard traffic operations (%0.2f K/s)",
						currload);
					SendUmode(OPERUMODE_Y,
						"*** Resuming standard traffic operations (%0.2f K/s)",
						currload);
				}
			}

			for (htm_count = 0; htm_count < HTM_RATIO; htm_count++)
				if (!ReadHub())
					return;

		} /* if (HTM) */
	#ifndef HAVE_PTHREADS
		else
	#endif

#endif /* HIGHTRAFFIC_MODE */

		#ifndef HAVE_PTHREADS

			if (last_ts != curr_ts)
			{
				if (curr_ts > (last_ts + 1))
				{
					time_t	delta;
					/*
					 * we skipped at least 1 time interval - make up for it
					 * by calling DoTimer() how ever many intervals we missed
					 * (ie: if the TS went from 5 to 8, call DoTimer() twice
					 * since we missed 6 and 7)
					 */
					for (delta = (last_ts + 1); delta < curr_ts; delta++)
						DoTimer(delta);
				}
				last_ts = curr_ts;
				DoTimer(curr_ts);
			}

		#endif /* HAVE_PTHREADS */

		FD_ZERO(&readfds);
		FD_ZERO(&writefds);
		TimeOut.tv_sec = 1;
		TimeOut.tv_usec = 0L;

		if (HubSock != NOSOCKET)
			FD_SET(HubSock, &readfds);

		for (tempconn = connections; tempconn; tempconn = tempconn->next)
		{
			FD_SET(tempconn->socket, &readfds);
			if (tempconn->authfd != NOSOCKET)
			{
				FD_SET(tempconn->authfd, &readfds);
				if (tempconn->flags & SOCK_WRID)
					FD_SET(tempconn->authfd, &writefds);
			}
		}

		for (tempport = portlist; tempport; tempport = tempport->next)
			if (tempport->socket != NOSOCKET)
				FD_SET(tempport->socket, &readfds);

		SelectResult = select(FD_SETSIZE, &readfds, &writefds, NULL, &TimeOut);
		if (SelectResult > 0)
		{
			/* something happened */

			if (HubSock != NOSOCKET)
			{
				/* This is for info coming from the hub server */
				if (FD_ISSET(HubSock, &readfds))
				{
					if (!ReadHub())
						return; /* something bad happened */
				} /* if (FD_ISSET(HubSock, &readfds)) */
			} /* if (HubSock != NOSOCKET) */

			for (tempport = portlist; tempport; tempport = tempport->next)
			{
				if (tempport->socket == NOSOCKET)
					continue;

				if (FD_ISSET(tempport->socket, &readfds))
				{
				#ifdef HAVE_PTHREADS
					pthread_t		clientid;
					pthread_attr_t	attr;
				#endif

					/*
					 * a client has connected on one of our listening ports
					 * (P: lines) - accept their connection, and perform
					 * ident etc
					 */

				#ifdef HAVE_PTHREADS
					/* this way gethostbyaddr() will be non-blocking */
					pthread_attr_init(&attr);
					pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
					pthread_create(&clientid, &attr, (void *) &ConnectClient, (void *) tempport);
				#else
					ConnectClient(tempport);
				#endif
				}
			}

			/* this is for info coming from a dcc/telnet/tcm client */
			for (tempconn = connections; tempconn; tempconn = tempconn->next)
			{
				if (tempconn->socket == NOSOCKET)
					continue;

				if (tempconn->authfd != NOSOCKET)
				{
					if (FD_ISSET(tempconn->authfd, &writefds) &&
							(tempconn->flags & SOCK_WRID))
						writeauth(tempconn); /* send ident request */

					if (FD_ISSET(tempconn->authfd, &readfds))
						readauth(tempconn); /* read ident reply */
				}

				if (tempconn->flags & SOCK_NEEDID)
					continue;

				if (FD_ISSET(tempconn->socket, &readfds))
				{
					/* read and parse any data from the socket */
					if ((ReadSock(tempconn) == (-1)) || DccError)
					{
						DccError = 0;
						break;
					}
				} /* if (FD_ISSET(tempconn->socket, &readfds)) */
			} /* for (tempconn = connections; tempconn; tempconn = tempconn->next) */
		} /* if (SelectResult > 0) */
		else
		{
			if ((SelectResult == (-1)) && (errno != EINTR))
			{
			#ifdef DEBUGMODE			
				fprintf(stderr, "Connection closed: %s\n", strerror(errno));
			#endif
				putlog(LOG1, "Lost connection to %s:%d (%s)",
					currenthub->hostname,
					currenthub->port,
					strerror(errno));
				return;
			}
		}
	} /* for (;;) */
} /* ReadSocketInfo() */

/*
ReadHub()
	Read and parse any data from the hub server - return 1 if successful,
0 if not; return 2 if the socket is ok, but there's no data to be read
*/

static int
ReadHub()

{
	int length; /* number of bytes we read */
	register char *ch;
	register char *linech;
#ifdef EXTREMEDEBUG
	FILE *fp;
#endif

	if (HubSock == NOSOCKET)
		return 0;

	/* read in a line */
	length = recv(HubSock, buffer, BUFSIZE, 0);

	if ((length == (-1)) && ((errno == EWOULDBLOCK) || (errno == EAGAIN)))
		return 2; /* no error - there's just nothing to read */

	if (length <= 0)
	{
		putlog(LOG1, "Read error from server: %s",
			strerror(errno));
		return 0; /* the connection was closed */
	}

	Network->RecvB += length;

	/*
	 * buffer may not be big enough to read the whole last line
	 * so it may contain only part of it
	 */
	buffer[length] = '\0';

#ifdef EXTREMEDEBUG
	if ((fp = fopen("hybserv.hubinfo", "a+")))
	{
		fprintf(fp, "%s", buffer);
		fclose(fp);
	}
#endif

	/*
	 * buffer could possibly contain several lines of info,
	 * especially if this is the inital connect burst, so go
	 * through, and record each line (until we hit a \n) and
	 * process it separately
	 */

	ch = buffer;
	linech = spill + offset;

	/*
	 * The following routine works something like this:
	 * buffer may contain several full lines, and then
	 * a partial line. If this is the case, loop through
	 * buffer, storing each character in 'spill' until
	 * we hit a \n or \r.  When we do, process the line.
	 * When we hit the end of buffer, spill will contain
	 * the partial line that buffer had, and offset will
	 * contain the index of spill where we left off, so the
	 * next time we recv() from the hub, the beginning
	 * characters of buffer will be appended to the end of
	 * spill, thus forming a complete line.
	 * If buffer does not contain a partial line, then
	 * linech will simply point to the first index of 'spill'
	 * (offset will be zero) after we process all of buffer's
	 * lines, and we can continue normally from there.
	 */

	while (*ch)
	{
		register char tmp;
	#ifdef DEBUGMODE
		int		ii;
	#endif

		tmp = *ch;
		if (IsEOL(tmp))
		{
			*linech = '\0';

			if (nextparam)
			{
				/*
				 * It is possible nextparam will not be NULL here
				 * if there is a line like:
				 * PASS password :TS
				 * where the text after the colon does not have
				 * any spaces, so we reach the \n and do not
				 * execute the code below which sets the next
				 * index of param[] to nextparam. Do it here.
				 */
				param[paramc++] = nextparam;
			}

		#ifdef DEBUGMODE
			for (ii = 0; ii < paramc; ii++)
				fprintf(stderr, "%s ", param[ii]);
			fprintf(stderr, "\n");
		#endif

			/*
			 * Make sure paramc is non-zero, because if the line
			 * starts with a \n, we will immediately come here,
			 * without initializing param[0]
			 */
			if (paramc)
			{
				/* process the line */
				ProcessInfo(paramc, param);
			}

			linech = spill;
			offset = 0;
			paramc = 0;
			nextparam = (char *) NULL;

			/*
			 * If the line ends in \r\n, then this algorithm would
			 * have only picked up the \r. We don't want an entire
			 * other loop to do the \n, so advance ch here.
			 */
			if (IsEOL(*(ch + 1)))
				ch++;
		}
		else
		{
			/* make sure we don't overflow spill[] */
			if (linech >= (spill + (sizeof(spill) - 1)))
			{
				ch++;
				continue;
			}

			*linech++ = tmp;
			if (tmp == ' ')
			{
				/*
				 * Only set the space character to \0 if this is
				 * the very first parameter, or if nextparam is
				 * not NULL. If nextparam is NULL, then we've hit
				 * a parameter that starts with a colon (:), so
				 * leave it as a whole parameter.
				 */
				if (nextparam || (paramc == 0))
					*(linech - 1) = '\0';

				if (paramc == 0)
				{
					/*
					 * Its the first parameter - set it to the beginning
					 * of spill
					 */
					param[paramc++] = spill;
					nextparam = linech;
				}
				else
				{
					if (nextparam)
					{
						param[paramc++] = nextparam;
						if (*nextparam == ':')
						{
							/*
							 * We've hit a colon, set nextparam to NULL,
							 * so we know not to set any more spaces to \0
							 */
							nextparam = (char *) NULL;

							/*
							 * Unfortunately, the first space has already
							 * been set to \0 above, so reset to to a
							 * space character
							 */
							*(linech - 1) = ' ';
						}
						else
							nextparam = linech;

						if (paramc >= MAXPARAM)
							nextparam = (char *) NULL;
					}
				}
			}
			offset++;
		}

		/*
		 * Advance ch to go to the next letter in the buffer
		 */
		ch++;
	}

	return 1;
} /* ReadHub() */

/*
ReadSock()
	Similar to ReadHub(), but read a dcc/telnet/tcm socket instead and
parse any data
  return 1 if successful, 0 if not; return 2 if the socket is ok,
but there's no data to be read
*/

static int
ReadSock(DccUser *connptr)

{
	int		length;
	register char *ch;
	register char *linech;

	if (!connptr)
		return 0;

	length = recv(connptr->socket, buffer, BUFSIZE, 0);

	if ((length == (-1)) && ((errno == EWOULDBLOCK) || (errno == EAGAIN)))
		return 2; /* no error - there's just nothing to read */

	if (length <= 0)
	{
		/* DCC client closed connection */
		CloseConnection(connptr);
		/*DccError = 0;*/
		return 0;
	}

	/*
   * recv() may have cut off the line in the middle
	 * if buffer wasn't big enough
	 */
	buffer[length] = '\0';

	if (telnet_check(connptr->socket, buffer))
		return 1;

	ch = buffer;
	linech = dccspill + dccoffset;

	while (*ch)
	{
		register char tmp;

		tmp = *ch;
		if (IsEOL(tmp))
		{
			*linech = '\0';

			if (*dccspill && IsEOL(*dccspill))
			{
				ch++;
				continue;
			}

			/* process the line */
			if (connptr->flags & SOCK_TCMBOT)
				BotProcess(connptr, dccspill); /* process line from bot */
			else
				DccProcess(connptr, dccspill); /* process line from client */

			linech = dccspill;
			dccoffset = 0;

			if (DccError)
				break;

			/*
			 * If the line ends in \r\n, advance past the \n
			 */
			if (IsEOL(*(ch + 1)))
				ch++;
		}
		else
		{
			/* make sure we don't overflow dccspill */
			if (linech < (dccspill + (sizeof(dccspill) - 1)))
			{
				*linech++ = tmp;
				dccoffset++;
			}
		}
		ch++;
	}

	return 1;
} /* ReadSock() */

/*
CycleServers()
 Go through server list trying to get a connection to a hub
server
*/

void
CycleServers()

{
	if (HubSock == NOSOCKET)
	{
		if (currenthub)
		{
			if (!currenthub->next)
				currenthub = serverlist;
			else
				currenthub = currenthub->next;
		}
		else
			currenthub = serverlist;

		SendUmode(OPERUMODE_Y,
			"*** Cycling server list");

		while (currenthub)
		{
			/* Attempt connection to server */
			if ((HubSock = ConnectToServer(currenthub->hostname, currenthub->port)) >= 0) 
			{
				set_nonblocking(HubSock);
				currenthub->connect_ts = time(NULL);
				signon();

				SendUmode(OPERUMODE_Y,
					"*** Connected to %s:%d",
					currenthub->hostname,
					currenthub->port);

				putlog(LOG1,
					"Connected to %s:%d",
					currenthub->hostname,
					currenthub->port);

				break;
			}
			else
			{
				SendUmode(OPERUMODE_Y,
					"*** Unable to connect to %s:%d (%s)",
					currenthub->hostname,
					currenthub->port,
					strerror(errno));

				close(HubSock);
				HubSock = NOSOCKET;
				currenthub = currenthub->next;
			}
		} /* while (currenthub) */
	} /* if (HubSock == NOSOCKET) */
} /* CycleServers() */

int
main(int argc, char *argv[])

{
#ifndef DEBUGMODE
	int			pid; /* pid of this process */
#endif
	FILE		*pidfile; /* to write our pid */
	struct tm *tm_gmt;
	uid_t		uid, /* real user id */
					euid; /* effective user id */
#ifdef HAVE_PTHREADS
	pthread_attr_t	attr;
	pthread_t	selectid,
						signalid;
#endif

	TimeStarted = time(NULL);

	fprintf(stderr,
		"\nHybServ TS services version %s by Patrick Alken (wnder@underworld.net)\n",
		hVersion);

	/*
	 * Load SETPATH (settings.conf) - this must be done
	 * before the config file is loaded, and before any
	 * putlog() calls are made, since LogFile is specified
	 * in settings.conf
	 */
	if (LoadSettings(0) == 0)
	{
		fprintf(stderr, "Fatal errors encountered and logged parsing %s, exiting\n",
			SETPATH);
		return (0);
	}

	/*
	 * If they run ./shownicks or ./showchans rather than ./hybserv
	 * display nicknames/channels
	 */
	if (strstr(argv[0], "shownicks"))
	{
	#ifdef NICKSERVICES
		ShowNicknames(argc, argv);
	#endif /* NICKSERVICES */
		return (0);
	}
	else if (strstr(argv[0], "showchans"))
	{
	#if defined(NICKSERVICES) && defined(CHANNELSERVICES)
		ShowChannels(argc, argv);
	#endif /* defined(NICKSERVICES) && defined(CHANNELSERVICES) */
		return (0);
	}

	putlog(LOG1, "HybServ TS services started");

	/*
	 * Get the offset from GMT (London time)
	 */
	tm_gmt = gmtime(&TimeStarted);
	gmt_offset = TimeStarted - mktime(tm_gmt);

	uid = getuid(); /* the user id of the user who ran the process */
	euid = geteuid(); /* the effective id (different if setuid) */

	if (!uid || !euid)
	{
		/*
		 * It is setuid root or running as root
		 */
		printf("Please do not run HybServ as root\n");
		putlog(LOG1, "Detected running with root uid, exiting");
		exit (0);
	}

	if (chdir(HPath) != 0)
	{
		fprintf(stderr,
			"HPath is an invalid directory, please check %s\n",
			SETPATH);
		putlog(LOG1, "Invalid HPath (%s), shutting down", HPath);
		exit(0);
	}

	/*
	 * the Network list must be initialized before the config
   * file is loaded
   */
	InitLists();

	/* load server, jupe, gline, user, admin info */
	LoadConfig(ConfigFile);

	/* load nick/chan/memo/stat databases */
	LoadData();

	if (LocalHostName)
	{
		struct	hostent		*hptr;

		/* set up virtual hostname */
		if ((hptr = gethostbyname(LocalHostName)) == (struct hostent *) NULL)
		{
			fprintf(stderr,
				"Unable to resolve virtual host [%s]: Unknown hostname\n",
				LocalHostName);
			MyFree(LocalHostName);
			LocalHostName = (char *) NULL;
		}
		else
		{
			memset((void *) &LocalAddr, 0, sizeof(struct sockaddr_in));
			LocalAddr.sin_family = AF_INET;
			LocalAddr.sin_port = 0;
			memcpy((void *) &LocalAddr.sin_addr, (void *) hptr->h_addr,
				hptr->h_length);
			printf("Using virtual host %s[%s]\n",
				LocalHostName,
				inet_ntoa(LocalAddr.sin_addr));
		}
	}

#ifndef DEBUGMODE
	pid = fork();
	if (pid == (-1))
	{
		printf("Unable to fork(), exiting.\n");
		exit(0);
	}
	if (pid != 0)
	{
		printf("Running in background (pid: %d)\n\n", pid);
		exit(0);
	}
	close(0);
	close(1);
	close(2);
#else
	printf("Entering debug mode\n");	
#endif


	/*
	 * Signals must be set up after fork(), since the parent
	 * exits
	 */
#ifdef HAVE_PTHREADS

	/* pthreads and signal() do not go well together */
	pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
	pthread_create(&signalid, &attr, p_CheckSignals, NULL);

#else

	/* setup signal hooks */
	signal(SIGHUP, ProcessSignal);
	signal(SIGSEGV, ProcessSignal);
	signal(SIGBUS, ProcessSignal);
	signal(SIGTERM, ProcessSignal);
	signal(SIGINT, ProcessSignal);
	signal(SIGCHLD, ProcessSignal);
	signal(SIGPIPE, SIG_IGN);

#endif /* HAVE_PTHREADS */

	srandom(time(NULL));

	/* write our pid to a file */
	if ((pidfile = fopen(PidFile, "w")) == (FILE *) NULL)
		putlog(LOG1, "Unable to open %s", PidFile);
	else
	{
		fprintf(pidfile, "%i\n", (int) getpid());
		fclose(pidfile);
	}

	/* initialize tcm/user listening ports */
	InitListenPorts();

	/* initialize hash tables */
	ClearHashes(1);

#ifdef BLOCK_ALLOCATION
	InitHeaps();
#endif

	HubSock = NOSOCKET;
	currenthub = (Servlist *) NULL;
	CycleServers();

#ifdef HAVE_PTHREADS

	/*
	 * This thread will call DoTimer() every second
	 */
	pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
	pthread_create(&timerid, &attr, p_CheckTime, NULL);

#endif

	while (1)
	{
		/* enter loop waiting for server info */

	#ifdef HAVE_PTHREADS

		pthread_create(&selectid, NULL, (void *) &ReadSocketInfo, NULL);
		pthread_join(selectid, NULL);

	#else

		ReadSocketInfo();

	#endif /* HAVE_PTHREADS */

		if (Me.hub)
			SendUmode(OPERUMODE_Y,
				"*** Disconnected from %s",
				Me.hub->name);
		else
			SendUmode(OPERUMODE_Y,
				"*** Disconnected from hub server");

		if (currenthub)
			if (currenthub->realname)
			{
				MyFree(currenthub->realname);
				currenthub->realname = (char *) NULL;
			}

		close(HubSock); /* There was an error */
		HubSock = NOSOCKET;
		currenthub->connect_ts = 0;

		/* 
		 * whenever HybServ connects/reconnects to a server, clear
		 * users, servers, and chans
		 */
		ClearUsers();
		ClearChans();
		ClearServs();
		/*
		 * ClearHashes() must be called AFTER ClearUsers(),
		 * or StatServ's unique client counts will be off since
		 * cloneTable[] would be NULL while it was trying to find
		 * clones
		 */
		ClearHashes(0);

		PostCleanup();
	} /* while (1) */

	return (0);
} /* main() */
