/*
 * HybServ TS Services, Copyright (C) 1998-1999 Patrick Alken
 * This program comes with absolutely NO WARRANTY
 *
 * Should you choose to use and/or modify this source code, please
 * do so under the terms of the GNU General Public License under which
 * this program is distributed.
 */

static char *rcsid = "$Id: init.c,v 1.20 1998/08/14 20:49:10 wnder Exp $";

#include	"headers.h"
#include  "settings.h"

#include  <sys/stat.h>
#include	<sys/wait.h>
#include	<sys/socket.h>
#include	<netinet/in.h>
#include  <fcntl.h>

extern  char  *ConfigFile;
extern  char  *LocalHostName; /* virtual hostname */

static Luser *introduce(char *, char *, char *);

/*
Rehash()
	args: none
	purpose: clear all old jupe, gline, and user lists and parse
					 ConfigFile for new ones.
	return: 0 if successful rehash, 1 if not
*/

int
Rehash()

{
	int		RehashError;
	FILE	*fp;
	Chanlist	*tempchan;
	Servlist	*tempserv;
	PortInfo	*pptr;

	/*
	 * We have to try to open the configuration file before
	 * calling ParseConf() in case we no longer have access to
	 * it, or it is no longer there, in which case do not
	 * kill the old conf lists. Unfortunately this means we'll
	 * have to call fopen() on it twice (once again in ParseConf()).
	 */
	if (!(fp = fopen(ConfigFile, "r")))
		return (1);

	fclose(fp);

	/* Kill old	conf lists */
	ClearConfLines();

	/* 
	 * mark all monitored channels with the delete flag, and "undelete"
	 * them as we read in C: lines from the conf - any left over,
	 * permanently delete
	 */
	for (tempchan = mychanlist; tempchan; tempchan = tempchan->next)
		tempchan->flags |= CHAN_DELETE;

	/*
	 * mark servers with delete flag and undelete them as we read
	 * in S: lines from conf - if we clear the serverlist, currenthub
	 * will point to garbage, so this makes sure that doesn't happen
	 */
	for (tempserv = serverlist; tempserv; tempserv = tempserv->next)
		tempserv->flags |= SERV_DELETE;

	/*
	 * do the same with the P: lines
	 */
	for (pptr = portlist; pptr; pptr = pptr->next)
		pptr->type = PRT_DELETE;

	HubCount = 0;

	/*
	 * Reload configuration files
	 */
	ParseConf(ConfigFile, 1);

	if (GlineFile)
		ParseConf(GlineFile, 1);
	if (JupeFile)
		ParseConf(JupeFile, 1);

	RehashError = 0;

	if (!userlist)
	{
		RehashError = 1;
		putlog(LOG1, "Rehash: No O: lines found in %s",
			ConfigFile);
	}

	if (!serverlist)
	{
		RehashError = 1;
		putlog(LOG1, "Rehash: No S: lines found in %s",
			ConfigFile);
	}

	/*
	 * PART any channels that have been removed from conf file
	 * and JOIN any new channels
	 */
	CheckChans();

	/*
	 * Delete server structures that no longer have an S: line
	 */
	CheckServers();

	/*
	 * Close any ports who no longer have a P: line, and start
	 * listening on new listen ports
	 */
	CheckListenPorts();

#ifdef ALLOW_JUPES
	/*
	 * SQUIT any psuedo-servers who no longer have a J: line
	 */
	CheckJupes();
#endif

	/*
	 * All users usermodes will have been erased - reload them
	 */
	if (os_loaddata() == (-2))
		RehashError = 1;

	/*
	 * Rehash settings.conf
	 */
	if (LoadSettings(1) == 0)
		RehashError = 1;

	return (RehashError);
} /* Rehash() */

/*
ServReboot()
	Close DCC connections, reconnect to the network
*/

void
ServReboot()

{
	DccUser	*tempconn;

	while (connections)
	{
		tempconn = connections->next;
		close(connections->socket);
		connections->socket = NOSOCKET;

		if (connections->nick)
			MyFree(connections->nick);
		if (connections->username)
			MyFree(connections->username);
		if (connections->hostname)
			MyFree(connections->hostname);
		MyFree(connections);

		connections = tempconn;
	}
	connections = (DccUser *) NULL;

	toserv("SQUIT %s :restarting\n",
		currenthub->realname ? currenthub->realname : "*");

	/* kill old connection and clear out user/chan lists etc */
	close(HubSock);

	ClearUsers();
	ClearChans();
	ClearServs();

	/*
	 * This has to come after ClearUsers(), or RECORD_SPLIT_TS
	 * will not work correctly
	 */
	ClearHashes(0);

	HubSock = NOSOCKET;

	currenthub->connect_ts = 0;
} /* ServReboot() */

/*
CopyFile()
 Copy 'oldfile' to 'newfile'
Return:
  0  - successful copy
  -1 - cannot open old file
  -2 - cannot open new file
  -3 - old file isn't normal
  -4 - ran out of disk space
*/

int
CopyFile(char *oldfile, char *newfile)

{
	int oldfd,
	    newfd;
	struct stat fst;
	char buffer[MAXLINE];
	int bytes;

	if ((oldfd = open(oldfile, O_RDONLY, 0)) < 0)
		return (-1);

	fstat(oldfd, &fst);
	if (!(fst.st_mode & S_IFREG))
		return (-3);

	if ((newfd = creat(newfile, (int) (fst.st_mode & 0700))) < 0)
	{
		close(oldfd);
		return (-2);
	}

	/*
	 * Now start copying
	 */
	for (bytes = 1; bytes > 0; )
	{
		bytes = read(oldfd, buffer, sizeof(buffer));
		if (bytes > 0)
		{
			if (write(newfd, buffer, bytes) < bytes)
			{
				/* disk space is full */
				close(oldfd);
				close(newfd);
				unlink(newfile); /* delete the file we created */
				return (-4);
			}
		}
		/*
		 * When bytes == 0 (EOF), we will drop out of the loop
		 */
	}

	/* successful copy */
	close(oldfd);
	close(newfd);
	return (0);
} /* CopyFile() */

/*
ProcessSignal()
	args: int sig
	purpose: process system signal specified by 'sig'
	return: none
*/

void
ProcessSignal(int sig)

{
	switch (sig)
	{
		/* rehash configuration file */
		case SIGHUP:
		{
			BroadcastDcc(DCCOPS,
				"Received signal SIGHUP, rehashing configuration file and databases\n");
			putlog(LOG1, "Received signal SIGHUP, rehashing configuration file and databases");

			Rehash();

			if (ReloadDbsOnHup)
				ReloadData();

			signal(SIGHUP, ProcessSignal); /* reset the signal */
			break;
		}

		/* restart services */
		case SIGINT:
		{
		#ifdef DEBUGMODE
			exit(1);
		#else
			BroadcastDcc(DCCOPS,
				"Received signal SIGINT, restarting services\n");
			putlog(LOG1, "Received signal SIGINT, restarting");
			ServReboot();
			signal(SIGINT, ProcessSignal);
		#endif
			break;
		}

		case SIGPIPE:
		{
			putlog(LOG1, "Received signal SIGPIPE, ignoring");
			/*abort();*/
			break;
		}

		/* 
		 * this is required to prevent a child process from becoming a
		 * zombie which just sits out there taking up fds
		 */
		case SIGCHLD:
		{
			wait(NULL);
			signal(SIGCHLD, ProcessSignal);
			break;
		}

		/* something really died */
		case SIGSEGV:
		case SIGBUS:
		case SIGTERM:
		{
			if (sig == SIGSEGV)
			{
				toserv(":%s QUIT :%s\n:%s QUIT :%s\n",
					n_OperServ,
					"ACK! SIGSEGV!",
					n_ChanServ,
					"ACK! SIGSEGV!");
				putlog(LOG1,
					"Received signal SIGSEGV, shutting down (databases not saved)");
			}
			else
				putlog(LOG1, "Received signal %s, shutting down",
					(sig == SIGBUS) ? "SIGBUS" : "SIGTERM");

			if (sig != SIGTERM)
				abort(); /* make a core file to work with */

			if (sig == SIGSEGV)
			{
				BroadcastDcc(DCCOPS,
					"*** Received SIGSEGV, shutting down\n");
				/*
				 * Don't write databases on SIGSEGV, because there's
				 * a good chance they could get corrupted
				 */
				/* do_shutdown((char *) NULL, "SIGSEGV Received"); */
				exit(1);
			}
			else
			{
				BroadcastDcc(DCCOPS,
					"*** Received SIGTERM, shutting down\n");
				do_shutdown((char *) NULL, "SIGTERM Received");
			}
		}
	}
} /* ProcessSignal() */

/*
ParseConf()
 Backend for LoadConfig().
 Args: filename - file name to parse
       rehash - set to 1 if this is a rehash
*/

void
ParseConf(char *filename, int rehash)

{
	FILE *fileptr;
	char line[MAXLINE];
	char *key;
	int scnt; /* S: line count */

	if (!filename)
		return;

	if ((fileptr = fopen(filename, "r")) == NULL)
	{
		fprintf(stderr, "Unable to open %s\n", filename);
		putlog(LOG1,
			"Unable to open configuration file: %s %s",
			filename,
			(filename == ConfigFile) ? "(FATAL)" : "");

		/*
		 * If the file is the top-level config file, exit.
		 * If this is a rehash, do not exit, because Rehash()
		 * would NOT have erased the old conf lists, so we
		 * can continue normally.
		 */
		if (!rehash && (filename == ConfigFile))
			exit(0);

		return;
	}

	scnt = 0;

	while (fgets(line, MAXLINE - 1, fileptr))
	{
		/*
		 * you may add comments to config file using a '#'
		 */
		if (line[0] == '#')
			continue;         

		key = getfield(line);
		if (key == (char *) NULL)
			continue;

		switch (*key)
		{
			case 'a':
			case 'A':
			{
				char *temp;

				/*
				 * If this is a rehash, Me.admin should already be
				 * set, so make sure it gets freed.
				 */
				if (Me.admin)
					MyFree(Me.admin);

				if (!(temp = getfield(NULL)))
					break;

				if (strlen(temp) > REALLEN)
				{
					Me.admin = (char *) MyMalloc(REALLEN);
					strncpy(Me.admin, temp, REALLEN - 1);
				}
				else
					Me.admin = MyStrdup(temp);

				break;
			} /* case 'A' */

			case 's':
			case 'S':
			{
				char *pass;
				char *host;
				char *port;
				Servlist *servptr;
				
				pass = getfield(NULL);
				host = getfield(NULL);
				port = getfield(NULL);
				if (!host || !pass)
					continue;

				if ((servptr = IsServLine(host, port)))
				{
					servptr->flags &= ~SERV_DELETE;
					if (servptr->password)
					{
						/*
						 * This must be a rehash, so in case the password
						 * was changed, set servptr->password to the new one.
						 */
						MyFree(servptr->password);
						servptr->password = MyStrdup(pass);
					}
				}
				else
				{
					AddServ(host, pass, port ? atoi(port) : DefaultHubPort);
				}

				scnt++;
				break;
			} /* case 'S' */

	#ifdef ALLOW_JUPES
			case 'j':
			case 'J':
			{
				char *name;
				char *reason;
				char *who;

				name = getfield(NULL);
				reason = getfield(NULL);
				who = getfield(NULL);
				if (!name || !reason || !who)
					continue;

				AddJupe(name, reason, who);

				break;	
			} /* case 'J' */
	#endif /* ALLOW_JUPES */

	#ifdef ALLOW_GLINES
			case 'g':
			case 'G':
			{
				char *host;
				char *reason;
				char *who;

				host = getfield(NULL);
				reason = getfield(NULL);
				who = getfield(NULL);

				if (!host || !who)
					continue;

				AddGline(host, reason, who, 0);

				break; 		
			} /* case 'G': */
	#endif /* ALLOW_GLINES */

			case 'c':
			case 'C':
			{
				char *cname;
				Chanlist *chanptr;

				cname = getfield(NULL);
				if (!cname)
					continue;

				if (*cname != '#')
					continue;

				if ((chanptr = IsChannel(cname)))
				{
					/*
					 * Since cname is already in the channel list,
					 * this must be a rehash. Remove the deletion
					 * flag from chanptr since the C: line was
					 * not deleted prior to the rehash
					 */
					chanptr->flags &= ~CHAN_DELETE;
				}
				else
				{
					/* a new chan was added to the conf */
					AddMyChan(cname);
				}

				break;
			} /* case 'C' */

			case 'o':
			case 'O':
			{
				char *nick;
				char *host;
				char *pass;
				char *flags;

				host = getfield(NULL);
				pass = getfield(NULL);
				nick = getfield(NULL);
				flags = getfield(NULL);
				if (!host || !pass || !nick || !flags)
					continue;

				AddUser(host, pass, nick, flags);

				break;		
			} /* case 'O' */

			case 'n':
			case 'N':
			{
				char *name;
				char *info;

				name = getfield(NULL);
				info = getfield(NULL);
				if (!name || !info)
					continue;

				if (strlen(name) > REALLEN)
				{
					Me.name = (char *) MyMalloc(REALLEN);
					strncpy(Me.name, name, REALLEN - 1);
				}
				else
					Me.name = MyStrdup(name);

				if (strlen(info) > REALLEN)
				{
					Me.info = (char *) MyMalloc(REALLEN);
					strncpy(Me.info, info, REALLEN - 1);
				}
				else
					Me.info = MyStrdup(info);

				break;
			} /* case 'N' */

			case 'b':
			case 'B':
			{
				char *name;
				char *host;
				char *pass;
				char *port;

				host = getfield(NULL);
				name = getfield(NULL);
				pass = getfield(NULL);
				port = getfield(NULL);
				if (!name || !host || !pass)
					continue;

				AddBot(name, host, pass, port ? atoi(port) : DefaultTcmPort);

				break;
			} /* case 'B' */

			case 'l':
			case 'L':
			{
				char *name;
				char *host;
				char *pass;

				host = getfield(NULL);
				name = getfield(NULL);
				pass = getfield(NULL);
				if (!name || !host || !pass)
					continue;

				AddRemoteBot(name, host, pass);

				break;
			} /* case 'L' */

			case 'i':
			case 'I':
			{
				char *host, *num;

				host = getfield(NULL);
				num = getfield(NULL);
				if (!host || !num)
					continue;

				AddHostLimit(host, atoi(num));

				break;
			} /* case 'I' */

			case 'p':
			case 'P':
			{
				char *port, *type, *host;

				host = getfield(NULL);
				port = getfield(NULL);
				type = getfield(NULL);

				if (!port)
					continue;
				if (!IsNum(port))
					continue;

				AddPort(atoi(port), host, type);

				break;
			} /* case 'P' */

			case 'v':
			case 'V':
			{
				char *vhost;

				vhost = getfield(NULL);
				if (!vhost)
					continue;

				if (!LocalHostName)
					LocalHostName = MyStrdup(vhost);

				break;
			} /* case 'V' */

			case '.':
			{
				char *start, *end;
				char *filename;

				if (!strncasecmp(line, ".include", 8))
				{
					/*
					 * It is a .include statement - meaning they
					 * want to include another file to be parsed
					 * as a regular configuration file. Normal
					 * syntax is: .include "filename"
					 */

					if ((start = strchr(line, '"')))
						filename = start + 1;
					else
					{
						/*
						 * There are no quotes in the line - bad syntax
						 */
						putlog(LOG1,
							"Incorrect config file syntax: %s",
							line);
						continue;
					}

					/*
					 * "filename" should now point past the first quote
					 * (ie: file") so now set the end quote to a \0
					 */
					if ((end = strchr(filename, '"')))
						*end = '\0';
					else
					{
						putlog(LOG1,
							"Incorrect config file syntax: %s",
							line);
						continue;
					}

					/*
					 * Now recursively call ParseConf() to parse the new
					 * configuration file
					 */
					ParseConf(filename, 0);
				} /* if (!strncasecmp(line, ".include", 8)) */

				break;
			} /* case '.' */

			default:
				break;
		}
	}

	HubCount += scnt;

	fclose(fileptr);
} /* ParseConf() */

/*
LoadConfig()
	args: none
	purpose: parse config file and store server, admin, user, and 
	         jupe/gline information
	return: none
*/

void
LoadConfig()

{
	int ConfError;

	ParseConf(ConfigFile, 0);

	/*
	 * If settings.conf specified a Gline/Jupe file, parse
	 * them as well
	 */
	if (GlineFile)
		ParseConf(GlineFile, 0);
	if (JupeFile)
		ParseConf(JupeFile, 0);

	ConfError = 0;

	if (!Me.admin)
	{
		fprintf (stderr,
			"No administrative contact information (A:) found in %s\n",
			ConfigFile);
		ConfError = 1;
	}

	if (!Me.name)
	{
		fprintf(stderr,
			"No server name (N:) specified in %s\n",
			ConfigFile);
		ConfError = 1;
	}

	if (!Me.info)
	{
		fprintf (stderr,
			"No server info (N:) specified in %s\n",
			ConfigFile);
		ConfError = 1;
	}

	if (!serverlist)
	{
		fprintf (stderr,
			"No hub servers (S:) specified in %s\n",
			ConfigFile);
		ConfError = 1;
	}
	
	if (!userlist)
	{
		fprintf (stderr,
			"No users (O:) specified for access to services in %s (at least one user should be added)\n",
			ConfigFile);
		ConfError = 1;
	}

	if (ConfError)
		exit(1);
} /* LoadConfig() */

/*
getfield()
	Get next field separated by a : in the conf file
	** This function is from standard ircd source
*/

char *
getfield (char *newline)

{
	static char *line = (char *) NULL;
	char	*end, *field;
	
	if (newline)
		line = newline;

	if (line == (char *) NULL)
		return ((char *) NULL);

	field = line;
	if ((end = strchr(line, ':')) == NULL)
	{
		line = (char *) NULL;
		if ((end = strchr(field, '\n')) == (char *) NULL)
			end = field + strlen(field);
	}
	else
		line = end + 1;

	*end = '\0';
	if (strlen(field))
		return (field);
	else
		return ((char *) NULL);
} /* getfield() */

/*
do_listen()
*/

void
do_listen(PortInfo *portptr)

{
	struct sockaddr_in socketname;

	memset((void *) &socketname, 0, sizeof(struct sockaddr));
	socketname.sin_family = AF_INET;
	socketname.sin_addr.s_addr = INADDR_ANY;
	socketname.sin_port = htons(portptr->port);
	
	portptr->tries++;

	if ((portptr->socket = socket(PF_INET, SOCK_STREAM, 6)) < 0)
	{
		putlog(LOG1, "Unable to create stream socket: %s",
			strerror(errno));
		close(portptr->socket);
		portptr->socket = NOSOCKET;
		return;
	}

	/* set various socket options */
	SetSocketOptions(portptr->socket);

	if (bind(portptr->socket, (struct sockaddr *)&socketname, sizeof(socketname)) < 0)
	{
		putlog(LOG1, "Unable to bind port %d: %s",
			portptr->port,
			strerror(errno));
		close(portptr->socket);
		portptr->socket = NOSOCKET;
		return;
	}

	if (listen(portptr->socket, 4) < 0)
	{
		putlog(LOG1, "Unable to listen on port %d: %s",
			portptr->port,
			strerror(errno));
		close(portptr->socket);
		portptr->socket = NOSOCKET;
		return;
	}
} /* do_listen() */

/* 
InitListenPorts()
	args: none
	purpose: initialize ports for listening
	return: none
*/

void
InitListenPorts()

{
	PortInfo	*pptr;

	for (pptr = portlist; pptr; pptr = pptr->next)
		do_listen(pptr);
} /* InitListenPorts() */

/*
signon()
	args: none
	purpose: send a PASS and SERVER string to make a connection 
	return: void
*/

void
signon()

{
	toserv("PASS %s :TS\nSERVER %s 1 :%s\n", 
		currenthub->password, 
		Me.name, 
		Me.info); 
} /* signon() */

/*
InitLists()
	args: none
	purpose: initiate linked lists
	return: none
*/

void
InitLists()

{
#ifdef NICKSERVICES
	int		ii;
#endif

	ClientList = (Luser *) NULL;
	ChannelList = (Channel *) NULL;
	ServerList = (Server *) NULL;

	Me.name = (char *) NULL;
	Me.info = (char *) NULL;
	Me.admin = (char *) NULL;
	Me.sptr = (Server *) NULL;
	Me.hub = (Server *) NULL;

	Me.osptr = (Luser *) NULL;
#ifdef NICKSERVICES

	Me.nsptr = (Luser *) NULL;

#ifdef CHANNELSERVICES
	Me.csptr = (Luser *) NULL;
#endif

#ifdef MEMOSERVICES
	Me.msptr = (Luser *) NULL;
#endif

#endif /* NICKSERVICES */

#ifdef STATSERVICES
	Me.ssptr = (Luser *) NULL;
#endif

#ifdef HELPSERVICES
	Me.hsptr = (Luser *) NULL;
#endif

#ifdef ALLOW_JUPES
	jupelist = (Jupe *) NULL;
#endif

#ifdef ALLOW_GLINES
	glinelist = (Gline *) NULL;
#endif

	DccError = 0;
	SafeConnect = 0;
	HubCount = 0;

	rhostlist = (rHost *) NULL;
	userlist = (Userlist *) NULL;
	serverlist = (Servlist *) NULL;
	mychanlist = (Chanlist *) NULL;
	botlist = (Botlist *) NULL;
	remotebots = (Botlist *) NULL;
	connections = (DccUser *) NULL;
	portlist = (PortInfo *) NULL;
	LocalHostName = (char *) NULL;

#ifdef HIGHTRAFFIC_MODE
	HTM = HTM_ts = 0;
	ReceiveLoad = RECVLOAD;
#endif

#ifdef HAVE_PTHREADS
	GoodTimer = 1;
#endif

	FloodTable[0] = FloodTable[1] = 0;
	FloodHits = 0;
	ActiveFlood = 0;

#ifdef NICKSERVICES

	for (ii = 0; ii < NICKLIST_MAX; ii++)
		nicklist[ii] = (NickInfo *) NULL;

	#ifdef CHANNELSERVICES
		for (ii = 0; ii < CHANLIST_MAX; ii++)
			chanlist[ii] = (ChanInfo *) NULL;
	#endif

	#ifdef MEMOSERVICES
		for (ii = 0; ii < MEMOLIST_MAX; ii++)
			memolist[ii] = (MemoInfo *) NULL;
	#endif

#endif /* NICKSERVICES */

	ignorelist = (Ignore *) NULL;

	/* List for network stuff - oper/user/chan count etc */
	Network = (aNetwork *) MyMalloc(sizeof(aNetwork));
	Network->TotalUsers = 0;
	Network->TotalServers = 0;
	Network->TotalOperators = 0;
	Network->TotalChannels = 0; 
	Network->TotalOperKills = 0;
	Network->TotalServKills = 0;
	Network->TotalJupes = 0;
	Network->TotalGlines = 0;
	Network->TotalConns = 0;
	Network->MyChans = 0;
	Network->flags = 0;
	Network->RecvB = 0;
	Network->CheckRecvB = 0;
	Network->LastRecvB = 0;
	Network->SendB = 0;

#ifdef NICKSERVICES
	Network->TotalNicks = 0;

	#ifdef CHANNELSERVICES
		Network->TotalChans = 0;
	#endif

	#ifdef MEMOSERVICES
		Network->TotalMemos = 0;
	#endif

#endif /* NICKSERVICES */

#ifdef STATSERVICES

	Network->MaxUsers = Network->MaxUsers_ts = 0;
	Network->MaxServers = Network->MaxServers_ts = 0;
	Network->MaxOperators = Network->MaxOperators_ts = 0;
	Network->MaxChannels = Network->MaxChannels_ts = 0;

	Network->MaxUsersT = Network->MaxUsersT_ts = 0;
	Network->MaxServersT = Network->MaxServersT_ts = 0;
	Network->MaxOperatorsT = Network->MaxOperatorsT_ts = 0;
	Network->MaxChannelsT = Network->MaxChannelsT_ts = 0;

	Network->OperKillsT = Network->ServKillsT = 0;

	Network->Identd = Network->NonIdentd = 0;
	Network->ResHosts = 0;

#endif /* STATSERVICES */

	/*
	 * Create a generic oper structure for any oper who doesn't
	 * have an O: line in hybserv.conf
	 */
	GenericOper = (Userlist *) MyMalloc(sizeof(Userlist));
	GenericOper->nick = MyStrdup("");
	GenericOper->username = MyStrdup("");
	GenericOper->hostname = MyStrdup("");
	GenericOper->password = MyStrdup("");
	GenericOper->flags = (PRIV_OPER | PRIV_JUPE | PRIV_GLINE);

#ifdef RECORD_SPLIT_TS
	GenericOper->split_ts = GenericOper->whensplit = 0;
#endif

} /* InitLists() */

/*
PostCleanup()
  Called after we are disconnected from a hub server, to
reset some linked lists
*/

void
PostCleanup()

{
	Me.sptr = (Server *) NULL;
	Me.hub = (Server *) NULL;

	Me.osptr = (Luser *) NULL;

#ifdef NICKSERVICES

	Me.nsptr = (Luser *) NULL;

	#ifdef CHANNELSERVICES
		Me.csptr = (Luser *) NULL;
	#endif

	#ifdef MEMOSERVICES
		Me.msptr = (Luser *) NULL;
	#endif

#endif /* NICKSERVICES */

#ifdef STATSERVICES
	Me.ssptr = (Luser *) NULL;
#endif

#ifdef HELPSERVICES
	Me.hsptr = (Luser *) NULL;
#endif
} /* PostCleanup() */

/*
introduce()
	Introduce 'nick' to the network with 'info'
*/

static Luser *
introduce(char *nick, char *ident, char *info)

{
	char	sendstr[MAXLINE];
	time_t CurrTime = time(NULL);
	char	**av;
	Luser *lptr;
	
  sprintf(sendstr, "NICK %s 1 %ld %s %s %s %s :%s\n", 
 			nick,
 			(long) CurrTime,
 			ServiceUmodes,
 			ident,
 			Me.name,
 			Me.name,
 			info);
  toserv(sendstr);

	SplitBuf(sendstr, &av);
	lptr = AddClient(av); /* Add 'nick' to user list */
	MyFree(av);

	return (lptr);
} /* introduce() */

/*
InitServs()
  args: Luser *servptr
  purpose: introduce *SERV_NICKs to the network after a successful
           connection - if servptr specified, introduce only
           that service
  return: none
*/

void
InitServs(Luser *servptr)

{
	if (servptr)
	{
		/*
		 * A service nick was killed, determine which one it was
		 * and re-introduce them. Now, the service will have
		 * been removed from the luser linked list already if
		 * it was a kill. However, s_kill() will have called
		 * GetService(), which returns a pointer to Me.*sptr,
		 * depending on which *Serv was killed. Therefore,
		 * 'servptr' will still correctly point to a Me.*sptr,
		 * even though it really points to garbage. So it is
		 * still safe to compare servptr to Me.*sptr's.
		 */

		if (servptr == Me.osptr)
			Me.osptr = introduce(n_OperServ, id_OperServ, desc_OperServ);

	#ifdef NICKSERVICES

		else if (servptr == Me.nsptr)
			Me.nsptr = introduce(n_NickServ, id_NickServ, desc_NickServ);

	#ifdef CHANNELSERVICES

		else if (servptr == Me.csptr)
			Me.csptr = introduce(n_ChanServ, id_ChanServ, desc_ChanServ);

	#endif /* CHANNELSERVICES */

	#ifdef MEMOSERVICES

		else if (servptr == Me.msptr)
			Me.msptr = introduce(n_MemoServ, id_MemoServ, desc_MemoServ);

	#endif /* MEMOSERVICES */

	#endif /* NICKSERVICES */

	#ifdef STATSERVICES

		else if (servptr == Me.ssptr)
			Me.ssptr = introduce(n_StatServ, id_StatServ, desc_StatServ);

	#endif /* STATSERVICES */

	#ifdef HELPSERVICES

		else if (servptr == Me.hsptr)
			Me.hsptr = introduce(n_HelpServ, id_HelpServ, desc_HelpServ);

	#endif /* HELPSERVICES */

		return;
	}

	/*
	 * Services probably just connected to the network,
	 * introduce all service nicks
	 */

	Me.osptr = introduce(n_OperServ, id_OperServ, desc_OperServ);

#ifdef NICKSERVICES

	/* introduce NickServ/ChanServ to the network */

	Me.nsptr = introduce(n_NickServ, id_NickServ, desc_NickServ);

	#ifdef CHANNELSERVICES

		Me.csptr = introduce(n_ChanServ, id_ChanServ, desc_ChanServ);

	#endif

	#ifdef MEMOSERVICES

		Me.msptr = introduce(n_MemoServ, id_MemoServ, desc_MemoServ);

	#endif

#endif /* NICKSERVICES */

#ifdef STATSERVICES

	Me.ssptr = introduce(n_StatServ, id_StatServ, desc_StatServ);

#endif

#ifdef HELPSERVICES

	Me.hsptr = introduce(n_HelpServ, id_HelpServ, desc_HelpServ);

#endif

} /* InitServs() */
