/*
 * HybServ TS Services, Copyright (C) 1998-1999 Patrick Alken
 * This program comes with absolutely NO WARRANTY
 *
 * Should you choose to use and/or modify this source code, please
 * do so under the terms of the GNU General Public License under which
 * this program is distributed.
 */

static char *rcsid = "$Id: jupe.c,v 1.00 1999/08/07 21:35:23 wnder Exp $";

#include	"headers.h"

#ifdef ALLOW_JUPES

/* 
AddJupe() 
	Add jupe 'name' to list with 'reason'
*/

void
AddJupe(char *name, char *reason, char *who)

{
	Jupe	*tempjupe;
	int		ii,
				nickjupe = 1;

	for (ii = 0; ii < strlen(name); ii++)
		if ((name[ii] == '*') || (name[ii] == '?') || (name[ii] == '.'))
		{
			nickjupe = 0;
			break;
		}

	tempjupe = (Jupe *)MyMalloc(sizeof(Jupe));
	tempjupe->name = MyStrdup(name);
	tempjupe->reason = MyStrdup(reason);
	tempjupe->who = MyStrdup(who);
	tempjupe->isnick = nickjupe;
	tempjupe->next = jupelist;
	tempjupe->prev = (Jupe *) NULL;
	if (tempjupe->next)
		tempjupe->next->prev = tempjupe;
	jupelist = tempjupe;

	Network->TotalJupes++;
} /* AddJupe() */

/*
DeleteJupe()
	Remove jptr from jupe list
*/

void
DeleteJupe(Jupe *jptr)

{
	if (!jptr)
		return;

	MyFree(jptr->name);
	MyFree(jptr->reason);
	MyFree(jptr->who);

	if (jptr->prev)
		jptr->prev->next = jptr->next;
	else
		jupelist = jptr->next;

	if (jptr->next)
		jptr->next->prev = jptr->prev;

	MyFree(jptr);

	Network->TotalJupes--;
} /* DeleteJupe() */

/*
CheckJupes()
 Called after rehash to check if there are any psuedo-servers
which no longer have a J: line. SQUIT any we find
*/

void
CheckJupes()

{
	Server *tempserv, *next;

	if (!Me.sptr)
		return;

	for (tempserv = ServerList; tempserv; )
	{
		/*
		 * Make sure tempserv is using us as an uplink, which
		 * would qualify it as a psuedo-server. But also make
		 * sure it's not OUR current hub server - we never
		 * want to squit our hub.
		 */
		if ((tempserv->uplink == Me.sptr) &&
				(Me.hub != tempserv))
		{
			/*
			 * We found a server who has us as a hub - check if there
			 * is a J: line for it
			 */
			if (!IsJupe(tempserv->name))
			{
				toserv("SQUIT %s :Unjuped\n",
					tempserv->name);

				next = tempserv->next;
				DeleteServer(tempserv);
				tempserv = next;
			}
			else
				tempserv = tempserv->next;
		}
		else
			tempserv = tempserv->next;
	}
} /* CheckJupes() */

/*
DoJupeSquit()
	SQUIT all servers matching serv and create a fake server to take its
place
*/

void
DoJupeSquit(char *serv, char *reason, char *who)

{
	Server *tempserv, *prev;
	char sendstr[MAXLINE], **arv;
	int acnt;

	for (tempserv = ServerList; tempserv; )
	{
		if (match(serv, tempserv->name))
		{
			/* squit the server */
			toserv("SQUIT %s :Juped: %s (%s)\n",
				tempserv->name, 
				reason, 
				who);

			prev = tempserv->next;
			DeleteServer(tempserv); /* remove server from list */
			tempserv = prev;
		}
		else
			tempserv = tempserv->next;
	}

	/* add a fake server to replace it */
	sprintf(sendstr, ":%s SERVER %s 2 :Jupitered\n",
		Me.name,
		serv);

	toserv(sendstr);

	acnt = SplitBuf(sendstr, &arv);
	AddServer(acnt, arv);

	MyFree(arv);
} /* DoJupeSquit() */

/*
CheckJuped()
	args: char *name
	purpose: check if 'name' is "juped" (ie, banned from the
					 network).  If so, SQUIT/KILL it.
	return: 1 if 'name' matches a jupe, 0 if not
*/

int
CheckJuped(char *name)

{
	Jupe	*tempjupe;
	Server	*tempserv;
	char	sendstr[MAXLINE], **arv;

	for (tempjupe = jupelist; tempjupe; tempjupe = tempjupe->next)
	{
		if (match(tempjupe->name, name))
		{
			if (tempjupe->isnick)
			{
				Luser		*lptr;

				if (!(lptr = FindClient(name)))
					return 0;

				/* its a nick jupe, not a server jupe */

				/* collide the nickname */
				sprintf(sendstr,
					"NICK %s 1 %ld +i juped juped.com %s :Jupitered Nickname\n",
					tempjupe->name,
				#ifdef NICKSERVICES
					(long) (lptr->nick_ts - 1),
				#else
					(long) (lptr->since - 1),
				#endif
					Me.name);
				toserv(sendstr);

				DeleteClient(lptr);

				SplitBuf(sendstr, &arv);
				AddClient(arv);

				if (Me.sptr)
					Me.sptr->numoperkills++;
				Network->TotalOperKills++;
			#ifdef STATSERVICES
				if (Network->TotalOperKills > Network->OperKillsT)
					Network->OperKillsT = Network->TotalOperKills;
			#endif
			}
			else
			{
				toserv("SQUIT %s :Juped: %s (%s)\n",
					name,
					tempjupe->reason,
					tempjupe->who);

				tempserv = FindServer(name);
				DeleteServer(tempserv);

				/* replace it with fake server */
				sprintf(sendstr, ":%s SERVER %s 2 :Jupitered\n",
					Me.name,
					name);
				toserv(sendstr);
				SplitBuf(sendstr, &arv);

				AddServer(5, arv);
			}
			return 1;
		}
	}
	return 0;
} /* CheckJuped */

/*
IsJupe()
	Determine if 'hostname' has a J: line, and return a pointer
to the structure if so
*/

Jupe *
IsJupe(char *hostname)

{
	Jupe	*tempjupe;

	if (!hostname)
		return ((Jupe *) NULL);

	for (tempjupe = jupelist; tempjupe; tempjupe = tempjupe->next)
		if (match(tempjupe->name, hostname))
			return (tempjupe);

	return ((Jupe *) NULL);
} /* IsJupe() */

/*
InitJupes()
	Called when we successfully connect to a hub - introduce juped
servers and nicknames so they cannot connect
*/

void
InitJupes()

{
	Jupe *tmpjupe;
	char sendstr[MAXLINE];
	char **av;

	for (tmpjupe = jupelist; tmpjupe; tmpjupe = tmpjupe->next)
	{
		if (tmpjupe->isnick)
		{
			/* collide the nickname */
			sprintf(sendstr,
				"NICK %s 1 1 +i juped juped.com %s :Jupitered Nickname\n",
				tmpjupe->name,
				Me.name);
			toserv(sendstr);

			SplitBuf(sendstr, &av);
			AddClient(av);
		}
		else
		{
			sprintf(sendstr,
				":%s SERVER %s 2 :Jupitered",
				Me.name,
				tmpjupe->name);
			toserv(":%s SQUIT %s :%s (%s)\n%s\n",
				Me.name,
				tmpjupe->name,
				tmpjupe->reason,
				tmpjupe->who,
				sendstr);

			SplitBuf(sendstr, &av);
			AddServer(5, av);
		}
	}
} /* InitJupes() */

#endif /* ALLOW_JUPES */
