/*
 * HybServ TS Services, Copyright (C) 1998-1999 Patrick Alken
 * This program comes with absolutely NO WARRANTY
 *
 * Should you choose to use and/or modify this source code, please
 * do so under the terms of the GNU General Public License under which
 * this program is distributed.
 */

static char *rcsid = "$Id: list.c,v 1.00 1998/08/05 14:17:15 wnder Exp $";

#include	"headers.h"

/*
 * The first half of this file includes linked list manipulation
 * routines (add/delete items). The second half includes
 * routines that locate items in lists.
 */

/*
AddServer()
	args: int argcnt, char **line
	purpose: add server info contained in 'line'
	return: pointer to newly created server structure
*/

Server *
AddServer(int argcnt, char **line)

{
	Server	*tempserv;

#ifdef BLOCK_ALLOCATION
	tempserv = (Server *) BlockSubAllocate(ServerHeap);
#else
	tempserv = (Server *) MyMalloc(sizeof(Server));
#endif

	tempserv->flags = 0;

	/*
	 * check if the server is the hub - if so, the format will
	 * be SERVER server.com .., not :hub SERVER server.com ..
	 */
	if (argcnt == 4)
	{
	#ifdef BLOCK_ALLOCATION
		strncpy(tempserv->name, line[1], SERVERLEN);
	#else
		tempserv->name = MyStrdup(line[1]);
	#endif

		tempserv->flags |= SERV_MYHUB;
		if (!Me.hub)
			Me.hub = tempserv;

		/*
		 * services won't be in the server list yet, so don't
		 * increment our hub's server count - it will be set
		 * to 1 anyway below
		 */
		tempserv->uplink = (Server *) NULL;
	}
	else
	{
		if (argcnt < 5)
			return (NULL);

	#ifdef BLOCK_ALLOCATION
		strncpy(tempserv->name, line[2], SERVERLEN);
	#else
		tempserv->name = MyStrdup(line[2]);
	#endif

		/*
		 * If the server we're adding is services itself,
		 * have "Me.sptr" point to it
		 */
		if (!Me.sptr)
			if (!strcasecmp(tempserv->name, Me.name))
				Me.sptr = tempserv;

		/*
		 * A new server is being added, increment the server count
		 * of it's hub; if the new server is services itself, don't
		 * increment our hub's server count because it will already
		 * be 1 from when it was added
		 */
		if ((tempserv->uplink = FindServer((line[0][0] == ':') ? line[0] + 1 : line[0])))
		{
			/*
			 * Make sure this server doesn't match services
			 */
			if (tempserv != Me.sptr)
			{
				tempserv->uplink->numservs++;

			#ifdef STATSERVICES
				if (tempserv->uplink->numservs > tempserv->uplink->maxservs)
				{
					tempserv->uplink->maxservs = tempserv->uplink->numservs;
					tempserv->uplink->maxservs_ts = time(NULL);
				}
			#endif
			}
		}

		if (SafeConnect)
			SendUmode(OPERUMODE_L,
				"*** New Server: %s (hub: %s)",
				tempserv->name,
				tempserv->uplink ? tempserv->uplink->name : "*unknown*");
	}

	tempserv->numusers = 0;
	tempserv->numopers = 0;

	tempserv->connect_ts = time(NULL);
	tempserv->numservs = 1;

#ifdef STATSERVICES
	tempserv->maxusers = tempserv->maxopers = 0;
	tempserv->maxusers_ts = tempserv->maxopers_ts = 0;
	tempserv->maxservs = 1;
	tempserv->maxservs_ts = tempserv->connect_ts;

	tempserv->numidentd = 0;
	tempserv->numreshosts = 0;

	tempserv->maxping = tempserv->minping = 0.0;
	tempserv->maxping_ts = tempserv->minping_ts = 0;
	tempserv->ping = 0.0;
	tempserv->lastping_sec = tempserv->lastping_usec = 0;
#endif /* STATSERVICES */

	tempserv->numoperkills = 0;
	tempserv->numservkills = 0;
	tempserv->next = ServerList;
	tempserv->prev = (Server *) NULL;
	if (tempserv->next)
		tempserv->next->prev = tempserv;
	ServerList = tempserv;

	(void) HashAddServer(tempserv);

	Network->TotalServers++;
#ifdef STATSERVICES
	if (Network->TotalServers > Network->MaxServers)
	{
		Network->MaxServers = Network->TotalServers;
		Network->MaxServers_ts = time(NULL);

		if ((Network->MaxServers % 5) == 0)
		{
			/* notify of new max server count */
			SendUmode(OPERUMODE_Y,
				"*** New Max Server Count: %ld",
				Network->MaxServers);
		}
	}
	if (Network->TotalServers > Network->MaxServersT)
	{
		Network->MaxServersT = Network->TotalServers;
		Network->MaxServersT_ts = time(NULL);
	}
#endif

	return (tempserv);
} /* AddServer() */

/*
DeleteServer()
	args: Server *sptr
	purpose: delete 'sptr' from server list
	return: none
*/

void
DeleteServer(Server *sptr)

{
	Server		*tmpserv;

	if (!sptr)
		return;

	/*
	 * Delete server from hash table
	 */
	(void) HashDelServer(sptr);

	if (sptr->uplink)
		sptr->uplink->numservs--;

	SendUmode(OPERUMODE_L,
		"*** Netsplit: %s (hub: %s)",
		sptr->name,
		sptr->uplink ? sptr->uplink->name : "*unknown*");

	/*
	 * go through server list, and check if any leafs had sptr
	 * as their hub, if so, nullify uplink
	 */
	for (tmpserv = ServerList; tmpserv; tmpserv = tmpserv->next)
	{
		if (tmpserv->uplink == sptr)
			tmpserv->uplink = (Server *) NULL;
	}

#ifndef BLOCK_ALLOCATION
	MyFree(sptr->name);
#endif

	if (sptr->prev)
		sptr->prev->next = sptr->next;
	else
		ServerList = sptr->next;

	if (sptr->next)
		sptr->next->prev = sptr->prev;

#ifdef BLOCK_ALLOCATION

	BlockSubFree(ServerHeap, sptr);

#else

	MyFree(sptr);

#endif /* BLOCK_ALLOCATION */

	Network->TotalServers--;
} /* DeleteServer() */

/*
AddHostLimit()
	Add 'host' to restricted hosts list
*/

void
AddHostLimit(char *host, int hostnum)

{
	rHost *ptr;
	char *tmp;

	ptr = (rHost *) MyMalloc(sizeof(rHost));

	if ((tmp = strchr(host, '@')))
	{
		*tmp++ = '\0';
		ptr->username = MyStrdup(host);
		ptr->hostname = MyStrdup(tmp);
	}
	else
	{
		ptr->username = MyStrdup("*");
		ptr->hostname = MyStrdup(host);
	}

	ptr->hostnum = hostnum;
	ptr->next = rhostlist;
	rhostlist = ptr;
} /* AddHostLimit() */

/*
AddUser()
	Add 'host' to O: line userlist with 'pass' and 'flags'
*/

void
AddUser(char *host, char *pass, char *nick, char *flags)

{
	Userlist *ptr;
	int ii;
	char *tmp;

	ptr = (Userlist *) MyMalloc(sizeof(Userlist));
	ptr->nick = MyStrdup(nick);

	if ((tmp = strchr(host, '@')))
	{
		*tmp++ = '\0';
		ptr->username = MyStrdup(host);
		ptr->hostname = MyStrdup(tmp);
	}
	else
	{
		ptr->username = MyStrdup("*");
		ptr->hostname = MyStrdup(host);
	}

	ptr->password = MyStrdup(pass);
	ptr->umodes = OPERUMODE_INIT;

	ptr->flags = 0;
	for (ii = 0; ii < strlen(flags); ii++)
	{
		switch (flags[ii])
		{
			case 's':
			case 'S':
			{
				ptr->flags |= (PRIV_SADMIN | PRIV_ADMIN | PRIV_OPER | PRIV_FRIEND | PRIV_CHAT | PRIV_EXCEPTION);
				break;
			}
			case 'a':
			case 'A':
			{
				ptr->flags |= (PRIV_ADMIN | PRIV_OPER | PRIV_FRIEND | PRIV_CHAT);
				break;
			}
			case 'o':
			case 'O':
			{
				ptr->flags |= (PRIV_OPER | PRIV_FRIEND);
				break;
			}
			case 'j':
			case 'J':
			{
				ptr->flags |= PRIV_JUPE;
				break;
			}
			case 'g':
			case 'G':
			{
				ptr->flags |= PRIV_GLINE;
				break;
			}
			case 'd':
			case 'D':
			{
				ptr->flags |= PRIV_CHAT;
				break;
			}
			case 'e':
			case 'E':
			{
				ptr->flags |= PRIV_EXCEPTION;
				break;
			}
			case 'f':
			case 'F':
			{
				ptr->flags |= PRIV_FRIEND;
				break;
			}
			default:
				break;
		} /* switch (flags[ii]) */
	}

#ifdef RECORD_SPLIT_TS
	ptr->split_ts = ptr->whensplit = 0;
#endif

	ptr->next = userlist;
	userlist = ptr;
} /* AddUser() */

/*
AddServ()
	Add 'hostname', 'password', and 'port' to list containing S: line
info
*/

void
AddServ(char *hostname, char *password, int port)

{
	Servlist	*ptr;

	ptr = (Servlist *) MyMalloc(sizeof(Servlist));
	ptr->hostname = MyStrdup(hostname);
	ptr->password = MyStrdup(password);
	ptr->realname = (char *) NULL;
	ptr->connect_ts = 0;
	ptr->port = port;
	ptr->flags = 0;

	ptr->next = serverlist;
	serverlist = ptr;
} /* AddServ() */

/*
AddMyChan()
	Add 'channel' to list of C: line channels
*/

void
AddMyChan(char *channel)

{
	Chanlist	*ptr;

	ptr = (Chanlist *) MyMalloc(sizeof(Chanlist));
	ptr->name = MyStrdup(channel);
	ptr->flags = 0;
	ptr->next = mychanlist;
	mychanlist = ptr;

	Network->MyChans++;
} /* AddMyChan() */

/*
AddBot()
	Add new bot info to list of B: line bots
*/

void
AddBot(char *nick, char *host, char *pass, int port)

{
	Botlist	*ptr;
	char *tmp;

	ptr = (Botlist *) MyMalloc(sizeof(Botlist));
	ptr->name = MyStrdup(nick);

	if ((tmp = strchr(host, '@')))
	{
		*tmp++ = '\0';
		ptr->username = MyStrdup(host);
		ptr->hostname = MyStrdup(tmp);
	}
	else
	{
		ptr->username = MyStrdup("*");
		ptr->hostname = MyStrdup(host);
	}

	ptr->password = MyStrdup(pass);
	ptr->port = port;
	ptr->next = botlist;
	botlist = ptr;
} /* AddBot() */

/*
AddRemoteBot()
	Add new bot info to list of L: line bots
*/

void
AddRemoteBot(char *nick, char *host, char *pass)

{
	Botlist *ptr;
	char *tmp;

	ptr = (Botlist *) MyMalloc(sizeof(Botlist));
	ptr->name = MyStrdup(nick);

	if ((tmp = strchr(host, '@')))
	{
		*tmp++ = '\0';
		ptr->username = MyStrdup(host);
		ptr->hostname = MyStrdup(tmp);
	}
	else
	{
		ptr->username = MyStrdup("*");
		ptr->hostname = MyStrdup(host);
	}

	ptr->password = MyStrdup(pass);
	ptr->port = 0; /* we don't need a port for remote bot lines */
	ptr->next = remotebots;
	remotebots = ptr;
} /* AddRemoteBot() */

/*
AddPort()
	Add port info
*/

void
AddPort(int port, char *host, char *type)

{
	PortInfo	*ptr;

	if ((ptr = IsPort(port)))
	{
		/*
		 * The port is already in the list, so this routine was probably
		 * called from Rehash() - just change the hostmask and type
		 * in case they were changed for the rehash
		 */
		if (ptr->host)
			MyFree(ptr->host);
	}
	else
	{
		ptr = (PortInfo *) MyMalloc(sizeof(PortInfo));
		ptr->port = port;
		ptr->socket = NOSOCKET;
		ptr->tries = 0;
		ptr->next = portlist;
		portlist = ptr;
	}

	if (host)
	{
		if (strlen(host))
			ptr->host = MyStrdup(host);
		else
			ptr->host = (char *) NULL;
	}
	else
		ptr->host = (char *) NULL;

	if (!type)
		ptr->type = PRT_USERS;
	else
	{
		ptr->type = 0;

		if (!strcasecmp(type, "TCM"))
			ptr->type = PRT_TCM;
		if (!strcasecmp(type, "USERS"))
			ptr->type = PRT_USERS;

		if (!ptr->type)
			ptr->type = PRT_USERS;
	}
} /* AddPort() */

/*
ClearConfLines()
 Called from Rehash() to clear old conf list
*/

void
ClearConfLines()

{
#ifdef ALLOW_JUPES
	Jupe		*tempjupe;
#endif
#ifdef ALLOW_GLINES
	Gline *tempgline;
#endif
	Userlist	*tempuser;
/*	Servlist	*tempserv; */
	Botlist	*tempbot;
	rHost		*temphost;

#ifdef ALLOW_JUPES
	while (jupelist)
	{
		tempjupe = jupelist->next;
		DeleteJupe(jupelist);
		jupelist = tempjupe;
	}
#endif

#ifdef ALLOW_GLINES
	while (glinelist)
	{
		tempgline = glinelist->next;
		DeleteGline(glinelist);
		glinelist = tempgline;
	}
#endif

	while (userlist)
	{
		tempuser = userlist->next;
		MyFree(userlist->nick);
		MyFree(userlist->username);
		MyFree(userlist->hostname);
		MyFree(userlist->password);
		MyFree(userlist);
		userlist = tempuser;
	}

/*
 * Don't clear the serverlist during a rehash because currenthub
 * will point to garbage
	while (serverlist)
	{
		MyFree(serverlist->hostname);
		MyFree(serverlist->password);
		if (serverlist->realname)
			MyFree(serverlist->realname);
		tempserv = serverlist->next;
		MyFree(serverlist);
		serverlist = tempserv;
	}
*/

	while (botlist)
	{
		MyFree(botlist->name);
		MyFree(botlist->username);
		MyFree(botlist->hostname);
		MyFree(botlist->password);
		tempbot = botlist->next;
		MyFree(botlist);
		botlist = tempbot;
	}

	while (remotebots)
	{
		MyFree(remotebots->name);
		MyFree(remotebots->username);
		MyFree(remotebots->hostname);
		MyFree(remotebots->password);
		tempbot = remotebots->next;
		MyFree(remotebots);
		remotebots = tempbot;
	}

	while (rhostlist)
	{
		MyFree(rhostlist->username);
		MyFree(rhostlist->hostname);
		temphost = rhostlist->next;
		MyFree(rhostlist);
		rhostlist = temphost;
	}
} /* ClearConfLines() */

/*
 * List item locating routines
 */

/*
IsChannel()
	args: char *chan
	purpose: determine if 'chan' is currently being monitored
	return: pointer to channel in monitored channel list
*/

Chanlist *
IsChannel(char *chan)

{
	Chanlist	*tempchan;

	for (tempchan = mychanlist; tempchan; tempchan = tempchan->next)
		if (!strcasecmp(tempchan->name, chan))
			return(tempchan);

	return ((Chanlist *) NULL);
} /* IsChannel() */

/*
IsServLine()
	args: char *name, int port
	purpose: determine if there is an S: line for 'name' and 'port'
	return: pointer to server struct in server list
*/

Servlist *
IsServLine(char *name, char *port)

{
	Servlist	*tempserv;
	int		portnum;

	if (!name)
		return ((Servlist *) NULL);

	if (!port)
		portnum = DefaultHubPort;
	else
		portnum = atoi(port);

	for (tempserv = serverlist; tempserv; tempserv = tempserv->next)
		if ((tempserv->port == portnum) &&
				(!strcasecmp(tempserv->hostname, name)))
			return (tempserv);

	return ((Servlist *) NULL);
} /* IsServLine() */

/*
IsListening()
	Determine if we are currently accepting connections on 'portnum'
*/

PortInfo *
IsListening(int portnum)

{
	PortInfo	*pptr;

	for (pptr = portlist; pptr; pptr = pptr->next)
		if ((pptr->port == portnum) &&
				(pptr->socket != NOSOCKET))
			return (pptr);

	return ((PortInfo *) NULL);
} /* IsListening() */

/*
IsPort()
	Determine if 'portnum' has a P: line in hybserv.conf
*/

PortInfo *
IsPort(int portnum)

{
	PortInfo	*pptr;

	for (pptr = portlist; pptr; pptr = pptr->next)
		if (pptr->port == portnum)
			return (pptr);

	return ((PortInfo *) NULL);
} /* IsPort() */

/*
IsBot()
	args: char *bname
	purpose: determine if 'bname' has a B: line 
	return: pointer to bot structure
*/

Botlist *
IsBot(char *bname)

{
	Botlist *tempbot;

	if (!bname)
		return ((Botlist *) NULL);

	tempbot = botlist;
	while (tempbot && (strcasecmp(tempbot->name, bname) != 0))
		tempbot = tempbot->next;

	return (tempbot);
} /* IsBot() */
