/*
 * HybServ TS Services, Copyright (C) 1998-1999 Patrick Alken
 * This program comes with absolutely NO WARRANTY
 *
 * Should you choose to use and/or modify this source code, please
 * do so under the terms of the GNU General Public License under which
 * this program is distributed.
 */

static char *rcsid = "$Id: log.c,v 1.00 1999/08/16 20:08:04 wnder Exp $";

#include "headers.h"
#include <dirent.h>

/*
putlog()
	args: int level, char *fmt, va_alist
	purpose: log 'fmt' to log file
	return: none
*/

void
putlog(int level, char *format, ...)

{
	FILE *fp;
	time_t CurrTime;
	char buf[MAXLINE];
	va_list args;

	/*
	 * Do not log anything if LogLevel equals 0 - also,
	 * if the given level is greater than LogLevel, do not
	 * log it
	 */
	if ((LogLevel == 0) || (LogLevel < level))
		return;

	if (!LogFile)
		return;

	if ((fp = fopen(LogFile, "a+")) == (FILE *) NULL)
	{
	#ifdef DEBUGMODE
		printf("Unable to open log file: %s\n", LogFile);
	#endif
		return;
	}

	CurrTime = time(NULL);
	strcpy(buf, ctime(&CurrTime));

	/*
	 * Erase the \n character that ctime() puts on the end, in order
	 * to concat the string we want to log
	 */
	buf[strlen(buf) - 1] = ' ';

	fprintf(fp, "%s", buf);

	va_start(args, format);

	/* write the string to the log file */
	vfprintf(fp, format, args);
	fprintf(fp, "\n");

	#ifdef DEBUGMODE
		fprintf(stderr, "Logged> ");
		vfprintf(stderr, format, args);
		fprintf(stderr, "\n");
	#endif

	fclose(fp);

	va_end(args);
} /* putlog() */

/*
CheckLogs()
 Called at midnight each day to rename the current log file
to: LogFile.YYYYMMDD, and also check if we have more than
MaxLogs log files. If so, erase the oldest one, to make
room for the newest one.
*/

void
CheckLogs(time_t unixtime)

{
	char tmplog[MAXLINE],
	     olddate[MAXLINE];
	char *currdate;
	struct tm *log_tm;
	time_t oldts;
	struct dirent **loglist;
	int lcnt,
	    lmatches,
	    ii,
	    len;

	if (MaxLogs)
	{
		/*
		 * We must now check if there are MaxLogs log files
		 * in HPath/. If so, delete the oldest one to make
		 * room for the current one.
		 */

		lcnt = scandir(HPath, &loglist, 0, alphasort);

		if (lcnt >= 0)
		{
			sprintf(tmplog, "%s.",
				LogFile);
			len = strlen(tmplog);

			lmatches = 0;
			olddate[0] = '\0';
			currdate = (char *) NULL;

			/*
			 * Go through all the files in the directory and
			 * pick out the ones that match "LogFile."
			 */
			for (ii = 0; ii < lcnt; ii++)
			{
				if (!strncasecmp(loglist[ii]->d_name, tmplog, len))
				{
					++lmatches;

					/*
					 * Now check the date on the log file to see
					 * if its the oldest.
					 */
					if (!*olddate)
						strcpy(olddate, loglist[ii]->d_name + len);
					else
					{
						/*
						 * This really isn't needed because
						 * the alphasort routine would have made
						 * sure the oldest file came first, but
						 * just in case..
						 */
						currdate = loglist[ii]->d_name + len;
						if (atol(olddate) > atol(currdate))
							strcpy(olddate, currdate);
					}
				}
			}

			if ((lmatches >= MaxLogs) && *olddate)
			{
				/*
				 * There are too many log files in the directory,
				 * delete the oldest one - it will be: LogFile.olddate
				 */
				sprintf(tmplog, "%s/%s.%s",
					HPath,
					LogFile,
					olddate);
				unlink(tmplog);
			}
		} /* if (lcnt >= 0) */
		else
		{
			putlog(LOG1, "Error reading log directory: %s",
				strerror(errno));
		}
	} /* if (MaxLogs) */

	/*
	 * Now rename the current log file. Use the TS of one
	 * second ago since this log file is actually yesterday's,
	 * because it just turned midnight.
	 */

	oldts = unixtime - 1;
	log_tm = localtime(&oldts);
	sprintf(tmplog, "%s.%d%02d%02d",
		LogFile,
		1900 + log_tm->tm_year,
		log_tm->tm_mon + 1,
		log_tm->tm_mday);

	rename(LogFile, tmplog);
} /* CheckLogs() */
