/*
 * HybServ TS Services, Copyright (C) 1998-1999 Patrick Alken
 * This program comes with absolutely NO WARRANTY
 *
 * Should you choose to use and/or modify this source code, please
 * do so under the terms of the GNU General Public License under which
 * this program is distributed.
 */

static char *rcsid = "$Id: memoserv.c,v 1.00 1998/11/12 18:59:58 wnder Exp $";

#include	"headers.h"

#if defined(NICKSERVICES) && defined(MEMOSERVICES)

/* hash table of memos */
MemoInfo *memolist[MEMOLIST_MAX];

static void AddMemoList(MemoInfo *);
static void AddMemo(MemoInfo *, Memo *);

static void m_send(Luser *, NickInfo *, int, char **);
static void m_list(Luser *, NickInfo *, int, char **);
static void m_read(Luser *, NickInfo *, int, char **);
static void m_help(Luser *, NickInfo *, int, char **);
static void m_del(Luser *, NickInfo *, int, char **);
static void m_undel(Luser *, NickInfo *, int, char **);
static void m_forward(Luser *, NickInfo *, int, char **);
static void m_reply(Luser *, NickInfo *, int, char **);
static void m_purge(Luser *, NickInfo *, int, char **);

static Command	memocmds[] = {
	{ "SEND", m_send, LVL_IDENT },
	{ "LIST", m_list, LVL_IDENT },
	{ "READ", m_read, LVL_IDENT },
	{ "HELP", m_help, LVL_NONE },
	{ "DEL", m_del, LVL_IDENT },
	{ "UNDEL", m_undel, LVL_IDENT },
	{ "FORWARD", m_forward, LVL_IDENT },
	{ "REPLY", m_reply, LVL_IDENT },
	{ "PURGE", m_purge, LVL_IDENT },

	{ (char *) 0, (void (*)()) 0, 0 }
};

/*
ms_process()
  Process command coming from 'nick' directed towards n_MemoServ
*/

void
ms_process(char *nick, char *command)

{
	int		acnt;
	char	**arv;
	Command	*mptr;
	Luser	*lptr;
	NickInfo	*nptr, *master;

	if (!command || !(lptr = FindClient(nick)))
		return;

	if (Network->flags & NET_OFF)
	{
		notice(n_MemoServ, lptr->nick,
			"Services are currently \002disabled\002");
		return;
	}

	acnt = SplitBuf(command, &arv);
	if (acnt == 0)
	{
		MyFree(arv);
		return;
	}

	mptr = GetCommand(memocmds, arv[0]);

	if (!mptr || (mptr == (Command *) -1))
	{
		notice(n_MemoServ, lptr->nick,
			"%s command [%s]",
			(mptr == (Command *) -1) ? "Ambiguous" : "Unknown",
			arv[0]);
		MyFree(arv);
		return;
	}

	/*
	 * Check if the command is for admins only - if so,
	 * check if they match an admin O: line.  If they do,
	 * check if they are EITHER oper'd, or registered with OperServ,
	 * if either of these is true, allow the command
	 */
	if ((mptr->level == LVL_ADMIN) && !(IsValidAdmin(lptr)))
	{
		notice(n_MemoServ, lptr->nick, "Unknown command [%s]",
			arv[0]);
		MyFree(arv);
		return;
	}

	nptr = FindNick(lptr->nick);
	master = GetMaster(nptr);

	if (!nptr && !master && (mptr->level != LVL_NONE))
	{
		/* the command requires a registered nickname */

		notice(n_MemoServ, lptr->nick,
			"Your nickname is not registered");
		notice(n_MemoServ, lptr->nick,
			ERR_MORE_INFO,
			n_NickServ,
			"REGISTER");
		MyFree(arv);
		return;
	}

	if (nptr)
	{
		if (nptr->flags & NS_FORBID)
		{
			notice(n_MemoServ, lptr->nick,
				"Cannot execute commands for forbidden nicknames");
			MyFree(arv);
			return;
		}

		if (mptr->level != LVL_NONE)
		{
			if (!(nptr->flags & NS_IDENTIFIED))
			{
				notice(n_MemoServ, lptr->nick,
					"Password identification is required for [\002%s\002]",
					mptr->cmd);
				notice(n_MemoServ, lptr->nick, 
					"Type \002/msg %s IDENTIFY <password>\002 and retry",
					n_NickServ);
				MyFree(arv);
				return;
			}
		}
	} /* if (nptr) */

	/* call mptr->func to execute command */
	(*mptr->func)(lptr, master, acnt, arv);

	MyFree(arv);

	return;
} /* ms_process() */

/*
ms_loaddata()
	Load MemoServ database - return 1 if successful, -1 if not, and -2
if errors are unrecoverable
*/

int
ms_loaddata()

{
	FILE	*fp;
	char	line[MAXLINE], **av;
	char	*keyword;
	int		ac, ret = 1, cnt;
	MemoInfo	*mi = (MemoInfo *) NULL;
	NickInfo	*nickptr;

	if (!(fp = fopen(MemoServDB, "r")))
	{
		/* MemoServ data file doesn't exist */
		return -1;
	}

	cnt = 0;
	/* load data into list */
	while (fgets(line, MAXLINE - 1, fp))
	{
		cnt++;
		ac = SplitBuf(line, &av);
		if (!ac)
		{
			/* probably a blank line */
			MyFree(av);
			continue;
		}

		if (av[0][0] == ';')
		{
			/* its a comment */
			MyFree(av);
			continue;
		}

		if (!strncasecmp("->", av[0], 2))
		{
			/* 
			 * check if there are enough args
			 */
			if (ac < 5)
			{
				fatal(1, "%s:%d Invalid database format (FATAL)",
					MemoServDB,
					cnt);
				ret = -2;
				MyFree(av);
				continue;
			}

			/* check if there is no nickname before it */
			if (!mi)
			{
				fatal(1, "%s:%d No nickname associated with data",
					MemoServDB,
					cnt);
				if (ret > 0)
					ret = -1;
				MyFree(av);
				continue;
			}

			keyword = av[0] + 2;
			if (!strncasecmp(keyword, "TEXT", 4))
			{
				Memo	*memoptr;

				memoptr = (Memo *) MyMalloc(sizeof(Memo));
				memoptr->sender = MyStrdup(av[1]);
				memoptr->sent = atol(av[2]);
				memoptr->flags = atol(av[3]);
				if (!(memoptr->flags & MS_READ))
					mi->newmemos++;
				memoptr->text = MyStrdup(av[4] + 1);
				memoptr->index = ++mi->memocnt;
				AddMemo(mi, memoptr);
			}

		} /* if (!strncasecmp("->", keyword, 2)) */
		else
		{
			if (mi)
			{
				if (!mi->memos)
				{
					fatal(1, "%s:%d No memos for entry [%s] (skipping)",
						MemoServDB,
						cnt,
						mi->name);
					MyFree(mi->name);
					MyFree(mi);
					mi = (MemoInfo *) NULL;
					if (ret > 0)
						ret = -1;
				}
				else
					AddMemoList(mi);
			}

			/*
			 * make sure there are enough args on the line:
			 * <nickname>
			 */
			if (ac < 1)
			{
				fatal(1, "%s:%d Invalid database format (FATAL)",
					MemoServDB,
					cnt);
				ret = -2;
				mi = (MemoInfo *) NULL;
				MyFree(av);
				continue;
			}

			if (!(nickptr = FindNick(av[0])))
			{
				fatal(1, "%s:%d Memo entry [%s] is not a registered nickname (skipping)",
					MemoServDB,
					cnt,
					av[0]);
				if (ret > 0)
					ret = -1;
				mi = (MemoInfo *) NULL;
				MyFree(av);
				continue;
			}

		#ifdef LINKED_NICKNAMES
			if (nickptr->master)
			{
				/*
				 * nickptr is a leaf nickname - they should not have
				 * memo entries
				 */
				fatal(1, "%s:%d Memo entry [%s] is not a master nickname (skipping)",
					MemoServDB,
					cnt,
					av[0]);
				if (ret > 0)
					ret = (-1);
				mi = (MemoInfo *) NULL;
				MyFree(av);
				continue;
			}
		#endif /* LINKED_NICKNAMES */

			mi = (MemoInfo *) MyMalloc(sizeof(MemoInfo));
			mi->name = MyStrdup(av[0]);
			mi->newmemos = 0;
			mi->memocnt = 0;
			mi->memos = mi->lastmemo = (Memo *) NULL;
			mi->flags = 0;
		}

		MyFree(av);
	} /* while (fgets(line, MAXLINE - 1, fp)) */

	if (mi)
	{
		if (!mi->memos)
		{
			fatal(1, "%s:%d No memos for entry [%s] (skipping)",
				MemoServDB,
				cnt,
				mi->name);
			MyFree(mi->name);
			MyFree(mi);
			if (ret > 0)
				ret = -1;
		}
		else
			AddMemoList(mi);
	}

	fclose (fp);
	return (ret);
} /* ms_loaddata() */

/*
DelMemo()
	Remove 'mptr' from memolist[]
*/

void
DelMemo(MemoInfo *mptr)

{
	Memo *memoptr, *next;
	unsigned int hashv;

	if (!mptr)
		return;

	hashv = MSHashMemo(mptr->name);

	/* kill all their memos (if there are any) */
	for (memoptr = mptr->memos; memoptr; )
	{
		next = memoptr->next;
		MyFree(memoptr->sender);
		MyFree(memoptr->text);
		MyFree(memoptr);
		memoptr = next;
	}
	mptr->memos = (Memo *) NULL;

	if (!mptr->prev)
	{
		if (!mptr->next)
			memolist[hashv] = (MemoInfo *) NULL;
		else
		{
			mptr->next->prev = (MemoInfo *) NULL;
			memolist[hashv] = mptr->next;
		}
	}
	else
	{
		if (!mptr->next)
			mptr->prev->next = (MemoInfo *) NULL;
		else
		{
			mptr->next->prev = mptr->prev;
			mptr->prev->next = mptr->next;
		}
	}

	MyFree(mptr->name);
	MyFree(mptr);
	mptr = (MemoInfo *) NULL;

	Network->TotalMemos--;
} /* DelMemo() */

/*
PurgeMemos()
	Go through mptr->memos and remove any marked for deletion
*/

int
PurgeMemos(MemoInfo *mptr)

{
	int		ret = 0;
	Memo	*memoptr, *prev;

	if (!mptr)
		return 0;

	if (!mptr->memos)
		return 0;

	prev = mptr->memos;
	for (memoptr = mptr->memos->next; memoptr; )
	{
		if (memoptr->flags & MS_DELETE)
		{
			ret++;
			if (!memoptr->next)
				mptr->lastmemo = prev;
			prev->next = memoptr->next;
			MyFree(memoptr->sender);
			MyFree(memoptr->text);
			MyFree(memoptr);
			memoptr = prev->next;
		}
		else
		{
			prev = memoptr;
			memoptr = memoptr->next;
		}
	}

	if (mptr->memos)
	{
		if (mptr->memos->flags & MS_DELETE)
		{
			ret++;
			prev = mptr->memos->next;
			if (!prev)
			{
				/* its the only entry in the list - update lastmemo */
				mptr->lastmemo = prev;
			}
			MyFree(mptr->memos->sender);
			MyFree(mptr->memos->text);
			MyFree(mptr->memos);
			mptr->memos = prev;
		}
	}

	mptr->memocnt -= ret;

	if (!mptr->memocnt)
	{
		/*
		 * mptr has no more memos - delete it from the memolist[]
		 * array
		 */
		DelMemo(mptr);
		return (ret);
	}
	else
		if (ret)
		{
			int		cnt;

			/*
			 * its possible that the indices of the memos are now
			 * messed up - if you have 3 memos, and you delete the 2nd one,
			 * you'll have 2 memos, one with an index of 1, and the other
			 * with an index of 3 - check if consecutive indices are
			 * one higher than the previous - if not, fix it
			 */

			cnt = 0;
			for (memoptr = mptr->memos; memoptr; memoptr = memoptr->next)
			{
				if (memoptr->index != (cnt + 1))
					memoptr->index = cnt + 1;
				cnt++;
			}
		}

	return (ret);
} /* PurgeMemos() */

/*
ExpireMemos()
 Delete any memos that have expired
*/

void
ExpireMemos(time_t unixtime)

{
	int ii;
	MemoInfo *mptr, *mtemp;
	Memo *memoptr;

	if (!MemoExpire)
		return;

	for (ii = 0; ii < MEMOLIST_MAX; ii++)
	{
		for (mptr = memolist[ii]; mptr; )
		{
			for (memoptr = mptr->memos; memoptr; memoptr = memoptr->next)
				if ((unixtime - memoptr->sent) >= MemoExpire)
					memoptr->flags |= MS_DELETE;

			mtemp = mptr->next;
			PurgeMemos(mptr);
			mptr = mtemp;
		}
	}
} /* ExpireMemos() */

/*
PurgeCheck()
 Go through all memos and delete any that have been
marked for deletion - should be called every "MemoPurgeFreq"
seconds from DoTimer()
*/

void
PurgeCheck()

{
	int		ii;
	MemoInfo	*mi, *next;

	for (ii = 0; ii < MEMOLIST_MAX; ii++)
	{
		for (mi = memolist[ii]; mi; )
		{
			/*
			 * Record the next ptr, in case mi has every memo
			 * marked for deletion, in which case the struct itself
			 * will be removed from the list
			 */
			next = mi->next;
			PurgeMemos(mi);
			mi = next;
		}
	}
} /* PurgeCheck() */

/*
AddMemoList()
	Insert 'mptr' into memolist[]
*/

static void
AddMemoList(MemoInfo *mptr)

{
	unsigned int hashv;

	hashv = MSHashMemo(mptr->name);

	mptr->next = memolist[hashv];
	mptr->prev = (MemoInfo *) NULL;
	if (mptr->next)
		mptr->next->prev = mptr;
	memolist[hashv] = mptr;

	Network->TotalMemos++;
} /* AddMemoList() */

/*
AddMemo()
	Insert 'memoptr' into mptr->memos
*/

static void
AddMemo(MemoInfo *mptr, Memo *memoptr)

{
	/*
	 * Add memoptr onto the end of mptr->memos because we want
	 * the most recent memos last for the LIST command
	 */
	memoptr->next = (Memo *) NULL;
	if (!mptr->memos)
	{
		mptr->memos = mptr->lastmemo = memoptr;
	}
	else
	{
		mptr->lastmemo->next = memoptr;
		mptr->lastmemo = mptr->lastmemo->next;
	}
} /* AddMemo() */

/*
FindMemoList
	Return a pointer to structure containing memolist for 'name'
*/

MemoInfo *
FindMemoList(char *name)

{
	MemoInfo *mi;
	unsigned int hashv;

	if (!name)
		return ((MemoInfo *) NULL);

	hashv = MSHashMemo(name);
	for (mi = memolist[hashv]; mi; mi = mi->next)
	{
		if (!strcasecmp(mi->name, name))
			return (mi);
	}

	return ((MemoInfo *) NULL);
} /* FindMemoList() */

/*
sendmemo()
	Record 'text' as a memo for 'target' - assume target is a valid
nickname or channel; return new memo index
*/

int
sendmemo(char *target, char *text, Luser *lptr)

{
	MemoInfo	*mi;
	Memo		*memoptr;

	if (!target || !text || !lptr)
		return (0);

	if (!(mi = FindMemoList(target)))
	{
		mi = (MemoInfo *) MyMalloc(sizeof(MemoInfo));
		mi->name = MyStrdup(target);
		mi->memocnt = 0;
		mi->newmemos = 0;
		mi->memos = mi->lastmemo = (Memo *) NULL;
		mi->flags = 0;
		AddMemoList(mi);
	}

	if (MaxMemos && (mi->memocnt >= MaxMemos))
	{
		notice(n_MemoServ, lptr->nick,
			"%s has reached the maximum memo limit, and cannot receive more",
			target);
		return (0);
	}

	mi->memocnt++;
	mi->newmemos++;

	memoptr = (Memo *) MyMalloc(sizeof(Memo));
	memoptr->next = (Memo *) NULL;
	memoptr->sender = MyStrdup(lptr->nick);
	memoptr->sent = time(NULL);
	memoptr->index = mi->memocnt;
	memoptr->flags = 0;
	memoptr->text = MyStrdup(text);
	AddMemo(mi, memoptr);

	return (memoptr->index);
} /* sendmemo() */

/*
MoveMemos()
  Copy all of source's memos to dest's memo list, and delete
source's memos when done.
*/

void
MoveMemos(NickInfo *source, NickInfo *dest)

{
	MemoInfo	*srcmptr, /* pointer to source's memos */
						*dstmptr; /* pointer to dest's memos */
	Memo	*memoptr, *mtmp;

	if (!source || !dest)
		return;

	if (!(srcmptr = FindMemoList(source->nick)))
		return; /* source has no memos to copy */

	if (!(dstmptr = FindMemoList(dest->nick)))
	{
		/*
		 * Create a memo structure for dest
		 */
		dstmptr = (MemoInfo *) MyMalloc(sizeof(MemoInfo));
		dstmptr->name = MyStrdup(dest->nick);
		dstmptr->memocnt = 0;
		dstmptr->newmemos = 0;
		dstmptr->memos = dstmptr->lastmemo = (Memo *) NULL;
		dstmptr->flags = 0;
		AddMemoList(dstmptr);
	}

	/*
	 * Go through source's memo list and create duplicate
	 * structures in dest's list
	 */
	for (memoptr = srcmptr->memos; memoptr; memoptr = memoptr->next)
	{
		mtmp = (Memo *) MyMalloc(sizeof(Memo));
		mtmp->next = (Memo *) NULL;
		mtmp->sender = MyStrdup(memoptr->sender);
		mtmp->sent = memoptr->sent;
		mtmp->index = ++dstmptr->memocnt;
		mtmp->flags = memoptr->flags;
		mtmp->text = MyStrdup(memoptr->text);

		/*
		 * Now add mtmp, which is a duplicate of memoptr, to dest's
		 * memo list
		 */
		AddMemo(dstmptr, mtmp);

		if (!(mtmp->flags & MS_READ))
			dstmptr->newmemos++;

		/*
		 * Mark all of source's memos for deletion
		 */
		memoptr->flags |= MS_DELETE;
	}

	/*
	 * Purge source's memos
	 */
	PurgeMemos(srcmptr);
} /* MoveMemos() */

/*
m_send()
	Send memo av[2]- to av[1]
*/

static void
m_send(Luser *lptr, NickInfo *nptr, int ac, char **av)

{
	char *memotext, /* memo text */
	     *to; /* who the memo is sent to */
	int	index;
	NickInfo *ni = (NickInfo *) NULL;
#ifdef CHANNELSERVICES
	ChanInfo *ci = (ChanInfo *) NULL;
#endif

	if (ac < 3)
	{
		notice(n_MemoServ, lptr->nick,
			"Syntax: SEND <nick/channel> <text>");
		notice(n_MemoServ, lptr->nick,
			ERR_MORE_INFO,
			n_MemoServ,
			"SEND");
		return;
	}

	if (av[1][0] == '#')
	{
		#ifndef CHANNELSERVICES
			notice(n_MemoServ, lptr->nick,
				"Channel services are disabled");
			return;
		#else
			if (!(ci = FindChan(av[1])))
			{
				notice(n_MemoServ, lptr->nick,
					ERR_CH_NOT_REGGED,
					av[1]);
				return;
			}

			if (!HasAccess(ci, lptr, CA_AUTOOP))
			{
				notice(n_MemoServ, lptr->nick,
					"AutoOp access is required to send a memo to [\002%s\002]",
					ci->name);
				return;
			}

			to = ci->name;

		#endif
	}
	else
	{
		/* it was sent to a nickname */

		if (!(ni = GetLink(av[1])))
		{
			notice(n_MemoServ, lptr->nick,
				ERR_NOT_REGGED,
				av[1]);
			return;
		}

		if (!(ni->flags & NS_MEMOS))
		{
			notice(n_MemoServ, lptr->nick,
				"[\002%s\002] is rejecting all memos",
				av[1]);
			return;
		}

		to = ni->nick;
	}

	/* get the actual memo text now */
	if (ac < 3)
		memotext = MyStrdup("");
	else
		memotext = GetString(ac - 2, av + 2);

	index = sendmemo(to, memotext, lptr);

	if (index)
	{
		notice(n_MemoServ, lptr->nick,
			"Memo has been recorded for [\002%s\002]",
			to);

		if (ni)
		{
			/*
			 * It was sent to a nickname - check if they are online
			 * and notify them
			 */
			if ((ni->flags & NS_MEMONOTIFY) && FindClient(to))
			{
				notice(n_MemoServ, to,
					"You have a new memo from \002%s\002 (#%d)",
					lptr->nick,
					index);
				notice(n_MemoServ, to,
					"Type \002/msg %s READ %d\002 to read it",
					n_MemoServ,
					index);
			}
		} /* if (ni) */
	#ifdef CHANNELSERVICES
		else
		{
			Channel		*chptr;
			ChanAccess	*ca;
			char		*newtext;

			/*
			 * It was sent to a channel - notify every AOP or higher
			 */
			if ((ci) && (chptr = FindChannel(to)))
			{
				ChanUser	*cu;
				NickInfo	*tmpn;

				for (cu = chptr->firstuser; cu; cu = cu->next)
				{
					if (FindService(cu->lptr))
						continue;

					tmpn = GetLink(cu->lptr->nick);

					if (HasAccess(ci, cu->lptr, CA_AUTOOP))
					{
						if (tmpn)
						{
							/* don't notify people who don't want memos */
							if (!(tmpn->flags & NS_MEMOS) ||
									!(tmpn->flags & NS_MEMONOTIFY))
								continue;
						}

						notice(n_MemoServ, cu->lptr->nick,
							"New channel memo from \002%s\002 (#%d)",
							lptr->nick,
							index);
						notice(n_MemoServ, cu->lptr->nick,
							"Type \002/msg %s READ %s %d\002 to read it",
							n_MemoServ,
							chptr->name,
							index);
					}
				}
			}

			newtext = (char *) MyMalloc(strlen(memotext) + strlen(ci->name) + 4);
			sprintf(newtext,
				"(%s) %s",
				ci->name,
				memotext);
			for (ca = ci->access; ca; ca = ca->next)
			{
				if (ca->nptr)
					sendmemo(ca->nptr->nick, newtext, lptr);
			}

			MyFree(newtext);

		} /* else */
	#endif /* CHANNELSERVICES */
	} /* if (index) */

	MyFree (memotext);
} /* m_send() */

/*
m_list()
	List the memos for lptr->nick, or channel av[1] if specified
*/

static void
m_list(Luser *lptr, NickInfo *nptr, int ac, char **av)

{
	MemoInfo	*mi = (MemoInfo *) NULL;
	Memo		*memoptr;
	char		status[4];

	if (!nptr)
		return;

	if (ac >= 2)
	{
		if (av[1][0] == '#')
		{
		#ifdef CHANNELSERVICES
			ChanInfo	*ci;
		#endif

		#ifndef CHANNELSERVICES
			notice(n_MemoServ, lptr->nick,
				"Channel services are disabled");
			return;
		#else
			if (!(ci = FindChan(av[1])))
			{
				notice(n_MemoServ, lptr->nick,
					ERR_CH_NOT_REGGED,
					av[1]);
				return;
			}

			if (!HasAccess(ci, lptr, CA_AUTOOP))
			{
				notice(n_MemoServ, lptr->nick,
					"AutoOp access is required to list memos for [\002%s\002]",
					ci->name);
				return;
			}

			mi = FindMemoList(ci->name);
			if (!mi)
			{
				notice(n_MemoServ, lptr->nick,
					"There are no recorded memos for [\002%s\002]",
					ci->name);
				return;
			}
		#endif /* CHANNELSERVICES */

		}
	}

	if (!mi)
	{
		if (!(mi = FindMemoList(nptr->nick)))
		{
			notice(n_MemoServ, lptr->nick,
				"You have no recorded memos");
			return;
		}
	}

	notice(n_MemoServ, lptr->nick,
		"-- Listing memos for [\002%s\002] --",
		mi->name);
	notice(n_MemoServ, lptr->nick,
		"    Idx Sender             Time Sent");
	for (memoptr = mi->memos; memoptr; memoptr = memoptr->next)
	{
		if (memoptr->flags & MS_DELETE)
			strcpy(status, "(D)");
		else
			if (mi->name[0] != '#')
			{
				if (memoptr->flags & MS_READ)
					strcpy(status, "(R)");
				else
					strcpy(status, "(N)");
			}
			else
				status[0] = '\0';

		notice(n_MemoServ, lptr->nick,
			"%3s %-3d %-18s %s ago",
			status,
			memoptr->index,
			memoptr->sender,
			timeago(memoptr->sent, 1));
	}
	notice(n_MemoServ, lptr->nick,
		"-- End of list --");
} /* m_list() */

/*
m_read()
	Read a memo index av[1] or av[2] if av[1] is a channel
*/

static void
m_read(Luser *lptr, NickInfo *nptr, int ac, char **av)

{
	MemoInfo	*mi;
	int		index;
	Memo	*memoptr;
	char	istr[10];

	if (!nptr)
		return;

	if (ac < 2)
	{
		notice(n_MemoServ, lptr->nick,
			"Syntax: READ [channel] <index|all>");
		notice(n_MemoServ, lptr->nick,
			ERR_MORE_INFO,
			n_MemoServ,
			"READ");
		return;
	}

	if (ac >= 3)
	{
		#ifdef CHANNELSERVICES
			ChanInfo	*ci;
		#endif

		#ifndef CHANNELSERVICES
			notice(n_MemoServ, lptr->nick,
				"Channel services are disabled");
			return;
		#else
			if (!(ci = FindChan(av[1])))
			{
				notice(n_MemoServ, lptr->nick,
					ERR_CH_NOT_REGGED,
					av[1]);
				return;
			}

			if (!HasAccess(ci, lptr, CA_AUTOOP))
			{
				notice(n_MemoServ, lptr->nick,
					"AutoOp access is required to read memos for [\002%s\002]",
					ci->name);
				return;
			}

			if (!(mi = FindMemoList(ci->name)))
			{
				notice(n_MemoServ, lptr->nick,
					"There are no recorded memos for [\002%s\002]",
					ci->name);
				return;
			}
			index = IsNum(av[2]);

		#endif /* CHANNELSERVICES */
	}
	else
	{
		if (!(mi = FindMemoList(nptr->nick)))
		{
			notice(n_MemoServ, lptr->nick,
				"You have no recorded memos");
			return;
		}
		index = IsNum(av[1]);
	}

	if ((index < 0) || (index > mi->memocnt) || 
			(!index && (strcasecmp(av[(ac >= 3) ? 2 : 1], "ALL") != 0)))
	{
		notice(n_MemoServ, lptr->nick,
			"[\002%s\002] is an invalid index",
			av[(ac >= 3) ? 2 : 1]);
		return;
	}

	for (memoptr = mi->memos; memoptr; memoptr = memoptr->next)
	{
		if (!index || (memoptr->index == index))
		{
			notice(n_MemoServ, lptr->nick,
				"Memo #%d from %s (sent %s ago):",
				memoptr->index,
				memoptr->sender,
				timeago(memoptr->sent, 1));
			notice(n_MemoServ, lptr->nick,
				memoptr->text);
			if (ac < 3)
			{
				/* only mark nickname memos as read - not channel memos */
				if (!(memoptr->flags & MS_READ))
					mi->newmemos--;
				memoptr->flags |= MS_READ;
			}
		}
	}

	if (index)
		sprintf(istr, "%d", index);
	else
		strcpy(istr, "ALL");

	if (ac >= 3)
		notice(n_MemoServ, lptr->nick,
			"To delete, type \002/msg %s DEL %s %s",
			n_MemoServ,
			av[1],
			istr);
	else
		notice(n_MemoServ, lptr->nick,
			"To delete, type \002/msg %s DEL %s",
			n_MemoServ,
			istr);
		
} /* m_read() */

/*
m_help()
	Give help to lptr->nick
*/

static void
m_help(Luser *lptr, NickInfo *nptr, int ac, char **av)

{
	if (ac >= 2)
	{
		char	str[MAXLINE];
		Command		*cptr;

		for (cptr = memocmds; cptr->cmd; cptr++)
			if (!strcasecmp(av[1], cptr->cmd))
				break;

		if (cptr->cmd)
			if ((cptr->level == LVL_ADMIN) &&
					!(IsValidAdmin(lptr)))
			{
				notice(n_MemoServ, lptr->nick,
					"No help available on \002%s\002",
					av[1]);
				return;
			}

		sprintf(str, "%s", av[1]);

		GiveHelp(n_MemoServ, lptr->nick, str, NODCC);
	}
	else
		GiveHelp(n_MemoServ, lptr->nick, (char *) NULL, NODCC);
} /* m_help() */

/*
m_del()
	Mark memo index av[1] for deletion - DoTimer() does the actual
deleting every hour on the hour
*/

static void
m_del(Luser *lptr, NickInfo *nptr, int ac, char **av)

{
	MemoInfo	*mi;
	Memo		*memoptr;
	char	istr[MAXLINE];
	int		index;

	if (!nptr)
		return;

	if (ac < 2)
	{
		notice(n_MemoServ, lptr->nick,
			"Syntax: DEL [channel] <index|all>");
		notice(n_MemoServ, lptr->nick,
			ERR_MORE_INFO,
			n_MemoServ,
			"DEL");
		return;
	}

	if (ac >= 3)
	{
		#ifdef CHANNELSERVICES
			ChanInfo	*ci;
		#endif

		#ifndef CHANNELSERVICES
			notice(n_MemoServ, lptr->nick,
				"Channel services are disabled");
			return;
		#else
			if (!(ci = FindChan(av[1])))
			{
				notice(n_MemoServ, lptr->nick,
					ERR_CH_NOT_REGGED,
					av[1]);
				return;
			}

			if (!HasAccess(ci, lptr, CA_SUPEROP))
			{
				notice(n_MemoServ, lptr->nick,
					"SuperOp access is required to delete memos for [\002%s\002]",
					ci->name);
				return;
			}

			if (!(mi = FindMemoList(ci->name)))
			{
				notice(n_MemoServ, lptr->nick,
					"There are no recorded memos for [\002%s\002]",
					ci->name);
				return;
			}
			index = IsNum(av[2]);

		#endif /* CHANNELSERVICES */

	}
	else
	{
		if (!(mi = FindMemoList(nptr->nick)))
		{
			notice(n_MemoServ, lptr->nick,
				"You have no recorded memos");
			return;
		}
		index = IsNum(av[1]);
	}

	if ((index < 0) || (index > mi->memocnt) || 
			(!index && (strcasecmp(av[(ac >= 3) ? 2 : 1], "ALL") != 0)))
	{
		notice(n_MemoServ, lptr->nick,
			"[\002%s\002] is an invalid index",
			av[(ac >= 3) ? 2 : 1]);
		return;
	}

	for (memoptr = mi->memos; memoptr; memoptr = memoptr->next)
	{
		if (!index || (memoptr->index == index))
			memoptr->flags |= MS_DELETE;
	}

	if (index)
		sprintf(istr, "Memo #%d has", index);
	else
		strcpy(istr, "All memos have");

	if (ac >= 3)
		notice(n_MemoServ, lptr->nick,
			"%s been marked for deletion for [\002%s\002]",
			istr,
			av[1]);
	else
		notice(n_MemoServ, lptr->nick,
			"%s been marked for deletion",
			istr);
} /* m_del() */

/*
m_undel()
	Undelete index av[1] or av[2] if av[1] is a channel
*/

static void
m_undel(Luser *lptr, NickInfo *nptr, int ac, char **av)

{
	MemoInfo	*mi;
	Memo		*memoptr;
	char	istr[MAXLINE];
	int		index;

	if (!nptr)
		return;

	if (ac < 2)
	{
		notice(n_MemoServ, lptr->nick,
			"Syntax: UNDEL [channel] <index|all>");
		notice(n_MemoServ, lptr->nick,
			ERR_MORE_INFO,
			n_MemoServ,
			"UNDEL");
		return;
	}

	if (ac >= 3)
	{
		#ifdef CHANNELSERVICES
			ChanInfo	*ci;
		#endif

		#ifndef CHANNELSERVICES
			notice(n_MemoServ, lptr->nick,
				"Channel services are disabled");
			return;
		#else
			if (!(ci = FindChan(av[1])))
			{
				notice(n_MemoServ, lptr->nick,
					ERR_CH_NOT_REGGED,
					av[1]);
				return;
			}

			if (!HasAccess(ci, lptr, CA_SUPEROP))
			{
				notice(n_MemoServ, lptr->nick,
					"SuperOp access is required to undelete memos for [\002%s\002]",
					ci->name);
				return;
			}

			if (!(mi = FindMemoList(ci->name)))
			{
				notice(n_MemoServ, lptr->nick,
					"There are no recorded memos for [\002%s\002]",
					ci->name);
				return;
			}
			index = IsNum(av[2]);
		#endif /* CHANNELSERVICES */

	}
	else
	{
		if (!(mi = FindMemoList(nptr->nick)))
		{
			notice(n_MemoServ, lptr->nick,
				"You have no recorded memos");
			return;
		}
		index = IsNum(av[1]);
	}

	if ((index < 0) || (index > mi->memocnt) || 
			(!index && (strcasecmp(av[(ac >= 3) ? 2 : 1], "ALL") != 0)))
	{
		notice(n_MemoServ, lptr->nick,
			"[\002%s\002] is an invalid index",
			av[(ac >= 3) ? 2 : 1]);
		return;
	}

	for (memoptr = mi->memos; memoptr; memoptr = memoptr->next)
	{
		if (!index || (memoptr->index == index))
			memoptr->flags &= ~MS_DELETE;
	}

	if (index)
		sprintf(istr, "Memo #%d has", index);
	else
		strcpy(istr, "All memos have");

	if (ac >= 3)
		notice(n_MemoServ, lptr->nick,
			"%s been unmarked for deletion for [\002%s\002]",
			istr,
			av[1]);
	else
		notice(n_MemoServ, lptr->nick,
			"%s been unmarked for deletion",
			istr);
} /* m_undel() */

/*
m_forward()
	Forward a memo to a nick/channel
*/

static void
m_forward(Luser *lptr, NickInfo *nptr, int ac, char **av)

{
	MemoInfo	*from, *target;
	Memo		*memoptr = (Memo *) NULL;
	Memo		*fromptr;
	char	*to; /* who the memo is sent to */
	int		index, cnt;
	char	buf[MAXLINE];
	NickInfo	*ni = (NickInfo *) NULL;
#ifdef CHANNELSERVICES
	ChanInfo	*ci = (ChanInfo *) NULL;
#endif

	if (!nptr)
		return;

	if (ac < 3)
	{
		notice(n_MemoServ, lptr->nick,
			"Syntax: FORWARD <index|ALL> <nick/channel>");
		notice(n_MemoServ, lptr->nick,
			ERR_MORE_INFO,
			n_MemoServ,
			"FORWARD");
		return;
	}

	if (!(from = FindMemoList(nptr->nick)))
	{
		notice(n_MemoServ, lptr->nick,
			"You have no recorded memos");
		return;
	}

	index = IsNum(av[1]);
	if ((index < 0) || (index > from->memocnt) || 
			(!index && (strcasecmp(av[1], "ALL") != 0)))
	{
		notice(n_MemoServ, lptr->nick,
			"[\002%s\002] is an invalid index",
			av[1]);
		return;
	}

	if (av[2][0] == '#')
	{
		#ifndef CHANNELSERVICES
			notice(n_MemoServ, lptr->nick,
				"Channel services are disabled");
			return;
		#else
			if (!(ci = FindChan(av[2])))
			{
				notice(n_MemoServ, lptr->nick,
					ERR_CH_NOT_REGGED,
					av[2]);
				return;
			}

			if (!HasAccess(ci, lptr, CA_AUTOOP))
			{
				notice(n_MemoServ, lptr->nick,
					"AutoOp access is required to forward a memo to [\002%s\002]",
					ci->name);
				return;
			}

			to = ci->name;

		#endif
	}
	else
	{
		/* it was sent to a nickname */

		if (!(ni = GetLink(av[2])))
		{
			notice(n_MemoServ, lptr->nick,
				ERR_NOT_REGGED,
				av[2]);
			return;
		}

		if (!(ni->flags & NS_MEMOS))
		{
			notice(n_MemoServ, lptr->nick,
				"[\002%s\002] is rejecting all memos",
				ni->nick);
			return;
		}

		to = ni->nick;
	}

	if (!(target = FindMemoList(to)))
	{
		target = (MemoInfo *) MyMalloc(sizeof(MemoInfo));
		target->name = MyStrdup(to);
		target->memocnt = 0;
		target->newmemos = 0;
		target->memos = target->lastmemo = (Memo *) NULL;
		target->flags = 0;
		AddMemoList(target);
	}
	else
	{
		if (from == target)
		{
			/*
			 * If we let someone forward memos to themselves,
			 * the below loop will never end, eventually malloc()'ing too
			 * much and crashing - head it off at the pass
			 */
			notice(n_MemoServ, lptr->nick,
				"You cannot forward memos to yourself");
			return;
		}
	}

	if (MaxMemos && (target->memocnt >= MaxMemos))
	{
		notice(n_MemoServ, lptr->nick,
			"%s has reached the maximum memo limit, and cannot receive more",
			to);
		return;
	}

	cnt = 0;
	for (fromptr = from->memos; fromptr; fromptr = fromptr->next)
	{
		if (!index || (fromptr->index == index))
		{
			memset(&buf, 0, MAXLINE);

			target->memocnt++;
			target->newmemos++;
			cnt++;

			memoptr = (Memo *) MyMalloc(sizeof(Memo));
			memoptr->next = (Memo *) NULL;
			memoptr->sender = MyStrdup(lptr->nick);
			memoptr->sent = time(NULL);
			memoptr->index = target->memocnt;
			memoptr->flags = 0;

			strcpy(buf, "[Fwd]: ");
			strncat(buf, fromptr->text, MAXLINE - 8);
			memoptr->text = MyStrdup(buf);

			AddMemo(target, memoptr);

			if (MaxMemos && (target->memocnt >= MaxMemos))
				break;
		}
	}

	if (!index)
		sprintf(buf, "All memos have");
	else
		sprintf(buf, "Memo #%d has", index);

	notice(n_MemoServ, lptr->nick,
		"%s been forwarded to [\002%s\002]",
		buf,
		target->name);

	if (ni)
	{
		/*
		 * It was sent to a nickname - check if they are online
		 * and notify them
		 */
		if ((ni->flags & NS_MEMONOTIFY) && FindClient(target->name))
		{
			notice(n_MemoServ, target->name,
				"You have %d new forwarded memo%s from \002%s\002",
				cnt,
				(cnt == 1) ? "" : "s",
				memoptr->sender);
			notice(n_MemoServ, target->name,
				"Type \002/msg %s LIST\002 to view %s",
				n_MemoServ,
				(cnt == 1) ? "it" : "them");
		}
	} /* if (ni) */
#ifdef CHANNELSERVICES
	else
	{
		Channel		*chptr;

		/*
		 * It was sent to a channel - notify every AOP or higher
		 */
		if ((ci) && (chptr = FindChannel(target->name)))
		{
			ChanUser	*cu;
			NickInfo	*tmpn;

			for (cu = chptr->firstuser; cu; cu = cu->next)
			{
				if (FindService(cu->lptr))
					continue;

				tmpn = GetLink(cu->lptr->nick);

				if (HasAccess(ci, cu->lptr, CA_AUTOOP))
				{
					if (tmpn)
						if (!(tmpn->flags & NS_MEMOS) ||
								!(tmpn->flags & NS_MEMONOTIFY))
							continue;

					notice(n_MemoServ, cu->lptr->nick,
						"%d new forwarded channel memo%s from \002%s\002",
						cnt,
						(cnt == 1) ? "" : "s",
						memoptr->sender);
					notice(n_MemoServ, cu->lptr->nick,
						"Type \002/msg %s LIST %s\002 to view %s",
						n_MemoServ,
						chptr->name,
						(cnt == 1) ? "it" : "them");
				}
			}
		}
	} /* else */
#endif /* CHANNELSERVICES */
} /* m_forward() */

/*
m_reply()
	Reply to memo number av[1] with text av[2]-
*/

static void
m_reply(Luser *lptr, NickInfo *nptr, int ac, char **av)

{
	MemoInfo	*from;
	Memo			*memoptr;
	NickInfo	*ni;
	char		*memotext;
	int		index, ii;

	if (!nptr)
		return;

	if (ac < 3)
	{
		notice(n_MemoServ, lptr->nick,
			"Syntax: REPLY <index> <text>");
		notice(n_MemoServ, lptr->nick,
			ERR_MORE_INFO,
			n_MemoServ,
			"REPLY");
		return;
	}

	if (!(from = FindMemoList(nptr->nick)))
	{
		notice(n_MemoServ, lptr->nick,
			"You have no recorded memos");
		return;
	}

	index = IsNum(av[1]);
	if ((index < 0) || (index > from->memocnt))
	{
		notice(n_MemoServ, lptr->nick,
			"[\002%s\002] is an invalid index",
			av[1]);
		return;
	}

	ni = (NickInfo *) NULL;
	for (memoptr = from->memos; memoptr; memoptr = memoptr->next)
	{
		if (memoptr->index == index)
		{
			if (!(ni = GetLink(memoptr->sender)))
			{
				notice(n_MemoServ, lptr->nick,
					ERR_NOT_REGGED,
					av[2]);
				return;
			}
			if (!(ni->flags & NS_MEMOS))
			{
				notice(n_MemoServ, lptr->nick,
					"[\002%s\002] is rejecting all memos",
					ni->nick);
				return;
			}
			break;
		}
	}

	if (ac >= 3)
	{
		memotext = (char *) MyMalloc(strlen("[Re]: ") + strlen(av[2]) + 1);
		sprintf(memotext, "[Re]: %s", av[2]);
		ii = 3;
		while (ii < ac)
		{
			memotext = (char *) MyRealloc(memotext, strlen(memotext) + strlen(av[ii]) + (2 * CHARSIZE));
			strcat(memotext, " ");
			strcat(memotext, av[ii]);
			ii++;
		}
	}
	else
		memotext = MyStrdup("");

	index = sendmemo(ni->nick, memotext, lptr);

	if (index)
	{
		notice(n_MemoServ, lptr->nick,
			"Memo has been recorded for [\002%s\002]",
			ni->nick);

		/*
		 * It was sent to a nickname - check if they are online
		 * and notify them
		 */
		if ((ni->flags & NS_MEMONOTIFY) && FindClient(ni->nick))
		{
			notice(n_MemoServ, ni->nick,
				"You have a new memo from \002%s\002 (#%d)",
				lptr->nick,
				index);
			notice(n_MemoServ, ni->nick,
				"Type \002/msg %s READ %d\002 to read it",
				n_MemoServ,
				index);
		}
	} /* if (index) */

	MyFree(memotext);
} /* m_reply() */

/*
m_purge()
	Delete all memos that were marked for deletion via the DEL command
*/

static void
m_purge(Luser *lptr, NickInfo *nptr, int ac, char **av)

{
	MemoInfo	*mi;
	int		cnt;

	if (!nptr)
		return;

	if (ac >= 2)
	{
		#ifdef CHANNELSERVICES
			ChanInfo	*ci;
		#endif

		#ifndef CHANNELSERVICES
			notice(n_MemoServ, lptr->nick,
				"Channel services are disabled");
			return;
		#else
			if (!(ci = FindChan(av[1])))
			{
				notice(n_MemoServ, lptr->nick,
					ERR_CH_NOT_REGGED,
					av[1]);
				return;
			}

			if (!HasAccess(ci, lptr, CA_SUPEROP))
			{
				notice(n_MemoServ, lptr->nick,
					"SuperOp access is required to purge memos for [\002%s\002]",
					ci->name);
				return;
			}

			if (!(mi = FindMemoList(ci->name)))
			{
				notice(n_MemoServ, lptr->nick,
					"There are no recorded memos for [\002%s\002]",
					ci->name);
				return;
			}
		#endif /* CHANNELSERVICES */
	}
	else
	{
		if (!(mi = FindMemoList(nptr->nick)))
		{
			notice(n_MemoServ, lptr->nick,
				"You have no recorded memos");
			return;
		}
	}

	cnt = PurgeMemos(mi);

	if (mi->name[0] == '#')
		notice(n_MemoServ, lptr->nick,
			"Memos marked for deletion on [\002%s\002] have been purged (%d found)",
			mi->name,
			cnt);
	else
		notice(n_MemoServ, lptr->nick,
			"Your memos marked for deletion have been purged (%d found)",
			cnt);
} /* m_purge() */

#endif /* defined(NICKSERVICES) && defined(MEMOSERVICES) */
