/*
 * HybServ TS Services, Copyright (C) 1998-1999 Patrick Alken
 * This program comes with absolutely NO WARRANTY
 *
 * Should you choose to use and/or modify this source code, please
 * do so under the terms of the GNU General Public License under which
 * this program is distributed.
 */

static char *rcsid = "$Id: mystring.c,v 1.00 1999/02/26 12:12:32 wnder Exp $";

#include	"headers.h"

/*
Snprintf()
 Similar to snprintf(), but log any buffer overflows that occur; also
make sure the string ends with a \0
*/

int
Snprintf(char *dest, size_t size, char *format, ...)

{
	int		ret;
	va_list	args;

	va_start(args, format);

#ifdef HAVE_VSNPRINTF
	ret = vsnprintf(dest, size - 1, format, args);
	dest[size - 1] = '\0';
#else
	ret = vsprintf(dest, format, args);

	/*
	 * add 1 to ret because vsprintf()'s return value does not
	 * include the trailing \0
	 */
	ret++;

	if (ret > size)
	{
		putlog(LOG1,
			"Snprintf: Buffer overflow in vsprintf() (tried to write %d bytes to a %d byte string)",
			ret,
			size);
		exit(1);
	}
#endif /* HAVE_VSNPRINTF */

	va_end(args);

	return (ret);
} /* Snprintf() */

/*
StrToupper()
	args: char *str
	purpose: convert the string 'str' to uppercase
	return: the uppercase string
*/

char *
StrToupper(char *str)

{
	int		ii;
	static	char	retstr[MAXLINE];

	if (!str)
		return ((char *) NULL);

	retstr[0] = '\0';
	for (ii = 0; ii < strlen(str); ii++)
		retstr[ii] = ToUpper(str[ii]);

	retstr[ii] = '\0';

	return (retstr);
} /* StrToupper() */

/*
StrTolower()
	args: char *str
	purpose: convert the string 'str' to lowercase
	return: the lowercase string
*/

char *
StrTolower(char *str)

{
	int ii;
	static char retstr[MAXLINE];

	if (!str)
		return ((char *) NULL);

	retstr[0] = '\0';
	for (ii = 0; ii < strlen(str); ii++)
		retstr[ii] = ToLower(str[ii]);

	retstr[ii] = '\0';

	return (retstr);
} /* StrTolower() */

/*
GetString()
 Reverse the array av back into a normal string assuming there
are 'ac' indices in the array. Space is allocated for the new
string.
*/

char *
GetString(int ac, char **av)

{
	char *final;
	char temp[MAXLINE];
	int ii;

	final = (char *) MyMalloc(sizeof(char));
	*final = '\0';

	ii = 0;
	while (ii < ac)
	{
		if ((ii + 1) >= ac)
		{
			/*
			 * This is the last arguement, so don't put an extra
			 * space on the end
			 */
			sprintf(temp, "%s",	av[ii]);
		}
		else
			sprintf(temp, "%s ", av[ii]);

		final = (char *) MyRealloc(final,
		                  strlen(final) + strlen(temp) + sizeof(char));
		strcat(final, temp);

		++ii;
	}

	return (final);
} /* GetString() */
