/*
 * HybServ TS Services, Copyright (C) 1998-1999 Patrick Alken
 * This program comes with absolutely NO WARRANTY
 *
 * Should you choose to use and/or modify this source code, please
 * do so under the terms of the GNU General Public License under which
 * this program is distributed.
 */

static char *rcsid = "$Id: operserv.c,v 1.01 1999/08/19 22:41:47 wnder Exp $";

#include	"headers.h"
#include  "settings.h"

#include	<sys/resource.h>
#include	<sys/utsname.h>

#ifdef ALLOW_FUCKOVER
/* array to hold current o_fuckover() targets */
static	Process		**fprocs = NULL;

/* number of indexes in fprocs array */
static	int				fprocs_num = 0;
#endif /* ALLOW_FUCKOVER */

static char *onick, /* nickname of current user giving command */
            *ouser, /* username of current user giving command */
            *ohost; /* hostname of current user giving command */

static void os_notice(Luser *, int, char *, ...);
static void o_RecordCommand(int, char *, ...);
static o_Command *GetoCommand(o_Command *, char *);

static	void	o_identify(Luser *, int, char **, int);
static	void	o_kill(Luser *, int, char **, int);
static	void	o_rehash(Luser *, int, char **, int);
static	void	o_restart(Luser *, int, char **, int);

#ifdef ALLOW_JUPES
static	void	o_jupe(Luser *, int, char **, int);
static	void	o_unjupe(Luser *, int, char **, int);
#endif

static	void	o_omode(Luser *, int, char **, int);
static	void	o_secure(Luser *, int, char **, int);

#ifdef ALLOW_DIE
static	void	o_die(Luser *, int, char **, int);
#endif

static	void	o_status(Luser *, int, char **, int);

#ifdef ALLOW_GLINES
static	void	o_gline(Luser *, int, char **, int);
static	void	o_ungline(Luser *, int, char **, int);
#endif

static	void	o_help(Luser *, int, char **, int);
static	void	o_join(Luser *, int, char **, int);
static	void	o_part(Luser *, int, char **, int);
static	void	o_clones(Luser *, int, char **, int);

#ifdef ALLOW_FUCKOVER
static	void	o_fuckover(Luser *, int, char **, int);
#endif

static	void	o_stats(Luser *, int, char **, int);
static	void	o_trace(Luser *, int, char **, int);
static	void	o_channel(Luser *, int, char **, int);
static	void	o_on(Luser *, int, char **, int);
static	void	o_off(Luser *, int, char **, int);
static	void	o_jump(Luser *, int, char **, int);

static	void	o_save(Luser *, int, char **, int);
static	void	o_reload(Luser *, int, char **, int);

#ifdef NICKSERVICES
static	void	o_ignore(Luser *, int, char **, int);
static	void	o_ignore_add(Luser *, int, char **, int);
static	void	o_ignore_del(Luser *, int, char **, int);
static	void	o_ignore_list(Luser *, int, char **, int);
#endif

#ifdef 		ALLOW_KILLCHAN
static	void	o_killchan(Luser *, int, char **, int);
#endif

#ifdef 		ALLOW_KILLHOST
static	void	o_killhost(Luser *, int, char **, int);
#endif

#ifdef		HIGHTRAFFIC_MODE
static	void	o_htm(Luser *, int, char **, int);
static	void	o_htm_on(Luser *, int, char **, int);
static	void	o_htm_off(Luser *, int, char **, int);
static	void	o_htm_max(Luser *, int, char **, int);
#endif

static	void	o_hub(Luser *, int, char **, int);
static	void	o_umode(Luser *, int, char **, int);

#ifdef ALLOW_DUMP
static	void	o_dump(Luser *, int, char **, int);
#endif

static  void  o_set(Luser *, int, char **, int);

static	void	o_who(Luser *, int, char **, int);
static	void	o_boot(Luser *, int, char **, int);
static	void	o_quit(Luser *, int, char **, int);
static	void	o_motd(Luser *, int, char **, int);
static	void	o_link(Luser *, int, char **, int);
static	void	o_unlink(Luser *, int, char **, int);

static void TakeOver(Channel *cptr);
static float CalcMem(char *, int socket);

#ifdef ALLOW_FUCKOVER
static void init_fprocess(char *from, char *ftarget);
#endif

static void show_trace(Luser *, Luser *, int, int);
static void show_channel(Luser *, Channel *, int);

#ifdef STATSERVICES
static void show_trace_info(Luser *, Luser *, int);
#endif

static void DisplaySettings(Luser *, int);

static struct UmodeInfo *FindUmode(char);

static o_Command opercmds[] = {

	/*
	 * Regular user commands
	 */

	{ "HELP", o_help, 0, 0 },
	{ "IDENTIFY", o_identify, 0, 0 },
	{ "REGISTER", o_identify, 0, 0 },

	/*
	 * Operator commands
	 */

	{ "CLONES", o_clones, 0, 'o' },

#ifdef HIGHTRAFFIC_MODE
	{ "HTM", o_htm, 0, 'o' },
#endif

	{ "HUB", o_hub, 0, 'o' },
	{ "KILL", o_kill, 0, 'o' },
	{ "STATS", o_stats, 0, 'o' },
	{ "STATUS", o_status, 0, 'o' },
	{ "UMODE", o_umode, 0, 'o' },
	{ "USERMODE", o_umode, 0, 'o' },

	/*
	 * Administrator commands
	 */

	{ "CHANNEL", o_channel, 0, 'a' },
	{ "JOIN", o_join, 0, 'a' },
	{ "JUMP", o_jump, 0, 'a' },

#ifdef NICKSERVICES
	{ "IGNORE", o_ignore, 0, 'a' },
#endif

#ifdef ALLOW_KILLCHAN
	{ "KILLCHAN", o_killchan, 0, 'a' },
#endif

#ifdef ALLOW_KILLHOST
	{ "KILLHOST", o_killhost, 0, 'a' },
#endif

	{ "MODE", o_omode, 0, 'a' },
	{ "OMODE", o_omode, 0, 'a' },
	{ "PART", o_part, 0, 'a' },
	{ "REROUTE", o_jump, 0, 'a' },
	{ "SECURE", o_secure, 0, 'a' },
	{ "TAKEOVER", o_secure, 0, 'a' },
	{ "TRACE", o_trace, 0, 'a' },
	{ "USER", o_trace, 0, 'a' },
	{ "WHOIS", o_trace, 0, 'a' },

	/*
	 * Services Administrator commands
	 */

#ifdef ALLOW_DIE
	{ "DIE", o_die, 0, 's' },
#endif

#ifdef ALLOW_DUMP
	{ "DUMP", o_dump, 0, 's' },
	{ "QUOTE", o_dump, 0, 's' },
	{ "RAW", o_dump, 0, 's' },
#endif /* ALLOW_DUMP */

	{ "DISABLE", o_off, 0, 's' },
	{ "ENABLE", o_on, 0, 's' },

/*
 * After a lot of thinking, and a lot of input from "network gurus",
 * I've decided to disable this command, since it could be
 * considered a DoS attack.  The flood itself doesn't actually last
 * longer than 5 seconds at most, but it can be argued that during
 * that period, it could interfere with other internet activities
 * that the user is engaged in.  That is illegal.  Although
 * originally coded for use on ednet, a small group of friends, for
 * fun, if it were to be used on the wrong person, it could become
 * an excuse to sue and/or accuse the administrator of the services
 * of malicious DoS activities (especially if used often and without
 * good reason).
 *                   -Patrick Alken 10/21/1998
 */

 /*
  * "fuckover" re-enabled under #define ALLOW_FUCKOVER due to
  * user requests (undefined by default)
  * 03/07/1999
  */

#ifdef ALLOW_FUCKOVER
	{ "FUCKOVER", o_fuckover, 0, 's' },
#endif

	{ "OFF", o_off, 0, 's' },
	{ "ON", o_on, 0, 's' },
	{ "REHASH", o_rehash, 0, 's' },
	{ "RELOAD", o_reload, 0, 's' },
	{ "RESTART", o_restart, 0, 's' },
	{ "SAVE", o_save, 0, 's' },
	{ "SET", o_set, 0, 's' },
	{ "SHUTDOWN", o_off, 0, 's' },

	/*
	 * Miscellaneous
	 */

#ifdef ALLOW_JUPES
	{ "JUPE", o_jupe, 0, 'j' },
	{ "UNJUPE", o_unjupe, 0, 'j' },
#endif

#ifdef ALLOW_GLINES
	{ "GLINE", o_gline, 0, 'g' },
	{ "AKILL", o_gline, 0, 'g' },
	{ "UNGLINE", o_ungline, 0, 'g' },
#endif

	/*
	 * Dcc Only
	 */

	{ "WHO", o_who, 1, 0 },
	{ "BOOT", o_boot, 1, 'a' },
	{ "QUIT", o_quit, 1, 0 },
	{ "MOTD", o_motd, 1, 0 },
	{ "LINK", o_link, 1, 'o' },
	{ "UNLINK", o_unlink, 1, 'o' },

	{ (char *) 0, (void (*)()) 0, 0, 0 }
};

#ifdef NICKSERVICES

/* sub-commands for OperServ IGNORE */
static o_Command ignorecmds[] = {
	{ "ADD", o_ignore_add, 0, 0 },
	{ "DEL", o_ignore_del, 0, 0 },
	{ "LIST", o_ignore_list, 0, 0 },

	{ 0, (void (*)()) 0, 0, 0 }
};

#endif /* NICKSERVICES */

#ifdef HIGHTRAFFIC_MODE

/* sub-commands for OperServ HTM */
static o_Command htmcmds[] = {
	{ "ON", o_htm_on, 0, 0 },
	{ "OFF", o_htm_off, 0, 0 },
	{ "MAX", o_htm_max, 0, 0 },

	{ 0, (void (*)()) 0, 0, 0 }
};

#endif /* HIGHTRAFFIC_MODE */

/*
 * OperServ usermodes
 */
static struct UmodeInfo
{
	char usermode;
	int flag;
} Usermodes[] = {
	{ 'b', OPERUMODE_B },     /* see tcm bot connect attempts etc */
	{ 'c', OPERUMODE_C },     /* see lone connections */
	{ 'C', OPERUMODE_CAPC },  /* see client connections */
	{ 'd', OPERUMODE_D },     /* debug information */
	{ 'e', OPERUMODE_E },     /* see clone exits */
	{ 'E', OPERUMODE_CAPE },  /* see lient exits */
	{ 'j', OPERUMODE_J },     /* see channel joins */
	{ 'k', OPERUMODE_K },     /* see channel kicks */
	{ 'l', OPERUMODE_L },     /* see netsplits/netjoins */
	{ 'm', OPERUMODE_M },     /* see channel modes */
	{ 'n', OPERUMODE_N },     /* see network activity */
	{ 'o', OPERUMODE_O },     /* see new operators */
	{ 'O', OPERUMODE_CAPO },  /* see operator kills */
	{ 'p', OPERUMODE_P },     /* see channel parts */
	{ 's', OPERUMODE_S },     /* see usages of service bots */
	{ 'S', OPERUMODE_CAPS },  /* see server kills */
	{ 't', OPERUMODE_T },     /* see channel topic changes */
	{ 'y', OPERUMODE_Y },     /* see various DoTimer() events */

	{ 0, 0 }
};

/*
os_process()
  Process command coming from 'nick' directed towards n_OperServ
*/

void
os_process(char *nick, char *command, int sockfd)

{
	DccUser		*dccptr;
	Luser		*lptr;
	int		acnt;
	int		bad;
	char	**arv;
	o_Command	*cptr;
	Userlist	*cmduser = (Userlist *) NULL;

	if (!command)
		return;

	lptr = (Luser *) NULL;
	if (sockfd == NODCC)
		if (!(lptr = FindClient(nick)))
			return;

	if ((Network->flags & NET_OFF) &&
			(strncasecmp(command, "ON", 2) != 0) &&
			(strncasecmp(command, "REGISTER", 8) != 0) &&
			(strncasecmp(command, "IDENTIFY", 8) != 0))
	{
		os_notice(lptr, sockfd,
			"Services are currently \002disabled\002");
		return;
	}

	dccptr = IsDccSock(sockfd);

	if (!dccptr)
	{
		/*
		 * It is a /msg command - give argument 0 so it MUST
		 * match hostnames, not just nicknames - otherwise if
		 * someone manages to get an oper's password, all they
		 * have to do is come on with any hostname and just use
		 * the oper's nickname - make sure that doesn't happen
		 */
		cmduser = GetUser(0, lptr->nick, lptr->username, lptr->hostname);
	}

	if (cmduser || dccptr)
	{

		onick = ouser = ohost = (char *) NULL;

		if (!dccptr)
		{
			/*
			 * As a last resort, check if there is a client on the network
			 * with the nick
			 */
			if (lptr)
			{
				onick = MyStrdup(lptr->nick);
				ouser = MyStrdup(lptr->username);
				ohost = MyStrdup(lptr->hostname);
			}
		}
		else
		{
			onick = MyStrdup(dccptr->nick);
			ouser = MyStrdup(dccptr->username);
			ohost = MyStrdup(dccptr->hostname);
		}

		if (!onick || !ouser || !ohost)
			return;

		if (dccptr)
			cmduser = DccGetUser(dccptr);

		if (!cmduser)
		{
			MyFree(onick);
			MyFree(ouser);
			MyFree(ohost);
			os_notice(lptr, sockfd, "Access Denied");
			return;
		}

		if (OpersHaveAccess)
		{
			/*
			 * GetUser() returned GenericOper, meaning this person
			 * is an operator, but does not have an O: line. Mark them
			 * as registered so they can perform Operator commands
			 * on OperServ
			 */
			if (lptr && (cmduser == GenericOper))
				lptr->flags |= L_OSREGISTERED;
		} /* if (OpersHaveAccess) */

		acnt = SplitBuf(command, &arv);
		if (acnt == 0)
		{
			MyFree(arv);
			MyFree(onick);
			MyFree(ouser);
			MyFree(ohost);
			return;
		}

		cptr = GetoCommand(opercmds, arv[0]);

		if (cptr == (o_Command *) -1)
		{
			os_notice(lptr, sockfd, "Ambiguous command [%s]",
				arv[0]);
			MyFree(arv);
			MyFree(onick);
			MyFree(ouser);
			MyFree(ohost);
			return;
		}

		bad = 0;

		if (!cptr)
			bad = 1;
		else
		{
			/* 
		 	 * check if the command is dcc only - if so, make sure it didn't
		 	 * come from a privmsg
     	 */
			if ((cptr->dcconly) && (sockfd == NODCC))
				bad = 1;

			if ((cptr->flag) &&
					!IsRegistered(lptr, sockfd) &&
					!IsValidAdmin(lptr))
				bad = 1;
		}

		if (bad)
		{
			os_notice(lptr, sockfd, "Unknown command [%s]",
				arv[0]);

			if (sockfd == NODCC)
			{
				int		ii;

				/* send the msg to the partyline */
				BroadcastDcc(DCCOPS, "[%s!%s@%s]",
					lptr->nick,
					lptr->username,
					lptr->hostname);
				for (ii = 0; ii < acnt; ii++)
					BroadcastDcc(DCCOPS, " %s", arv[ii]);
				BroadcastDcc(DCCOPS, "\n");
			}
		}
		else
		{
			/* check if the user has authorization to exec command */
			if (CheckAccess(cmduser, cptr->flag))
			{
				/* call cptr->func to execute command */
				(*cptr->func)(lptr, acnt, arv, sockfd);
				MyFree(arv);
				MyFree(onick);
				MyFree(ouser);
				MyFree(ohost);
				return;
			}
			else
				os_notice(lptr, sockfd, ERR_NOPRIVS);
		}
		MyFree(arv);
		MyFree(onick);
		MyFree(ouser);
		MyFree(ohost);
		return;
	} /* if (cmduser || dccptr) */
	else
		os_notice(lptr, NODCC, "Access Denied");

	return;
} /* os_process() */

/*
os_loaddata()
	Load OperServ database - return 1 if successful, -1 if not, and -2
if errors are unrecoverable
*/

int
os_loaddata()

{
	FILE	*fp;
	char	line[MAXLINE], **av;
	int		ac,
				ret = 1,
				cnt,
				found;
	Userlist	*tempuser;

	if (!(fp = fopen(OperServDB, "r")))
	{
		/* OperServ data file doesn't exist */
		return (-1);
	}

	/*
	 * If this is being called from ReloadData(), make sure
	 * all users' umodes are zeroed, or OperServ will
	 * log a warning about multiple entries
	 */
	for (tempuser = userlist; tempuser; tempuser = tempuser->next)
		tempuser->umodes = 0;

	cnt = 0;
	while (fgets(line, MAXLINE - 1, fp))
	{
		cnt++;

		ac = SplitBuf(line, &av);
		if (!ac)
		{
			/* blank line probably */
			MyFree(av);
			continue;
		}

		if (av[0][0] == ';')
		{
			/* its a comment */
			MyFree(av);
			continue;
		}

		/*
		 * OperServDB format should be:
		 * Nickname <umodes>
		 * where <umodes> is an integer number corresponding to
		 * the OPERUMODE_* flags
		 */

		if (ac < 2)
		{
			fatal(1, "%s:%d Invalid database format (FATAL)",
				OperServDB,
				cnt);
			ret = (-2);
			MyFree(av);
			continue;
		}

		/*
		 * Go through user list and try to find a matching nickname
		 * with av[0]
		 */
		found = 0;
		for (tempuser = userlist; tempuser; tempuser = tempuser->next)
		{
			if (!strcasecmp(av[0], tempuser->nick))
			{
				found = 1;
				if (tempuser->umodes && (tempuser->umodes != OPERUMODE_INIT))
				{
					/*
					 * we've already seen a line with this nickname in
					 * OperServDB, so there must be multiple entries
					 */
					fatal(1, "%s:%d OperServ entry [%s] has multiple occurences (using first)",
						OperServDB,
						cnt,
						tempuser->nick);
					if (ret > 0)
						ret = (-1);
					break;
				}
				else
					tempuser->umodes = atol(av[1]);
			} /* if (!strcasecmp(av[0], tempuser->nick)) */
		} /* for (tempuser = userlist; tempuser; tempuser = tempuser->next) */

		if (!found)
		{
			/*
			 * No O: line found matching the nickname
			 */
			fatal(1, "%s:%d No O: line entry for nickname [%s] (ignoring)",
				OperServDB,
				cnt,
				av[0]);
			if (ret > 0)
				ret = (-1);
		}

		MyFree(av);
	} /* while (fgets(line, MAXLINE - 1, fp)) */

	fclose(fp);

	/*
	 * Now go through the list again and anyone who's
	 * usermodes are set to OPERUMODE_INIT did not have
	 * an entry in oper.db - therefore give them
	 * default usermodes (+cby)
	 */
	for (tempuser = userlist; tempuser; tempuser = tempuser->next)
	{
		if (tempuser->umodes == OPERUMODE_INIT)
			tempuser->umodes = (OPERUMODE_C | OPERUMODE_B | OPERUMODE_Y);
	}

	return (ret);
} /* os_loaddata() */

/*
GetoCommand()
 Return a pointer to the index of opercmds[] containing the command
'name'
Returns NULL if command not found, or (o_Command *) -1 if multiple
matches found
*/

static o_Command *
GetoCommand(o_Command *cmdlist, char *name)

{
	o_Command *cmdptr, *tmp;
	int		matches, /* number of matches we've had so far */
				clength;

	if (!cmdlist || !name)
		return ((o_Command *) NULL);

	tmp = (o_Command *) NULL;
	matches = 0;
	clength = strlen(name);
	for (cmdptr = cmdlist; cmdptr->cmd; cmdptr++)
	{
		if (!strncasecmp(name, cmdptr->cmd, clength))
		{
			if (clength == strlen(cmdptr->cmd))
			{
				/*
				 * name and cmdptr->cmd are the same length, so it
				 * must be an exact match, don't search any further
				 */
				matches = 0;
				break;
			}
			tmp = cmdptr;
			matches++;
		}
	}

	/*
	 * If matches > 1, name is an ambiguous command, so the
	 * user needs to be more specific
	 */
	if ((matches == 1) && (tmp))
		cmdptr = tmp;

	if (cmdptr->cmd)
		return (cmdptr);

	if (matches == 0)
		return ((o_Command *) NULL); /* no matches found */
	else
		return ((o_Command *) -1); /* multiple matches found */
} /* GetoCommand() */

/*
os_notice()
	args: Luser *lptr, int sockfd, char *msg
	purpose: send a NOTICE to 'lptr->nick' with 'msg'
	return: none
*/

static void
os_notice(Luser *lptr, int sockfd, char *format, ...)

{
	va_list		args;
	char	finstr[MAXLINE];

	va_start(args, format);

#ifdef HAVE_VSNPRINTF
	vsnprintf(finstr, sizeof(finstr) - 1, format, args);
	finstr[sizeof(finstr) - 1] = '\0';
#else
	vsprintf(finstr, format, args);
#endif /* HAVE_VSNPRINTF */

	va_end(args);

	if (sockfd == NODCC)
	{
		if (lptr)
			toserv(":%s NOTICE %s :%s\n",
				(ServerNotices) ? Me.name : n_OperServ,
				lptr->nick,
				finstr);
	}
	else
	{
		writesocket(sockfd, finstr);
		writesocket(sockfd, "\n");
	}
} /* os_notice() */

/*
os_join()
 Have OperServ join a channel
*/

void
os_join(Channel *cptr)

{
	char sendstr[MAXLINE];
	char **av;

	if (!cptr)
		return;

	sprintf(sendstr,
		":%s SJOIN %ld %s + :@%s\n",
		Me.name,
		(long) cptr->since,
		cptr->name,
		n_OperServ);

	toserv(sendstr);

	SplitBuf(sendstr, &av);

	/* Add OperServ to channel nick list etc */
	AddChannel(av, 0, (char **) NULL);

	MyFree(av);
} /* os_join() */

/*
os_join_ts_minus_1
 Similar to os_join((), but have OperServ join the channel
with TS - 1
*/

void
os_join_ts_minus_1(Channel *cptr)

{
	char sendstr[MAXLINE];
	char **av;
	int ac;

	if (!cptr)
		return;

	sprintf(sendstr,
		":%s SJOIN %ld %s + :@%s\n",
		Me.name,
		(cptr->since != 0) ? (long) cptr->since - 1 : 0,
		cptr->name,
		n_OperServ);

	toserv(sendstr);
	ac = SplitBuf(sendstr, &av);
	s_sjoin(ac, av);
	MyFree(av);
} /* os_join_ts_minus_1() */

/*
os_part()
 Have OperServ part a channel
*/

void
os_part(Channel *chptr)

{
	if (!chptr)
		return;

	toserv(":%s PART %s\n",
		n_OperServ,
		chptr->name);

	RemoveFromChannel(chptr, Me.osptr);
} /* os_part() */

/*
o_RecordCommand()
 Called when an OperServ command is executed - log it and
send it to users with a +s usermode
*/

static void
o_RecordCommand(int sockfd, char *format, ...)

{
	va_list args;
	char buffer[MAXLINE];

	va_start(args, format);

#ifdef HAVE_VSNPRINTF
	vsnprintf(buffer, sizeof(buffer) - 1, format, args);
	buffer[sizeof(buffer) - 1] = '\0';
#else
	vsprintf(buffer, format, args);
#endif /* HAVE_VSNPRINTF */

	va_end(args);

	/* log the command */
	putlog(LOG2,
		"%s: %s!%s@%s %s",
		n_OperServ,
		onick,
		ouser,
		ohost,
		buffer);

	/* send it to opers with usermode +s */
	if (sockfd == NODCC)
	{
		/*
		 * If they're using the command via privmsg,
		 * use the format: OperServ: nick!user@host COMMAND ...
		 */
		SendUmode(OPERUMODE_S,
			"%s: %s!%s@%s %s",
			n_OperServ,
			onick,
			ouser,
			ohost,
			buffer);
	}
	else
	{
		/*
		 * If they're using the command via dcc chat,
		 * use the format: #nick# COMMAND ...
		 */
		SendUmode(OPERUMODE_S,
			"#%s# %s",
			onick,
			buffer);
	}
} /* o_RecordCommand() */

/*
o_identify()
	Checks if lptr->nick is allowed to register with password av[1]
	ac: param count
*/

static void
o_identify(Luser *lptr, int ac, char **av, int sockfd)

{
	Userlist *uptr;

	if (ac < 2)
	{
		os_notice(lptr, sockfd, "Syntax: \002IDENTIFY <password>\002");
		return;
	}

	if (IsRegistered(lptr, sockfd))
	{
		os_notice(lptr, sockfd, "You are already identified");
		return;
	}

	uptr = GetUser(1, onick, ouser, ohost);
	if (!uptr)
		return;

	if (operpwmatch(uptr->password, av[1]))
	{
		DccUser *dccptr;

		os_notice(lptr, sockfd, "You are now identified");
		if (lptr)
			lptr->flags |= L_OSREGISTERED;

		if ((dccptr = IsDccSock(sockfd)))
			dccptr->flags &= ~SOCK_PENDING;

		o_RecordCommand(sockfd,
			"IDENTIFY");
	}
	else 
	{
		os_notice(lptr, sockfd, "Access Denied");

		o_RecordCommand(sockfd,
			"failed IDENTIFY");
	}
} /* o_identify() */

/*
o_rehash()
	Rehash the configuration file
*/

static void
o_rehash(Luser *lptr, int ac, char **av, int sockfd)

{
	if (Rehash() == 0)
	{
		os_notice(lptr, sockfd, "Rehash complete");

		o_RecordCommand(sockfd,
			"REHASH");
	}
	else
	{
		os_notice(lptr, sockfd, "Rehash failed");

		o_RecordCommand(sockfd,
			"REHASH (failed)");
	}
} /* o_rehash() */

/*
o_restart()
	Restart services
*/

static void
o_restart(Luser *lptr, int ac, char **av, int sockfd)

{
	o_RecordCommand(sockfd,
		"RESTART");

	ServReboot(); /* do the actual restart */

	DccError = 1;
} /* o_restart() */

#ifdef ALLOW_DIE

/*
o_die()
	Terminates HybServ
*/

static void
o_die(Luser *lptr, int ac, char **av, int sockfd)

{
	char *reason;

	if (ac < 2)
		reason = MyStrdup("");
	else
		reason = GetString(ac - 1, av + 1);

	o_RecordCommand(sockfd,
		"DIE [%s]",
		reason);

	os_notice(lptr, sockfd, "Shutting down");

	do_shutdown(onick, reason);
} /* o_die() */

#endif /* ALLOW_DIE */

/*
o_status()
	Displays various statistics such as jupes, glines, users etc
*/

static void
o_status(Luser *lptr, int ac, char **av, int sockfd)

{
	float mem;
	struct utsname un;
	struct rusage rus;

	o_RecordCommand(sockfd,
		"STATUS");

#if defined(NICKSERVICES) || defined(STATSERVICES)
	os_notice(lptr, sockfd, "Expire settings");

#ifdef NICKSERVICES

	#ifdef CHANNELSERVICES
		os_notice(lptr, sockfd, "            Channels: \002%s\002",
			timeago(ChannelExpire, 3));
	#endif

	os_notice(lptr, sockfd, "           Nicknames: \002%s\002",
		timeago(NickNameExpire, 3));

	#ifdef MEMOSERVICES
		os_notice(lptr, sockfd, "               Memos: \002%s\002",
			timeago(MemoExpire, 3));
	#endif

#endif /* NICKSERVICES */

#ifdef STATSERVICES
	os_notice(lptr, sockfd, "       StatServ data: \002%s\002",
		timeago(StatExpire, 3));
#endif

	os_notice(lptr, sockfd, "\002\002");

#endif /* defined(NICKSERVICES) || defined(STATSERVICES) */

#ifdef NICKSERVICES

	os_notice(lptr, sockfd, "Databases");

	os_notice(lptr, sockfd, "       Save interval: \002%s\002",
		timeago(DatabaseSync, 3));

	os_notice(lptr, sockfd, "           Next save: \002%s\002",
		timeago((DatabaseSync - (time(NULL) % DatabaseSync)), 3));

	os_notice(lptr, sockfd, " Password encryption: \002%s\002",
	#ifdef CRYPT_PASSWORDS
		"on");
	#else
		"off");
	#endif

	os_notice(lptr, sockfd, "\002\002");

#endif /* NICKSERVICES */

	os_notice(lptr, sockfd, "Security settings");

	os_notice(lptr, sockfd, "  O: line encryption: \002%s\002",
	#ifdef CRYPT_OPER_PASSWORDS
		"on");
	#else
		"off");
	#endif

	os_notice(lptr, sockfd, "    Flood Protection: \002%s\002",
		(FloodProtection) ? "on" : "off");

	if (FloodProtection)
	{
		os_notice(lptr, sockfd, "       Flood trigger: \002%d\002 messages in \002%s\002",
			FloodCount,
			timeago(FloodTime, 3));

		os_notice(lptr, sockfd, "         Ignore time: \002%s\002",
			timeago(IgnoreTime, 3));
	
		os_notice(lptr, sockfd, "    Permanent Ignore: \002%d\002 offenses",
			IgnoreOffenses);
	}

#ifdef NICKSERVICES

	#ifdef CHANNELSERVICES
		os_notice(lptr, sockfd, "   Maximum AutoKicks: \002%d\002 per channel",
			MaxAkicks);
	#endif

	os_notice(lptr, sockfd, "  Nicknames held for: \002%s\002",
		timeago(NSReleaseTimeout, 3));

#endif /* NICKSERVICES */

	if (MaxClones)
	{
		os_notice(lptr, sockfd, "      Maximum clones: \002%d\002 (autokill: \002%s\002)",
			MaxClones,
			(AutoKillClones) ? "on" : "off");
	} /* if (MaxClones) */

	os_notice(lptr, sockfd, "   High-Traffic Mode: \002%s\002",
	#ifdef HIGHTRAFFIC_MODE
		(HTM == 1) ? "on" : "off");
	#else
		"disabled");
	#endif

	os_notice(lptr, sockfd, "   Restricted Access: \002%s\002",
		(RestrictedAccess) ? "on" : "off");

	os_notice(lptr, sockfd, "\002\002");
	os_notice(lptr, sockfd, "General Status");

	if (uname(&un) == 0)
	{
		os_notice(lptr, sockfd, "          Running on: \002%s %s\002",
			un.sysname,
			un.release);
	}
	else
		os_notice(lptr, sockfd, "          Running on: \002*unknown*\002");

	mem = CalcMem((char *) NULL, NODCC);
	os_notice(lptr, sockfd, "        Memory Usage: \002%0.2f\002 kb (\002%0.2f\002 mb)",
		mem / 1024,
		(mem / 1024) / 1024);

	if (getrusage(RUSAGE_SELF, &rus) == -1)
		putlog(LOG1, "o_status: getrusage() error: %s",
			strerror(errno));
	else
	{
		time_t	seconds;

		seconds = rus.ru_utime.tv_sec + rus.ru_stime.tv_sec;
		if (seconds == 0)
			seconds = 1;
		os_notice(lptr, sockfd, "            Cpu Time: \002%02d:%02d\002 (mins:secs)",
			seconds / 60,
			seconds % 60);
	}

	if (HubSock != NOSOCKET)
	{
		os_notice(lptr, sockfd, "         Current Hub: \002%s\002 (connected for [\002%s\002])",
			currenthub->realname ? currenthub->realname : currenthub->hostname,
			timeago(currenthub->connect_ts, 0));
	}
	else
		os_notice(lptr, sockfd, "         Current Hub: \002none\002");

	os_notice(lptr, sockfd, "Hub Connect Interval: \002%s\002",
		timeago(ConnectFrequency, 3));

	if (AutoLinkFreq)
		os_notice(lptr, sockfd, "   Tcm Link Interval: \002%s\002",
			timeago(AutoLinkFreq, 3));
} /* o_status() */

/*
o_kill()
	Issues a KILL for av[1] if lptr->nick is authorized
*/

static void
o_kill(Luser *lptr, int ac, char **av, int sockfd)

{
	char *reason;
	int ii;
	Luser *kptr;

	if (ac < 2)
	{
		os_notice(lptr, sockfd, "Syntax: \002KILL <nick> [reason]\002");
		return;
	}

	kptr = FindClient(av[1]);

	if (kptr == (Luser *) NULL)
	{
		os_notice(lptr, sockfd, "No such user: %s", av[1]);
		return;
	}

	if (FindService(kptr))
	{
		os_notice(lptr, sockfd, "%s!%s@%s is a Service",
			kptr->nick,
			kptr->username,
			kptr->hostname);
		return;
	}

	if (ac < 3)
		reason = MyStrdup("");
	else
		reason = GetString(ac - 2, av + 2);

	if (kptr->flags & L_OSREGISTERED)
		ii = 1;
	else
		ii = 0;

	/* Check if the user has an exception flag */
	if (IsProtected(GetUser(ii, kptr->nick, kptr->username, kptr->hostname)))
	{
		/* The user is protected */
		os_notice(lptr, sockfd, "%s!%s@%s is protected", 
			kptr->nick,
			kptr->username,
			kptr->hostname);

		putlog(LOG2,
			"%s!%s@%s attempted to KILL protected user %s [%s]",
			onick,
			ouser,
			ohost,
			av[1],
			reason);

		MyFree(reason);
		return;
	}

	o_RecordCommand(sockfd,
		"KILL %s [%s]",
		kptr->nick,
		reason);

	toserv(":%s KILL %s :%s!%s (%s (Requested by %s))\n",
		n_OperServ,
		av[1],
		Me.name,
		n_OperServ,
		reason,
		onick);

	/* remove the killed nick from list */
	DeleteClient(kptr);

	os_notice(lptr, sockfd,
		"%s has been killed [%s]", av[1], reason);

	MyFree(reason);

	if (Me.sptr)
		Me.sptr->numoperkills++;

	Network->TotalOperKills++;
#ifdef STATSERVICES
	if (Network->TotalOperKills > Network->OperKillsT)
		Network->OperKillsT = Network->TotalOperKills;
#endif
} /* o_kill() */

#ifdef ALLOW_JUPES

/*
o_jupe()
	Jupes server av[1] with reason av[2]-.  (Prevents a server from
	connecting to the network)
*/

static void
o_jupe(Luser *lptr, int ac, char **av, int sockfd)

{
	char *reason; /* reason for jupe */
	char sendstr[MAXLINE], whostr[MAXLINE], **arv;
	Jupe *tempjupe;
	FILE *fp;
	int ii,
	    nickjupe = 1;
	struct tm *jupe_tm;
	time_t CurrTime;
	Luser *jptr;

	if (ac < 2)
	{
		os_notice(lptr, sockfd, "Syntax: \002JUPE <server/nick> [reason]\002");
		return;	
	}

	/*
	 * Make sure we don't jupe our hub server
	 * (thanks good!)
	 */
	if (currenthub && currenthub->realname)
	{
		if (match(av[1], currenthub->realname))
		{
			os_notice(lptr, sockfd,
				"[\002%s\002] matches services' current hub server, cancelling jupe",
				av[1]);
			return;
		}
	}

	for (ii = 0; ii < strlen(av[1]); ii++)
		if ((av[1][ii] == '*') || (av[1][ii] == '?') || (av[1][ii] == '.'))
		{
			nickjupe = 0;
			break;
		}

	/* Check if jupe already exists */
	for (tempjupe = jupelist; tempjupe; tempjupe = tempjupe->next)
		if (match(av[1], tempjupe->name))
		{
			os_notice(lptr, sockfd, "A jupe already exists for %s [%s]",
				tempjupe->name,
				tempjupe->reason);
			return;	
		}	

	if (JupeFile)
		fp = fopen(JupeFile, "a+");
	else
		fp = fopen(ConfigFile, "a+");

	if (fp == (FILE *) NULL)
	{
		os_notice(lptr, sockfd, "Unable to open config file");
		return;
	}

	if (!nickjupe)
	{
		if (ac < 3)
			reason = MyStrdup("");
		else
			reason = GetString(ac - 2, av + 2);
	}
	else
		reason = MyStrdup("Jupitered Nick");

	CurrTime = time(NULL);
	jupe_tm = localtime(&CurrTime);
	sprintf(whostr, "%d/%02d/%02d %s@%s",
		1900 + jupe_tm->tm_year,
		jupe_tm->tm_mon + 1,
		jupe_tm->tm_mday,
		onick,
		n_OperServ);

	fprintf(fp, "J:%s:%s:%s\n", av[1], reason, whostr);
	fclose(fp);

	AddJupe(av[1], reason, whostr); /* add jupe to list */

	os_notice(lptr, sockfd,
		"Jupe for %s [%s] has been successfully added to config file",
		av[1],
		reason);

	o_RecordCommand(sockfd,
		"JUPE %s [%s]",
		av[1],
		reason);

	if (!nickjupe)
		DoJupeSquit(av[1], reason, whostr);
	else
	{
		jptr = FindClient(av[1]);
		if (jptr != (Luser *) NULL)
		{
			/* kill the juped nick */
			toserv(":%s KILL %s :%s!%s (%s)\n",
				n_OperServ,
				av[1],
				Me.name,
				n_OperServ,
				reason);
			DeleteClient(jptr);

			if (Me.sptr)
				Me.sptr->numoperkills++;
			Network->TotalOperKills++;
		#ifdef STATSERVICES
			if (Network->TotalOperKills > Network->OperKillsT)
				Network->OperKillsT = Network->TotalOperKills;
		#endif
		}
		/* replace nick with a fake user */
		sprintf(sendstr,
			"NICK %s 1 %ld +i juped juped.com %s :Jupitered Nickname\n",
			av[1],
			(long) CurrTime,
			Me.name);
		toserv(sendstr);
		SplitBuf(sendstr, &arv);
		AddClient(arv);
	}
	
	MyFree(reason);
} /* o_jupe() */

/*
o_unjupe()
	Removes jupe on server av[1]
*/

static void
o_unjupe(Luser *lptr, int ac, char **av, int sockfd)

{
	int		ii,
				nickjupe;
	int		jcnt; /* number of matching jupes found */
	char	line[MAXLINE], linetemp[MAXLINE];
	char	tmpfile[MAXLINE];
	char *configname;
	char	*key;
	char	*jhost;
	Jupe	*tempjupe, *next;
	FILE	*configfp;
	FILE	*fp; /* random file to store all lines except the ones */
						 /* which match the jupes 												 */

	if (ac < 2)
	{
		os_notice(lptr, sockfd, "Syntax: \002UNJUPE <server>\002");
		return;
	}

	o_RecordCommand(sockfd,
		"UNJUPE %s",
		av[1]);

	/* remove J: line from config file */
	if (JupeFile)
		configname = JupeFile;
	else
		configname = ConfigFile;

	if ((configfp = fopen(configname, "r")) == NULL)
	{
		os_notice(lptr, sockfd, "Unable to open config file");
		return;
	}

	sprintf(tmpfile, "%s.tmp",
		configname);

	fp = fopen(tmpfile, "w"); 
	jcnt = 0;
	while (fgets(line, MAXLINE - 1, configfp))
	{
		strcpy(linetemp, line);
		key = strtok(line, ":");
		if ((*key == 'J') || (*key == 'j'))
		{
			jhost = strtok((char *) NULL, ":"); 
			if (match(av[1], jhost) == 0)
				fputs(linetemp, fp);
			else
				jcnt++;
		}
		else
			fputs(linetemp, fp);
	} /* while () */
	fclose(configfp);
	fclose(fp);

	rename(tmpfile, configname);

	if (jcnt == 0)
	{
		os_notice(lptr, sockfd, "No jupes matching [%s] found", 
			av[1]);
		return;
	}

	for (tempjupe = jupelist; tempjupe; )
	{
		if (match(av[1], tempjupe->name))
		{
			nickjupe = 1;
			for (ii = 0; ii < strlen(tempjupe->name); ii++)
			{
				if ((tempjupe->name[ii] == '*') ||
						(tempjupe->name[ii] == '?') ||
						(tempjupe->name[ii] == '.'))
				{
					nickjupe = 0;
					break;
				}
			}

			if (nickjupe)
			{
				Luser *tempuser;
				
				/* if theres a psuedo client holding the nick, kill it */
				if ((tempuser = FindClient(tempjupe->name)) != (Luser *) NULL)
				{
					toserv(":%s QUIT :UnJuped\n",
						tempuser->nick);
					DeleteClient(tempuser);
				}
			}
			else
			{
				/* squit all psuedo servers we created while they were juped */
				do_squit(tempjupe->name, "UnJuped");
			}

			next = tempjupe->next;

			DeleteJupe(tempjupe); /* remove jupe from list */

			tempjupe = next;
		} /* if (match(av[1], tempjupe->name)) */
		else
			tempjupe = tempjupe->next;
	}

	os_notice(lptr, sockfd,
		"Jupes matching [%s] removed (%d matches)", 
		av[1], 
		jcnt);
} /* o_unjupe() */

#endif /* ALLOW_JUPES */

/*
o_omode()
	Sets modes av[2]- on channel av[1] (requires "a" flag)
*/

static void
o_omode(Luser *lptr, int ac, char **av, int sockfd)

{
	char *omodes;
	Channel *chptr;

	if (ac < 3) 
	{
		os_notice(lptr, sockfd, "Syntax: \002OMODE <#channel> <modes>\002");
		return;
	}

	omodes = GetString(ac - 2, av + 2);

	o_RecordCommand(sockfd,
		"OMODE %s %s",
		av[1],
		omodes);

	if (!(chptr = FindChannel(av[1])))
	{
		os_notice(lptr, sockfd,
			"No such channel: %s",
			av[1]);
		return;
	}

	/* set the modes */
	DoMode(chptr, omodes, YES);

	MyFree(omodes);
} /* o_omode() */

/*
o_secure()
	Secures channel av[1] by deopping non "f" users, oping "f" users,
and unbanning administrators
*/

static void
o_secure(Luser *lptr, int ac, char **av, int sockfd)

{
	Channel	*cptr;

	if (ac < 2)
	{
		os_notice(lptr, sockfd, "Syntax: \002SECURE <#channel>\002");
		return;
	}

	o_RecordCommand(sockfd,
		"SECURE %s",
		av[1]);

	cptr = FindChannel(av[1]);
	if (cptr != (Channel *) NULL)
	{
		TakeOver(cptr);
		os_notice(lptr, sockfd, "%s has been secured",
			cptr->name);
	}
	else
		os_notice(lptr, sockfd, "No such channel: %s",
			av[1]);
} /* o_secure() */

#ifdef ALLOW_GLINES

/*
o_gline()
	Glines user av[1] from the network with reason av[2]- (bans them
	from entire network)
*/

static void
o_gline(Luser *lptr, int ac, char **av, int sockfd)

{
	char whostr[MAXLINE];
	char *reason, *hostmask;
	char *username,
	     *hostname;
	int ii, charcnt, sidx;
	FILE *fp;
	Luser *user, *prev;
	Gline *gptr;
	struct tm *gline_tm;
	time_t CurrTime;
	long expires;
	char uhost[UHOSTLEN + 2];

	if (ac < 2)
	{
		os_notice(lptr, sockfd,
			"Syntax: \002GLINE [time] <user@host> [reason]\002");
		return;
	}

	if ((expires = timestr(av[1])) > 0)
	{
		sidx = 3;
		hostmask = av[2];
	}
	else
	{
		sidx = 2;
		hostmask = av[1];
	}

	if (NonStarChars)
	{
		/* Check if a valid user@host was given */
		charcnt = 0;
		for (ii = strlen(hostmask) - 1; ii > 0; ii--)
		{
			if (IsWildChar(hostmask[ii]))
				continue;

			charcnt++;
		}

		if (NonStarChars > charcnt)
		{
			os_notice(lptr, sockfd,
				"[%s] is too general, please include at least %d non (*,?,@,.) characters",
        hostmask,
        NonStarChars);
      return;
		}
	} /* if (NonStarChars) */

	if (!(hostname = strchr(hostmask, '@')))
	{
		username = (char *) NULL;
		hostname = hostmask;
	}
	else
	{
		username = hostmask;
		*hostname++ = '\0';
	}

	/* Check if hostmask matches an exception user */
	if (IsProtectedHost(username, hostname))
	{
		os_notice(lptr, sockfd, "%s@%s matches a protected host, unable to gline", 
			username ? username : "*",
			hostname);
		return;
	}

	/* Check if gline already exists */
	if ((gptr = IsGline(username ? username : "*", hostname)))
	{
		os_notice(lptr, sockfd,
			"Gline already exists for %s@%s [%s]", 
			gptr->username,
			gptr->hostname, 
			gptr->reason);
		return;	
	}

	if (GlineFile)
		fp = fopen(GlineFile, "a+");
	else
		fp = fopen(ConfigFile, "a+");

	if (fp == (FILE *) NULL)
	{
		os_notice(lptr, sockfd, "Unable to open config file");
		return;
	}

	if (ac < (sidx + 1))
		reason = MyStrdup("");
	else
		reason = GetString(ac - sidx, av + sidx);

	CurrTime = time(NULL);
	gline_tm = localtime(&CurrTime);
	sprintf(whostr, "%d/%02d/%02d %s@%s",
		1900 + gline_tm->tm_year,
		gline_tm->tm_mon + 1,
		gline_tm->tm_mday,
		onick,
		n_OperServ);

	/* don't add temp glines to the config file */
	if (!expires)
		fprintf(fp, "G:%s@%s:%s:%s\n",
			username ? username : "*",
			hostname,
			reason,
			whostr);

	fclose(fp);

	sprintf(uhost, "%s@%s",
		username ? username : "*",
		hostname);

	/* add gline to list */
	AddGline(uhost, reason, whostr, expires);

	if (!expires)
	{
		os_notice(lptr, sockfd, "Gline for %s@%s [%s] has been activated", 
			username ? username : "*",
			hostname,
			reason);
	}
	else
	{
		os_notice(lptr, sockfd,
			"Temporary gline [%s] for %s@%s [%s] activated",
			av[1],
			username ? username : "*",
			hostname,
			reason);
	}

	if (!expires)
		o_RecordCommand(sockfd,
			"GLINE %s@%s [%s]",
			username ? username : "*",
			hostname,
			reason);
	else
		o_RecordCommand(sockfd,
			"GLINE %s@%s [%s] (%s)",
			username ? username : "*",
			hostname,
			reason,
			av[1]);

	/* Check if any users on the network match the new gline */
	prev = (Luser *) NULL;
	for (user = ClientList; user; )
	{
		if (user->server == Me.sptr)
		{
			user = user->next;
			continue;
		}

		if (username)
			if (!match(username, user->username))
			{
				user = user->next;
				continue;
			}

		if (match(hostname, user->hostname))
		{
			toserv(":%s KILL %s :%s!%s (Glined: %s (%s))\n",
				n_OperServ,
				user->nick,
				Me.name,
				n_OperServ,
				reason,
				whostr);

			if (Me.sptr)
				Me.sptr->numoperkills++;

			Network->TotalOperKills++;
		#ifdef STATSERVICES
			if (Network->TotalOperKills > Network->OperKillsT)
				Network->OperKillsT = Network->TotalOperKills;
		#endif

			if (prev)
			{
				DeleteClient(user);
				user = prev;
			}
			else
			{
				DeleteClient(user);
				user = (Luser *) NULL;
			}
		}

		prev = user;

		if (user)
			user = user->next;
		else
			user = ClientList;
	}

#ifdef HYBRID_GLINES
	ExecuteGline(username, hostname, reason);
#endif /* HYBRID_GLINES */

	MyFree(reason);
} /* o_gline() */

/*
o_ungline()
	Removes gline av[1] from gline list
*/

static void
o_ungline(Luser *lptr, int ac, char **av, int sockfd)

{
	int gcnt;
	char line[MAXLINE], linetemp[MAXLINE];
	char tmpfile[MAXLINE];
	char *configname;
	char *key;
	char *ghost;
	char *user,
	     *host;
	Gline *gptr, *next;
	FILE *configfp;
	/*
	 * random file to store all lines except the ones
	 * which match the glines
	 */
	FILE *fp;

	if (ac < 2)
	{
		os_notice(lptr, sockfd, "Syntax: \002UNGLINE <user@host>\002");
		return;
	}

	o_RecordCommand(sockfd,
		"UNGLINE %s",
		av[1]);

	if (!(host = strchr(av[1], '@')))
	{
		user = (char *) NULL;
		host = av[1];
	}
	else
	{
		user = av[1];
		*host++ = '\0';
	}

	gcnt = 0;
	for (gptr = glinelist; gptr; )
	{
		if (user)
			if (!match(user, gptr->username))
			{
				gptr = gptr->next;
				continue;
			}

		if (match(host, gptr->hostname))
		{
			gcnt++;

			next = gptr->next;
			DeleteGline(gptr); /* remove gline from list */
			gptr = next;
		}
		else
			gptr = gptr->next;
	}

	if (gcnt == 0)
	{
		os_notice(lptr, sockfd, "No glines matching %s found", 
			av[1]);
		return;
	}

	if (GlineFile)
		configname = GlineFile;
	else
		configname = ConfigFile;

	if ((configfp = fopen(configname, "r")) == (FILE *) NULL)
	{
		os_notice(lptr, sockfd, "Unable to open config file");
		return;
	}

	sprintf(tmpfile, "%s.tmp",
		configname);

	fp = fopen(tmpfile, "w"); 
	while (fgets(line, MAXLINE - 1, configfp))
	{
		strcpy(linetemp, line);
		key = strtok(line, ":");
		if ((strcmp(key, "g") == 0) || (strcmp(key, "G") == 0))
		{
			ghost = strtok((char *) NULL, ":");
			if (match(av[1], ghost) == 0)
				fputs(linetemp, fp);
		}
		else
			fputs(linetemp, fp);
	} /* while () */
	fclose(configfp);
	fclose(fp);

	rename(tmpfile, configname);

	os_notice(lptr, sockfd, "Glines for %s removed [%d matches]", 
		av[1], 
		gcnt);
} /* o_ungline() */

#endif /* ALLOW_GLINES */

/*
o_help()
	Lets give this user some help :-)
*/

static void
o_help(Luser *lptr, int ac, char **av, int sockfd)

{
	if (ac >= 2)
	{
		o_Command		*cptr;

		for (cptr = opercmds; cptr->cmd; cptr++)
			if (!strcasecmp(av[1], cptr->cmd))
				break;

		if (cptr->cmd && cptr->flag)
			if (!CheckAccess(GetUser(1, onick, ouser, ohost), cptr->flag) || 
					!IsRegistered(lptr, sockfd))
			{
				os_notice(lptr, sockfd,
					ERR_NO_HELP,
					av[1],
					"");
				return;
			}

		o_RecordCommand(sockfd,
			"HELP %s",
			av[1]);

		GiveHelp(n_OperServ, lptr ? lptr->nick : (char *) NULL, av[1], sockfd);
	}
	else
	{
		o_RecordCommand(sockfd,
			"HELP");

		GiveHelp(n_OperServ, lptr ? lptr->nick : (char *) NULL, (char *) NULL, sockfd);
	}
} /* o_help() */

/*
o_join()
	Adds channel av[1] to list of monitored channels
*/

static void
o_join(Luser *lptr, int ac, char **av, int sockfd)

{
	FILE	*fp;
	Chanlist	*tempchan;

	if (ac < 2)
	{
		os_notice(lptr, sockfd, "Syntax: \002JOIN <#channel>\002");
		return;
	}

	o_RecordCommand(sockfd,
		"JOIN %s",
		av[1]);

	if ((tempchan = IsChannel(av[1])) != (Chanlist *) NULL)
	{
		os_notice(lptr, sockfd, "%s is already being monitored",
			tempchan->name);
		return;
	}

	AddMyChan(av[1]);
	os_join(FindChannel(av[1]));

	os_notice(lptr, sockfd, "Now monitoring %s", 
		av[1]);

	if ((fp = fopen(ConfigFile, "a")) == (FILE *) NULL)
	{
		os_notice(lptr, sockfd, "Unable to open config file");
		return;
	}
	fprintf(fp, "C:%s\n", av[1]);
	fclose(fp);
} /* o_join() */

/*
o_part()
	Removes channel av[1] to list of monitored channels
*/

static void
o_part(Luser *lptr, int ac, char **av, int sockfd)

{
	char line[MAXLINE], linetemp[MAXLINE];
	char *key, *ptemp;
	FILE *fp;
	FILE *configfp;
	char tmpfile[MAXLINE];
	Chanlist *chanptr, *tempchan, *prev;
	Channel *chptr;

	if (ac < 2)
	{
		os_notice(lptr, sockfd, "Syntax: \002PART <#channel>\002");
		return;
	}

	chanptr = IsChannel(av[1]);
	chptr = FindChannel(av[1]);

	if (!chanptr || !chptr)
	{
		os_notice(lptr, sockfd, "%s is not being monitored", 
			av[1]);
		return;
	}

	o_RecordCommand(sockfd,
		"PART %s",
		chanptr->name);

	os_part(chptr);

	/*
	 * Remove channel from mychanlist
	 */

	prev = (Chanlist *) NULL;

	for (tempchan = mychanlist; tempchan; )
	{
		if (tempchan == chanptr)
		{
			MyFree(tempchan->name);

			if (prev)
			{
				prev->next = tempchan->next;
				MyFree(tempchan);
				tempchan = prev;
			}
			else
			{
				mychanlist = tempchan->next;
				MyFree(tempchan);
				tempchan = (Chanlist *) NULL;
			}

			Network->MyChans--;

			/*
			 * There should be only one match, so break
			 */
			break;
		}

		prev = tempchan;

		if (tempchan)
			tempchan = tempchan->next;
		else
			tempchan = mychanlist;
	}

	/* Remove channel from config file */
	if ((configfp = fopen(ConfigFile, "r")) == NULL)
	{
		os_notice(lptr, sockfd, "Unable to open config file");
		return;
	}

	sprintf(tmpfile, "%s.tmp",
		ConfigFile);
	fp = fopen(tmpfile, "w"); 

	while (fgets(line, MAXLINE - 1, configfp))
	{
		strcpy(linetemp, line);
		key = strtok(line, ":");
		if ((strcmp(key, "c") == 0) || (strcmp(key, "C") == 0))
		{
			ptemp = strtok((char *) NULL, "\r\n");
			if (strncasecmp(ptemp, av[1], strlen(av[1])) != 0)
				fputs(linetemp, fp);
		}
		else
			fputs(linetemp, fp);
	} /* while () */

	fclose(fp);
	fclose(configfp);

	rename(tmpfile, ConfigFile);

	os_notice(lptr, sockfd, "No longer monitoring %s", 
		av[1]);
} /* o_part() */

/*
o_clones()
	Displays current clones on the network
*/

static void
o_clones(Luser *lptr, int ac, char **av, int sockfd)

{
	int ii,
	    doneset,
	    cnt,
	    clcnt;
	Luser *tempuser;

	o_RecordCommand(sockfd,
		"CLONES");

	os_notice(lptr, sockfd, "-- Processing --");

	clcnt = 0;
	for (ii = 0; ii < HASHCLIENTS; ii++)
	{
		if (cloneTable[ii].list == (void *) NULL)
			continue;

		for (tempuser = cloneTable[ii].list; tempuser; tempuser = tempuser->cnext)
		{
			/* we know there must be at least 1 user in the list */
			if (tempuser->cnext)
			{
				if (!CloneMatch(tempuser, tempuser->cnext))
					continue;

				doneset = 0;

				cnt = 1;
				while (1)
				{
					if (!doneset)
					{
						/*
						 * Only print out the heading (user@hostname) once per
						 * batch of clones. The next time we get here,
						 * doneset will be 1, so we won't print out the
						 * heading again
						 */
						os_notice(lptr, sockfd, "%s@%s",
							tempuser->username[0] == '~' ? tempuser->username + 1 : tempuser->username,
							tempuser->hostname);
						os_notice(lptr, sockfd, "  %d) %-10s [%s]", 
							cnt++, 
							tempuser->nick, 
							tempuser->server ? tempuser->server->name : "*unknown*");
						doneset = 1;

						++clcnt;
					} /* if (!doneset) */

					os_notice(lptr, sockfd, "  %d) %-10s [%s]", 
						cnt++, 
						tempuser->cnext->nick, 
						tempuser->cnext->server ? tempuser->cnext->server->name : "*unknown*");
					
					++clcnt;

					if (tempuser->cnext->cnext)
					{
						if (CloneMatch(tempuser->cnext, tempuser->cnext->cnext))
						{
							/*
							 * We have yet another clone, continue the loop
							 */
							tempuser = tempuser->cnext;
						}
						else
						{
							/*
							 * All clones are grouped one after another. Since
							 * we found two clients that do not match, there
							 * can be no more possible clones of tempuser,
							 * break out of loop
							 */
							break;
						}
					} /* if (tempuser->cnext->cnext) */
					else
					{
						/*
						 * next user is NULL, break out of while loop
						 */
						break;
					}
				} /* while (1) */
			} /* if (tempuser->cnext) */
		} /* for (tempuser = cloneTable[ii].list; tempuser; tempuser = tempuser->cnext) */
	} /* for (ii = 0; ii < HASHCLIENTS; ii++) */

	os_notice(lptr, sockfd,
		"-- End of list (%d total clones found) --",
		clcnt);
} /* o_clones() */

#ifdef ALLOW_FUCKOVER

/*
o_fuckover()
	flood client av[1] with ircd codes until it pings out or sendq
is overflowed
*/

static void
o_fuckover(Luser *lptr, int ac, char **av, int sockfd)

{
	int 	ii;
	Luser	*fptr;

	if (ac < 2)
	{
		os_notice(lptr, sockfd, "Syntax: \002FUCKOVER <nickname>\002");
		return;
	}

	fptr = FindClient(av[1]);
	if (fptr == (Luser *) NULL)
	{
		os_notice(lptr, sockfd, "No such nick [%s]",
			av[1]);
		return;
	}

	o_RecordCommand(sockfd,
		"FUCKOVER %s",
		fptr->nick);

	if (fptr->server == Me.sptr)
	{
		os_notice(lptr, sockfd, "[%s] is a psuedo nickname, canceling flood",
			fptr->nick);
		return;
	}

	if (fptr->flags & L_OSREGISTERED)
		ii = 1;
	else
		ii = 0;

	/* Check if the user has an exception flag */
	if (IsProtected(GetUser(ii, fptr->nick, fptr->username, fptr->hostname)))
	{
		/* The user is protected */
		os_notice(lptr, sockfd, "%s!%s@%s is protected", 
			fptr->nick,
			fptr->username,
			fptr->hostname);

		putlog(LOG2, "%s!%s@%s attempted to FUCKOVER protected user %s",
			onick,
			ouser,
			ohost,
			av[1]);

		return;
	}

	/* check if av[1] is already being flooded */
	for (ii = 0; ii < fprocs_num; ii++)
		if (fprocs[ii])
			if (!strcasecmp(av[1], fprocs[ii]->target))
			{
				os_notice(lptr, sockfd, "Server flood already in progress for [%s]",
					av[1]);
				return;
			}

	init_fprocess(onick, fptr->nick); /* start the flood */

	SendUmode(OPERUMODE_Y,
		"*** Server flood activated for [%s]", 
		av[1]);
} /* o_fuckover() */

/*
init_fprocess()
	backend for o_fuckover - fork() a process to flood fptr->nick,
and add it to process table fprocs
*/

static void
init_fprocess(char *from, char *ftarget)

{
	int		pid;

	if (!from || !ftarget)
		return; /* shouldn't happen */

	pid = fork();
	switch (pid)
	{
		case -1: return;

		/* 
		 * Child: flood the target
		 */
		case 0:
		{
			time_t fstart, timecheck;
			int		ii, stop = 0;

			fstart = time(NULL);
			while (!stop)
			{
				for (ii = 200; ii < 512; ii++)
				{
					/*
					 * stop flood after 20 secs in case the client can handle it
					 * *laff*
					 */
					timecheck = time(NULL);
					if ((timecheck - fstart) > 20)
					{
						SendUmode(OPERUMODE_Y,
							"*** Time limit expired, canceling flood");

						stop = 1;
						break;
					}

					/* send null string to target */
					toserv(":%s %d %s :\n", 
						Me.name,
						ii,
						ftarget);
				}
			}

			exit(-1);
			break;
		} /* case 0: */

		/* 
		 * Parent: add sub-process to fprocs table
		 */
		default:
		{
			int		ii;
			Process	*newproc = (Process *) MyMalloc(sizeof(Process));

			/* Initialize table if necessary */
			if (!fprocs)
			{
				fprocs = (Process **) MyMalloc(sizeof(Process *));
				fprocs_num = 1;
				fprocs[0] = NULL;
			}

			/* Find first empty slot */
			for (ii = 0; ii < fprocs_num; ii++)
				if (!fprocs[ii])
				{
					fprocs[ii] = newproc;
					break;
				}

			/* no available slots? make a new one */
			if (ii == fprocs_num)
			{
				fprocs_num++;
				fprocs = MyRealloc(fprocs, sizeof(Process *) * fprocs_num);
				fprocs[ii] = newproc;
			}

			newproc->who = MyStrdup(from);
			newproc->target = MyStrdup(ftarget);
			newproc->pid = pid;
			newproc->index = ii;

			break;
		}
	} /* switch (pid) */
} /* init_fprocess() */

/*
check_ftarget()
	Called when a user quits, or changes their nick (newnick != NULL) to
see if they are a target of o_fuckover() - if so, halt the process
(and if newnick != NULL, redirect the flood to the new nick).
*/

void
check_ftarget(Luser *fptr, char *newnick)

{
	int ii,
      killret;

	if (!fptr)
		return;

	for (ii = 0; ii < fprocs_num; ii++)
		if (fprocs[ii]) /* make sure its not null */
		{
			if (!strcasecmp(fptr->nick, fprocs[ii]->target))
			{
				/* we have a match - kill the process id */
				killret = kill(fprocs[ii]->pid, SIGKILL);

				/* if its a nick change, start flooding the new nick */
				if (newnick != (char *) NULL)
				{
					SendUmode(OPERUMODE_Y,
						"*** Redirecting flood to [%s]",
						newnick);
					init_fprocess(fprocs[ii]->who, newnick);
				}
				else if (killret == 0)
					SendUmode(OPERUMODE_Y,
						"*** Flood successful");

				MyFree(fprocs[ii]->who);
				MyFree(fprocs[ii]->target);
				fprocs[ii] = NULL;

				break;
			}
		}
} /* check_ftarget() */

#endif /* ALLOW_FUCKOVER */

/*
o_trace()
	Displays user information on all users matching a certain
pattern, as well as certain options, if specified

Options:
  -ops      Limit matches to irc operators
  -nonops   Limit matches to non irc operators
  -clones   Limit matches to clones
  -kill     Kill all matches
  -info     Display statistical information for matches
  -realname Limit matches to those who match the given realname
  -server   Limit matches to those who are on the given server
  -proxy    Limit matches to those running insecure proxies
  -msg      Send all matches the given message
  -long     Display in long format (default to short)
  -nolimit  Ignore MaxTrace limit
*/

static void
o_trace(Luser *lptr, int ac, char **av, int sockfd)

{
	char *target;
	char chkstr[MAXLINE];
	char *argbuf;
	int argmem,
	    alen;
	Luser *lptr2, *prev;
	int	cnt,
			bad,
			exactmatch; /* 1 if target is a user's nick, not a hostmask */
	int ops, /* 1 if -ops, 2 if -nonops */
			showinfo, /* show statistics about matches */
			clones, /* show only clones */
			kill, /* kill matches */
			showlong, /* show matches in long format */
			nolimit; /* ignore MaxTrace limit */
	int showmatches;

	struct timeval TimeOut;
	int selret; /* select() return value */
	fd_set writefds;

#ifdef HAVE_PTHREADS
	int proxy; /* show only insecure proxy people */
#endif

	Server *servptr;
	char *realname;
	char msgbuf[MAXLINE];
		
	target = (char *) NULL;
	servptr = (Server *) NULL;
	realname = (char *) NULL;
	*msgbuf = '\0';
	ops = 0;
	showinfo = 0;
	clones = 0;
	kill = 0;
	showlong = 0;
	nolimit = 0;

#ifdef HAVE_PTHREADS
	proxy = 0;
#endif

	showmatches = 1;

	for (cnt = 1; cnt < ac; cnt++)
	{
		alen = strlen(av[cnt]);

		if (!strncasecmp(av[cnt], "-ops", alen))
			ops = 1;
		else if (!strncasecmp(av[cnt], "-nonops", alen))
			ops = 2;
		else if (!strncasecmp(av[cnt], "-clones", alen))
			clones = 1;
		else if (!strncasecmp(av[cnt], "-kill", alen))
			kill = 1;
		else if (!strncasecmp(av[cnt], "-long", alen))
			showlong = 1;
		else if (!strncasecmp(av[cnt], "-nolimit", alen))
			nolimit = 1;
	#ifdef HAVE_PTHREADS
		else if (!strncasecmp(av[cnt], "-proxy", alen))
			proxy = 1;
	#endif /* HAVE_PTHREADS */
		else if (!strncasecmp(av[cnt], "-info", alen))
		{
		#ifndef STATSERVICES
			os_notice(lptr, sockfd, "Stat Services are not enabled");
			return;
		#endif

			showinfo = 1;
		}
		else if (!strncasecmp(av[cnt], "-realname", alen))
		{
			cnt++;
			if (cnt >= ac)
			{
				os_notice(lptr, sockfd, "No realname specified");
				return;
			}
			realname = av[cnt];
		}
		else if (!strncasecmp(av[cnt], "-server", alen))
		{
			cnt++;
			if (cnt >= ac)
			{
				os_notice(lptr, sockfd, "No server specified");
				return;
			}

			if (!(servptr = FindServer(av[cnt])))
			{
				os_notice(lptr, sockfd, "No such server: %s",
					av[cnt]);
				return;
			}
		}
		else if (!strncasecmp(av[cnt], "-msg", alen))
		{
			cnt++;
			if (cnt >= ac)
			{
				os_notice(lptr, sockfd, "No message specified");
				return;
			}

			while (cnt < ac)
			{
				strncat(msgbuf, av[cnt++], sizeof(msgbuf) - strlen(msgbuf) - 20);
				strcat(msgbuf, " ");
			}
			msgbuf[strlen(msgbuf) - 1] = '\0';

			break;
		}
		else
		{
			/*
			 * it wasn't an -option; assume it is the pattern they
			 * want to trace
			 */
			if (!target)
				target = av[cnt];
		}
	}

	if (!target)
	{
		os_notice(lptr, sockfd, "No pattern specified");
		os_notice(lptr, sockfd,
			"Syntax: \002TRACE <pattern> [options]\002");
		return;
	}

	argmem = 200;
	if (realname)
		argmem += strlen(realname) + 1;
	if (servptr)
		argmem += strlen(servptr->name) + 1;
	if (*msgbuf)
		argmem += strlen(msgbuf) + 1;

	argbuf = (char *) MyMalloc(argmem + 4);

	sprintf(argbuf, "[%s] ",
		target);

	if (showlong)
		strcat(argbuf, "-long ");

	if (nolimit)
		strcat(argbuf, "-nolimit ");

	if (ops == 1)
		strcat(argbuf, "-ops ");
	else if (ops == 2)
		strcat(argbuf, "-nonops ");

	if (clones)
		strcat(argbuf, "-clones ");

	if (showinfo)
		strcat(argbuf, "-info ");

#ifdef HAVE_PTHREADS
	if (proxy)
		strcat(argbuf, "-proxy ");
#endif

	if (realname)
	{
		strcat(argbuf, "-realname ");
		strcat(argbuf, realname);
		strcat(argbuf, " ");
	}

	if (servptr)
	{
		strcat(argbuf, "-server ");
		strcat(argbuf, servptr->name);
		strcat(argbuf, " ");
	}

	if (kill)
		strcat(argbuf, "-kill ");

	if (*msgbuf)
	{
		strcat(argbuf, "-msg ");
		strcat(argbuf, msgbuf);
	}

	o_RecordCommand(sockfd,
		"TRACE %s",
		argbuf);

	MyFree(argbuf);

	if (sockfd != NODCC)
	{
		/*
		 * If it is a telnet connection, use select() to determine
		 * if the socket is writeable. Otherwise, if they do
		 * a trace -nolimit, and error could possibly occur if
		 * we fill up the sendq buffer.
		 */
		TimeOut.tv_sec = 1;
		TimeOut.tv_usec = 0;
		FD_ZERO(&writefds);

		FD_SET(sockfd, &writefds);
	}

	exactmatch = 0;

	if ((lptr2 = FindClient(target)))
		exactmatch = 1;

	cnt = 0;
	bad = 0;
	prev = (Luser *) NULL;

	if (!exactmatch)
		lptr2 = ClientList;

	for ( ; lptr2; )
	{
		if ((ops == 1) && (!IsOperator(lptr2)))
			bad = 1; /* -ops */
		else if ((ops == 2) && (IsOperator(lptr2)))
			bad = 1; /* -nonops */
		else if ((servptr) && (lptr2->server) && (servptr != lptr2->server))
			bad = 1; /* -server */
		else if ((clones == 1) && !IsClone(lptr2))
			bad = 1; /* -clones */
	#ifdef HAVE_PTHREADS
		else if ((proxy == 1) && !IsProxy(lptr2))
			bad = 1; /* -proxy */
	#endif /* HAVE_PTHREADS */
		else if ((realname) && (lptr2->realname))
		{
			if (match(realname, lptr2->realname) == 0)
				bad = 1; /* -realname */
		}

		if (bad)
		{
			/* this user is not a good match */

			if (exactmatch)
			{
				/*
				 * The given target was the nickname of a valid client,
				 * not a userhost mask - since the client didn't pass
				 * the [options] test, stop the loop
				 */
				break;
			}

			bad = 0;
			lptr2 = lptr2->next;
			continue;
		}

		sprintf(chkstr, "%s!%s@%s",
			lptr2->nick,
			lptr2->username,
			lptr2->hostname);

		if (exactmatch || (match(target, chkstr)))
		{
			cnt++;
			if (*msgbuf) /* -msg */
			{
				/*
				 * Make sure we don't send a message to a service
				 * nick - ircd would assume a ghost and issue a KILL
				 * if that were to happen
				 */
				if (lptr2->server == Me.sptr)
				{
					cnt--;

					if (exactmatch)
						break;

					lptr2 = lptr2->next;
					continue;
				}

				if (!nolimit && (cnt > MaxTrace))
				{
					os_notice(lptr, sockfd,
						"-- More than %d matches found, halting messages --", 
						MaxTrace);
					showmatches = 0;
					break;
				}

				os_notice(lptr, sockfd,
					"** Noticing %s",
					lptr2->nick);

				/* send 'msgbuf' to matches through a NOTICE */
				toserv(":%s NOTICE %s :%s\n",
					n_OperServ,
					lptr2->nick,
					msgbuf);
			}
			else if (kill) /* -kill */
			{
				/* make sure we don't kill a service nick */
				if (lptr2->server == Me.sptr)
				{
					cnt--;

					if (exactmatch)
						break;

					lptr2 = lptr2->next;
					continue;
				}

				if (!nolimit && (cnt > MaxKill))
				{
					os_notice(lptr, sockfd,
						"-- More than %d matches found, halting kills --",
						MaxKill);
					showmatches = 0;
					break;
				}

				os_notice(lptr, sockfd,
					"** Killing %s",
					lptr2->nick);

				toserv(":%s KILL %s :%s!%s (#%d (%s@%s))\n",
					n_OperServ,
					lptr2->nick,
					Me.name,
					n_OperServ,
					cnt,
					onick,
					n_OperServ);

				DeleteClient(lptr2);

				if (exactmatch)
					break;

				if (prev)
					lptr2 = prev;
				else
				{
					lptr2 = ClientList;
					continue;
				}
			}
			else
			{
				if (!nolimit && (cnt > MaxTrace))
				{
					os_notice(lptr, sockfd,
						"-- More than %d matches found, truncating list --", 
						MaxTrace);
					showmatches = 0;
					break;
				}

				if (sockfd != NODCC)
				{
					selret = select(FD_SETSIZE, NULL, &writefds, NULL, &TimeOut);
					if (selret < 0)
					{
						os_notice(lptr, sockfd,
							"-- Error occurred during write: %s --",
							strerror(errno));
						break;
					}
				}

				os_notice(lptr, sockfd, "-----");

				if (showinfo) /* -info */
				{
				#ifdef STATSERVICES
					show_trace_info(lptr, lptr2, sockfd);
				#endif
				}
				else
				{
					/* display normally */
					show_trace(lptr, lptr2, sockfd, showlong);
				}
			}
		} /* if (match(target, chkstr)) */

		if (exactmatch)
			break;

		prev = lptr2;
		lptr2 = lptr2->next;
	} /* for ( ; lptr2; ) */

	if (showmatches)
	{
		os_notice(lptr, sockfd, "%d match%s found",
			cnt,
			(cnt == 1) ? "" : "es");
	}
} /* o_trace() */

/*
show_trace()
	Show o_trace info
*/

static void
show_trace(Luser *lptr, Luser *target, int sockfd, int showlong)

{
	UserChan	*tempchan;
	char	tmp[MAXLINE];

	if (!target)
		return;

	if (!showlong)
	{
		/*
		 * Display in short format: nick!user@host (realname) [server]
		 */
		os_notice(lptr, sockfd,
			"%s!%s@%s (%s) [%s]",
			target->nick,
			target->username,
			target->hostname,
			target->realname,
			target->server ? target->server->name : "*unknown*");
		return;
	}

	os_notice(lptr, sockfd,
		"User:      %s (%s@%s)",
		target->nick,
		target->username,
		target->hostname);
	os_notice(lptr, sockfd,
		"Realname:  %s",
		target->realname);
	os_notice(lptr, sockfd,
		"Server:    %s",
		target->server ? target->server->name : "*unknown*");

	strcpy(tmp, "+");
	if (target->umodes & UMODE_I)
		strcat(tmp, "i");
	if (target->umodes & UMODE_S)
		strcat(tmp, "s");
	if (target->umodes & UMODE_W)
		strcat(tmp, "w");
	if (target->umodes & UMODE_O)
		strcat(tmp, "o");
	os_notice(lptr, sockfd,
		"Usermodes: %s",
		tmp);

	strcpy(tmp, ctime((time_t *) &target->since));
	tmp[strlen(tmp) - 1] = '\0';
	os_notice(lptr, sockfd,
		"Signon:    %s",
		tmp);

	if (target->firstchan)
	{
		if (sockfd == NODCC)
			toserv(":%s NOTICE %s :Channels:  ",
				(ServerNotices) ? Me.name : n_OperServ,
				lptr->nick);
		else
			writesocket(sockfd, "Channels:  ");

		for (tempchan = target->firstchan; tempchan; tempchan = tempchan->next)
		{
			if (tempchan->flags & CH_OPPED)
			{
				if (sockfd == NODCC)
					toserv("@");
				else
					writesocket(sockfd, "@");
			}
			if (tempchan->flags & CH_VOICED)
			{
				if (sockfd == NODCC)
					toserv("+");
				else
					writesocket(sockfd, "+");
			}
			if (sockfd == NODCC)
				toserv("%s ",
					tempchan->chptr->name);
			else
			{
				writesocket(sockfd, tempchan->chptr->name);
				writesocket(sockfd, " ");
			}
		} /* for (tempchan = target->firstchan; tempchan; tempchan = tempchan->next) */

		if (sockfd == NODCC)
			toserv("\n");
		else
			writesocket(sockfd, "\n");
	}
} /* show_trace() */

#ifdef STATSERVICES

static void
show_trace_info(Luser *lptr, Luser *target, int sockfd)

{
	os_notice(lptr, sockfd,
		"User:         %s (%s@%s)",
		target->nick,
		target->username,
		target->hostname);

	os_notice(lptr, sockfd,
		"Ops:          %ld",
		target->numops);
	os_notice(lptr, sockfd,
		"DeOps:        %ld",
		target->numdops);
	os_notice(lptr, sockfd,
		"Voices:       %ld",
		target->numvoices);
	os_notice(lptr, sockfd,
		"DeVoices:     %ld",
		target->numdvoices);
	os_notice(lptr, sockfd,
		"Kicks:        %ld",
		target->numkicks);
	os_notice(lptr, sockfd,
		"Nick Changes: %ld",
		target->numnicks);
	os_notice(lptr, sockfd,
		"Kills:        %ld",
		target->numkills);
} /* show_trace_info() */

#endif /* STATSERVICES */

/*
o_channel() 
 Displays channel information on channels matching <pattern>
and [options]

Options:
 -minimum   - limit matches to channels with at least the given
              number of users
 -maximum   - limit matches to channels with at most the given
              number of users
 -banmatch  - limit matches to channels who have the given
              hostmask on their ban list
 -exmatch   - limit matches to channels who have the given
              hostmask on their ban exception list
 -nolimit   - Ignore MaxChannel limit
*/

static void
o_channel(Luser *lptr, int ac, char **av, int sockfd)

{
	Channel	*cptr;
	char	argbuf[MAXLINE];
	int		cnt, /* number of matches */
				bad,
				exactmatch;
	char  *target;
	int   min, /* -min */
	      max; /* -max */
	char  *banmatch, /* -banmatch */
	      *exmatch; /* -exmatch */
	int nolimit; /* -nolimit */
  char  temp[MAXLINE];
  int alen,
      showmatches;
		
	if (ac < 2)
	{
		os_notice(lptr, sockfd, "Syntax: \002CHANNEL <pattern> [options]\002");
		return;
	}

	target = (char *) NULL;
	min = 0;
	max = 0;
	banmatch = (char *) NULL;
	exmatch = (char *) NULL;
	nolimit = 0;

	showmatches = 1;

	for (cnt = 1; cnt < ac; cnt++)
	{
		alen = strlen(av[cnt]);

		if (!strncasecmp(av[cnt], "-nolimit", alen))
			nolimit = 1;
		if (!strncasecmp(av[cnt], "-minimum", alen))
		{
			cnt++;
			if (cnt >= ac)
			{
				os_notice(lptr, sockfd, "No minimum value specified");
				return;
			}

			min = atoi(av[cnt]);
		}
		else if (!strncasecmp(av[cnt], "-maximum", alen))
		{
			cnt++;
			if (cnt >= ac)
			{
				os_notice(lptr, sockfd, "No maximum value specified");
				return;
			}

			max = atoi(av[cnt]);
		}
		else if (!strncasecmp(av[cnt], "-banmatch", alen))
		{
			cnt++;
			if (cnt >= ac)
			{
				os_notice(lptr, sockfd, "No hostmask specified");
				return;
			}

			banmatch = av[cnt];
		}
		else if (!strncasecmp(av[cnt], "-exmatch", alen))
		{
			cnt++;
			if (cnt >= ac)
			{
				os_notice(lptr, sockfd, "No hostmask specified");
				return;
			}

			exmatch = av[cnt];
		}
		else
		{
			/*
			 * it wasn't an -option; assume it is the pattern they
			 * want to use
			 */
			if (!target)
				target = av[cnt];
		}
	}

	if (!target)
	{
		os_notice(lptr, sockfd, "No pattern specified");
		os_notice(lptr, sockfd,
			"Syntax: \002CHANNEL <pattern> [options]\002");
		return;
	}

	sprintf(argbuf, "[%s] ",
		target);

	if (nolimit)
		strcat(argbuf, "-nolimit ");

	if (min)
	{
		strcat(argbuf, "-minimum");
		sprintf(temp, "%s %d ",
			argbuf,
			min);
		strcpy(argbuf, temp);
	}

	if (max)
	{
		strcat(argbuf, "-maximum");
		sprintf(temp, "%s %d ",
			argbuf,
			max);
		strcpy(argbuf, temp);
	}

	if (banmatch)
	{
		strcat(argbuf, "-banmatch ");
		strcat(argbuf, banmatch);
		strcat(argbuf, " ");
	}

	if (exmatch)
	{
		strcat(argbuf, "-exmatch ");
		strcat(argbuf, exmatch);
		strcat(argbuf, " ");
	}

	o_RecordCommand(sockfd,
		"CHANNEL %s",
		argbuf);

	cnt = 0;
	bad = 0;
	exactmatch = 0;

	if ((cptr = FindChannel(target)))
		exactmatch = 1;

	if (!exactmatch)
		cptr = ChannelList;

	for (; cptr; cptr = cptr->next)
	{
		if (exactmatch || (match(target, cptr->name)))
		{
			if ((min) && (min > cptr->numusers)) /* -minimum */
				bad = 1;
			else if ((max) && (max < cptr->numusers)) /* -maximum */
				bad = 1;
			else if ((banmatch) && !MatchBan(cptr, banmatch)) /* -banmatch */
				bad = 1;
			else if ((exmatch) && !MatchException(cptr, exmatch)) /* -exmatch */
				bad = 1;

			if (bad)
			{
				/*
				 * cptr did not pass the [options] screening
				 */
				if (exactmatch)
					break;

				bad = 0;
				continue;
			}

			cnt++;

			if (!nolimit && (cnt > MaxChannel))
			{
				os_notice(lptr, sockfd,
					"--- More than %d matches found, truncating list ---", 
					MaxChannel);
				showmatches = 0;
				break;
			}

			show_channel(lptr, cptr, sockfd);
		} /* if (match(target, tempchan->name)) */

		if (exactmatch)
			break;
	} /* for (; cptr; cptr = cptr->next) */

	if (showmatches)
	{
		os_notice(lptr, sockfd, "%d match%s found",
			cnt,
			(cnt == 1) ? "" : "es");
	}
} /* o_channel() */

/*
show_channel()
	Show o_channel info
*/

static void
show_channel(Luser *lptr, Channel *cptr, int sockfd)

{
	BanInfo	*tempban;
	Except	*tempe;
	ChanUser	*tempuser;
	char	*btime; /* time ban was set */
	char	modes[MAXLINE]; /* channel modes */
	int		bcnt = 0, /* ban count */
	      ecnt = 0, /* exception count */
				ocnt = 0, /* chan ops */
				vcnt = 0, /* chan voices */
				icnt = 0, /* chan ircops */
				ncnt = 0; /* nonops */

	os_notice(lptr, sockfd, "Channel Information for %s:", 
		cptr->name);
	strcpy(modes, "+");
	if (cptr->modes & MODE_S)
		strcat(modes, "s");
	if (cptr->modes & MODE_P)
		strcat(modes, "p");
	if (cptr->modes & MODE_N)
		strcat(modes, "n");
	if (cptr->modes & MODE_T)
		strcat(modes, "t");
	if (cptr->modes & MODE_M)
		strcat(modes, "m");
	if (cptr->modes & MODE_I)
		strcat(modes, "i");
	if (cptr->limit)
		strcat(modes, "l");
	if (cptr->key && *cptr->key)
		strcat(modes, "k");

	if (cptr->limit)
	{
		char	temp[MAXLINE];

		sprintf(temp, "%s %d", modes, cptr->limit);
		strcpy(modes, temp);
	}

	if (cptr->key && *cptr->key)
	{
		char	temp[MAXLINE];

		sprintf(temp, "%s %s", modes, cptr->key);
		strcpy(modes, temp);
	}

	os_notice(lptr, sockfd, 
		"Modes:     %s",
		modes);
	strcpy(modes, ctime(&cptr->since));
	modes[strlen(modes) - 1] = '\0';
	os_notice(lptr, sockfd,
		"Created:   %s",
		modes);

	ocnt = ncnt = vcnt = icnt = bcnt = 0;
	for (tempuser = cptr->firstuser; tempuser; tempuser = tempuser->next)
	{
		if (tempuser->flags & CH_OPPED)
			ocnt++;
		else
			ncnt++;
		if (tempuser->flags & CH_VOICED)
			vcnt++;
		if (tempuser->lptr->umodes & UMODE_O)
			icnt++;
	}

	os_notice(lptr, sockfd, 
		"Ops:       %d",
		ocnt);
	os_notice(lptr, sockfd,
		"Voices:    %d",
		vcnt);
	os_notice(lptr, sockfd,
		"NonOps:    %d",
		ncnt);
	os_notice(lptr, sockfd,
		"IrcOps:    %d",
		icnt);
	os_notice(lptr, sockfd,
		"Total:     %d",
		cptr->numusers);

	if (sockfd == NODCC)
		toserv(":%s NOTICE %s :Nicks:     ",
			(ServerNotices) ? Me.name : n_OperServ,
			lptr->nick);
	else
		writesocket(sockfd, "Nicks:     ");
	for (tempuser = cptr->firstuser; tempuser; tempuser = tempuser->next)
	{
		if (tempuser->flags & CH_OPPED)
		{
			if (sockfd == NODCC)
				toserv("@");
			else
				writesocket(sockfd, "@");
		}
		if (tempuser->flags & CH_VOICED)
		{
			if (sockfd == NODCC)
				toserv("+");
			else
				writesocket(sockfd, "+");
		}
		if (sockfd == NODCC)
			toserv("%s ",
				tempuser->lptr->nick);
		else
		{
			writesocket(sockfd, tempuser->lptr->nick);
			writesocket(sockfd, " ");
		}
	}
	if (sockfd == NODCC)
		toserv("\n");
	else
		writesocket(sockfd, "\n");

	if (cptr->firstban != (BanInfo *) NULL)
	{
		os_notice(lptr, sockfd, "Bans:");
		for (tempban = cptr->firstban; tempban; tempban = tempban->next)
		{
			bcnt++;
			btime = (char *) ctime(&tempban->when);
			btime[strlen(btime) - 1] = '\0';
			os_notice(lptr, sockfd, " [%2d] [%-10s] [%-10s] [%-15s]",
				bcnt,
				tempban->mask,
				tempban->who ? tempban->who : "unknown",
				btime);
		}
	}

	if (cptr->exceptlist)
	{
		os_notice(lptr, sockfd, "Ban Exceptions:");
		for (tempe = cptr->exceptlist; tempe; tempe = tempe->next)
		{
			ecnt++;
			btime = (char *) ctime(&tempe->when);
			btime[strlen(btime) - 1] = '\0';
			os_notice(lptr, sockfd, " [%2d] [%-10s] [%-10s] [%-15s]",
				ecnt,
				tempe->mask,
				tempe->who ? tempe->who : "unknown",
				btime);
		}
	}
} /* show_channel() */

/*
o_on()
	Reactivate Services
*/

static void
o_on(Luser *lptr, int ac, char **av, int sockfd)

{
	if (!(Network->flags & NET_OFF))
	{
		os_notice(lptr, sockfd,
			"Services are not disabled");
		return;
	}

	Network->flags &= ~NET_OFF;
	os_notice(lptr, sockfd,
		"Services have been \002reactivated\002");

	o_RecordCommand(sockfd,
		"ON");
} /* o_on() */

/*
o_off()
	Deactivate services
*/

static void
o_off(Luser *lptr, int ac, char **av, int sockfd)

{
	if (Network->flags & NET_OFF)
	{
		os_notice(lptr, sockfd,
			"Services are already disabled");
		return;
	}

	Network->flags |= NET_OFF;
	os_notice(lptr, sockfd,
		"Services have been \002disabled\002");

	o_RecordCommand(sockfd,
		"OFF");
} /* o_off() */

/*
o_jump()
	Reroute to another hub server (av[1])
*/

static void
o_jump(Luser *lptr, int ac, char **av, int sockfd)

{
	int		port, tempsock;
	Servlist	*temp;

	if (ac < 2)
	{
		os_notice(lptr, sockfd,
			"Syntax: \002JUMP {server} [port]\002");
		return;
	}

	for (temp = serverlist; temp; temp = temp->next)
		if (match(av[1], temp->hostname))
			break;

	if (!temp)
	{
		os_notice(lptr, sockfd,
			"[\002%s\002] is an invalid hub server",
			av[1]);
		return;
	}

	if (ac >= 3)
	{
		if (!(port = IsNum(av[2])))
		{
			os_notice(lptr, sockfd, "Invalid port [\002%s\002]",
				av[2]);
			return;
		}
	}
	else
		port = temp->port;

	o_RecordCommand(sockfd,
		"JUMP %s %d",
		temp->hostname,
		port);

	/* Attempt connection to server */
	if ((tempsock = ConnectToServer(temp->hostname, port)) >= 0) 
	{

	#ifdef HAVE_PTHREADS
		/*
		 * Set this to 0, so p_CheckTimer won't call DoTimer()
		 * which might cycle the server list
		 */
		/*GoodTimer = 0;*/
	#endif

		toserv("SQUIT %s :ReRouting\n",
			currenthub->realname ? currenthub->realname : "*");

		/* kill old connection and clear out user/chan lists etc */
		close(HubSock);
		ClearUsers();
		ClearChans();
		ClearServs();

		/*
		 * This has to come after ClearUsers(), or RECORD_SPLIT_TS
		 * will not work correctly
		 */
		ClearHashes(0);

		HubSock = tempsock;

		currenthub->connect_ts = 0;
		temp->connect_ts = time(NULL);
		currenthub = temp;

		/* send PASS/SERVER combo */
		signon();

	#ifdef HAVE_PTHREADS
		/*GoodTimer = 1;*/
	#endif

	}
	else
	{
		close(tempsock);
		os_notice(lptr, sockfd,
			"Unable to connect to port \002%d\002 of \002%s\002: %s",
			port,
			temp->hostname,
			strerror(errno));
		return;
	}
} /* o_jump() */

/*
o_save()
	Save nickname/channel databases or settings.conf
depending on arguements
Valid Options:
  -dbs    : Save all databases
  -sets   : Save all settings
  -all    : Save everything
*/

static void
o_save(Luser *lptr, int ac, char **av, int sockfd)

{
	int goodsave,
	    ii,
	    alen;
	int savesets, /* -sets */
	    savedbs; /* -dbs */
	char argbuf[MAXLINE];

	savesets = 0;
	savedbs = 0;

	if (ac >= 2)
	{
		for (ii = 1; ii < ac; ii++)
		{
			alen = strlen(av[ii]);
			if (!strncasecmp(av[ii], "-dbs", alen))
				savedbs = 1;
			else if (!strncasecmp(av[ii], "-sets", alen))
				savesets = 1;
			else if (!strncasecmp(av[ii], "-all", alen))
				savedbs = savesets = 1;
		}
	}
	else
		savedbs = 1; /* default to database save */

	argbuf[0] = '\0';
	if (savedbs && savesets)
		strcat(argbuf, "-all ");
	else if (savedbs)
		strcat(argbuf, "-databases ");
	else if (savesets)
		strcat(argbuf, "-settings ");

	o_RecordCommand(sockfd,
		"SAVE %s",
		argbuf);

	if (savedbs)
	{
		goodsave = WriteDatabases();

		if (goodsave)
		{
			os_notice(lptr, sockfd,
				"Databases have been saved");
		}
		else
		{
			os_notice(lptr, sockfd,
				"Database save failed");

			/* 
			 * we don't need to log the failure, since WriteDatabases()
			 * would have done so
			 */
		}
	} /* if (savedbs) */

	if (savesets)
	{
		goodsave = SaveSettings();

		if (goodsave)
		{
			os_notice(lptr, sockfd,
				"Settings have been saved");
		}
		else
		{
			os_notice(lptr, sockfd,
				"Settings save failed");
		}
	} /* if (savesets) */
} /* o_save() */

/*
o_reload()
	Reload nickname/channel/memo/stat databases
*/

static void
o_reload(Luser *lptr, int ac, char **av, int sockfd)

{
	int		goodreload;

	goodreload = ReloadData();

	if (goodreload)
		os_notice(lptr, sockfd, "Reload successful");
	else
		os_notice(lptr, sockfd,
			"Reload contained errors, continuing to use old databases");

	o_RecordCommand(sockfd,
		"RELOAD %s",
		goodreload ? "" : "(failed)");
} /* o_reload() */

/*
o_set()
 Configure setting av[1] to value av[2]. If av[1] is "list",
list all settings.
*/

static void
o_set(Luser *lptr, int ac, char **av, int sockfd)

{
	Directive *dptr;
	char sendstr[MAXLINE],
	     tmp[MAXLINE];
	int ii,
	    pcnt;

	if (ac < 2)
	{
		os_notice(lptr, sockfd,
			"Syntax: \002SET <item | list> [value]\002");
		return;
	}

	if (!strncasecmp(av[1], "list", strlen(av[1])))
	{
		o_RecordCommand(sockfd,
			"SET LIST");

		DisplaySettings(lptr, sockfd);

		return;
	}

	if (!(dptr = FindDirective(av[1])))
	{
		os_notice(lptr, sockfd,
			"Unknown setting: %s",
			av[1]);
		return;
	}

	if (ac < 3)
	{
		/*
		 * They didn't give a new value, display the current one
		 */

		sendstr[0] = '\0';
		for (ii = 0; ii < PARAM_MAX; ii++)
		{
			if (!dptr->param[ii].type)
				break;

			switch (dptr->param[ii].type)
			{
				case PARAM_STRING:
				{
					sprintf(tmp, "\"%s\" ",
						*(char **) dptr->param[ii].ptr);
					strcat(sendstr, tmp);
					break;
				}

				case PARAM_TIME:
				{
					sprintf(tmp, "%s ",
						timeago(*(long *) dptr->param[ii].ptr, 2));
					strcat(sendstr, tmp);
					break;
				}

				case PARAM_INT:
				case PARAM_SET:
				case PARAM_PORT:
				{
					sprintf(tmp, "%d ",
						*(int *) dptr->param[ii].ptr);
					strcat(sendstr, tmp);
					break;
				}
			} /* switch (dptr->param[ii].type) */
		}
	
		o_RecordCommand(sockfd,
			"SET [%s]",
			dptr->name);

		os_notice(lptr, sockfd,
			"[\002%s\002] is set to: %s",
			dptr->name,
			sendstr);

		return;
	}

	if (dptr->flag == D_NORUNTIME)
	{
		os_notice(lptr, sockfd,
			"The directive [\002%s\002] is not run-time configurable",
			dptr->name);
		return;
	}

	/*
	 * Starting position of our first arguement
	 */
	pcnt = 2;

	sendstr[0] = '\0';
	for (ii = 0; ii < PARAM_MAX; ii++, pcnt++)
	{
		if (!dptr->param[ii].type)
			break;

		/*
		 * Now assign our variable (dptr->param[ii].ptr) to the
		 * corresponding av[] slot
		 */
		if ((dptr->param[ii].type != PARAM_SET) && (pcnt >= ac))
		{
			os_notice(lptr, sockfd,
				"Not enough arguements to [%s] directive",
				dptr->name);
			return;
		}

		switch (dptr->param[ii].type)
		{
			case PARAM_STRING:
			{
				char *strptr;

				if ((*(char **) dptr->param[ii].ptr) != (char *) NULL)
				{
					/*
					 * make sure we free the old string before allocating
					 * a new one.
					 */
					MyFree(*(char **) dptr->param[ii].ptr);
				}

				if (*av[pcnt] == ':')
					strptr = av[pcnt] + 1;
				else
					strptr = av[pcnt];

				*(char **) dptr->param[ii].ptr = MyStrdup(strptr);

				sprintf(tmp, "\"%s\" ",
					strptr);
				strcat(sendstr, tmp);

				break;
			}

			case PARAM_INT:
			{
				int value;

				value = IsNum(av[pcnt]);
				if (!value)
				{
					os_notice(lptr, sockfd,
						"Invalid integer: %s",
						av[pcnt]);
					return;
				}

				/*
				 * We have to call atoi() anyway since IsNum() would
				 * have returned 1 if av[pcnt] was "0", but IsNum()
				 * is a good way of making sure we have a valid integer
				 */
				value = atoi(av[pcnt]);
				*(int *) dptr->param[ii].ptr = value;

				sprintf(tmp, "%d ",
					value);
				strcat(sendstr, tmp);

				break;
			}

			case PARAM_TIME:
			{
				long value;

				value = timestr(av[pcnt]);
				if (!value)
				{
					os_notice(lptr, sockfd,
						"Invalid time format: %s",
						av[pcnt]);
					return;
				}

				*(long *) dptr->param[ii].ptr = value;

				sprintf(tmp, "%s ",
					timeago(value, 2));
				strcat(sendstr, tmp);

				break;
			}

			case PARAM_SET:
			{
				int value;

				value = IsNum(av[pcnt]);
				if (!value)
				{
					os_notice(lptr, sockfd,
						"Invalid integer (must be 1 or 0): %s",
						av[pcnt]);
					return;
				}

				value = atoi(av[pcnt]);

				if (value)
				{
					*(int *) dptr->param[ii].ptr = 1;
					strcat(sendstr, "1 ");
				}
				else
				{
					*(int *) dptr->param[ii].ptr = 0;
					strcat(sendstr, "0 ");
				}

				break;
			}

			case PARAM_PORT:
			{
				int value;

				value = IsNum(av[pcnt]);
				if (!value)
				{
					os_notice(lptr, sockfd,
						"Invalid port number (must be between 1 and 65535): %s",
						av[pcnt]);
					return;
				}

				value = atoi(av[pcnt]);

				if ((value < 1) || (value > 65535))
				{
					os_notice(lptr, sockfd,
						"Invalid port number (must be between 1 and 65535): %s",
						av[pcnt]);
					return;
				}

				*(int *) dptr->param[ii].ptr = value;

				sprintf(tmp, "%d ",
					value);
				strcat(sendstr, tmp);

				break;
			}

			default:
			{
				/* we shouldn't get here */
				break;
			}
		} /* switch (dptr->param[ii].type) */
	} /* for (ii = 0; ii < PARAM_MAX; ii++, pcnt++) */

	o_RecordCommand(sockfd,
		"SET [%s] %s",
		dptr->name,
		sendstr);

	os_notice(lptr, sockfd,
		"The value of [\002%s\002] has been set to: %s",
		dptr->name,
		sendstr);
} /* o_set() */

/*
DisplaySettings()
 Output a list of all settings and their values
*/

static void
DisplaySettings(Luser *lptr, int sockfd)

{
	Directive *dptr;
	int ii,
	    cnt;
	char sendstr[MAXLINE],
	     tmp[MAXLINE];

	os_notice(lptr, sockfd,
		"-- Listing settings --");

	cnt = 0;
	for (dptr = directives; dptr->name; ++dptr, ++cnt)
	{
		sendstr[0] = '\0';
		for (ii = 0; ii < PARAM_MAX; ii++)
		{
			if (!dptr->param[ii].type)
				break;

			switch (dptr->param[ii].type)
			{
				case PARAM_STRING:
				{
					sprintf(tmp, "\"%s\" ",
						*(char **) dptr->param[ii].ptr);
					strcat(sendstr, tmp);
					break;
				}

				case PARAM_TIME:
				{
					sprintf(tmp, "%s ",
						timeago(*(long *) dptr->param[ii].ptr, 2));
					strcat(sendstr, tmp);
					break;
				}

				case PARAM_INT:
				case PARAM_SET:
				case PARAM_PORT:
				{
					sprintf(tmp, "%d ",
						*(int *) dptr->param[ii].ptr);
					strcat(sendstr, tmp);
					break;
				}
			} /* switch (dptr->param[ii].type) */
		}

		os_notice(lptr, sockfd,
			"%-20s %-30s",
			dptr->name,
			sendstr);
	}

	os_notice(lptr, sockfd,
		"-- End of list (%d settings) --",
		cnt);
} /* DisplaySettings() */

#ifdef NICKSERVICES

/*
o_ignore()
	Modify the services ignorance list
*/

static void
o_ignore(Luser *lptr, int ac, char **av, int sockfd)

{
	o_Command	*cptr;

	if (ac < 2)
	{
		os_notice(lptr, sockfd,
			"Syntax: \002IGNORE {ADD|DEL|LIST} [nick|mask|index]\002");
		return;
	}

	cptr = GetoCommand(ignorecmds, av[1]);

	if (cptr && (cptr != (o_Command *) -1))
	{
		/* call cptr->func to execute command */
		(*cptr->func)(lptr, ac, av, sockfd);
	}
	else
	{
		/* the option they gave was not valid */
		os_notice(lptr, sockfd,
			"%s switch [\002%s\002]",
			(cptr == (o_Command *) -1) ? "Ambiguous" : "Unknown",
			av[1]);
		os_notice(lptr, sockfd,
			"Syntax: \002IGNORE {ADD|DEL|LIST} [mask|nickname|index]\002");
	}
} /* o_ignore() */

static void
o_ignore_add(Luser *lptr, int ac, char **av, int sockfd)

{
	char hostmask[MAXLINE];
	time_t expire = (time_t) NULL;

	if (ac < 3)
	{
		os_notice(lptr, sockfd,
			"Syntax: \002IGNORE ADD <nickname|hostmask> [time]\002");
		return;
	}

	if (match("*!*@*", av[2]))
		strcpy(hostmask, av[2]);
	else
		if (match("*!*", av[2]))
		{
			strcpy(hostmask, av[2]);
			strcat(hostmask, "@*");
		}
		else
			if (match("*@*", av[2]))
			{
				strcpy(hostmask, "*!");
				strcat(hostmask, av[2]);
			}
			else
				if (match("*.*", av[2]))
				{
					strcpy(hostmask, "*!*@");
					strcat(hostmask, av[2]);
				}
				else
				{
					Luser *ptr;
					char *mask;

					/* it must be a nickname - try to get hostmask */
					if ((ptr = FindClient(av[2])))
					{
						mask = HostToMask(ptr->username, ptr->hostname);
						strcpy(hostmask, "*!");
						strcat(hostmask, mask);
						MyFree(mask);
					}
					else
					{
						strcpy(hostmask, av[2]);
						strcat(hostmask, "!*@*");
					}
				}

	if (OnIgnoreList(hostmask))
	{
		os_notice(lptr, sockfd,
			"The hostmask [\002%s\002] is already on the ignorance list",
			hostmask);
		return;
	}

	if (ac >= 4)
		expire = timestr(av[3]);

	o_RecordCommand(sockfd,
		"IGNORE ADD %s %s",
		hostmask,
		(ac < 4) ? "(perm)" : av[3]);

	AddIgnore(hostmask, expire);

	if (expire)
		os_notice(lptr, sockfd,
			"[\002%s\002] has been added to the ignorance list with an expire of %s",
			hostmask,
			timeago(expire, 3));
	else
		os_notice(lptr, sockfd,
			"[\002%s\002] has been added to the ignorance list with no expiration",
			hostmask);
} /* o_ignore_add() */

static void
o_ignore_del(Luser *lptr, int ac, char **av, int sockfd)

{
	Ignore	*temp;
	char	*host = (char *) NULL;
	int		idx;

	if (ac < 3)
	{
		os_notice(lptr, sockfd,
			"Syntax: \002IGNORE DEL <hostmask | index>\002");
		return;
	}

	if ((idx = IsNum(av[2])))
	{
		int		cnt = 0;

		for (temp = ignorelist; temp; temp = temp->next, cnt++)
			if (idx == (cnt + 1))
			{
				host = MyStrdup(temp->hostmask);
				break;
			}

		if (!host)
		{
			os_notice(lptr, sockfd,
				"[\002%d\002] is not a valid index",
				idx);
			return;
		}
	}
	else
		host = MyStrdup(av[2]);

	o_RecordCommand(sockfd,
		"IGNORE DEL %s",
		host);

	if (!DelIgnore(host))
	{
		os_notice(lptr, sockfd,
			"[\002%s\002] was not found on the ignorance list",
			host);
		MyFree(host);
		return;
	}

	os_notice(lptr, sockfd,
		"[\002%s\002] has been removed from the ignorance list",
		host);

	MyFree(host);
} /* o_ignore_del() */

static void
o_ignore_list(Luser *lptr, int ac, char **av, int sockfd)

{
	Ignore	*tmp;
	int		idx;
	char	*mask = (char *) NULL;
	time_t	currtime = time(NULL);

	if (ac >= 3)
		mask = av[2];

	o_RecordCommand(sockfd,
		"IGNORE LIST %s",
		mask ? mask : "");

	if (ignorelist)
	{
		os_notice(lptr, sockfd,
			"-- Ignorance List --");
		os_notice(lptr, sockfd,
			"Num Expire (min) Hostmask");
		os_notice(lptr, sockfd,
			"--- ------------ --------");
		idx = 1;
		for (tmp = ignorelist; tmp; tmp = tmp->next, idx++)
		{
			if (mask)
				if (match(mask, tmp->hostmask) == 0)
					continue;
			os_notice(lptr, sockfd,
				"%-3d %-12.1f %-35s",
				idx,
				tmp->expire ? ((float)(tmp->expire - currtime) / 60) : (float)0,
				tmp->hostmask);
		}
		os_notice(lptr, sockfd,
			"-- End of list --");
	}
	else
		os_notice(lptr, sockfd,
			"The ignorance list is empty");
} /* o_ignore_list() */

#endif /* NICKSERVICES */

/*
AddIgnore()
	Add 'hostmask' to ignore list with 'expire'
*/

void
AddIgnore(char *hostmask, time_t expire)

{
	Ignore	*ptr;

	ptr = (Ignore *) MyMalloc(sizeof(Ignore));
	ptr->hostmask = MyStrdup(hostmask);

	if (!expire)
		ptr->expire = (time_t) 0; /* no expiration */
	else
		ptr->expire = (time(NULL) + expire);

	ptr->next = ignorelist;
	ignorelist = ptr;
} /* AddIgnore() */

/*
OnIgnoreList()
	Return ptr to ignore structure containing 'mask'
*/

Ignore *
OnIgnoreList(char *mask)

{
	Ignore	*tmp;

	if (!mask)
		return ((Ignore *) NULL);

	for (tmp = ignorelist; tmp; tmp = tmp->next)
	{
		if (match(tmp->hostmask, mask))
			return (tmp);
	}

	return ((Ignore *) NULL);
} /* OnIgnoreList() */

/*
DelIgnore()
	Delete 'mask' from ignore list; return number of matches
*/

int
DelIgnore(char *mask)

{
	Ignore	*temp, *prev;
	int		cnt;

	if (!mask)
		return 0;

	if (!ignorelist)
		return 0;

	cnt = 0;
	prev = ignorelist;
	for (temp = ignorelist->next; temp; )
	{
		if (match(mask, temp->hostmask))
		{
			cnt++;
			prev->next = temp->next;
			MyFree(temp->hostmask);
			MyFree(temp);
			temp = prev->next;
		}
		else
		{
			prev = temp;
			temp = temp->next;
		}
	}
	if (ignorelist)
	{
		if (match(mask, ignorelist->hostmask))
		{
			cnt++;
			prev = ignorelist->next;
			MyFree(ignorelist->hostmask);
			MyFree(ignorelist);
			ignorelist = prev;
		}
	}

	return (cnt);
} /* DelIgnore() */

/*
ExpireIgnores()
 Check if any ignores have expired - if so, remove them
from the list
*/

void
ExpireIgnores(time_t unixtime)

{
	Ignore	*tmp, *prv;

	prv = (Ignore *) NULL;
	for (tmp = ignorelist; tmp; )
	{
		if (tmp->expire && (unixtime >= tmp->expire))
		{
			SendUmode(OPERUMODE_Y,
				"*** Expired ignore: %s",
				tmp->hostmask);

			MyFree(tmp->hostmask);
			if (prv)
			{
				prv->next = tmp->next;
				MyFree(tmp);
				tmp = prv;
			}
			else
			{
				ignorelist = tmp->next;
				MyFree(tmp);
				tmp = (Ignore *) NULL;
			}
		}

		prv = tmp;

		if (tmp)
			tmp = tmp->next;
		else
			tmp = ignorelist;
	}
} /* ExpireIgnores() */

/*
o_who()
	Displays list of users/bots on partyline
*/

static void
o_who(Luser *lptr, int ac, char **av, int sockfd)

{
	char idle[MAXLINE];
	char prefix;
	int AreBots = 0, mins;
	Userlist *tempuser;
	DccUser *tempconn;
	char uhost[UHOSTLEN + 2];

	o_RecordCommand(sockfd,
		"WHO");

	os_notice(lptr, sockfd,
		"Current users: (%% = Services Admin, * = Admin, + = Oper)");

	for (tempconn = connections; tempconn; tempconn = tempconn->next)
	{
		if ((tempconn->flags & SOCK_PENDING) &&
				!(tempconn->flags & SOCK_DCC))
			continue;
		
		if (tempconn->flags & SOCK_NEEDID)
			continue;

		if (tempconn->flags & SOCK_TCMBOT)
		{
			AreBots = 1;
			continue;
		}

		tempuser = DccGetUser(tempconn);

		if (IsServicesAdmin(tempuser))
			prefix = '%';
		else if (IsAdmin(tempuser))
			prefix = '*';
		else if (IsOper(tempuser))
			prefix = '+';
		else
			prefix = ' ';

		mins = ((time(NULL) - tempconn->idle) / 60) % 60;
		if (mins >= 5)
		{
			strcpy(idle, "idle: ");
			strcat(idle, timeago(tempconn->idle, 0));
		}
		else
			idle[0] = '\0';

		sprintf(uhost, "%s@%s",
			tempconn->username,
			tempconn->hostname);

		os_notice(lptr, sockfd, "  %c%-10s %-25s %s",
			prefix,
			tempconn->nick,
			uhost,
			idle);
	}

	if (AreBots)
	{
		os_notice(lptr, sockfd, "\r\nBot connections:");
		for (tempconn = connections; tempconn; tempconn = tempconn->next)
		{
			if (!(tempconn->flags & SOCK_TCMBOT))
				continue;
			if (tempconn->flags & SOCK_BOTHUB)
				os_notice(lptr, sockfd, " -> %s [%s@%s]",
					tempconn->nick,
					tempconn->username,
					tempconn->hostname);
			else
				os_notice(lptr, sockfd, " <- %s [%s@%s]",
					tempconn->nick,
					tempconn->username,
					tempconn->hostname);
		}
	} /* if (AreBots) */
} /* o_who() */

/*
o_boot()
	Remove user av[1] from the partyline with message av[2]-
*/

static void
o_boot(Luser *lptr, int ac, char **av, int sockfd)

{
	char *reason;
	DccUser *dccptr;

	if (ac < 2)
	{
		os_notice(lptr, sockfd, "Syntax: \002BOOT <nick> [reason]\002");
		return;
	}

	if (!(dccptr = IsOnDcc(av[1])))
	{
		os_notice(lptr, sockfd, "%s is not on the partyline\n", 
			av[1]);
		return;
	}

	if (ac < 3)
		reason = MyStrdup("");
	else
		reason = GetString(ac - 2, av + 2);

	o_RecordCommand(sockfd,
		"BOOT %s %s",
		av[1],
		reason);

	BroadcastDcc(DCCALL, "%s has been booted by %s [%s]\n",
		av[1],
		onick,
		reason);

	os_notice(lptr, dccptr->socket,
		"You have been booted by %s [%s]",
		onick,
		reason);

	CloseConnection(dccptr);

	MyFree(reason);
} /* o_boot() */

/*
o_quit()
	Removes lptr->nick from the partyline with message av[1]-
*/

static void
o_quit(Luser *lptr, int ac, char **av, int sockfd)

{
	char *reason;
	DccUser *dccptr;

	if (!(dccptr = IsDccSock(sockfd)))
		return; /* shouldn't happen */

	if (ac < 2)
		reason = MyStrdup("");
	else
		reason = GetString(ac - 1, av + 1);


	os_notice(lptr, sockfd, "Closing connection: %s", 
		reason);

	BroadcastDcc(DCCALL, "%s (%s@%s) has left the partyline [%s]\n",
		onick,
		ouser,
		ohost,
		reason);

	CloseConnection(dccptr);

	MyFree(reason);
} /* o_quit() */

/*
o_motd()
	Display motd
*/

static void
o_motd(Luser *lptr, int ac, char **av, int sockfd)

{
	o_RecordCommand(sockfd,
		"MOTD");

	SendMotd(sockfd);
} /* o_motd() */

/*
o_link()
	Attempt to link to TCM bot av[1]
*/

static void
o_link(Luser *lptr, int ac, char **av, int sockfd)

{
	Botlist *bptr;

	if (ac < 2)
	{
		os_notice(lptr, sockfd, "Syntax: \002LINK <botnick>\002");
		return;
	}

	if ((bptr = IsBot(av[1])) == (Botlist *) NULL)
	{
		os_notice(lptr, sockfd, "No such bot: %s", 
			av[1]);
		return;
	}

	o_RecordCommand(sockfd,
		"LINK %s",
		bptr->name);

	os_notice(lptr, sockfd, "Attempting to link to %s", 
		bptr->name);
	os_notice(lptr, sockfd, "Making connection to %s:%d",
		bptr->hostname,
		bptr->port);

	ConnectToTCM(onick, bptr);
} /* o_link() */

/*
o_unlink()
	Unlink TCM bot av[1] with reason av[2]-
*/

static void
o_unlink(Luser *lptr, int ac, char **av, int sockfd)

{
	char sendstr[MAXLINE];
	char *reason;
	DccUser *botptr, *tmp;

	if (ac < 2)
	{
		os_notice(lptr, sockfd,
			"Syntax: \002UNLINK <bot name> [reason]\002");
		return;
	}

	if (!(botptr = GetBot(av[1])))
	{
		os_notice(lptr, sockfd, "No such bot: %s",
			av[1]);
		return;
	}

	if (ac < 3)
		reason = MyStrdup("");
	else
		reason = GetString(ac - 2, av + 2);

	o_RecordCommand(sockfd,
		"UNLINK %s %s",
		botptr->nick,
		reason);

	if (reason[0] == '\0')
	{
		sprintf(sendstr, "(%s)",
			onick);
		MyFree(reason);
		reason = MyStrdup(sendstr);
	}
	else
	{
		sprintf(sendstr, "(%s: %s)",
			onick,
			reason);
		MyFree(reason);
		reason = MyStrdup(sendstr);
	}

	for (tmp = connections; tmp; tmp = tmp->next)
	{
		if (tmp == botptr)
			continue;

		if (tmp->flags & SOCK_TCMBOT)
		{
			sprintf(sendstr, "(%s) Unlinked from %s %s\n",
				n_OperServ,
				botptr->nick,
				reason);
			writesocket(tmp->socket, sendstr);
		}
		else
		{
			sprintf(sendstr, "*** Unlinked from %s %s\n",
				botptr->nick,
				reason);
			writesocket(tmp->socket, sendstr);
		}
	}

	CloseConnection(botptr);

	MyFree(reason);
} /* o_unlink() */

/*
o_stats()
 Displays various stats depending on switch

Switches:
   O : O:lines
   B : bot lines
   S : server list lines
   U : uptime
   M : memory info
   I : restricted host lines
   G : glines
   J : jupes
   P : port lines
   C : monitored channels
   ? : recv/send info
*/

static void
o_stats(Luser *lptr, int ac, char **av, int sockfd)

{
	if (ac < 2)
	{
		os_notice(lptr, sockfd, "Syntax: \002STATS <switch>\002");
		return;
	}

	o_RecordCommand(sockfd,
		"STATS %c",
		ToUpper(*av[1]));

	switch (av[1][0])
	{
		case 'b':
		case 'B':
		{
			Botlist *tempbot;
			char uh[UHOSTLEN + 2];
		
			if (botlist || remotebots)
			{
				os_notice(lptr, sockfd, "-- TCM Bot Lines --");
				os_notice(lptr, sockfd, "    [Nickname       ] [Hostname                 ] [Port ]");

				for (tempbot = botlist; tempbot; tempbot = tempbot->next)
				{
					os_notice(lptr, sockfd, "[B] [%-15s] [%-25s] [%-5d]",
						tempbot->name,
						tempbot->hostname,
						tempbot->port);
				}

				for (tempbot = remotebots; tempbot; tempbot = tempbot->next)
				{
					sprintf(uh, "%s@%s",
						tempbot->username,
						tempbot->hostname);

					os_notice(lptr, sockfd, "[L] [%-15s] [%-25s]",
						tempbot->name,
						uh);
				}

				os_notice(lptr, sockfd, "-- End of list --");
			}
			else
				os_notice(lptr, sockfd, "No bot lines");
			break;
		} /* case 'B' */

		case 's':
		case 'S':
		{
			Servlist	*tempserv;

			os_notice(lptr, sockfd, "-- Server List --");
			os_notice(lptr, sockfd,
				"[Hostname                      ] [Port ]");
			for (tempserv = serverlist; tempserv; tempserv = tempserv->next)
			{
				os_notice(lptr, sockfd, "[%-30s] [%-5d]",
					tempserv->hostname,
					tempserv->port);
			}
			os_notice(lptr, sockfd, "-- End of list --");

			break;
		} /* case 'S' */

		case 'o':
		case 'O':
		{
			char flags[MAXLINE];
			char uhost[UHOSTLEN + 2];
			Userlist *tempuser;
			int cnt = 0;

			os_notice(lptr, sockfd, "-- Authorized Users --");
			os_notice(lptr, sockfd,
				"[Nickname      ] [Hostmask                 ] [Flags     ]");
			for (tempuser = userlist; tempuser; tempuser = tempuser->next)
			{
				cnt++;

				flags[0] = '\0';
				if (tempuser->flags & PRIV_ADMIN)
					strcat(flags, "a");
				if (tempuser->flags & PRIV_CHAT)
					strcat(flags, "d");
				if (tempuser->flags & PRIV_EXCEPTION)
					strcat(flags, "e");
				if (tempuser->flags & PRIV_FRIEND)
					strcat(flags, "f");
				if (tempuser->flags & PRIV_GLINE)
					strcat(flags, "g");
				if (tempuser->flags & PRIV_JUPE)
					strcat(flags, "j");
				if (tempuser->flags & PRIV_OPER)
					strcat(flags, "o");
				if (tempuser->flags & PRIV_SADMIN)
					strcat(flags, "S");

				sprintf(uhost, "%s@%s",
					tempuser->username,
					tempuser->hostname);

				os_notice(lptr, sockfd, "[%-14s] [%-25s] [%-10s]",
					tempuser->nick,
					uhost,
					flags);
			}

			os_notice(lptr, sockfd,
				"-- End of list (%d users found) --",
				cnt);

			break;
		}
		case 'u':
		case 'U':
		{
			os_notice(lptr, sockfd,
				"%s has been up for %s",
				Me.name,
				timeago(TimeStarted, 1));
			break;
		} /* case 'U' */
		case 'm':
		case 'M':
		{
			CalcMem(onick, sockfd);
			break;
		} /* case 'M' */
		case 'i':
		case 'I':
		{
			rHost *temphost;
			char uhost[UHOSTLEN + 2];

			os_notice(lptr, sockfd, "-- Restricted Hostmasks --");
			os_notice(lptr, sockfd, "[Hostmask                 ] [Max connections]");
			for (temphost = rhostlist; temphost; temphost = temphost->next)
			{
				sprintf(uhost, "%s@%s",
					temphost->username,
					temphost->hostname);

				os_notice(lptr, sockfd, "[%-25s] [%-15d]",
					uhost,
					temphost->hostnum);
			}

			os_notice(lptr, sockfd, "-- End of list --");
			break;
		} /* case 'I' */

		case 'g':
		case 'G':
		{
			Gline *gptr;
			time_t currtime = time(NULL);
			char expstr[MAXLINE];
			char uh[UHOSTLEN + 2];

		#ifdef ALLOW_GLINES
			os_notice(lptr, sockfd, "-- Listing Glines --");
			os_notice(lptr, sockfd,
				"[Hostmask                 ] [Reason                   ] [Expire]");
			for (gptr = glinelist; gptr; gptr = gptr->next)
			{
				if (gptr->expires)
				{
					sprintf(expstr, "%-6.1f",
						(((float) (gptr->expires - currtime)) / 60));
				}
				else
					sprintf(expstr, "never ");

				sprintf(uh, "%s@%s",
					gptr->username,
					gptr->hostname);

				os_notice(lptr, sockfd,
					"[%-25s] [%-25s] [%s]",
					uh,
					gptr->reason,
					expstr);
			}

			os_notice(lptr, sockfd, "-- End of list (%d gline%s) --",
				Network->TotalGlines,
				(Network->TotalGlines == 1) ? "" : "s");
		#else

			os_notice(lptr, sockfd, "Glines are disabled");

		#endif /* ALLOW_GLINES */

			break;
		} /* case 'G' */

		case 'j':
		case 'J':
		{
			Jupe		*tempjupe;

		#ifdef ALLOW_JUPES
			os_notice(lptr, sockfd, "-- Listing Jupes --");
			os_notice(lptr, sockfd,
				"[Server/Nickname          ] [Reason                        ]");
			for (tempjupe = jupelist; tempjupe; tempjupe = tempjupe->next)
			{
				os_notice(lptr, sockfd, "[%-25s] [%-30s]", 
					tempjupe->name,
					tempjupe->reason);
			}
			os_notice(lptr, sockfd, "-- End of list (%d jupe%s) --", 
				Network->TotalJupes,
				(Network->TotalJupes == 1) ? "" : "s");
		#else
			os_notice(lptr, sockfd, "Jupes are disabled");
		#endif

			break;
		} /* case 'J' */

		case 'p':
		case 'P':
		{
			PortInfo	*tempport;

			os_notice(lptr, sockfd, "-- Port List --");
			os_notice(lptr, sockfd,
				"[Port #] [Type    ] [Hostmask                 ]");
			for (tempport = portlist; tempport; tempport = tempport->next)
			{
				os_notice(lptr, sockfd,
					"[%-6d] [%-8s] [%-25s]",
					tempport->port,
					(tempport->type == PRT_TCM) ? "TCM" : "Users",
					tempport->host ? tempport->host : "*.*");
			}
			os_notice(lptr, sockfd, "-- End of list --");

			break;
		} /* case 'P' */

		case 'c':
		case 'C':
		{
			Chanlist	*tempchan;

			os_notice(lptr, sockfd, "-- Listing Channels --");
			os_notice(lptr, sockfd,
				"[Channel name             ]");
			for (tempchan = mychanlist; tempchan; tempchan = tempchan->next)
			{
				os_notice(lptr, sockfd, "[%-25s]", 
					tempchan->name);
			}

			os_notice(lptr, sockfd, "-- End of list (%d channel%s) --", 
				Network->MyChans,
				(Network->MyChans == 1) ? "" : "s");

			break;
		} /* case 'C' */
		
		case '?':
		{
			time_t	uptime = time(NULL) - TimeStarted;

			os_notice(lptr, sockfd,
				"Total Sent %10.2f %s (%4.1f K/s)",
				GMKBv((float)Network->SendB),
				GMKBs((float)Network->SendB),
				(float)((float)((float)Network->SendB / (float)1024.0) / (float)(uptime)));
			os_notice(lptr, sockfd,
				"Total Recv %10.2f %s (%4.1f K/s)",
				GMKBv((float)Network->RecvB),
				GMKBs((float)Network->RecvB),
				(float)((float)((float)Network->RecvB / (float)1024.0) / (float)(uptime)));

			break;
		} /* case '?' */

		default:
		{
			os_notice(lptr, sockfd, "Unknown switch [%c]",
				av[1][0]);
			break;
		}
	} /* switch (letter[0]) */
} /* o_stats() */

#ifdef ALLOW_KILLCHAN

/*
o_killchan()
 Kills all members of a channel who meet the criteria given
by [options]

Valid options:
 -ops       - Limit matches to channel ops
 -nonops    - Limit matches to non channel ops
 -voices    - Limit matches to channel voices
 -nonvoices - Limit matches to non channel voices
 -nonopers  - Do not kill IRC Operators

Does not affect Service Bots or O: line exception users
*/

static void
o_killchan(Luser *lptr, int ac, char **av, int sockfd)

{
	char *reason;
	Channel *chptr;
	ChanUser *tempuser, *prev = NULL;
	int ii,
	    bad,
	    cnt;
	int alen;
	int nonops, /* -nonops */
	    ops, /* -ops */
	    nonvoices, /* -nonvoices */
	    voices, /* -voices */
	    nonopers; /* -nonopers */
	char argbuf[MAXLINE];

	if (ac < 2)
	{
		os_notice(lptr, sockfd,
			"Syntax: \002KILLCHAN [options] <channel> [reason]\002");
		return;
	}

	chptr = (Channel *) NULL;
	reason = (char *) NULL;
	nonopers = 0;
	ops = 0;
	nonops = 0;
	voices = 0;
	nonvoices = 0;

	for (cnt = 1; cnt < ac; cnt++)
	{
		alen = strlen(av[cnt]);
		if (!strncasecmp(av[cnt], "-nonopers", alen))
			nonopers = 1;
		else if (!strncasecmp(av[cnt], "-ops", alen))
			ops = 1;
		else if (!strncasecmp(av[cnt], "-nonops", alen))
			nonops = 1;
		else if (!strncasecmp(av[cnt], "-voices", alen))
			voices = 1;
		else if (!strncasecmp(av[cnt], "-nonvoices", alen))
			nonvoices = 1;
		else
		{
			if (!(chptr = FindChannel(av[cnt++])))
			{
				os_notice(lptr, sockfd,
					"Invalid channel: %s",
					av[cnt - 1]);
				return;
			}

			if (ac >= (cnt + 1))
				reason = GetString(ac - cnt, av + cnt);
			else
				reason = MyStrdup("");
			break;
		}
	}

	if (!reason)
		reason = MyStrdup("");

	if (!chptr)
	{
		os_notice(lptr, sockfd,
			"No channel specified");
		MyFree(reason);
		return;
	}

	argbuf[0] = '\0';

	if (nonopers)
		strcat(argbuf, "-nonopers ");

	if (ops)
		strcat(argbuf, "-ops ");

	if (nonops)
		strcat(argbuf, "-nonops ");

	if (voices)
		strcat(argbuf, "-voices ");

	if (nonvoices)
		strcat(argbuf, "-nonvoices ");

	o_RecordCommand(sockfd,
		"KILLCHAN %s %s[%s]",
		chptr->name,
		argbuf,
		reason);

	for (tempuser = chptr->firstuser; tempuser; )
	{
		bad = 0;

		if (FindService(tempuser->lptr))
			bad = 1;
		else if (nonopers && IsOperator(tempuser->lptr))
			bad = 1;
		else if (ops && !(tempuser->flags & CH_OPPED))
			bad = 1;
		else if (nonops && (tempuser->flags & CH_OPPED))
			bad = 1;
		else if (voices && !(tempuser->flags & CH_VOICED))
			bad = 1;
		else if (nonvoices && (tempuser->flags & CH_VOICED))
			bad = 1;
		else
		{
			if (tempuser->lptr->flags & L_OSREGISTERED)
				ii = 1;
			else
				ii = 0;
			if (IsProtected(GetUser(ii, tempuser->lptr->nick, tempuser->lptr->username, tempuser->lptr->hostname)))
				bad = 1;
		}

		if (bad)
		{
			tempuser = tempuser->next;
			continue;
		}

		toserv(":%s KILL %s :%s!%s (%s (%s@%s))\n",
			n_OperServ,
			tempuser->lptr->nick,
			Me.name,
			n_OperServ,
			reason,
			onick,
			n_OperServ);

		prev = tempuser->next;
		DeleteClient(tempuser->lptr);
		tempuser = prev;
	}

	os_notice(lptr, sockfd, "Channel [%s] has been killed",
		chptr->name);

	MyFree(reason);
} /* o_killchan() */

#endif /* ALLOW_KILLCHAN */

#ifdef ALLOW_KILLHOST

/*
o_killhost()
	Kills all users with the user@host av[1] (unless they have an "e"
flag) with reason av[2]-
*/

static void
o_killhost(Luser *lptr, int ac, char **av, int sockfd)

{
	char *reason;
	int ii;
	Luser *tempuser, *tmp = NULL;
	char *user,
	     *host;

	if (ac < 2)
	{
		os_notice(lptr, sockfd, "Syntax: \002KILLHOST <hostmask>\002");
		return;
	}

	if (ac < 3)
		reason = MyStrdup("");
	else
		reason = GetString(ac - 2, av + 2);

	o_RecordCommand(sockfd,
		"KILLHOST %s [%s]",
		av[1],
		reason);

	if (!(host = strchr(av[1], '@')))
	{
		user = (char *) NULL;
		host = av[1];
	}
	else
	{
		user = av[1];
		*host++ = '\0';
	}

	if (IsProtectedHost(user ? user : "*", host))
	{
		os_notice(lptr, sockfd,
			"%s@%s matches a protected host, unable to kill",
			user ? user : "*",
			host);
		MyFree(reason);
		return;
	}
	
	ii = 0;
	for (tempuser = ClientList; tempuser; )
	{
		if (FindService(tempuser))
		{
			tempuser = tempuser->next;
			continue;
		}

		if (user)
			if (!match(user, tempuser->username))
			{
				tempuser = tempuser->next;
				continue;
			}

		if (match(host, tempuser->hostname))
		{
			ii++;
			if (ii > MaxKill)
			{
				os_notice(lptr, sockfd,
					"-- More than %d matches found, halting kills --",
					MaxKill);
				break;
			}

			toserv(":%s KILL %s :%s!%s (%s (%s@%s))\n",
				n_OperServ,
				tempuser->nick,
				Me.name,
				n_OperServ,
				reason,
				onick,
				n_OperServ);
			tmp = tempuser->next;
			DeleteClient(tempuser);
			tempuser = tmp;
		}
		else
			tempuser = tempuser->next;
	}

	if (ii <= MaxKill)
	{
		os_notice(lptr, sockfd,
			"%d match%s of [%s@%s] killed",
			ii,
			(ii == 1) ? "" : "es",
			user ? user : "*",
			host);
	}

	MyFree(reason);
} /* o_killhost() */

#endif /* ALLOW_KILLHOST */

#ifdef HIGHTRAFFIC_MODE

static void
o_htm(Luser *lptr, int ac, char **av, int sockfd)

{
	float currload;
	o_Command *cptr;

	o_RecordCommand(sockfd,
		"HTM %s %s",
		(ac < 2) ? "" : StrToupper(av[1]),
		(ac < 3) ? "" : av[2]);

	if (ac < 2)
	{
		currload = ((float) (Network->RecvB - Network->CheckRecvB) / 
								(float) 1024) / (float) HTM_INTERVAL;

		os_notice(lptr, sockfd,
			"High-Traffic mode is \002%s\002",
			(HTM) ? "ON" : "OFF");
		os_notice(lptr, sockfd,
			"Maximum load: \002%d.00\002 K/s",
			ReceiveLoad);
		os_notice(lptr, sockfd,
			"Current load: \002%0.2f\002 K/s",
			currload);

		return;
	}

	cptr = GetoCommand(htmcmds, av[1]);

	if (cptr && (cptr != (o_Command *) -1))
	{
		/* call cptr->func to execute command */
		(*cptr->func)(lptr, ac, av, sockfd);
	}
	else
	{
		/* the option they gave was not valid */
		os_notice(lptr, sockfd,
			"%s switch [\002%s\002]",
			(cptr == (o_Command *) -1) ? "Ambiguous" : "Unknown",
			av[1]);
	}
} /* o_htm() */

static void
o_htm_on(Luser *lptr, int ac, char **av, int sockfd)

{
	float		currload;

	if (HTM)
	{
		os_notice(lptr, sockfd, "High-Traffic mode is currently active");
		return;
	}

	currload = ((float) (Network->RecvB - Network->CheckRecvB) / 
							(float) 1024) / (float) HTM_INTERVAL;

	HTM = 1;
	HTM_ts = time(NULL);

	putlog(LOG1,
		"Entering high-traffic mode (%0.2f K/s): Forced by %s!%s@%s",
		currload,
		onick,
		ouser,
		ohost);

	SendUmode(OPERUMODE_Y,
		"*** Entering high-traffic mode (%0.2f K/s): Forced by %s!%s@%s\n",
		currload,
		onick,
		ouser,
		ohost);

	os_notice(lptr, sockfd, "High-Traffic mode is now ON");
} /* o_htm_on() */

static void
o_htm_off(Luser *lptr, int ac, char **av, int sockfd)

{
	float		currload;

	if (!HTM)
	{
		os_notice(lptr, sockfd, "High-Traffic mode is currently inactive");
		return;
	}

	currload = ((float) (Network->RecvB - Network->CheckRecvB) / 
							(float) 1024) / (float) HTM_INTERVAL;

	HTM = HTM_ts = 0;

	putlog(LOG1,
		"Resuming standard traffic operations (%0.2f K/s): Forced by %s!%s@%s",
		currload,
		onick,
		ouser,
		ohost);

	SendUmode(OPERUMODE_Y,
		"*** Resuming standard traffic operations (%0.2f K/s): Forced by %s!%s@%s\n",
		currload,
		onick,
		ouser,
		ohost);

	os_notice(lptr, sockfd, "High-Traffic mode is now OFF");
} /* o_htm_off() */

static void
o_htm_max(Luser *lptr, int ac, char **av, int sockfd)

{
	int		newmax;

	if (ac < 3)
	{
		os_notice(lptr, sockfd,
			"Syntax: HTM MAX <new rate>");
		return;
	}

	if (!(newmax = IsNum(av[2])))
	{
		os_notice(lptr, sockfd,
			"[%s] is an invalid number",
			av[2]);
		return;
	}

	ReceiveLoad = newmax;

	putlog(LOG1, "HTM rate changed to %d K/s by %s!%s@%s",
		newmax,
		onick,
		ouser,
		ohost);

	SendUmode(OPERUMODE_Y,
		"*** New HTM Rate: %d K/s (forced by %s!%s@%s)\n",
		newmax,
		onick,
		ouser,
		ohost);

	os_notice(lptr, sockfd,
		"The HTM Rate has been set to: %d K/s",
		newmax);
} /* o_htm_max() */

#endif /* HIGHTRAFFIC_MODE */

/*
o_hub()
 Go through server list and display any servers that have av[1]
for a hub
*/

static void
o_hub(Luser *lptr, int ac, char **av, int sockfd)

{
	Server *hubptr, *tmpserv;
	int cnt;

	o_RecordCommand(sockfd,
		"HUB %s",
		(ac < 2) ? "" : av[1]);

	if (ac < 2)
	{
		/*
		 * They didn't give a hub name..list all hubs
		 */

		os_notice(lptr, sockfd,
			"-- Listing current hubs --");
		os_notice(lptr, sockfd,
			"[Hub Name                      ] [# servers]");

		cnt = 0;
		for (tmpserv = ServerList; tmpserv; tmpserv = tmpserv->next)
		{
			if (tmpserv->numservs > 1)
			{
				cnt++;
				os_notice(lptr, sockfd,
					"[%-30s] [%-9ld]",
					tmpserv->name,
					tmpserv->numservs);
			}
		}

		os_notice(lptr, sockfd,
			"-- End of list (%d hub%s found) --",
			cnt,
			(cnt == 1) ? "" : "s");

		return;
	}

	if (!(hubptr = FindServer(av[1])))
	{
		os_notice(lptr, sockfd,
			"No such server: \002%s\002",
			av[1]);
		return;
	}

	if (hubptr->numservs <= 1)
	{
		os_notice(lptr, sockfd,
			"\002%s\002 appears to be a leaf",
			hubptr->name);
		return;
	}

	os_notice(lptr, sockfd, "-- \002%s\002 --",
		hubptr->name);
	os_notice(lptr, sockfd,
		"[Leaf Name                     ] [Time Connected      ]");

	/*
	 * go through server list and try to find servers who's uplink
	 * matches hubptr
	 */
	cnt = 0;
	for (tmpserv = ServerList; tmpserv; tmpserv = tmpserv->next)
	{
		if (tmpserv->uplink == hubptr)
		{
			cnt++;
			os_notice(lptr, sockfd,
				"[%-30s] [%-20s]",
				tmpserv->name,
				timeago(tmpserv->connect_ts, 0));
		}
	}

	os_notice(lptr, sockfd,
		"-- End of list (%d lea%s found) --",
		cnt,
		(cnt == 1) ? "f" : "ves");
} /* o_hub() */

/*
o_umode()
 Configure various user modes for operator lptr->nick.

Possible Usermodes:
  b - show all tcm bot link attempts, invalid connections attempts
      etc
  c - see all clone connections
  C - see all client connections
  d - debugging usermode - combination of all other usermodes
      plus debug() calls
  e - see all clone exits
      (Note: exits are only shown if STATSERVICES is defined,
       since HashDelClient() only finds clone matches to
       update StatServ unique client counts)
  E - see all client exits
  j - see channel joins
  k - see channel kicks
  l - show whenever a server links/delinks to/from the network
  m - see channel mode changes
  n - see nickname changes
  o - see new operators
  O - see operator kills
  p - see channel parts
  s - show commands executed through service nicks
  S - see server kills
  t - see channel topic changes
  y - show various events such as ignore expiration, database
      saves, gline expiration, hub server connections etc.
*/

static void
o_umode(Luser *lptr, int ac, char **av, int sockfd)

{
	struct UmodeInfo	*umodeptr;
	Userlist *userptr, *tempuser;
	char str[MAXLINE];
	char *tmpstr;
	int minus,
	    ii;
	long newmodes;

	o_RecordCommand(sockfd,
		"UMODE %s",
		(ac < 2) ? "" : av[1]);

	if (sockfd != NOSOCKET)
		userptr = DccGetUser(IsDccSock(sockfd));
	else
		userptr = GetUser(1, onick, ouser, ohost);

	if (!userptr)
		return;

	if (ac < 2)
	{
		/*
		 * They didn't give an arguement - simply display their
		 * current usermodes
		 */
		ii = 0;
		for (umodeptr = Usermodes; umodeptr->usermode; umodeptr++)
		{
			if (userptr->umodes & umodeptr->flag)
				str[ii++] = umodeptr->usermode;
		}
		str[ii] = '\0';

		os_notice(lptr, sockfd,
			"Your usermodes are currently [\002+%s\002]",
			str);

		return;
	}

	/*
	 * Parse the given usermodes
	 */
	minus = 0;
	for (tmpstr = av[1]; *tmpstr; tmpstr++)
	{
		if (*tmpstr == '+')
			minus = 0;
		else if (*tmpstr == '-')
			minus = 1;
		else
		{
			/* try to find the usermode in Usermodes[] */
			if ((umodeptr = FindUmode(*tmpstr)))
			{
				/*
				 * its a usermode, if minus is 1, remove the usermode,
				 * otherwise add it to userptr->umodes
				 */
				if (minus)
					userptr->umodes &= ~umodeptr->flag;
				else
					userptr->umodes |= umodeptr->flag;
			}
		}
	}

	/*
	 * Now go through userlist, and find all structs that have
	 * this nickname - update their umodes
	 */
	newmodes = userptr->umodes;
	for (tempuser = userlist; tempuser; tempuser = tempuser->next)
	{
		if (!strcasecmp(userptr->nick, tempuser->nick))
			tempuser->umodes = newmodes;
	}

	/*
	 * Show them their new usermodes
	 */
	ii = 0;
	for (umodeptr = Usermodes; umodeptr->usermode; umodeptr++)
	{
		if (userptr->umodes & umodeptr->flag)
			str[ii++] = umodeptr->usermode;
	}
	str[ii] = '\0';

	os_notice(lptr, sockfd,
		"Your usermodes are now [\002+%s\002]",
		str);
} /* o_umode() */

#ifdef ALLOW_DUMP

/*
o_dump()
 Send a raw string to the server
*/

static void
o_dump(Luser *lptr, int ac, char **av, int sockfd)

{
	char *dstr;
	char tempstr[MAXLINE];
	char **newav;
	int newac;

	if (ac < 2)
	{
		os_notice(lptr, sockfd,
			"Syntax: \002DUMP <string>\002");
		return;
	}

	dstr = GetString(ac - 1, av + 1);

	o_RecordCommand(sockfd,
		"DUMP [%s]",
		dstr);

	toserv("%s\n", dstr);

	/*
	 * Technically, this av almost the same as the arv[]
	 * passed to ProcessInfo earlier, sending it through
	 * the whole routine again could cause problems -
	 * make a new one
	 */
	strcpy(tempstr, dstr);
	newac = SplitBuf(tempstr, &newav);
	ProcessInfo(newac, newav);

	MyFree(dstr);
	MyFree(newav);
} /* o_dump() */

#endif /* ALLOW_DUMP */

/*
FindUmode()
 Attempt to find the usermode 'usermode' in the Usermodes[] array.
Return a pointer to the corresponding index
*/

static struct UmodeInfo *
FindUmode(char usermode)

{
	struct	UmodeInfo	*umodeptr;

	for (umodeptr = Usermodes; umodeptr->usermode; umodeptr++)
	{
		if (umodeptr->usermode == usermode)
			return (umodeptr);
	}

	return ((struct UmodeInfo *) NULL);
} /* FindUmode() */

/*
modestr()
	return number of 'mode's specified by 'num'
	ie: if mode = 'o' and num = 3, return 'ooo'
*/

char *
modestr(int num, char mode)

{
	int		ii;
	char	*done;

	if (!num)
		return ((char *) NULL);

	done = (char *) MyMalloc((CHARSIZE * num) + 1);
	done[0] = '\0';

	for (ii = 0; ii < num; ii++)
		done[ii] = mode;

	done[ii] = '\0';

	return done;
} /* modestr() */

/*
TakeOver
	args: Channel *cptr
	purpose: join 'cptr->name', deop all users who don't match "f",
					 op those who do, clear all conflicting modes (imslk),
					 and clear all bans matching an admin
	return: none
*/

static void
TakeOver(Channel *cptr)

{
	int ii, acnt, mcnt;
	char sendstr[MAXLINE], done[MAXLINE];
	char *opnicks, *dopnicks;
	char **av, *abans, *mtmp;
	BanInfo *tempban;
	Userlist *tempuser;
	ChanUser *tempnick;

	if (cptr == (Channel *) NULL)
		return;

#ifdef SAVE_TS
	if (IsChannel(cptr->name) == (Chanlist *) NULL)
		os_join_ts_minus_1(cptr);
#endif 
	/*
	 * we don't need to deop anyone if we join with TS - 1, since its
	 * already done for us
	 */

	/* deop non "f" flag users, op "f" flag users */
	opnicks = (char *) MyMalloc(CHARSIZE);
	opnicks[0] = '\0';
	dopnicks = (char *) MyMalloc(CHARSIZE);
	dopnicks[0] = '\0';
	for (tempnick = cptr->firstuser; tempnick; tempnick = tempnick->next)
	{
		if (FindService(tempnick->lptr))
			continue;

		if (tempnick->lptr->flags & L_OSREGISTERED)
			ii = 1;
		else
			ii = 0;

		tempuser = GetUser(ii, tempnick->lptr->nick, tempnick->lptr->username, tempnick->lptr->hostname);
		if (IsChannelOp(cptr, tempnick->lptr) && !IsFriend(tempuser))
		{
			dopnicks = (char *) MyRealloc(dopnicks, strlen(dopnicks) + strlen(tempnick->lptr->nick) + (2 * CHARSIZE));
			strcat(dopnicks, tempnick->lptr->nick);
			strcat(dopnicks, " ");
		}
		else if (!IsChannelOp(cptr, tempnick->lptr) && IsFriend(tempuser))
		{
			opnicks = (char *) MyRealloc(opnicks, strlen(opnicks) + strlen(tempnick->lptr->nick) + (2 * CHARSIZE));
			strcat(opnicks, tempnick->lptr->nick);
			strcat(opnicks, " ");
		}
	} /* for (tempnick ..) */
	
#ifdef SAVE_TS

	if ((IsChannel(cptr->name) != (Chanlist *) NULL) || (cptr->since == 0))
	{
		/*
		 * if OperServ is monitoring the channel, then it didn't join with
		 * TS-1 and noone deoped, therefore we must manually deop
		 * non "f" flag users - likewise if the channel's TS is 0,
		 * join_ts_minus_1() would not deop anyone
		 */

#else
	if (1)
	{
#endif

		acnt = SplitBuf(dopnicks, &av);
		memset(&done, 0, MAXLINE);
		mcnt = 1;
		for (ii = 0; ii < acnt; ii++)
		{
			strcat(done, av[ii]);
			strcat(done, " ");
			if (mcnt == MaxModes)
			{
				mcnt = 0;
				mtmp = modestr(MaxModes, 'o');
				sprintf(sendstr, "-%s %s", mtmp, done);
				DoMode(cptr, sendstr, NO);
				MyFree(mtmp);
				memset(&done, 0, MAXLINE);
			}
			mcnt++;
		}
		if (done[0] != '\0')
		{
			mtmp = modestr(mcnt - 1, 'o');
			sprintf(sendstr, "-%s %s", mtmp, done);
			DoMode(cptr, sendstr, NO);
			MyFree(mtmp);
		}
		MyFree(av);
	}
	else
	{
		if (cptr->since == 0)
		{
			/* if the channel's TS is 0, we can't join with TS-1, so kick all
			   the users */

			acnt = SplitBuf(dopnicks, &av);
			for (ii = 0; ii < acnt; ii++)
			{
				toserv(":%s KICK %s %s :\n",
					n_OperServ,
					cptr->name,
					av[ii]);
				RemoveFromChannel(cptr, FindClient(av[ii]));
			}
			MyFree(av);
		}
	}
	MyFree(dopnicks);

	/* clear bans matching an administrator */
	abans = (char *) MyMalloc(CHARSIZE);
	abans[0] = '\0';
	for (tempban = cptr->firstban; tempban; tempban = tempban->next)
	{
		if (MatchesAdmin(tempban->mask) == 1)
		{
			abans = (char *) MyRealloc(abans, strlen(abans) + strlen(tempban->mask) + (2 * CHARSIZE));
			strcat(abans, tempban->mask);
			strcat(abans, " ");
		}
	}

	memset(&done, 0, MAXLINE);
	acnt = SplitBuf(abans, &av);
	mcnt = 1;
	for (ii = 0; ii < acnt; ii++)
	{
		strcat(done, av[ii]);
		strcat(done, " ");
		if (mcnt == MaxModes)
		{
			mcnt = 0;
			mtmp = modestr(MaxModes, 'b');
			sprintf(sendstr, "-%s %s", mtmp, done);
			DoMode(cptr, sendstr, NO);
			MyFree(mtmp);
			memset(&done, 0, MAXLINE);
		}
		mcnt++;
	}
	if (done[0] != '\0')
	{
		mtmp = modestr(mcnt - 1, 'b');
		sprintf(sendstr, "-%s %s", mtmp, done);
		DoMode(cptr, sendstr, NO);
		MyFree(mtmp);
	}
	MyFree(abans);
	MyFree(av);

	/* clear all bad modes (smilk) */
	memset(&done, 0, MAXLINE);
	strcpy(done, "+ntp-");
	if (cptr->modes & MODE_S)
		strcat(done, "s");
	if (cptr->modes & MODE_M)
		strcat(done, "m");
	if (cptr->modes & MODE_I)
		strcat(done, "i");
	if (cptr->limit)
		strcat(done, "l");
	if (cptr->key)
	{
		strcat(done, "k ");
		strcat(done, cptr->key);
	}

	DoMode(cptr, done, NO);

	/* op "f" flag users */
	memset(&done, 0, MAXLINE);
	acnt = SplitBuf(opnicks, &av);
	mcnt = 1;
	for (ii = 0; ii < acnt; ii++)
	{
		strcat(done, av[ii]);
		strcat(done, " ");
		if (mcnt == MaxModes)
		{
			mcnt = 0;
			mtmp = modestr(MaxModes, 'o');
			sprintf(sendstr, "+%s %s", mtmp, done);
			DoMode(cptr, sendstr, NO);
			MyFree(mtmp);
			memset(&done, 0, MAXLINE);
		}
		mcnt++;
	}
	if (done[0] != '\0')
	{
		mtmp = modestr(mcnt - 1, 'o');
		sprintf(sendstr, "+%s %s", mtmp, done);
		DoMode(cptr, sendstr, NO);
		MyFree(mtmp);
	}
	MyFree(opnicks);
	MyFree(av);

#ifdef SAVE_TS
	if (IsChannel(cptr->name) == (Chanlist *) NULL)
		os_part(cptr);
#endif
} /* TakeOver() */

/*
CalcMem()
	Determine memory usage of HybServ and send the info to 'socket'
or 'nick'
*/

static float
CalcMem(char *nick, int socket)

{
	Luser		*tempuser;
	UserChan	*userc;
	Channel	*tempchan;
	ChanUser	*chanu;
	BanInfo	*tempban;
#ifndef BLOCK_ALLOCATION
	Server	*tempserv;
#endif
	Userlist	*tempuserlist;
	Servlist	*tempservlist;
	Chanlist	*tempchanlist;
	Botlist		*tempbot;
	rHost			*temprhost;
	PortInfo	*tempprt;

#ifdef ALLOW_JUPES
	Jupe		*tempjupe;
#endif

#ifdef ALLOW_GLINES
	Gline		*gptr;
#endif

#ifdef NICKSERVICES
	NickInfo	*nptr;

#ifdef CHANNELSERVICES
	ChanInfo	*cptr;
#endif

#ifdef MEMOSERVICES
	MemoInfo	*mptr;
#endif

#endif /* NICKSERVICES */

#ifdef STATSERVICES
	HostHash	*hosth;
#endif

#if defined(NICKSERVICES) || defined(STATSERVICES)
	int		ii;
#endif

	Ignore		*tmpig;

	DccUser		*tmpconn;

	char	sendstr[MAXLINE];

	float	clientm; /* memory used by clients */
	float	servm; /* memory used by servers */
	float	chanm; /* memory used by channels */
	float	chanbanm; /* memory used by channel bans */
	float confm; /* total memory used by conf lines */
	float connm; /* memory used by dcc connections */
	float hashc; /* client hash */
	float hashch; /* channel hash */
	float hashcl; /* clone hash */
	float hashs; /* server hash */
	unsigned long chanbanc; /* number of channel bans */
	unsigned long confc; /* total number of conf lines */

#ifdef NICKSERVICES
	float nsm; /* mem used by NickServ */

#ifdef CHANNELSERVICES
	float csm; /* mem used by ChanServ */
#endif

#ifdef MEMOSERVICES
	float	msm; /* mem used by MemoServ */
#endif

#endif /* NICKSERVICES */

#ifdef STATSERVICES
	float	ssm; /* mem used by StatServ */
#endif

	float igm; /* mem used for ignore list */
	unsigned long igc; /* number of ignore entries */

#ifdef BLOCK_ALLOCATION
	Heap	*heapptr;
	SubBlock	*sub;
#endif

	float	total; /* total memory */

	clientm = 0;
	for (tempuser = ClientList; tempuser; tempuser = tempuser->next)
	{
	#ifndef BLOCK_ALLOCATION
		clientm += (strlen(tempuser->nick) + strlen(tempuser->username) +
								strlen(tempuser->hostname) + strlen(tempuser->realname) +
								sizeof(Luser));
	#endif

		for (userc = tempuser->firstchan; userc; userc = userc->next)
			clientm += sizeof(UserChan);
	}

#ifdef BLOCK_ALLOCATION

	clientm += sizeof(Heap);
	heapptr = ClientHeap;
	for (sub = heapptr->base; sub; sub = sub->next)
		clientm += ((heapptr->ElementsPerBlock * heapptr->ElementSize) +
								sizeof(SubBlock));

#endif /* BLOCK_ALLOCATION */

	chanm = 0;
	chanbanm = 0;
	chanbanc = 0;
	for (tempchan = ChannelList; tempchan; tempchan = tempchan->next)
	{
	#ifndef BLOCK_ALLOCATION
		int 	keylen = 0;

		if (tempchan->key)
			keylen = strlen(tempchan->key);

		chanm += (strlen(tempchan->name) + keylen + sizeof(Channel));
	#endif /* BLOCK_ALLOCATION */

		for (chanu = tempchan->firstuser; chanu; chanu = chanu->next)
			chanm += sizeof(ChanUser);
		for (tempban = tempchan->firstban; tempban; tempban = tempban->next)
		{
			chanbanc++;
			chanbanm += (strlen(tempban->mask) + sizeof(BanInfo));
			if (tempban->who)
				chanbanm += strlen(tempban->who);
		}
	}

#ifdef BLOCK_ALLOCATION

	chanm += sizeof(Heap);
	heapptr = ChannelHeap;
	for (sub = heapptr->base; sub; sub = sub->next)
		chanm += ((heapptr->ElementsPerBlock * heapptr->ElementSize) +
							sizeof(SubBlock));

#endif /* BLOCK_ALLOCATION */

	servm = 0;

#ifdef BLOCK_ALLOCATION
	servm += sizeof(Heap);
	heapptr = ServerHeap;
	for (sub = heapptr->base; sub; sub = sub->next)
		servm += ((heapptr->ElementsPerBlock * heapptr->ElementSize) +
							sizeof(SubBlock));
#else
	for (tempserv = ServerList; tempserv; tempserv = tempserv->next)
		servm += (strlen(tempserv->name) + sizeof(Server));
#endif /* BLOCK_ALLOCATION */

	confm = (strlen(Me.admin) + strlen(Me.name) + strlen(Me.info));
	confc = 2;

	for (tempuserlist = userlist; tempuserlist; tempuserlist = tempuserlist->next)
	{
		confm += (strlen(tempuserlist->username) +
							strlen(tempuserlist->hostname) +
							strlen(tempuserlist->password) +
							sizeof(Userlist));
		confc++;
	}
	
	for (tempservlist = serverlist; tempservlist; tempservlist = tempservlist->next)
	{
		confm += (strlen(tempservlist->hostname) +
		          strlen(tempservlist->password) +
		          sizeof(Servlist));

		if (tempservlist->realname)
			confm += strlen(tempservlist->realname);

		confc++;
	}

	for (tempchanlist = mychanlist; tempchanlist; tempchanlist = tempchanlist->next)
	{
		confm += (strlen(tempchanlist->name) + sizeof(Chanlist));
		confc++;
	}

	for (tempbot = botlist; tempbot; tempbot = tempbot->next)
	{
		confm += (strlen(tempbot->name) +
		          strlen(tempbot->hostname) +
		          strlen(tempbot->password) +
		          sizeof(Botlist));
		if (tempbot->username)
			confm += strlen(tempbot->username);
		confc++;
	}

	for (tempbot = remotebots; tempbot; tempbot = tempbot->next)
	{
		confm += (strlen(tempbot->name) +
		          strlen(tempbot->username) +
		          strlen(tempbot->hostname) +
		          strlen(tempbot->password) +
		          sizeof(Botlist));
		confc++;
	}

	for (temprhost = rhostlist; temprhost; temprhost = temprhost->next)
	{
		confm += (strlen(temprhost->username) +
		          strlen(temprhost->hostname) +
		          sizeof(rHost));
		confc++;
	}

	for (tempprt = portlist; tempprt; tempprt = tempprt->next)
	{
		confm += sizeof(PortInfo);
		if (tempprt->host)
			confm += strlen(tempprt->host);
		confc++;
	}

#ifdef ALLOW_JUPES
	for (tempjupe = jupelist; tempjupe; tempjupe = tempjupe->next)
		confm += (strlen(tempjupe->name) +
		          strlen(tempjupe->reason) +
		          strlen(tempjupe->who) +
		          sizeof(Jupe));
	confc += Network->TotalJupes;
#endif

#ifdef ALLOW_GLINES
	for (gptr = glinelist; gptr; gptr = gptr->next)
	{
		confm += (strlen(gptr->username) +
		          strlen(gptr->hostname) +
		          strlen(gptr->reason) +
		          strlen(gptr->who) +
		          sizeof(Gline));
		/* 
		 * make sure the gline isn't temporary because it wouldn't have
		 * a conf line
		 */
		if (!(gptr->expires))
			confc++;
	}
#endif

	igc = 0;
	igm = 0;
	for (tmpig = ignorelist; tmpig; tmpig = tmpig->next)
	{
		igc++;
		igm += (strlen(tmpig->hostmask) + sizeof(Ignore));
	}

	connm = 0;
	for (tmpconn = connections; tmpconn; tmpconn = tmpconn->next)
	{
		connm += sizeof(DccUser);
		if (tmpconn->nick)
			connm += strlen(tmpconn->nick);
		if (tmpconn->username)
			connm += strlen(tmpconn->username);
		if (tmpconn->hostname)
			connm += strlen(tmpconn->hostname);
	}

#ifdef NICKSERVICES

	nsm = sizeof(nicklist);
	for (ii = 0; ii < NICKLIST_MAX; ii++)
	{
		if (nicklist[ii])
		{
			NickHost	*hptr;

			for (nptr = nicklist[ii]; nptr; nptr = nptr->next)
			{
				nsm += (strlen(nptr->nick) + sizeof(NickInfo));
				if (nptr->password)
					nsm += strlen(nptr->password);
				if (nptr->email)
					nsm += strlen(nptr->email);
				if (nptr->url)
					nsm += strlen(nptr->url);

				if (nptr->lastu)
					nsm += strlen(nptr->lastu);
				if (nptr->lasth)
					nsm += strlen(nptr->lasth);
				if (nptr->lastqmsg)
					nsm += strlen(nptr->lastqmsg);

				for (hptr = nptr->hosts; hptr; hptr = hptr->next)
					nsm += (strlen(hptr->hostmask) + sizeof(NickHost));
			}
		}
	} /* for (ii = 0; ii < NICKLIST_MAX; ii++) */
	
#ifdef CHANNELSERVICES

	csm = sizeof(chanlist);
	for (ii = 0; ii < CHANLIST_MAX; ii++)
	{
		AutoKick *ak;
		ChanAccess *ca;

		for (cptr = chanlist[ii]; cptr; cptr = cptr->next)
		{
			csm += (strlen(cptr->name) + sizeof(ChanInfo));
			if (cptr->password)
				csm += strlen(cptr->password);
			if (cptr->founder)
				csm += strlen(cptr->founder);
			if (cptr->successor)
				csm += strlen(cptr->successor);
			if (cptr->topic)
				csm += strlen(cptr->topic);
			if (cptr->key)
				csm += strlen(cptr->key);
			if (cptr->entrymsg)
				csm += strlen(cptr->entrymsg);
			if (cptr->email)
				csm += strlen(cptr->email);
			if (cptr->url)
				csm += strlen(cptr->url);
			for (ak = cptr->akick; ak; ak = ak->next)
			{
				csm += (strlen(ak->hostmask) + sizeof(AutoKick));
				if (ak->reason)
					csm += strlen(ak->reason);
			}
			for (ca = cptr->access; ca; ca = ca->next)
			{
				csm += sizeof(ChanAccess);
				if (ca->hostmask)
					csm += strlen(ca->hostmask);
			}
		}
	} /* for (ii = 0; ii < CHANLIST_MAX; ii++) */

#endif /* CHANNELSERVICES */

#ifdef MEMOSERVICES

	msm = sizeof(memolist);
	for (ii = 0; ii < MEMOLIST_MAX; ii++)
	{
		Memo *memoptr;

		for (mptr = memolist[ii]; mptr; mptr = mptr->next)
		{
			msm += (strlen(mptr->name) + sizeof(MemoInfo));
			for (memoptr = mptr->memos; memoptr; memoptr = memoptr->next)
				msm += (strlen(memoptr->sender) + strlen(memoptr->text) + sizeof(Memo));
		}
	} /* for (ii = 0; ii < MEMOLIST_MAX; ii++) */

#endif /* MEMOSERVICES */

#endif /* NICKSERVICES */

#ifdef STATSERVICES

	ssm = 0;
	for (ii = 0; ii < HASHCLIENTS; ii++)
	{
		if (hostTable[ii].list)
		{
			for (hosth = hostTable[ii].list; hosth; hosth = hosth->hnext)
				ssm += (strlen(hosth->hostname) + sizeof(HostHash) + sizeof(aHashEntry));
		}
		else
			ssm += sizeof(aHashEntry);
	}
	
#endif /* STATSERVICES */

	hashc = HASHCLIENTS * sizeof(aHashEntry);
	hashch = HASHCHANNELS * sizeof(aHashEntry);
	hashcl = HASHCLIENTS * sizeof(aHashEntry);
	hashs = HASHSERVERS * sizeof(aHashEntry);

	total = clientm +
	        chanm +
	        chanbanm +
	        servm +
	        confm +
	        connm +
	        hashc +
	        hashch +
	        hashcl +
	        hashs;

#ifdef NICKSERVICES
	total += nsm;

#ifdef CHANNELSERVICES
	total += csm;
#endif

#ifdef MEMOSERVICES
	total += msm;
#endif

#endif /* NICKSERVICES */

#ifdef STATSERVICES
	total += ssm;
#endif

	total += igm;

	if ((nick) || (socket != NODCC))
	{
		/* User structures usage */
		sprintf(sendstr,
			"Clients:    %5.0f (%10.0fb) (%10.2fkb)\n",
			Network->TotalUsers,
			clientm,
			clientm / 1024);
		if (socket == NODCC)
			toserv(":%s NOTICE %s :%s",
				n_OperServ,
				nick,
				sendstr);
		else
			writesocket(socket, sendstr);

		/* Server structures usage */
		sprintf(sendstr,
			"Servers:    %5.0f (%10.0fb) (%10.2fkb)\n",
			Network->TotalServers,
			servm,
			servm / 1024);
		if (socket == NODCC)
			toserv(":%s NOTICE %s :%s",
				n_OperServ,
				nick,
				sendstr);
		else
			writesocket(socket, sendstr);

		/* Channel structures usage */
		sprintf(sendstr,
			"Channels:   %5.0f (%10.0fb) (%10.2fkb)\n",
			Network->TotalChannels,
			chanm,
			chanm / 1024);
		if (socket == NODCC)
			toserv(":%s NOTICE %s :%s",
				n_OperServ,
				nick,
				sendstr);
		else
			writesocket(socket, sendstr);

		/* Channel ban mem usage */
		sprintf(sendstr,
			"Bans:       %5ld (%10.0fb) (%10.2fkb)\n",
			chanbanc,
			chanbanm,
			chanbanm / 1024);
		if (socket == NODCC)
			toserv(":%s NOTICE %s :%s",
				n_OperServ,
				nick,
				sendstr);
		else
			writesocket(socket, sendstr);

		sprintf(sendstr,
			"Conf lines: %5ld (%10.0fb) (%10.2fkb)\n",
			confc,
			confm,
			confm / 1024);
		if (socket == NODCC)
			toserv(":%s NOTICE %s :%s",
				n_OperServ,
				nick,
				sendstr);
		else
			writesocket(socket, sendstr);

		sprintf(sendstr,
			"Ignores: %8ld (%10.0fb) (%10.2fkb)\n",
			igc,
			igm,
			igm / 1024);
		if (socket == NODCC)
			toserv(":%s NOTICE %s :%s",
				n_OperServ,
				nick,
				sendstr);
		else
			writesocket(socket, sendstr);

		sprintf(sendstr,
			"Partyline: %6d (%10.0fb) (%10.2fkb)\n",
			Network->TotalConns,
			connm,
			connm / 1024);
		if (socket == NODCC)
			toserv(":%s NOTICE %s :%s",
				n_OperServ,
				nick,
				sendstr);
		else
			writesocket(socket, sendstr);

	#ifdef NICKSERVICES
		sprintf(sendstr,
			"NickServ:         (%10.0fb) (%10.2fkb)\n",
			nsm,
			nsm / 1024);
		if (socket == NODCC)
			toserv(":%s NOTICE %s :%s",
				n_OperServ,
				nick,
				sendstr);
		else
			writesocket(socket, sendstr);

	#ifdef CHANNELSERVICES
		sprintf(sendstr,
			"ChanServ:         (%10.0fb) (%10.2fkb)\n",
			csm,
			csm / 1024);
		if (socket == NODCC)
			toserv(":%s NOTICE %s :%s",
				n_OperServ,
				nick,
				sendstr);
		else
			writesocket(socket, sendstr);
	#endif

	#ifdef MEMOSERVICES
		sprintf(sendstr,
			"MemoServ:         (%10.0fb) (%10.2fkb)\n",
			msm,
			msm / 1024);
		if (socket == NODCC)
			toserv(":%s NOTICE %s :%s",
				n_OperServ,
				nick,
				sendstr);
		else
			writesocket(socket, sendstr);
	#endif

	#endif /* NICKSERVICES */

	#ifdef STATSERVICES
		sprintf(sendstr,
			"StatServ:         (%10.0fb) (%10.2fkb)\n",
			ssm,
			ssm / 1024);
		if (socket == NODCC)
			toserv(":%s NOTICE %s :%s",
				n_OperServ,
				nick,
				sendstr);
		else
			writesocket(socket, sendstr);
	#endif /* STATSERVICES */

		/* Client Hash */
		sprintf(sendstr,
			"Client Hash:      (%10.0fb) (%10.2fkb)\n",
			hashc,
			hashc / 1024);
		if (socket == NODCC)
			toserv(":%s NOTICE %s :%s",
				n_OperServ,
				nick,
				sendstr);
		else
			writesocket(socket, sendstr);

		/* Channel Hash */
		sprintf(sendstr,
			"Channel Hash:     (%10.0fb) (%10.2fkb)\n",
			hashch,
			hashch / 1024);
		if (socket == NODCC)
			toserv(":%s NOTICE %s :%s",
				n_OperServ,
				nick,
				sendstr);
		else
			writesocket(socket, sendstr);

		/* Clone Hash */
		sprintf(sendstr,
			"Clone Hash:       (%10.0fb) (%10.2fkb)\n",
			hashcl,
			hashcl / 1024);
		if (socket == NODCC)
			toserv(":%s NOTICE %s :%s",
				n_OperServ,
				nick,
				sendstr);
		else
			writesocket(socket, sendstr);

		/* Total memory used */
		sprintf(sendstr,
			"Total:            (%10.0fb) (%10.2fkb) (%10.2fmb)\n",
			total,
			total / 1024,
			(total / 1024) / 1024);
		if (socket == NODCC)
			toserv(":%s NOTICE %s :%s",
				n_OperServ,
				nick,
				sendstr);
		else
			writesocket(socket, sendstr);
	} /* if (nick || (socket != NODCC)) */

	return (total);
} /* CalcMem() */
