/*
 * HybServ TS Services, Copyright (C) 1998-1999 Patrick Alken
 * This program comes with absolutely NO WARRANTY
 *
 * Should you choose to use and/or modify this source code, please
 * do so under the terms of the GNU General Public License under which
 * this program is distributed.
 */

static char *rcsid = "$Id: server.c,v 2.30 1998/08/14 20:48:47 wnder Exp $";

#include	"headers.h"

#ifdef STATSERVICES

extern  aHashEntry hostTable[HASHCLIENTS];

#endif /* STATSERVICES */

/*
CheckChans()
	purpose: this function is called after a rehash, to check if any
					 new channels have been added to the config file, or any
					 old channels removed from the config file.  Any new
					 channels are join()'d and old channels no longer
					 being monitored are part()'d.
	return: none
*/

void
CheckChans()

{
	Chanlist *tempchan, *prev;
	Channel *cptr;

	/* 
	 * go through channel list and PART any channels that were removed
	 * from the conf
	 */
	prev = (Chanlist *) NULL;
	for (tempchan = mychanlist; tempchan; )
	{
		if (tempchan->flags & CHAN_DELETE)
		{
			putlog(LOG1, "%s: No longer monitoring %s",
				n_OperServ,
				tempchan->name);

			cptr = FindChannel(tempchan->name);
			if (IsChannelMember(cptr, Me.osptr))
				os_part(cptr);

			MyFree(tempchan->name);

			if (prev)
			{
				prev->next = tempchan->next;
				MyFree(tempchan);
				tempchan = prev;
			}
			else
			{
				mychanlist = tempchan->next;
				MyFree(tempchan);
				tempchan = (Chanlist *) NULL;
			}

			Network->MyChans--;
		} /* if (tempchan->flags & CHAN_DELETE) */

		prev = tempchan;

		if (tempchan)
			tempchan = tempchan->next;
		else
			tempchan = mychanlist;
	}

	/*
	 * go through channel list and JOIN any new channels that were added 
	 * to the conf
	 */
	for (tempchan = mychanlist; tempchan; tempchan = tempchan->next)
	{
		cptr = FindChannel(tempchan->name);
		if (cptr && !IsChannelMember(cptr, Me.osptr))
		{
			os_join(cptr);
			putlog(LOG1, "%s: Now monitoring channel %s",
				n_OperServ,
				tempchan->name);
		}
	}
} /* CheckChans() */

/*
CheckServers()
 Called after a rehash to remove any servers (S: lines) that are
marked for deletion
*/

void
CheckServers()

{
	Servlist	*tempserv, *prev;

	prev = (Servlist *) NULL;

	for (tempserv = serverlist; tempserv; )
	{
		if (tempserv->flags & SERV_DELETE)
		{
			if (tempserv == currenthub)
			{
				/*
				 * Someone deleted our current hub server out of
				 * hybserv.conf .. just keep it in the list, or it will
				 * screw a lot of stuff up :-)
				 */
				tempserv->flags &= ~SERV_DELETE;
				tempserv = tempserv->next;
				continue;
			}

			MyFree(tempserv->hostname);
			MyFree(tempserv->password);
			if (tempserv->realname)
				MyFree(tempserv->realname);

			if (prev)
			{
				prev->next = tempserv->next;
				MyFree(tempserv);
				tempserv = prev;
			}
			else
			{
				serverlist = tempserv->next;
				MyFree(tempserv);
				tempserv = (Servlist *) NULL;
			}
		}

		prev = tempserv;
		if (tempserv)
			tempserv = tempserv->next;
		else
			tempserv = serverlist;
	}
} /* CheckServers() */

/*
CheckListenPorts()
	Delete any ports who no longer have a P: line, and start listening
on new ports
*/

void
CheckListenPorts()

{
	PortInfo	*tempport, *prev;

	prev = (PortInfo *) NULL;
	for (tempport = portlist; tempport; )
	{
		if (tempport->type == PRT_DELETE)
		{
			putlog(LOG1, "No longer listening on port %d %s%s%s",
				tempport->port,
				tempport->host ? "[" : "",
				tempport->host ? tempport->host : "",
				tempport->host ? "]" : "");

			if (tempport->socket != NOSOCKET)
				close(tempport->socket);

			if (tempport->host)
				MyFree(tempport->host);

			if (prev)
			{
				prev->next = tempport->next;
				MyFree(tempport);
				tempport = prev;
			}
			else
			{
				portlist = tempport->next;
				MyFree(tempport);
				tempport = (PortInfo *) NULL;
			}
		} /* if (tempport->type == PRT_DELETE) */

		prev = tempport;
		if (tempport)
			tempport = tempport->next;
		else
			tempport = portlist;
	}

	/*
	 * Now go through and listen on new ports
	 */
	for (tempport = portlist; tempport; tempport = tempport->next)
		if (!IsListening(tempport->port))
			do_listen(tempport);

} /* CheckListenPorts() */

/*
do_squit()
	SQUIT all servers matching serv
*/

void
do_squit(char *serv, char *reason)

{
	Server	*tempserv, *prev;

	for (tempserv = ServerList; tempserv; )
	{
		/* squit the server */
		if (match(serv, tempserv->name))
		{
			toserv("SQUIT %s :%s\n",
				tempserv->name, 
				reason);

			prev = tempserv->next;

			DeleteServer(tempserv); /* remove server from list */

			tempserv = prev;
		}
		else
			tempserv = tempserv->next;
	}
} /* do_squit() */

/*
notice()
	purpose: send a NOTICE to 'nick' with 'msg'
	return: none
*/

void
notice(char *from, char *nick, char *format, ...)

{
	char	finalstr[MAXLINE];
	char	who[MAXLINE];
	va_list	args;

	va_start(args, format);

#ifdef HAVE_VSNPRINTF
	vsnprintf(finalstr, sizeof(finalstr) - 1, format, args);
	finalstr[sizeof(finalstr) - 1] = '\0';
#else
	vsprintf(finalstr, format, args);
#endif /* HAVE_VSNPRINTF */

	if (ServerNotices)
		strcpy(who, Me.name);
	else
		strcpy(who, from);

	toserv(":%s NOTICE %s :%s\n", 
		who,
		nick,
		finalstr);

	va_end(args);
} /* notice() */

/*
RecordCommand()
 Called when someone executes a NickServ/ChanServ/MemoServ/StatServ
command - log it and send it to opers with usermode +s
*/

void
RecordCommand(char *format, ...)

{
	va_list args;
	char buffer[MAXLINE];

	va_start(args, format);

#ifdef HAVE_VSNPRINTF
	vsnprintf(buffer, sizeof(buffer) - 1, format, args);
	buffer[sizeof(buffer) - 1] = '\0';
#else
	vsprintf(buffer, format, args);
#endif /* HAVE_VSNPRINTF */

	va_end(args);

	/* log the command */
	putlog(LOG2, buffer);

	/* send it to opers with usermode +s */
	SendUmode(OPERUMODE_S, buffer);
} /* RecordCommand() */

/*
Substitute()
	args: char *nick, char *str, int sockfd
	purpose: replace any formatting characters in 'str' with the
					 corresponding information.

		Formatting characters:
      %O - Nickname of n_OperServ
      %N - Nickname of n_NickServ
      %C - Nickname of n_ChanServ
      %M - Nickname of n_MemoServ
      %T - Nickname of n_StatServ
      %H - Nickname of n_HelpServ
      %S - Name of this server
      %A - Administrative info
      %V - current version
      %B - bold character
      %+<flag> - needs <flag> to read the line

	return: NULL if its a blank line, -1 if 'nick' doesn't have privs
					to read the line; otherwise a ptr to substituted string
*/

char *
Substitute(char *nick, char *str, int sockfd)

{
	char	tempstr[MAXLINE], key;
	char	*finalstr = (char *) NULL;
	int		tcnt, fcnt;
	Luser	*lptr;

	lptr = FindClient(nick);

	strcpy(tempstr, str);
	finalstr = (char *) MyMalloc(MAXLINE);
	memset(finalstr, 0, MAXLINE);
	fcnt = 0;
	tcnt = 0;
	while (tcnt < MAXLINE)
	{
		if (IsEOL(tempstr[tcnt]))
			break;

		if (tempstr[tcnt] == '%')
		{
			key = tempstr[++tcnt];
			switch (key)
			{
				case 'o':
				case 'O':
				{
					strcat(finalstr, n_OperServ);
					fcnt += strlen(n_OperServ) - 1;
					break;
				}
				case 'n':
				case 'N':
				{
					strcat(finalstr, n_NickServ);
					fcnt += strlen(n_NickServ) - 1;
					break;
				}
				case 'c':
				case 'C':
				{
					strcat(finalstr, n_ChanServ);
					fcnt += strlen(n_ChanServ) - 1;
					break;
				}
				case 'm':
				case 'M':
				{
					strcat(finalstr, n_MemoServ);
					fcnt += strlen(n_MemoServ) - 1;
					break;
				}
				case 't':
				case 'T':
				{
					strcat(finalstr, n_StatServ);
					fcnt += strlen(n_StatServ) - 1;
					break;
				}
				case 'h':
				case 'H':
				{
					strcat(finalstr, n_HelpServ);
					fcnt += strlen(n_HelpServ) - 1;
					break;
				}
				case 's':
				case 'S':
				{
					strcat(finalstr, Me.name);
					fcnt += strlen(Me.name) - 1;
					break;
				}
				case 'b':
				case 'B':
				{
					strcat(finalstr, "\002");
					break;
				}
				case 'v':
				case 'V':
				{
					strcat(finalstr, hVersion);
					fcnt += strlen(hVersion) - 1;
					break;
				}
				case 'a':
				case 'A':
				{
					strcat(finalstr, Me.admin);
					fcnt += strlen(Me.admin) - 1;
					break;
				}
				case '+':
				{
					char flag;
					char *cptr, *finstr;
					Userlist	*tempuser = (Userlist *) NULL;

					flag = tempstr[tcnt + 1];
					if (!nick)
						tempuser = DccGetUser(IsDccSock(sockfd));
					else if (lptr)
						tempuser = GetUser(1, lptr->nick, lptr->username, lptr->hostname);
					else
						tempuser = GetUser(1, nick, (char *) NULL, (char *) NULL);

					if ((CheckAccess(tempuser, flag)))
					{
						if (!IsRegistered(lptr, sockfd))
							return ((char *) -1);

						tcnt += 2;
						cptr = &tempstr[tcnt];
						finstr = Substitute(nick, cptr, sockfd);
						if (!finstr)
						{
							/* its a line like "%+a" with no text, so send a \r\n */
							MyFree(finstr);
							finstr = MyStrdup("\r\n");
						}
						return (finstr);
					}
					else
						return ((char *) -1); /* user doesn't have privs to read line */
					break;
				}

				default:
				{
					strcat(finalstr, "%");
					finalstr[++fcnt] = tempstr[tcnt];
					break;
				}
			} /* switch (key) */
		}
		else
			finalstr[fcnt] = tempstr[tcnt];

		fcnt++;
		tcnt++;
	}

	if (finalstr[0])
	{
		finalstr[fcnt++] = '\n';
		finalstr[fcnt] = '\0';
		return (finalstr);
	}
	else
	{
		MyFree(finalstr);
		return ((char *) NULL);
	}
} /* Substitute() */

/*
do_shutdown()
	Shut down services
*/

void
do_shutdown(char *who, char *reason)

{
	char	sendstr[MAXLINE];
	PortInfo	*pptr;

	putlog(LOG1, "Shutting down services");

	if (!WriteDatabases())
		putlog(LOG1, "Database update failed");

#if defined(NICKSERVICES) && defined(CHANNELSERVICES)
	toserv(":%s QUIT :%s\n",
		n_ChanServ,
		"Shutting Down");
#endif

	/* close listening sockets */
	for (pptr = portlist; pptr; pptr = pptr->next)
		if (pptr->socket != NOSOCKET)
			close(pptr->socket);

	if (reason)
	{
		if (who)
			sprintf(sendstr, "%s (authorized by %s)", reason, who);
		else
			strcpy(sendstr, reason);
	}
	else
		if (who)
			sprintf(sendstr, "Authorized by %s", who);
		else
			sendstr[0] = '\0';

	toserv(":%s QUIT :%s\nSQUIT %s :%s\n",
		n_OperServ,
		"Shutting Down",
		currenthub->realname ? currenthub->realname : currenthub->hostname,
		sendstr);

	exit (1);
} /* do_shutdown() */

/*
ClearUsers()
	args: none
	purpose: clear all users in linked list
	return: none
*/

void
ClearUsers()

{
	Luser	*next;

	while (ClientList)
	{
		next = ClientList->next;
		DeleteClient(ClientList);
		ClientList = next;
	}
} /* ClearUsers() */

/*
ClearServs()
	args: none
	purpose: delete all servers in list
	return: none
*/

void
ClearServs()

{
	Server	*next;

	while (ServerList)
	{
		next = ServerList->next;
		DeleteServer(ServerList);
		ServerList = next;
	}
} /* ClearServs() */

/*
ClearChans()
	args: none
	purpose: delete all channels in list
	return: none
*/

void
ClearChans()

{
	Channel	*next;

	while (ChannelList)
	{
		next = ChannelList->next;
		DeleteChannel(ChannelList);
		ChannelList = next;
	}
} /* ClearChans() */

/*
s_pass()
*/

void
s_pass(int ac, char **av)

{
	/* do nothing */
	return;
} /* s_pass() */

/*
s_ping()
 Respond to a ping request

 If a server is requesting a PING
  av[0] = "PING"
  av[1] = :servername

 If a client is requesting a PING
	av[0] = :nickname
	av[1] = "PING"
	av[2] = nickname
	av[3] = Me.name
*/

void
s_ping(int ac, char **av)

{
	char *who;

	if (ac < 2)
		return;

	if (!strcasecmp(av[0], "PING"))
		who = av[1];
	else
		who = av[0];

	if (*who == ':')
		who++;

	toserv(":%s PONG %s :%s\n", 
		Me.name, 
		Me.name, 
		who);
} /* s_ping() */

/*
s_server()
	Process a SERVER statement from hub

If hub is identifying itself:
  av[0] = "SERVER"
	av[1] = server name
	av[2] = hop count
	av[3] = description

If its another server:
	av[0] = hub name
	av[1] = "SERVER"
	av[2] = server name
	av[3] = hop count
	av[4] = description
*/

void
s_server(int ac, char **av)

{
	char sendstr[MAXLINE];
	char **line;
	Server *tempserv;

	if (ac < 4)
		return;

	/*
	 * make sure av[2] isn't in the server list - one reason it might
	 * be is if services squit'd it and created a fake server
	 * in the connect burst, but the hub would send the server line
	 * to us anyway, before processing our SQUIT
	 */
	if (FindServer(av[2]))
		return;

	tempserv = AddServer(ac, av);
	if (tempserv == Me.hub)
	{
		/*
		 * it's the SERVER line from the hub in the initial burst
		 */

		/*
		 * get the network server name of hub in case it
		 * doesn't match it's hostname 
	 	 */
		currenthub->realname = MyStrdup(av[1]);

		/*
		 * add the services (our) server as well
		 */
		sprintf(sendstr, ":%s SERVER %s 1 :%s",
			av[1],
			Me.name,
			Me.info);

		SplitBuf(sendstr, &line);

		AddServer(5, line);

		MyFree(line);

		/*
		 * Set our hub's uplink as services (we are the hub of our
		 * uplink)
		 */
		tempserv->uplink = Me.sptr;

		/* initialize service nicks */
		InitServs((Luser *) NULL);

	#ifdef ALLOW_JUPES
		/* Introduce juped nicks/servers to the network */
		InitJupes();
	#endif

	}
	else
	{
	#ifdef ALLOW_JUPES
		/* a new server connected to network, check if its juped */
		CheckJuped(av[2]);
	#endif
	}
} /* s_server() */

/*
s_nick()
  Process a NICK statement

	if new user:
	av[1] = nick name
	av[2] = hop count
	av[3] = TimeStamp
	av[4] = usermodes
	av[5] = username
	av[6] = hostname
	av[7] = server
	av[8] = real name

	if nick change:
	av[0] = who
	av[1] = "NICK"
	av[2] = new nick
	av[3] = TS
*/

void
s_nick(int ac, char **av)

{
	Luser	*lptr,
	      *serviceptr;
#ifdef NICKSERVICES
	NickInfo	*nptr, *newptr;
#endif

#ifdef RECORD_SPLIT_TS
	Userlist	*uptr;
#endif

	if (ac == 4)
	{
		char	*who;
	#if defined(BLOCK_ALLOCATION) || defined(NICKSERVICES)
		char	newnick[NICKLEN + 1];
	#endif

		/* nick change, not new user */

		if (av[0][0] == ':')
			who = av[0] + 1;
		else
			who = av[0];

		lptr = FindClient(who);
		if (lptr == (Luser *) NULL)
			return;

	#if defined(BLOCK_ALLOCATION) || defined(NICKSERVICES)
		/*memset(&newnick, 0, NICKLEN + 1);*/
		strncpy(newnick, av[2], NICKLEN);
		newnick[NICKLEN] = '\0';
	#endif

		SendUmode(OPERUMODE_N,
			"*** Nick Change: %s -> %s",
			lptr->nick,
			av[2]);

	#ifdef NICKSERVICES
		nptr = FindNick(lptr->nick);

		newptr = FindNick(newnick);
		if (newptr)
		{
			/*
			 * Un-Identify the new nickname in case it is registered -
			 * it could possibly be identified from the flags read
			 * from the userfile, if no-one has used this nickname
			 * yet.
			 */
			newptr->flags &= ~NS_IDENTIFIED;
		}

		if (nptr)
		{

		#ifdef LINKED_NICKNAMES
			/*
			 * One of the features of linked nicknames is if you
			 * identify for one nick in the link, you are identified
			 * for every nick. If lptr->nick is changing his/her nick
			 * to another nickname in the link (and is currently
			 * identified), make sure they get correctly identified
			 * for the new nickname.
			 */
			if (nptr->flags & NS_IDENTIFIED)
			{
				if (newptr && IsLinked(nptr, newptr))
					newptr->flags |= NS_IDENTIFIED;
			}
		#endif /* LINKED_NICKNAMES */
			
			/*
			 * Un-Identify the old nickname if it is registered
			 */
			nptr->flags &= ~NS_IDENTIFIED;
		} /* if (nptr) */
	#endif /* NICKSERVICES */

		/*
		 * If the user is a generic operator who registered themselves
		 * by sending OperServ a message, remove that flag now -
		 * This code is here for a very special case scenario: If a
		 * malicious operator registers him/herself with OperServ,
		 * by sending a generic operator command, and then changes
		 * their nickname to one which has administrator privileges
		 * in hybserv.conf, they will be registered without having
		 * to give the correct password, or even without having
		 * to match the correct hostmask. This code removes
		 * their registration flag to ensure that doesn't happen.
		 * Execute this code even if OpersHaveAccess is turned
		 * off, because an operator could have registered with
		 * OperServ, while it was turned on (before a rehash
		 * etc) and the operator could get administrator access
		 * that way.
		 */
		if (GetUser(1, lptr->nick, lptr->username, lptr->hostname) == GenericOper)
			lptr->flags &= ~L_OSREGISTERED;

	#ifdef ALLOW_FUCKOVER
		/* check if old nick was being flooded */
		check_ftarget(lptr, av[2]);
	#endif
		
		HashDelClient(lptr, 1);

		/* Erase old nick, and store new nick in it's position */

	#ifdef BLOCK_ALLOCATION
		strcpy(lptr->nick, newnick);
	#else
		MyFree(lptr->nick);
		lptr->nick = MyStrdup(av[2]);
	#endif /* BLOCK_ALLOCATION */

		HashAddClient(lptr, 1);

	#ifdef STATSERVICES
		lptr->numnicks++;
	#endif

	#ifdef NICKSERVICES
		/*
		 * Update nick_ts, so if the new nick is registered, we know
		 * when to do the KILL if they haven't registered within
		 * 1 minute
		 */
		lptr->nick_ts = atoi(av[3] + 1);

		/*
		 * check if new nick is stealing someone's nick - but only
		 * if who != av[2], because the user "aba" might change
		 * their nick to "aBa", in which case we don't need to
		 * check
		 */
		if (strcasecmp(who, av[2]) != 0)
			CheckNick(av[2]);
	#endif /* NICKSERVICES */

	#ifdef ALLOW_JUPES
		CheckJuped(av[2]); /* check if new nick is juped */
	#endif

		return;
	} /* if (ac == 4) */

	/* It is a new user */

	if (ac < 9)
		return;

	/*
	 * Check if someone is nick colliding a service nick
	 */
	if ((serviceptr = GetService(av[1])))
	{
		/*
		 * Make sure it is a valid nick collide before
		 * we send a kill. If this is the initial
		 * hub burst, the hub would not have processed
		 * our NICK statement yet, so let TS clean everything
		 * up
		 */
		if (IsNickCollide(serviceptr, av))
		{
			toserv(":%s KILL %s :%s\n",
				Me.name,
				av[1],
				"Attempt to Nick Collide Services");

			DeleteClient(FindClient(av[1]));

			InitServs(serviceptr);

			if (Me.sptr)
				Me.sptr->numservkills++;
			Network->TotalServKills++;

		#ifdef STATSERVICES
			if (Network->TotalServKills > Network->ServKillsT)
				Network->ServKillsT = Network->TotalServKills;
		#endif
		}

		/*
		 * Even if it is not a valid nick collide, do not
		 * add the user to our user list, since TS will
		 * have the user killed
		 */

		return;
	}

	/*
	 * Send new client connections to opers with a +C usermode
	 * in the format:
	 * nick!user@host (server)
	 */
	if (SafeConnect)
		SendUmode(OPERUMODE_CAPC,
			"*** Client connection: %s!%s@%s [%s]",
			av[1],
			av[5],
			av[6],
			av[7]);

	/* Add user to list */
	lptr = AddClient(av);

	if (!lptr)
	{
		/*
		 * Normally lptr will point to the newly created structure
		 * for av[1]. If it is NULL, one possibility is that
		 * HashAddClient() killed av[1] due to AutoKillClones, so
		 * go no furthur.
		 */
		return;
	}

#ifdef ALLOW_JUPES
	CheckJuped(av[1]);
#endif

#ifdef NICKSERVICES
	/*
	 * If the nick is registered, and was IDENTIFY'd at the time
	 * the userfile was written, the flags will still show they
	 * were identified, even if hybserv crashed etc, so when a
	 * new nick is introduced, make sure their flags don't contain
	 * NS_IDENTIFIED since theres no way they could have identify'd
	 * yet
	 */
	if ((nptr = FindNick(av[1])))
	{
		nptr->flags &= ~(NS_IDENTIFIED | NS_RELEASE);

	#ifdef RECORD_SPLIT_TS
		/*
		 * If nptr->split_ts is not 0, then the user QUIT while
		 * being identified, so if their current TS matches their
		 * split_ts, then they must have been on a valid netsplit,
		 * and we can allow them to continue being identified
		 */
		if (nptr->split_ts)
			if (nptr->split_ts == atoi(av[3]))
			{
				nptr->flags |= NS_IDENTIFIED;

				/*
				 * Only reset nptr->split_ts if a match was made, because
				 * if services sees someone QUIT during a split, and another
				 * client comes on services' side with the same nick, when
				 * the nick collide occurs, the old nick should still be
				 * able to be recognized with the same TS
				 */
				nptr->split_ts = nptr->whensplit = 0;
			}
	#endif /* RECORD_SPLIT_TS */
	}

	CheckNick(av[1]);

#endif /* NICKSERVICES */

#ifdef RECORD_SPLIT_TS
	/*
	 * Check if they were registered with OperServ during a split,
	 * if so, re-register them
	 */
	if ((uptr = GetUser(0, av[1], av[5], av[6])))
	{
		if (uptr->split_ts)
			if (uptr->split_ts == atoi(av[3]))
			{
				lptr->flags |= L_OSREGISTERED;
				uptr->split_ts = uptr->whensplit = 0;
			}
	}
#endif /* RECORD_SPLIT_TS */

#ifdef HAVE_PTHREADS
	/*
	 * Make sure we are safely connected to the hub so hopefully
	 * we won't be flooded with clients that we're supposed to
	 * do proxy checks on
	 */
	if (ProxyCheck && SafeConnect)
	{
		pthread_t clientid;
		pthread_attr_t attr;

		/*
		 * Peform an insecure proxy check on them
		 */
		pthread_attr_init(&attr);
		pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
		pthread_create(&clientid, &attr, (void *) &Proxy, (void *) lptr);
	}
#endif /* HAVE_PTHREADS */
} /* s_nick() */

/*
s_privmsg()
	Process a PRIVMSG statement
	av[0] = who
	av[1] = "PRIVMSG"
	av[2] = destination
	av[3] = actual message
*/

void
s_privmsg(int ac, char **av)

{
	char	*who, /* nick who gave privmsg */
				*command, /* contains the /msg that 'who' gave */
				*tmp;
	Luser *lptr,
				/*
				 * if it was sent to a service nick as opposed to a
				 * juped nick etc
				 */
	      *serviceptr;

	if (ac < 4)
		return;

	if (av[0][0] == ':')
		who = av[0] + 1;
	else
		who = av[0];

	if (!(lptr = FindClient(who)))
		return;

	if (RestrictedAccess)
	{
		/*
		 * Check if the user has an "o" flag - if not, don't allow
		 * the command
		 */
		if (!IsOper(GetUser(0, lptr->nick, lptr->username, lptr->hostname)))
			return;
	}

#ifndef EXTREMEDEBUG
	if (ignorelist)
	{
		char chkstr[MAXLINE];

		if (lptr)
		{
			sprintf(chkstr, "*!%s@%s",
				lptr->username,
				lptr->hostname);
			if (OnIgnoreList(chkstr))
				return;
		}
	}
#endif /* !EXTREMEDEBUG */

	if ((tmp = strchr(av[2], '@')))
	{
		if (!strcasecmp(tmp + 1, Me.name))
			*tmp = '\0';
	}

	serviceptr = GetService(av[2]);

	if (FloodProtection && serviceptr)
	{
		if (!IsValidAdmin(lptr))
		{
			lptr->messages++;
			if (!lptr->msgs_ts[0])
				lptr->msgs_ts[0] = time(NULL);
			else
			{
				lptr->msgs_ts[1] = time(NULL);
				if (lptr->messages == FloodCount)
				{
					lptr->messages = 0;
					if ((lptr->msgs_ts[1] - lptr->msgs_ts[0]) <= FloodTime)
					{
						char *mask = HostToMask(lptr->username, lptr->hostname);

						/*
						 * they sent too many messages in too little time
						 */
						lptr->flood_trigger++;
						if (lptr->flood_trigger >= IgnoreOffenses)
						{
							AddIgnore(mask, 0);
							notice(n_OperServ, lptr->nick,
								"Maximum flood offenses reached, you are on permanent ignore");
							putlog(LOG2,
								"PRIVMSG flood from %s!%s@%s (permanent ignore)",
								lptr->nick,
								lptr->username,
								lptr->hostname);

							SendUmode(OPERUMODE_Y,
								"*** Detected flood from %s!%s@%s (permanent ignore)",
								lptr->nick,
								lptr->username,
								lptr->hostname);
						}
						else
						{
							AddIgnore(mask, IgnoreTime);
							notice(n_OperServ, lptr->nick,
								"Received %d+ messages in <= %d seconds, you are being ignored for %s",
								FloodCount,
								FloodTime,
								timeago(IgnoreTime, 3));
							putlog(LOG2,
								"PRIVMSG flood from %s!%s@%s (%s ignore)",
								lptr->nick,
								lptr->username,
								lptr->hostname,
								timeago(IgnoreTime, 2));

							SendUmode(OPERUMODE_Y,
								"*** Detected flood from %s!%s@%s (%s ignore)",
								lptr->nick,
								lptr->username,
								lptr->hostname,
								timeago(IgnoreTime, 2));
						}

						MyFree(mask);
						return;
					}
					else
					{
						lptr->messages = 1;
						lptr->msgs_ts[0] = time(NULL);
					}
				}
				else
				{
					if ((lptr->msgs_ts[1] - lptr->msgs_ts[0]) > FloodTime)
					{
						/*
						 * FloodTime seconds has passed since their last message,
						 * reset everything
						 */
						lptr->messages = 1;
						lptr->msgs_ts[0] = time(NULL);
					}
				}
			}
		} /* if (!IsValidAdmin(lptr)) */
	} /* if (FloodProtection && GetService(av[2])) */

	/* remove the leading ":" from the command string */
	command = av[3] + 1;

	if (!serviceptr && (*command == '\001'))
	{
		Channel *chptr;

		chptr = FindChannel(av[2]);
		if (IsChannelMember(chptr, Me.osptr))
		{
			/* its a ctcp request to a channel that n_OperServ is on */
			onctcp(who, n_OperServ, command);
		}

	#if defined(NICKSERVICES) && defined(CHANNELSERVICES)
		if (IsChannelMember(chptr, Me.csptr))
		{
			/* its a ctcp request to a channel that n_ChanServ is on */
			onctcp(who, n_ChanServ, command);
		}
	#endif /* defined(NICKSERVICES) && defined(CHANNELSERVICES) */

		return;
	}

	if (serviceptr && (command[0] == '\001'))
	{
		/* process ctcp request */
		onctcp(who, serviceptr->nick, command);
		return;
	}

	if (serviceptr == Me.osptr)
		os_process(who, command, NODCC);

#ifdef NICKSERVICES

	else if (serviceptr == Me.nsptr)
		ns_process(who, command);

	#ifdef CHANNELSERVICES
		else if (serviceptr == Me.csptr)
			cs_process(who, command);
	#endif

	#ifdef MEMOSERVICES
		else if (serviceptr == Me.msptr)
			ms_process(who, command);
	#endif

#endif /* NICKSERVICES */

#ifdef STATSERVICES

	else if (serviceptr == Me.ssptr)
		ss_process(who, command);

#endif

#ifdef HELPSERVICES

	else if (serviceptr == Me.hsptr)
		hs_process(who, command);

#endif
} /* s_privmsg() */

/*
s_squit()
	Process an SQUIT statement
	av[0] = "SQUIT"
	av[1] = server name
	av[2] = reason
*/

void
s_squit(int ac, char **av)

{
	Server *sptr;

	if (ac == 3)
		sptr = FindServer(av[1]);
	else if (ac == 4)
		sptr = FindServer(av[2]);
	else
		sptr = NULL;

	DeleteServer(sptr); /* delete server from list */
} /* s_squit() */

/*
s_quit()
	Process a QUIT statement
	av[0] = who
	av[1] = "QUIT"
	av[2] = message
*/

void
s_quit(int ac, char **av)

{
	Luser *lptr;
#ifdef NICKSERVICES
	NickInfo	*nptr;
#endif

	if (ac < 3)
		return;

	if (av[0][0] == ':')
		lptr = FindClient(av[0] + 1);
	else
		lptr = FindClient(av[0]);

	if (!lptr)
	{
		putlog(LOG1,
			"DEBUG: s_quit(): Unable to locate client [%s]",
			av[0] + 1);
		return;
	}

#ifdef NICKSERVICES
	if (LastSeenInfo && (nptr = FindNick(lptr->nick)))
	{
		if (nptr->flags & NS_IDENTIFIED)
		{
			if (nptr->lastqmsg)
				MyFree(nptr->lastqmsg);
			nptr->lastqmsg = MyStrdup(av[2] + 1);
		}
	}
#endif /* NICKSERVICES */

	DeleteClient(lptr); /* delete user */
} /* s_quit() */

/*
s_kill()
	Process a KILL statement
	av[0] = who
	av[1] = "KILL"
	av[2] = victim
	av[3] = message
*/

void
s_kill(int ac, char **av)

{
	Luser *kuser,
	      *lptr,
	      *serviceptr;
	Server *sptr;
	char *who;

	if (ac < 4)
		return;

	kuser = FindClient(av[2]);
	if (!kuser)
		return;

	DeleteClient(kuser);

	if (av[0][0] == ':')
		who = av[0] + 1;
	else
		who = av[0];

	lptr = FindClient(who);

	if (lptr)
	{
		SendUmode(OPERUMODE_CAPO,
			"*** Operator Kill: [%s] by %s!%s@%s (%s)",
			av[2],
			lptr->nick,
			lptr->username,
			lptr->hostname,
			av[3] + 1);
	}
	else
	{
		SendUmode(OPERUMODE_CAPS,
			"*** Server Kill: [%s] by %s (%s)",
			av[2],
			who,
			av[3] + 1);
	}

	if (match("*.*", who))
	{
		sptr = FindServer(who);
		if (sptr != (Server *) NULL)
			sptr->numservkills++;
		Network->TotalServKills++;
	#ifdef STATSERVICES
		if (Network->TotalServKills > Network->ServKillsT)
			Network->ServKillsT = Network->TotalServKills;
	#endif
	}
	else
	{
		if (lptr)
		{
			if (lptr->server)
				lptr->server->numoperkills++;

		#ifdef STATSERVICES
			lptr->numkills++;
		#endif
		}

		Network->TotalOperKills++;
	#ifdef STATSERVICES
		if (Network->TotalOperKills > Network->OperKillsT)
			Network->OperKillsT = Network->TotalOperKills;
	#endif
	}

	if ((serviceptr = FindService(kuser)))
	{
		/* a service nick was killed - reintroduce them to the network */

		/*
		 * NOTE: serviceptr should ONLY be used for it's pointer value
		 * now, since the Luser structure it points to has been
		 * deleted by DeleteClient()
		 */

		if (lptr)
		{
			putlog(LOG1, "%s was killed by %s!%s@%s, re-initializing",
				av[2],
				who,
				lptr->username,
				lptr->hostname);
		}
		else
		{
			putlog(LOG1, "%s was killed by %s (nick collide), re-initializing",
				av[2],
				who);

			toserv(":%s KILL %s :%s\n",
				Me.name,
				av[2],
				"Attempt to Nick Collide Services");

			if (Me.sptr)
				Me.sptr->numservkills++;
			Network->TotalServKills++;
		#ifdef STATSERVICES
			if (Network->TotalServKills > Network->ServKillsT)
				Network->ServKillsT = Network->TotalServKills;
		#endif
		}

		/* re-introduce the killed *Serv to the network */
		InitServs(serviceptr);

		/*
		 * We can't do: if (serviceptr == Me.osptr) here, because
		 * if OperServ was killed, InitServs() would have given
		 * Me.osptr a new value, where serviceptr would be
		 * the old value
		 */
		if (!strcasecmp(av[2], n_OperServ))
		{
			Chanlist *tempchan;
			Channel *chptr;

			/*
			 * have OperServ rejoin all its chans, if other
			 * users are in them
			 */
			for (tempchan = mychanlist; tempchan; tempchan = tempchan->next)
				if ((chptr = FindChannel(tempchan->name)))
					os_join(chptr);
		}

	#if defined(NICKSERVICES) && defined(CHANNELSERVICES)

		/*
		 * If ChanServ was killed, rejoin all it's channels
		 */
		else if (!strcasecmp(av[2], n_ChanServ))
			cs_RejoinChannels();

	#endif

	}
} /* s_kill() */

/*
s_kick()
	Process a KICK statement
	av[0] = who
	av[1] = "KICK"
	av[2] = channel
	av[3] = nickname
	av[4] = reason
*/

void
s_kick(int ac, char **av)

{
	char	*who;
	Luser	*lptr,
	      *kptr;
	Channel *chptr;

	if (ac < 5)
		return;

	if (!(kptr = FindClient(av[3])))
		return;

	if (!(chptr = FindChannel(av[2])))
		return;

	if (av[0][0] == ':')
		who = av[0] + 1;
	else
		who = av[0];

	lptr = FindClient(who);

	RemoveFromChannel(chptr, kptr);

	if (FindService(kptr))
		FloodCheck(chptr, lptr, kptr, 1);

#ifdef STATSERVICES
	if (lptr)
		lptr->numkicks++;
#endif

	if (lptr)
	{
		putlog(LOG3, "%s: %s!%s@%s kicked %s [%s]",
			av[2],
			lptr->nick,
			lptr->username,
			lptr->hostname,
			av[3],
			av[4] + 1);

		SendUmode(OPERUMODE_K,
			"*** %s: Kick [%s] by %s!%s@%s (%s)",
			av[2],
			av[3],
			lptr->nick,
			lptr->username,
			lptr->hostname,
			av[4] + 1);
	}
	else
	{
		putlog(LOG3, "%s: %s kicked %s [%s]",
			av[2],
			who,
			av[3],
			av[4] + 1);

		SendUmode(OPERUMODE_K,
			"*** %s: Kick [%s] by %s (%s)",
			av[2],
			av[3],
			who,
			av[4] + 1);
	}
} /* s_kick() */

/*
s_mode()
	Process MODE statement
	av[0] = who
	av[1] = "MODE"
	av[2] = channel
	av[3] = modes
	av[4]- = optional args
*/

void
s_mode(int ac, char **av)

{
	char *who,
	     *modes;
	Luser	*lptr;
	Channel *chptr;

	if (ac < 4)
		return;

	if (av[0][0] == ':')
		who = av[0] + 1;
	else
		who = av[0];
	lptr = FindClient(who);

	modes = GetString(ac - 3, av + 3);

	if (!IsChanPrefix(*av[2]) && lptr)
	{
		/* it was a umode change, not channel mode */
		UpdateUserModes(lptr, (*modes == ':') ? modes + 1 : modes);
		MyFree(modes);
		return;
	}

	if (!(chptr = FindChannel(av[2])))
	{
		MyFree(modes);
		return;
	}

	if (lptr)
	{
		SendUmode(OPERUMODE_M,
			"*** %s: Mode [%s] by %s!%s@%s",
			chptr->name,
			modes,
			lptr->nick,
			lptr->username,
			lptr->hostname);
	}
	else
	{
		SendUmode(OPERUMODE_M,
			"*** %s: Mode [%s] by %s",
			chptr->name,
			modes,
			who);
	}

	/* update modes field in chan structure */
	UpdateChanModes(who, chptr, modes);

	if (lptr)
	{
		putlog(LOG3, "%s: mode change \"%s\" by %s!%s@%s",
			chptr->name,
			modes,
			lptr->nick,
			lptr->username,
			lptr->hostname);
	}
	else
	{
		putlog(LOG3, "%s: mode change \"%s\" by %s",
			chptr->name,
			modes,
			who);
	}

	MyFree(modes);
} /* s_mode() */

/*
s_sjoin()
	Process an SJOIN statement
	av[0] = server name
	av[1] = "SJOIN"
	av[2] = channel TS
	av[3] = channel
	av[4] = channel modes
	av[5] = nicks
*/

void
s_sjoin(int ac, char **av)

{
	char sendstr[MAXLINE],
       /* new modes if the SJOIN TS is lower than the
        * current TS
        */
	     modes[MAXLINE];
	char **line,
	     **nicks; /* array of SJOINing nicknames */
	time_t CurrTime; /* current TS if its a new channel */
	Channel *cptr, *oldptr;
	int ncnt, /* number of SJOINing nicks */
	    mcnt;

#if defined(NICKSERVICES) && defined(CHANNELSERVICES)
	ChanInfo *ci;
	char *currnick;
	int ii;
#endif

	if (!strcasecmp(av[1], "JOIN"))
	{
		char *chan;

		if (ac < 3)
			return;

		chan = (*av[2] == ':') ? av[2] + 1 : av[2];

		CurrTime = time(NULL);

		/* kludge for older ircds that don't use SJOIN */
		sprintf(sendstr, ":%s SJOIN %ld %s + :%s",
			currenthub->realname,
			(long) CurrTime,
			chan,
			av[0]);

		SplitBuf(sendstr, &line);
		SplitBuf(av[0], &nicks);

		cptr = AddChannel(line, 1, nicks);

		/*
		 * If the channel has a C: line, have OperServ join it
		 */
		if (cptr)
			if (IsChannel(cptr->name) && !IsChannelMember(cptr, Me.osptr))
				os_join(cptr);

	#if defined(NICKSERVICES) && defined(CHANNELSERVICES)

	#ifdef HYBRID_ONLY

		/*
		 * Since HYBRID_ONLY is defined, ChanServ won't join
		 * the channel from this call, but will call cs_CheckChan()
		 * on it
		 */
		if (cptr && (cptr->numusers <= 1))
			cs_join(FindChan(chan));

	#else

		if (cptr)
			if ((ci = FindChan(chan)) && !IsChannelMember(cptr, Me.csptr))
				if (!(ci->flags & CS_FORGET))
					cs_join(FindChan(chan));

	#endif /* HYBRID_ONLY */

	#endif

		MyFree(line);
		MyFree(nicks);

		return;
	}

	/*
	 * It is an SJOIN
	 */

	if (ac < 6)
		return;

	mcnt = 5; /* default position for channel nicks, if no limit/key */
	strcpy(modes, av[4]);

	/*
	 * names list *should* start with a :, if it doesn't, there's a
	 * limit and/or key 
	 */
	if (av[5][0] != ':')
	{
		strcat(modes, " ");
		strcat(modes, av[5]);
		mcnt++;
		if (av[6][0] != ':')
		{
			strcat(modes, " ");
			strcat(modes, av[6]);
			mcnt++;
		}
	}

	ncnt = SplitBuf(av[mcnt] + 1, &nicks);

	/*
	 * when a user joins the channel "0", they get parted from all their
	 * channels
	 */
	if (*av[3] == '0')
	{
		Luser *lptr;
		UserChan *tempc;
 
 		if (!(lptr = FindClient(nicks[0])))
 			return;

 		while (lptr->firstchan)
 		{
 			tempc = lptr->firstchan->next;
 			RemoveFromChannel(lptr->firstchan->chptr, lptr);
 			lptr->firstchan = tempc;
 		}

		MyFree(nicks);

		return;
	} /* if (*av[3] == '0') */

	oldptr = FindChannel(av[3]);

	if (SafeConnect)
	{
		if (!oldptr)
		{
			/*
			 * It is a new channel - notify all +j users
			 */
			SendUmode(OPERUMODE_J,
				"*** New channel: %s (created by %s)",
				av[3],
				GetNick(nicks[0]));
		} /* if (!oldptr) */
		else
			SendUmode(OPERUMODE_J,
				"*** Channel join: %s (%s)",
				GetNick(nicks[0]),
				oldptr->name);
	}

	cptr = AddChannel(av, ncnt, nicks);

	if (!cptr)
		return; /* should never happen */

	/*
	 * If oldptr is valid, it was an existing channel,
	 * so check TSora stuff to keep synched.
	 * Otherwise, it's a new channel, so check if
	 * OperServ should be monitoring it.
	 */

	if (oldptr)
	{
		if (atoi(av[2]) < cptr->since)
		{
			ChanUser *tempuser;
			UserChan *tempchan;
			BanInfo  *nextban;

			/*
			 * if the TS given in the SJOIN is less than the recorded 
			 * channelTS, everyone in the channel except the SJOIN'ing 
			 * nick(s) is deoped/devoiced, all previous modes are replaced 
			 * with the new modes, and all previous bans are unset
			 */

			cptr->since = atoi(av[2]); /* keep the older TS */
			/* keep older modes */
			cptr->modes = 0;
			UpdateChanModes((char *) NULL, cptr, modes);

			/* clear all bans */
			while (cptr->firstban)
			{
				nextban = cptr->firstban->next;
				if (cptr->firstban->who)
					MyFree(cptr->firstban->who);
				MyFree(cptr->firstban->mask);
				MyFree(cptr->firstban);
				cptr->firstban = nextban;
			}

			for (tempuser = cptr->firstuser; tempuser; tempuser = tempuser->next)
			{
				/* don't deop the nicks who just SJOIN'd */
				if (IsInNickArray(ncnt, nicks, tempuser->lptr->nick))
					continue;

				if (tempuser->flags & CH_OPPED)
				{
					tempuser->flags &= ~CH_OPPED;
					tempchan = FindChannelByUser(tempuser->lptr, cptr);
					if (tempchan)
						tempchan->flags &= ~CH_OPPED;
				}
				if (tempuser->flags & CH_VOICED)
				{
					tempuser->flags &= ~CH_VOICED;
					tempchan = FindChannelByUser(tempuser->lptr, cptr);
					if (tempchan)
						tempchan->flags &= ~CH_VOICED;
				}
			} /* for (tempuser .. ) */

			if (IsChannelMember(cptr, Me.osptr) &&
					!IsChannelOp(cptr, Me.osptr))
			{
				/* n_OperServ was deoped in the sjoin, must reop it */
				#ifdef SAVE_TS
					os_part(cptr);
					os_join(cptr);
				#else
					toserv(":%s MODE %s +o %s\n",
						Me.name,
						cptr->name,
						n_OperServ);
				#endif
					putlog(LOG2, "%s: %s attempted to deop %s", cptr->name, av[0] + 1, n_OperServ);
			}

		#if defined(NICKSERVICES) && defined(CHANNELSERVICES)
			if (IsChannelMember(cptr, Me.csptr) &&
					!IsChannelOp(cptr, Me.csptr))
			{
				/* n_ChanServ was deoped in the sjoin, must reop it */
				#ifdef SAVE_TS
					cs_part(cptr);
					cs_join(FindChan(cptr->name));
				#else
					toserv(":%s MODE %s +o %s\n",
						Me.name,
						cptr->name,
						n_ChanServ);
				#endif

					putlog(LOG2, "%s: %s attempted to deop %s",
						cptr->name,
						av[0] + 1,
						n_ChanServ);
			}
		#endif

		} /* if (atoi(av[2]) < cptr->since) */
	} /* if (oldptr) */
	else
	{
		/*
		 * It is a new channel - check if OperServ should be
		 * monitoring it
		 */

		if (IsChannel(cptr->name) && !IsChannelMember(cptr, Me.osptr))
			os_join(cptr);

	} /* !oldptr */

#if defined(NICKSERVICES) && defined(CHANNELSERVICES)

	ci = FindChan(cptr->name);

	if (ci)
	{
		/*
		 * Call cs_CheckSjoin() to check if any of the SJOINing
		 * nicks should be deopped. If oldptr is NULL,
		 * the channel did not exist before, so pass an arguement
		 * of 1 to cs_CheckSjoin()'s "newchan" parameter indicating
		 * the channel is brand new.
		 */
		cs_CheckSjoin(cptr, ci, ncnt, nicks, oldptr ? 0 : 1);

		for (ii = 0; ii < ncnt; ii++)
		{
			currnick = GetNick(nicks[ii]);
			if (!currnick)
				continue;

			/*
			 * Check if nick is on the AKICK list for chname
			 */
			cs_CheckJoin(cptr, ci, currnick);

			/*
			 * Check if nick has AutoOp access on channel, if so, op them
			 */
			cs_CheckOp(cptr, ci, currnick);
		}
	} /* if (ci) */
#endif /* defined(NICKSERVICES) && defined(CHANNELSERVICES) */

	MyFree(nicks);
} /* s_sjoin() */

/*
s_part()
	Process a PART statement
	av[0] = who
	av[1] = "PART"
	av[2] = channel
*/

void
s_part(int ac, char **av)

{
	char *who;

	if (ac < 3)
		return;

	if (av[0][0] == ':')
		who = av[0] + 1;
	else
		who = av[0];

	RemoveFromChannel(FindChannel(av[2]), FindClient(who));
} /* s_part() */

/*
s_error()
	Log the error that the hub sends
*/

void
s_error(int ac, char **av)

{
	char *serror;

	if (ac >= 3)
		serror = av[2] + 1;
	else if (ac >= 2)
		serror = av[1] + 1;
	else
		serror = "";

	putlog(LOG1, "Server Error: %s", serror);
} /* s_error() */

/*
s_whois()
	Reply to a remote whois query

	av[0] = who
	av[1] = "WHOIS"
	av[2] = nickname
	av[3] = nickname
*/

void
s_whois(int ac, char **av)

{
	char *who, *target, *isoper;
	Luser *serviceptr;

	if (ac < 4)
		return;

	if (av[0][0] == ':')
		who = av[0] + 1;
	else
		who = av[0];

	if (av[3][0] == ':')
		target = av[3] + 1;
	else
		target = av[3];

	SendUmode(OPERUMODE_Y,
		"*** Remote whois [%s] requested by %s",
		target,
		who);

	if (!(serviceptr = GetService(target)))
		return;

	isoper = strchr(ServiceUmodes, 'o');

	if (serviceptr == Me.osptr)
	{
		toserv(":%s 311 %s %s %s %s * :%s\n",
			Me.name,
			who,
			n_OperServ,
			id_OperServ,
			Me.name,
			desc_OperServ);
		toserv(":%s 312 %s %s %s :%s\n",
			Me.name,
			who,
			n_OperServ,
			Me.name,
			Me.info);
		if (isoper)
		{
			toserv(":%s 313 %s %s :is madly feared (is an IRC Operator)\n",
				Me.name,
				who,
				n_OperServ);
		}
		toserv(":%s 317 %s %s 0 %ld :seconds idle, signon time\n",
			Me.name,
			who,
			n_OperServ,
			TimeStarted);
		toserv(":%s 318 %s %s :End of /WHOIS list.\n",
			Me.name,
			who,
			n_OperServ);
	}

#ifdef NICKSERVICES

	if (serviceptr == Me.nsptr)
	{
		toserv(":%s 311 %s %s %s %s * :%s\n",
			Me.name,
			who,
			n_NickServ,
			id_NickServ,
			Me.name,
			desc_NickServ);
		toserv(":%s 312 %s %s %s :%s\n",
			Me.name,
			who,
			n_NickServ,
			Me.name,
			Me.info);
		if (isoper)
		{
			toserv(":%s 313 %s %s :is madly feared (is an IRC Operator)\n",
				Me.name,
				who,
				n_NickServ);
		}
		toserv(":%s 317 %s %s 0 %ld :seconds idle, signon time\n",
			Me.name,
			who,
			n_NickServ,
			TimeStarted);
		toserv(":%s 318 %s %s :End of /WHOIS list.\n",
			Me.name,
			who,
			n_NickServ);
	}

	#ifdef CHANNELSERVICES

		if (serviceptr == Me.csptr)
		{
			toserv(":%s 311 %s %s %s %s * :%s\n",
				Me.name,
				who,
				n_ChanServ,
				id_ChanServ,
				Me.name,
				desc_ChanServ);
			toserv(":%s 312 %s %s %s :%s\n",
				Me.name,
				who,
				n_ChanServ,
				Me.name,
				Me.info);
			if (isoper)
			{
				toserv(":%s 313 %s %s :is madly feared (is an IRC Operator)\n",
					Me.name,
					who,
					n_ChanServ);
			}
			toserv(":%s 317 %s %s 0 %ld :seconds idle, signon time\n",
				Me.name,
				who,
				n_ChanServ,
				TimeStarted);
			toserv(":%s 318 %s %s :End of /WHOIS list.\n",
				Me.name,
				who,
				n_ChanServ);
		}

	#endif /* CHANNELSERVICES */

	#ifdef MEMOSERVICES

		if (serviceptr == Me.msptr)
		{
			toserv(":%s 311 %s %s %s %s * :%s\n",
				Me.name,
				who,
				n_MemoServ,
				id_MemoServ,
				Me.name,
				desc_MemoServ);
			toserv(":%s 312 %s %s %s :%s\n",
				Me.name,
				who,
				n_MemoServ,
				Me.name,
				Me.info);
			if (isoper)
			{
				toserv(":%s 313 %s %s :is madly feared (is an IRC Operator)\n",
					Me.name,
					who,
					n_MemoServ);
			}
			toserv(":%s 317 %s %s 0 %ld :seconds idle, signon time\n",
				Me.name,
				who,
				n_MemoServ,
				TimeStarted);
			toserv(":%s 318 %s %s :End of /WHOIS list.\n",
				Me.name,
				who,
				n_MemoServ);
		}

	#endif /* MEMOSERVICES */

#endif /* NICKSERVICES */

#ifdef STATSERVICES

	if (serviceptr == Me.ssptr)
	{
		toserv(":%s 311 %s %s %s %s * :%s\n",
			Me.name,
			who,
			n_StatServ,
			id_StatServ,
			Me.name,
			desc_StatServ);
		toserv(":%s 312 %s %s %s :%s\n",
			Me.name,
			who,
			n_StatServ,
			Me.name,
			Me.info);
		if (isoper)
		{
			toserv(":%s 313 %s %s :is madly feared (is an IRC Operator)\n",
				Me.name,
				who,
				n_StatServ);
		}
		toserv(":%s 317 %s %s 0 %ld :seconds idle, signon time\n",
			Me.name,
			who,
			n_StatServ,
			TimeStarted);
		toserv(":%s 318 %s %s :End of /WHOIS list.\n",
			Me.name,
			who,
			n_StatServ);
	}

#endif /* STATSERVICES */

#ifdef HELPSERVICES

	if (serviceptr == Me.hsptr)
	{
		toserv(":%s 311 %s %s %s %s * :%s\n",
			Me.name,
			who,
			n_HelpServ,
			id_HelpServ,
			Me.name,
			desc_HelpServ);
		toserv(":%s 312 %s %s %s :%s\n",
			Me.name,
			who,
			n_HelpServ,
			Me.name,
			Me.info);
		if (isoper)
		{
			toserv(":%s 313 %s %s :is madly feared (is an IRC Operator)\n",
				Me.name,
				who,
				n_HelpServ);
		}
		toserv(":%s 317 %s %s 0 %ld :seconds idle, signon time\n",
			Me.name,
			who,
			n_HelpServ,
			TimeStarted);
		toserv(":%s 318 %s %s :End of /WHOIS list.\n",
			Me.name,
			who,
			n_HelpServ);
	}

#endif /* HELPSERVICES */
} /* s_whois() */

/*
s_trace()
	Reply to a remote trace query

	av[0] = who
	av[1] = "TRACE"
	av[2] = server name
*/

void
s_trace(int ac, char **av)

{
	char	*who, *isoper;

	if (ac < 3)
		return;

	if (strcasecmp(av[2] + 1, Me.name) != 0)
		return;

	isoper = strchr(ServiceUmodes, 'o');
	if (av[0][0] == ':')
		who = av[0] + 1;
	else
		who = av[0];

	SendUmode(OPERUMODE_Y,
		"*** Remote trace query requested by %s",
		who);

	toserv(":%s %d %s %s %d :%s[%s@%s]\n",
		Me.name,
		isoper ? 204 : 205,
		who,
		isoper ? "Oper" : "User",
		isoper ? 5 : 1,
		n_OperServ,
		id_OperServ,
		Me.name);

#ifdef NICKSERVICES

	toserv(":%s %d %s %s %d :%s[%s@%s]\n",
		Me.name,
		isoper ? 204 : 205,
		who,
		isoper ? "Oper" : "User",
		isoper ? 5 : 1,
		n_NickServ,
		id_NickServ,
		Me.name);

	#ifdef CHANNELSERVICES

		toserv(":%s %d %s %s %d :%s[%s@%s]\n",
			Me.name,
			isoper ? 204 : 205,
			who,
			isoper ? "Oper" : "User",
			isoper ? 5 : 1,
			n_ChanServ,
			id_ChanServ,
			Me.name);

	#endif /* CHANNELSERVICES */

	#ifdef MEMOSERVICES

		toserv(":%s %d %s %s %d :%s[%s@%s]\n",
			Me.name,
			isoper ? 204 : 205,
			who,
			isoper ? "Oper" : "User",
			isoper ? 5 : 1,
			n_MemoServ,
			id_MemoServ,
			Me.name);

	#endif /* MEMOSERVICES */

#endif /* NICKSERVICES */

#ifdef STATSERVICES

	toserv(":%s %d %s %s %d :%s[%s@%s]\n",
		Me.name,
		isoper ? 204 : 205,
		who,
		isoper ? "Oper" : "User",
		isoper ? 5 : 1,
		n_StatServ,
		id_StatServ,
		Me.name);

#endif /* STATSERVICES */

#ifdef HELPSERVICES

	toserv(":%s %d %s %s %d :%s[%s@%s]\n",
		Me.name,
		isoper ? 204 : 205,
		who,
		isoper ? "Oper" : "User",
		isoper ? 5 : 1,
		n_HelpServ,
		id_HelpServ,
		Me.name);

#endif /* HELPSERVICES */

	toserv(":%s 206 %s Serv 10 %1.0fS %1.0fC %s :AutoConn.!*@%s\n",
		Me.name,
		who,
		Network->TotalServers,
		Network->TotalUsers,
		currenthub->realname ? currenthub->realname : currenthub->hostname,
		Me.name);

	toserv(":%s 262 %s %s :End of TRACE\n",
		Me.name,
		who,
		Me.name);
} /* s_trace() */

/*
s_version()
	Reply to a remote version query

	av[0] = who
	av[1] = "VERSION"
	av[2] = server name
*/

void
s_version(int ac, char **av)

{
	char	*who;

	if (ac < 3)
		return;

	if (strcasecmp(Me.name, (av[2][0] == ':') ? av[2] + 1 : av[2]) != 0)
		return;

	if (av[0][0] == ':')
		who = av[0] + 1;
	else
		who = av[0];

	SendUmode(OPERUMODE_Y,
		"*** Remote version query requested by %s",
		who);

	toserv(":%s 351 %s HybServ-%s. %s :TS3\n",
		Me.name,
		who,
		hVersion,
		Me.name);
} /* s_version() */

/*
s_motd()
	Send global motd to av[0]

	av[0] = who
	av[1] = "MOTD"
	av[2] = server name
*/

void
s_motd(int ac, char **av)

{
	char	*who, *final;
	char	line[MAXLINE];
	FILE	*fp;

	if (ac < 3)
		return;

	if (strcasecmp(Me.name, (av[2][0] == ':') ? av[2] + 1 : av[2]) != 0)
		return;

	if (av[0][0] == ':')
		who = av[0] + 1;
	else
		who = av[0];

	SendUmode(OPERUMODE_Y,
		"*** Remote motd query requested by %s",
		who);

	if (!(fp = fopen(MotdFile, "r")))
	{
		toserv(":%s 422 %s :MOTD File is missing\n",
			Me.name,
			who);
		return;
	}

	toserv(":%s 375 %s :- %s Message of the Day -\n",
		Me.name,
		who,
		Me.name);

	while (fgets(line, MAXLINE - 1, fp))
	{
		if (IsEOL(*line))
		{
			toserv(":%s 372 %s :- \n",
				Me.name,
				who);
			continue;
		}

		final = Substitute((char *) NULL, line, NODCC);
		if (final && (final != (char *) -1))
		{
			toserv(":%s 372 %s :- %s\n",
				Me.name,
				who,
				final);
			MyFree(final);
		}
	}

	toserv(":%s 376 %s :End of /MOTD command\n",
		Me.name,
		who);

	fclose(fp);
} /* s_motd() */

#if defined(NICKSERVICES) && defined(CHANNELSERVICES)

/*
s_topic()
	Check if topic change conflicts with a locked topic on channel
av[2]

	av[0] = who
	av[1] = "TOPIC"
	av[2] = channel
	av[3]- = new topic
*/

void
s_topic(int ac, char **av)

{
	if (ac < 4)
		return;

	SendUmode(OPERUMODE_T,
		"*** %s: Topic by %s (%s)",
		av[2],
		av[0] + 1,
		av[3] + 1);

	cs_CheckTopic(av[0] + 1, av[2]);
} /* s_topic() */

#endif

#ifdef STATSERVICES

/*
s_pong()
 Process a server PONG

  av[0] = :servername
  av[1] = "PONG"
  av[2] = servername
  av[3] = Me.name
*/

void
s_pong(int ac, char **av)

{
	struct timeval pong_t;
	time_t currts;
	long deltasec, deltausec;
	Server *servptr;
	char *who;

	who = av[0];
	if (*who == ':')
		who++;

	if (!(servptr = FindServer(who)))
		return; /* server doesn't exist */

	/*
	 * If servptr->lastping_* is 0, we never pinged them, so
	 * they shouldn't be PONGing us
	 */
	if (!servptr->lastping_sec || !servptr->lastping_usec)
		return;

	GetTime(&pong_t);
	currts = time(NULL);

	/*
	 * Now set delta variables to the delta given by
	 * pong_t->tv_{u}sec - servptr->lastping_{u}sec.
	 * That will represent how much time has passed since we
	 * sent the ping requests.
	 */
	deltasec = pong_t.tv_sec - servptr->lastping_sec;
	deltausec = pong_t.tv_usec - servptr->lastping_usec;

	/*
	 * Finally, set the current ping variable to the float number:
	 * "deltasec.deltausec".
	 * Of course, we must divide the microseconds by one million
	 * since a microsecond is 10^-6 seconds.
	 */
	servptr->ping = (float) (deltasec + (deltausec / 1000000.0));

	/*
	 * Now determine if this current ping reply is the highest
	 * or lowest we've received from this server.
	 */
	if (servptr->ping > servptr->maxping)
	{
		servptr->maxping = servptr->ping;
		servptr->maxping_ts = currts;
	}

	if ((servptr->minping == 0.0) || (servptr->ping < servptr->minping))
	{
		servptr->minping = servptr->ping;
		servptr->minping_ts = currts;
	}

	/*
	 * Now reset the lastping fields so we can ping the server
	 * again in the future.
	 */
	servptr->lastping_sec = servptr->lastping_usec = 0;

	if (MaxPing)
	{
		/*
		 * Make sure servptr is our current hub. Our hub's uplink
		 * field will ALWAYS match services, even though we are
		 * not the actual hub of our hub.
		 */
		if ((servptr->uplink == Me.sptr) && (servptr->ping > (float) MaxPing))
		{
			/*
			 * servptr matches our current hub server, and it's ping
			 * reply is greater than MaxPing - attempt to switch
			 * hubs - make sure there's at least 1 other hub to
			 * switch to.
			 */
			if (HubCount > 1)
			{
				SendUmode(OPERUMODE_Y,
					"*** Ping response time from hub (%s) has exceeded [%s], rerouting",
					servptr->name,
					timeago(MaxPing, 3));

				toserv("SQUIT %s :ReRouting (current lag: %5.4f seconds)\n",
					servptr->name,
					servptr->ping);

				close(HubSock);
				ClearUsers();
				ClearChans();
				ClearServs();
				ClearHashes(0);

				HubSock = NOSOCKET;

				/*
				 * Now, try to cycle through server list for a new
				 * hub
				 */
				CycleServers();
			}
		}
	} /* if (MaxPing) */
} /* s_pong() */

#endif /* STATSERVICES */

/*
GetUser()
	Find O: line entry containing nickname.  If 'user' and 'host' are
given, try to find best match in case there are O: lines with the
same nickname; if hostname is given, but nickonly is 0, do not just
match nicknames - make sure the user@host matches as well
*/

Userlist *
GetUser(int nickonly, char *nickname, char *user, char *host)

{
	Userlist	*tempuser,
						/* we found a hostname match, but not a nick match */
						*hostmatch, 
						/* we found a nick match, but not a hostname match */
						*nickmatch;

	if (!nickname)
		return (Userlist *) NULL;
	
	hostmatch = nickmatch = (Userlist *) NULL;
	
	/*
	 * First, attempt to match both nickname and hostname, if possible.
	 * If we don't find a match, try to match just hostnames, because
	 * it is possible the user is simply using another nickname
	 */

	for (tempuser = userlist; tempuser; tempuser = tempuser->next)
	{
		if (!strcasecmp(tempuser->nick, nickname))
		{
			if (nickonly && (!user && !host))
			{
				/*
				 * We found a nickname match - nickonly is 1, and userhost
				 * is not supplied, return the pointer
				 */
				return(tempuser);
			}
			else
			{
				/*
				 * We have a nickname match, but nickonly is 0, so try
				 * to match userhost's as well
				 */
				if (match(tempuser->username, user) &&
						match(tempuser->hostname, host))
					return (tempuser);
			}

			/*
			 * Either nickonly is 0, and the userhosts don't match, or
			 * nickonly is 1, but a userhost was still given and does
			 * not match tempuser's userhost.  In either case, we have
			 * a nickname match, so store this user structure into
			 * 'nickmatch', so if we don't find a match later on,
			 * with a nickname and hostname, we will return 'nickmatch'
			 * but only if nickonly is 1, meaning we are allowed to
			 * match only nicknames
			 */
			if (!nickmatch)
				nickmatch = tempuser;
		}
		else
		{
			if (user && host)
				if (match(tempuser->username, user) &&
						match(tempuser->hostname, host))
				{
					/*
					 * The nicknames do not match, but the userhosts do.
					 * Record this user structure as a hostname match; in
					 * case we don't find an exact match with both
					 * nickname and hostname, this is still a pretty good
					 * match
					 */
					if (!hostmatch)
						hostmatch = tempuser;
				}
		}
	}

	if (hostmatch)
	{
		/*
		 * The given userhost matches an O: line, but the nickname
		 * does not - it is still a good match since the user
		 * may be using a different nickname than usual -
		 * return the hostname pointer
		 */
		return (hostmatch);
	}

	if (nickonly && nickmatch)
	{
		/*
		 * the given userhost doesn't have an O: line, but we did
		 * have a nickname match, and since nickonly is 1, we are
		 * allowed to disregard hostnames, so return the nickname
		 * match we found
		 */
		return (nickmatch);
	}

	if (OpersHaveAccess)
	{
		if (IsOperator(FindClient(nickname)))
			return (GenericOper);
	}

	return ((Userlist *) NULL);
} /* GetUser() */

/*
GetService
  Returns a pointer to luser structure containing n_*Nick
if 'name' matches a Service Bot
*/

Luser *
GetService(char *name)

{
	if (!name)
		return ((Luser *) NULL);

	if (!strcasecmp(name, n_OperServ))
		return (Me.osptr);

#ifdef NICKSERVICES

	if (!strcasecmp(name, n_NickServ))
		return (Me.nsptr);

	#ifdef CHANNELSERVICES
		if (!strcasecmp(name, n_ChanServ))
			return (Me.csptr);
	#endif

	#ifdef MEMOSERVICES
		if (!strcasecmp(name, n_MemoServ))
			return (Me.msptr);
	#endif

#endif /* NICKSERVICES */

#ifdef STATSERVICES

	if (!strcasecmp(name, n_StatServ))
		return (Me.ssptr);

#endif

#ifdef HELPSERVICES

	if (!strcasecmp(name, n_HelpServ))
		return (Me.hsptr);

#endif

	return ((Luser *) NULL);
} /* GetService() */

/*
FindService
  Similar to GetService(), except accept a Luser * pointer
and attempt to find a service match. Returns a pointer
to appropriate Luser structure if 'lptr' is a service nick
*/

Luser *
FindService(Luser *lptr)

{
	if (!lptr)
		return ((Luser *) NULL);

	if (lptr == Me.osptr)
		return (Me.osptr);

#ifdef NICKSERVICES

	if (lptr == Me.nsptr)
		return (Me.nsptr);

	#ifdef CHANNELSERVICES
		if (lptr == Me.csptr)
			return (Me.csptr);
	#endif

	#ifdef MEMOSERVICES
		if (lptr == Me.msptr)
			return (Me.msptr);
	#endif

#endif /* NICKSERVICES */

#ifdef STATSERVICES

	if (lptr == Me.ssptr)
		return (Me.ssptr);

#endif

#ifdef HELPSERVICES

	if (lptr == Me.hsptr)
		return (Me.hsptr);

#endif

	return ((Luser *) NULL);
} /* FindService() */

/*
pwmatch()
	Check if 'password' matches 'chkpass' - return 1 if so, 0 if not
Assumes you're trying to match passwords from *Serv databases
*/

int
pwmatch(char *password, char *chkpass)

{
#ifdef CRYPT_PASSWORDS
	char	*encr;
#endif

	if (!password || !chkpass)
		return 0;

#ifdef CRYPT_PASSWORDS

	encr = (char *) crypt(chkpass, password);
	if (!strcmp(encr, password))
		return 1;
	else
		return 0;

#else

	if (!strcmp(password, chkpass))
		return 1;
	else
		return 0;

#endif
} /* pwmatch() */

/*
operpwmatch()
	Check if 'password' matches 'chkpass' - return 1 if so, 0 if not
Assumes you're trying to match passwords from O: lines in hybserv.conf
*/

int
operpwmatch(char *password, char *chkpass)

{
#ifdef CRYPT_OPER_PASSWORDS
	char	*encr;
#endif

	if (!password || !chkpass)
		return 0;

#ifdef CRYPT_OPER_PASSWORDS

	encr = (char *) crypt(chkpass, password);

	if (!strcmp(encr, password))
		return 1;
	else
		return 0;

#else

	if (!strcmp(password, chkpass))
		return 1;
	else
		return 0;

#endif
} /* operpwmatch() */

/*
IsInNickArray()
 Determine if 'nickname' is in the array 'nicks'
Return 1 if so, 0 if not
*/

int
IsInNickArray(int nickcnt, char **nicks, char *nickname)

{
	int ii;
	char *ntmp;

	for (ii = 0; ii < nickcnt; ii++)
	{
		ntmp = GetNick(nicks[ii]);
		if (!ntmp)
			continue;

		if (!strcasecmp(ntmp, nickname))
			return (1);
	}

	return (0);
} /* IsInNickArray() */

/*
IsNum()
	Determine if 'str' is a number - return 0 if not, the number if it is
*/

int
IsNum(char *str)

{
	int result = 0;
	char	tmp[MAXLINE];
	char	*tmp2;

	if (!str)
		return 0;

	strcpy(tmp, str);
	tmp2 = tmp;
	while (*tmp2)
	{
		if (IsDigit(*tmp2))
		{
			result *= 10;
			result += ((*tmp2) & 0xF);
			tmp2++;
		}
		else
			return 0;
	}

	/* if 'str' actually contains the number 0, return 1 */
	if (result == 0)
		result = 1;

	return (result);
} /* IsNum() */

/*
GetCommand()
 Attempt to find the command "name" in the list "cmdlist".
Return a pointer to the index containing "name" if found,
otherwise NULL.  If the command is found, but there is
more than 1 match (ambiguous), return (Command *) -1.
*/

Command *
GetCommand(Command *cmdlist, char *name)

{
	Command *cmdptr, *tmp;
	int		matches, /* number of matches we've had so far */
				clength;

	if (!cmdlist || !name)
		return ((Command *) NULL);

	tmp = (Command *) NULL;
	matches = 0;
	clength = strlen(name);
	for (cmdptr = cmdlist; cmdptr->cmd; cmdptr++)
	{
		if (!strncasecmp(name, cmdptr->cmd, clength))
		{
			if (clength == strlen(cmdptr->cmd))
			{
				/*
				 * name and cmdptr->cmd are the same length, so it
				 * must be an exact match, don't search any further
				 */
				matches = 0;
				break;
			}
			tmp = cmdptr;
			matches++;
		}
	}

	/*
	 * If matches > 1, name is an ambiguous command, so the
	 * user needs to be more specific
	 */
	if ((matches == 1) && (tmp))
		cmdptr = tmp;

	if (cmdptr->cmd)
		return (cmdptr);

	if (matches == 0)
		return ((Command *) NULL); /* no matches found */
	else
		return ((Command *) -1); /* multiple matches found */
} /* GetCommand() */

/*
CheckAccess()
	args: Userlist *user, char flag
	purpose: determines whether O: line user specified by 'user' has
					 the flag 'flag'.
	return: >= 1 if the user has 'flag', 0 if not 				 
*/

int
CheckAccess(Userlist *user, char flag)

{
	if (user == (Userlist *) NULL)
		return 0;

	switch (flag)
	{
		case '\0':
			return (1);
		case 'a':
		case 'A':
			return (IsAdmin(user));
		case 'd':
		case 'D':
			return (CanChat(user));
		case 'e':
		case 'E':
			return (IsProtected(user));
		case 'f':
		case 'F':
			return (IsFriend(user));
		case 'g':
		case 'G':
			return (CanGline(user));
		case 'j':
		case 'J':
			return (CanJupe(user));
		case 'o':
		case 'O':
			return (IsOper(user));
		case 's':
		case 'S':
			return (IsServicesAdmin(user));

		default:
			return 0;
	}
} /* CheckAccess() */

/*
IsProtectedHost()
	Determines if userhost matches an exception O: line -
returns 1 if so, 0 if not
*/

int
IsProtectedHost(char *username, char *hostname)

{
	Userlist *tempuser;

	if (!username || !hostname)
		return 0;

	for (tempuser = userlist; tempuser; tempuser = tempuser->next)
	{
		if (!(tempuser->flags & PRIV_EXCEPTION))
			continue;

		if (match(username, tempuser->username) &&
				match(hostname, tempuser->hostname))
			return 1;
	}

	return 0;
} /* IsProtectedHost() */

/*
IsRestrictedHost()
	Return pointer to restricted host structure matching 'host'
*/

rHost *
IsRestrictedHost(char *user, char *host)

{
	rHost *temphost;

	for (temphost = rhostlist; temphost; temphost = temphost->next)
	{
		if (match(temphost->username, user) &&
		    match(temphost->hostname, host))
			return (temphost);
	}

	return (rHost *) NULL;
} /* IsRestrictedHost() */

int
MatchesAdmin(char *mask)

{
	Userlist *tempuser;
	char tempstr[MAXLINE];

	if (!mask)
		return (0);

	for (tempuser = userlist; tempuser; tempuser = tempuser->next)
	{
		if (IsAdmin(tempuser))
		{
			sprintf(tempstr, "*!%s@%s",
				tempuser->username,
				tempuser->hostname);

			if (match(mask, tempstr))
				return (1);
		}
	}

	return (0);
} /* MatchesAdmin() */

/* 
HostToMask()
	Convert a userhost like "user@host.name.org" into "*user@*.name.org"
or "user@1.2.3.4" into "*user@1.2.3.*"
*/

char *
HostToMask (char *username, char *hostname)

{
  char *realhost, /* user@host (without nick!) */
       *temp,
       *final, /* final product */
       *user, /* stores the username */
       *host, /* stores the hostname */
       /* 
        * first segment of hostname (ie: if host == "a.b.c",
        * topsegment == "c" (top-level domain)
        */
       *topsegment = NULL;
	char userhost[UHOSTLEN + 2];
	int ii, /* looping */
			cnt;

	if (!username || !hostname)
		return ((char *) NULL);

	sprintf(userhost, "%s@%s",
		username,
		hostname);

	final = (char *) MyMalloc(strlen(userhost) + 32);

	/* strip off a nick nick!user@host (if there is one) */
	realhost = (host = strchr(userhost, '!')) ? host + 1 : userhost;
	user = realhost;
	if ((host = strchr(realhost, '@')))
	{
		final[0] = '*';
		ii = 1;
		/* 
		 * only use the last 8 characters of the username if there 
		 * are more
		 */
		if ((host - realhost) > 10)
			user = host - 8;

		/* now store the username into 'final' */
		while (*user != '@')
			final[ii++] = *user++;

		final[ii++] = '@';

		/* host == "@host.name", so point past the @ */
		host++;
	}
	else
	{
		/* there's no @ in the hostname, just make it *@*.host.com */
		strcpy(final, "*@");
		ii = 2;
		host = userhost;
	}

	/* 
	 * ii now points to the offset in 'final' of where the 
	 * converted hostname should go
	 */

	realhost = strchr(host, '.');
	if (realhost)
		topsegment = strchr(realhost + 1, '.');
	if (!realhost || !topsegment)
	{
		/* 
		 * if realhost is NULL, then the hostname must be a top-level
		 * domain; and if topsegment is NULL, the hostname must be
		 * 2 parts (ie: blah.org), so don't strip off "blah"
		 */
		strcpy(final + ii, host);
	}
	else
	{
		/*
		 * topsegment now contains the top-level domain - if it's
		 * numerical, it MUST be an ip address, since there are
		 * no numerical TLD's =P
		 */

		/* advance to the end of topsegment */
		for (temp = topsegment; *temp; temp++);

		temp--; /* point to the last letter (number) of the TLD */
		if ((*temp >= '0') && (*temp <= '9'))
		{
			/* Numerical IP Address */
			while (*temp != '.')
				temp--;

			/* 
			 * copy the ip address (except the last .XXX) into the 
			 * right spot in 'final'
			 */
			strncpy(final + ii, host, temp - host);

			/* stick a .* on the end :-) */
			ii += (temp - host);
			strcpy(final + ii, ".*");
		}
		else
		{
			/* its a regular hostname with >= 3 segments */

			if (SmartMasking)
			{
				/*
				 * Pick up with temp from were we left off above.
				 * Temp now points to the very last charater of userhost.
				 * Go backwards, counting all the periods we encounter.
				 * If we find 3 periods, make the hostmask:
				 *   *.seg1.seg2.seg3
				 * Since some users may have extremely long hostnames
				 * because of some weird isp. Also, if they have
				 * a second TLD, such as xx.xx.isp.com.au, this
				 * routine will make their mask: *.isp.com.au, which
				 * is much better than *.xx.isp.com.au
				 */
				cnt = 0;
				while (temp--)
				{
					if (*temp == '.')
						if (++cnt >= 3)
							break;
				}

				if (cnt >= 3)
				{
					/*
					 * We have a hostname with more than 3 segments.
					 * Set topsegment to temp, so the final mask
					 * will be *user@*.seg1.seg2.seg3
					 */
					topsegment = temp;
				}
			} /* if (SmartMasking) */

			/*
			 * topsegment doesn't necessarily point to the TLD.
			 * It simply points one segment further than realhost.
			 * Check if there is another period in topsegment,
			 * and if so use it. Otherwise use realhost
			 */
			sprintf(final + ii, "*%s", 
				strchr(topsegment + 1, '.') ? topsegment : realhost);
		}
	}

	return (final);
} /* HostToMask() */
